/*
    Copyright 2008-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>
                        Matthias Nagl <mnagl@uni-goettingen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 3 of the License.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file gpl-3.0.txt. If not, see
    <http://www.gnu.org/licenses/>.
*/

#include "emea.h"
#include "emea.moc"

#include <iostream>

#include <QtCore/QDebug>
#include <QtCore/QTimer>

#include <QtGui/QApplication>
#include <QtGui/QMenuBar>
#include <QtGui/QMenu>
#include <QtGui/QAction>
#include <QtGui/QTreeView>
#include <QtGui/QFileDialog>
#include <QtGui/QScrollArea>
#include <QtGui/QCloseEvent>
#include <QtGui/QToolBar>

#include <thread_template.h>
#include <stackediconwidget.h>
#include <osciwidget_v3.h>
#include <qtreeitemmodel.h>
#include <qtreeitemmodeldelegate.h>
#include <resultviewer.h>
#include <spectrummodel.h>
//#include <tiparameter.h>
#include <netserver.h>

#include "valuemeter.h"

#include "epostreeview.h"

class EMEABase::Private
{
    public:
        Private( EMEABase* p ) {
            logger = new NSCore::StatisticsAndLogging();
            netparent = new NSCore::TiParameter("", QVariant::Invalid, 0);
            new NSCore::TiParameter("Statistics", QVariant::Invalid, netparent);
            netserver = new NSNetwork::NetServer(netparent, p);
        }
        ~Private() {
            delete netserver;
            delete netparent;
        }

        QList<NSCore::DataSetProducer*> producers;
        QList< QList<NSCore::DataSetChainMember*> > chain;

        QList<QThread*> threads;

        void updateChainConnections() {
            //qDebug() << "EMEABase::Private::updateChainConnections()";
            foreach( NSCore::DataSetProducer* producer, producers )
                producer->unregisterAllConsumers();
            foreach( QList<NSCore::DataSetChainMember*> list, chain )
                foreach( NSCore::DataSetProducer* producer, list )
                    producer->unregisterAllConsumers();
            if ( chain.size() > 0 ) {
                foreach( NSCore::DataSetProducer* producer, producers )
                    foreach( NSCore::DataSetChainMember* consumer, chain.at( 0 ) )
                        producer->registerConsumer( consumer );
            }
            for ( int i=1; i<chain.size(); ++i ) {
                foreach( NSCore::DataSetChainMember* producer, chain.at( i-1 ) ) {
                    foreach( NSCore::DataSetChainMember* consumer, chain.at( i ) ) {
                        //qDebug() << i << "\t connecting:" << producer << consumer;
                        producer->registerConsumer( consumer );
                    }
                }
            }
        }
        QList<NSCore::QTreeItem*> getAllRootItems() {
            QList<NSCore::QTreeItem*> items;
            foreach( NSCore::DataSetProducer* producer, producers )
                items.append( producer );
            foreach( QList<NSCore::DataSetChainMember*> level, chain )
                foreach( NSCore::DataSetChainMember* member, level )
                    items.append( member );
            return items;
        }

        void toggleTool( NSCore::DataSetChainMember* item, bool n ) {
            if ( ! chain.isEmpty() )
                foreach( NSCore::DataSetChainMember* member, chain.last() ) {
                    member->unregisterConsumer( item );
                    // Register at last members in chain
                    if ( n && chain.size() > 0 )
                        member->registerConsumer( item );
                }
            if ( ! producers.isEmpty() )
                foreach( NSCore::DataSetProducer* producer, producers ) {
                    producer->unregisterConsumer( item );
                    // Register at the producers directly
                    if ( n && chain.size() == 0 )
                        producer->registerConsumer( item );
                }
        }

        QTimer* timer;
        NSCore::StatisticsAndLogging* logger;
        QTime loggertime;
        QStringList loggeroldkeys;

        NSCore::TiParameter* netparent;
        NSNetwork::NetServer* netserver;

        QMap<QString, NSSpectrum::AbstractSpectrumModel*> spectra;
};

EMEABase::EMEABase( QObject* p )
    : QObject( p )
    , d( new EMEABase::Private( this ) )
{
    qDebug() << "EMEABase::EMEABase(" << p << ")";
    d->timer = new QTimer( this );
    connect( d->timer, SIGNAL( timeout() ), this, SLOT( statisticsTimeout() ) );
    connect( d->timer, SIGNAL( timeout() ), this, SLOT( checkStatus() ) );
    d->timer->setInterval( 2*1000 );
    d->timer->start();
    d->loggertime.start();
    connect( d->netserver, SIGNAL( setStatus(QString) ), this, SLOT( changeStatus(QString) ), Qt::QueuedConnection );
    connect( d->netserver, SIGNAL( saveXml(QString&) ), this, SLOT( saveXmlToString(QString&) ) );
    connect( d->netserver, SIGNAL( loadXml(QString) ), this, SLOT( loadXmlFromString(QString) ), Qt::QueuedConnection );
    connect( d->netserver, SIGNAL( getSpectrum(QString, QByteArray&) ),
            this, SLOT( saveSpectrum(QString, QByteArray&) ) );

    connect( this, SIGNAL( newSpectrum( NSSpectrum::AbstractSpectrumModel* ) ),
            this, SLOT( newSpectrumCreated( NSSpectrum::AbstractSpectrumModel* ) ), Qt::QueuedConnection );
}
EMEABase::~EMEABase() {
    qDebug() << "EMEABase::~EMEABase()";
    while ( ! d->threads.isEmpty() )
        delete d->threads.takeFirst();
}

void EMEABase::addToChainStart( NSCore::DataSetProducer* item ) {
    //qDebug() << "EMEABase::addToChainStart(" << item << ")";
    item->setDispatching( NSCore::DataSetProducer::DispatchEqual );
    d->producers.append( item );
    d->netparent->addChild( item );
    connect( this, SIGNAL( start() ), item, SLOT( start() ) );
    connect( this, SIGNAL( stop()  ), item, SLOT( stop()  ) );
    connect( this, SIGNAL( abort() ), item, SLOT( abort() ) );
    connect( this, SIGNAL( reset() ), item, SLOT( reset() ) );
    emit newTreeItem( item );
}

void EMEABase::addToChain( QList<NSCore::DataSetChainMember*> items, bool add ) {
    //qDebug() << "EMEABase::addToChain(" << items << ")";
    d->chain.append( items );
    if ( add )
        foreach( NSCore::DataSetChainMember* item, items ) {
            addToGui( item );
            connect( this, SIGNAL( abort() ), item, SLOT( abort() ) );
            connect( this, SIGNAL( reset() ), item, SLOT( reset() ) );
        }
}
void EMEABase::addToChain( NSCore::DataSetChainMember* item, bool addToGui ) {
    //qDebug() << "EMEABase::addToChain(" << item << ")";
    addToChain( QList<NSCore::DataSetChainMember*>() << item, addToGui );
}
void EMEABase::addToGui( NSCore::QTreeItem* item ) {
    d->netparent->addChild( item );
    emit newTreeItem( item );
}

void EMEABase::loadXmlFromFile( QString path ) {
    //qDebug() << "EMEABase::loadXmlFromFile(" << qPrintable( path ) << ")";
    QFile f( path );
    if ( f.open( QIODevice::ReadOnly | QIODevice::Text ) ) {
        QDomDocument doc( "eposxml" );
        doc.setContent( &f );
        f.close();
        loadXml( doc );
    }
}
void EMEABase::loadXml( QDomDocument doc ) {
    //qDebug() << "EMEABase::loadXml( doc )";
    QList<NSCore::QTreeItem*> items = d->netparent->childobjects<NSCore::QTreeItem>();
    QDomElement elem = doc.documentElement().firstChildElement();
    while ( ! elem.isNull() ) {
        foreach( NSCore::QTreeItem* item, items ) {
            //qDebug() << QString( " trying %1 on %2" ).arg( elem.nodeName() ).arg( qptrdiff( item ) );
            item->fromXml( elem );
        }
        elem = elem.nextSiblingElement();
    }
    sleep( 2 );
}
void EMEABase::saveXml( QDomNode parentnode, QDomDocument doc ) {
    QDomNode netp = d->netparent->toXml( doc );
    QDomNode tmp = netp.firstChild();
    while ( ! tmp.isNull() ) {
        QDomNode tmp2 = tmp.nextSibling();
        if ( parentnode.firstChildElement( tmp.nodeName() ).isNull() )
            parentnode.appendChild( tmp );
        tmp = tmp2;
    }
}

void EMEABase::addThread( QThread* thread ) {
    d->threads.append( thread );
}

void EMEABase::statisticsTimeout() {
    NSCore::StatisticsAndLogging::statisticsmap map;
    foreach ( NSCore::DataSetProducer* producer, d->producers ) {
        NSCore::StatisticsAndLogging::statisticsmap tmp = producer->logging()->getStatistics();
        QString prodname = QString("%1/%2").arg(producer->objectName());
        foreach ( QString name, tmp.keys() ) {
            map[prodname.arg(name)] += tmp[name];
        }
    }
    foreach ( QList<NSCore::DataSetChainMember*> members, d->chain ) {
        foreach ( NSCore::DataSetChainMember* member, members ) {
            NSCore::StatisticsAndLogging::statisticsmap tmp = member->logging()->getStatistics();
            QString prodname = QString("%1/%2").arg(member->objectName());
            foreach ( QString name, tmp.keys() ) {
                map[prodname.arg(name)] += tmp[name];
            }
        }
    }
    d->logger->addStatistics(map);
    QStringList keys = map.keys();
    if (d->loggeroldkeys != keys) {
        std::cout << "\"secs\" \"time\" ";
        foreach (QString name, keys) {
            std::cout << "\"" << qPrintable(name) << "\" ";
            NSCore::QTreeItem* p = d->netparent->child( "Statistics" );
            while ( name.contains( '/' ) ) {
                p = new NSCore::TiParameter( name.section('/', 0, 0), QVariant::Invalid, p );
                name = name.section( '/', 1 );
            }
            new NSCore::TiParameter( name, QVariant::Double, p );
        }
        std::cout << std::endl;
        d->loggeroldkeys = keys;
    }

    foreach (QString name, keys)
        d->netparent->getParameter( QString( "Statistics/%1" ).arg(name) )->setValue( map[name] );
    emit newLoggingValues( map );

    if ( d->loggertime.elapsed() > 59*1000 ) {
        std::cout << QDateTime::currentDateTime().toTime_t() << " "
            << qPrintable(QDateTime::currentDateTime().toString("yyyyMMdd-hhmmss")) << " ";
        foreach (QString name, keys) {
            std::cout << map[name] << " ";
            d->netparent->getParameter( QString( "Statistics/%1" ).arg(name) )->setValue( map[name] );
        }
        std::cout << std::endl;
        d->loggertime.restart();
    }
}
void EMEABase::changeStatus( QString status ) {
    qDebug() << "EMEABase::changeStatus(" << status << ")";

    if ( status.toLower() == "start" ) emit start();
    if ( status.toLower() == "stop" ) emit stop();
    if ( status.toLower() == "abort" ) emit abort();
    if ( status.toLower() == "reset" ) emit reset();

    checkStatus();
}
void EMEABase::saveXmlToString( QString& xml ) {
    qDebug() << "EMEABase::saveXmlToString()";
    QDomDocument doc( "eposxml" );
    QDomNode node = doc.createElement( "emea" );
    doc.appendChild( node );
    saveXml( node, doc );
    xml = doc.toString().replace("\n", "<br />");
}
void EMEABase::loadXmlFromString( QString xml ) {
    qDebug() << "EMEABase::loadXmlFromString()";
    QDomDocument doc( "eposxml" );
    doc.setContent( xml.replace("<br />", "\n" ) );
    loadXml( doc );
}
void EMEABase::checkStatus() {
    bool running = false;
    foreach ( NSCore::DataSetProducer* prod, d->producers ) {
        running |= prod->isRunning();
    }
    foreach ( QList<NSCore::DataSetChainMember*> members, d->chain ) {
        foreach ( NSCore::DataSetChainMember* member, members ) {
            running |= member->isRunning();
        }
    }
    d->netserver->updateStatus( running?"Running":"Idle" );
}
void EMEABase::newSpectrumCreated( NSSpectrum::AbstractSpectrumModel* spectrum ) {
    qDebug() << "EMEABase::newSpectrumCreated(" << spectrum << ")";
    connect( spectrum, SIGNAL( destroyed( QObject* ) ),
            this, SLOT( spectrumDeleted( QObject* ) ), Qt::DirectConnection );
    d->netserver->addToSpectrumList( spectrum->objectName() );
    d->spectra[ spectrum->objectName() ] = spectrum;
}
void EMEABase::spectrumDeleted( QObject* spectrum ) {
    qDebug() << "EMEABase::spectrumDeleted(" << &spectrum << ")" << spectrum->objectName(); // << dynamic_cast<NSSpectrum::AbstractSpectrumModel*>( spectrum );
    d->netserver->removeFromSpectrumList( spectrum->objectName() );
    d->spectra.take( spectrum->objectName() );
}
void EMEABase::saveSpectrum( QString name, QByteArray& data ) {
    qDebug() << "EMEABase::saveSpectrum(" << name << "," << data << ")";
    if ( ! d->spectra.contains( name ) )
        return;

    QBuffer buffer(&data, this);
    buffer.open(QIODevice::WriteOnly);
    d->spectra[name]->saveTo( &buffer );
    buffer.close();
    //qDebug() << data.data();
}



class EMEAWindow::Private
{
    public:
        Private( QMainWindow* p ) {
            stack = new StackedIconWidget( p );
            osci = new OsciWidgetV3( 0 );
            stack->addWidget( osci );
            viewer = new ResultViewer( 0 );
            QScrollArea* tmp = new QScrollArea( p );
            tmp->setObjectName( viewer->objectName() );
            tmp->setWindowIcon( viewer->windowIcon() );
            tmp->setWidget( viewer );
            tmp->setWidgetResizable( true );
            stack->addWidget( tmp );

            treeview = new EposTreeView( 0 );
            treeview->setObjectName( "Chain" );
            treeview->setWindowIcon( QIcon( ":/icons/kde/cr22-action-view_tree.png" ) );
            stack->addWidget( treeview );
            treeview->setItemDelegate( new QTreeItemModelDelegate( p ) );
            treeview->setEditTriggers(QAbstractItemView::AllEditTriggers);
            treeview->setAlternatingRowColors(true);

            spectrumguicreator = new NSSpectrum::SpectrumGuiCreator( p );
            p->connect( spectrumguicreator, SIGNAL( newSpectrumGui( QWidget* ) ), p, SLOT( addWidgetToStack( QWidget* ) ) );

            guimeter = new ValueMeters( p );
            stack->addWidget( guimeter );

        }

        EMEABase* base;

        StackedIconWidget* stack;
        OsciWidgetV3* osci;
        ResultViewer* viewer;

        QTreeView* treeview;

        QAction* aosciconnect;
        QAction* aresultviewerconnect;

        NSSpectrum::SpectrumGuiCreator* spectrumguicreator;

        ValueMeters* guimeter;

        QList<NSCore::QTreeItem*> getAllRootItems() {
            QList<NSCore::QTreeItem*> items;
            foreach( NSCore::DataSetProducer* producer, base->d->producers )
                items.append( producer );
            foreach( QList<NSCore::DataSetChainMember*> level, base->d->chain )
                foreach( NSCore::DataSetChainMember* member, level )
                    items.append( member );
            return items;
        }

};

EMEAWindow::EMEAWindow( EMEABase* base, QWidget* p )
    : QMainWindow( p )
    , d( new EMEAWindow::Private( this ) )
{
    d->base = base;
    connect( d->base, SIGNAL( newLoggingValues( QMap<QString, double> ) ),
            d->guimeter, SLOT( addLogging( QMap<QString, double> ) ) );
    qDebug() << "EMEAWindow::EMEAWindow(" << p << ")";

    // load window geometry
    QSettings s;
    if (s.contains("mainWindowSize"))
        resize(s.value("mainWindowSize").toSize());
    else
        resize(800, 600);

    NSCore::QTreeItemModel* model = new NSCore::QTreeItemModel( this );
    connect( base, SIGNAL( newTreeItem( NSCore::QTreeItem* ) ),
            model, SLOT( addToRootItem( NSCore::QTreeItem* ) ) );
    d->treeview->setModel( model );

    connect( d->base, SIGNAL( newSpectrum( NSSpectrum::AbstractSpectrumModel* ) ),
            d->spectrumguicreator, SLOT( newSpectrum( NSSpectrum::AbstractSpectrumModel* ) ) );

    setCentralWidget( d->stack );

    QMenu* filemenu = menuBar()->addMenu( "File" );
    QAction* aopen = new QAction( QIcon( ":/icons/kde/cr22-action-fileopen.png" ), "Open xml-file", this );
    connect( aopen, SIGNAL( triggered() ), this, SLOT( loadXmlFromFile() ) );
    filemenu->addAction( aopen );
    QAction* asave = new QAction( QIcon( ":/icons/kde/cr22-action-filesaveas.png" ), "Save to xml-file", this );
    connect( asave, SIGNAL( triggered() ), this, SLOT( saveXmlToFile() ) );
    filemenu->addAction( asave );
    filemenu->addSeparator();
    QAction* aquit = new QAction( QIcon( ":icons/kde/cr22-action-exit.png" ), "Quit", this );
    aquit->setShortcut( Qt::CTRL|Qt::Key_Q );
    connect( aquit, SIGNAL( triggered() ), this, SLOT( close() ) );
    filemenu->addAction( aquit );

    QMenu* runmenu = menuBar()->addMenu( "Run" );
    QToolBar* runtoolbar = addToolBar( "Run Evaluation" );
    QAction* astart = new QAction( QIcon( ":/icons/kde/cr22-action-player_play.png" ), "Start Evaluation", this );
    connect( astart, SIGNAL( triggered() ), d->base, SIGNAL( start() ) );
    runmenu->addAction( astart );
    runtoolbar->addAction( astart );
    QAction* astop = new QAction( QIcon( ":/icons/kde/cr22-action-player_stop.png" ), "Stop Evaluation", this );
    connect( astop, SIGNAL( triggered() ), d->base, SIGNAL( stop() ) );
    runmenu->addAction( astop );
    runtoolbar->addAction( astop );
    QAction* areset = new QAction( QIcon( ":/icons/kde/cr22-action-player_start.png" ), "Reset Evaluation", this );
    connect( areset, SIGNAL( triggered() ), d->base, SIGNAL( reset() ) );
    runmenu->addAction( areset );
    runtoolbar->addAction( areset );
    QAction* aabort = new QAction( QIcon( ":/icons/kde/cr22-action-stop.png" ), "Abort Evaluation", this );
    connect( aabort, SIGNAL( triggered() ), d->base, SIGNAL( abort() ) );
    runmenu->addAction( aabort );
    runtoolbar->addAction( aabort );

    QMenu* toolsmenu = menuBar()->addMenu( "Tools" );
    d->aosciconnect = new QAction( "Connect Osci", this );
    d->aosciconnect->setCheckable( true );
    connect( d->aosciconnect, SIGNAL( toggled( bool ) ), this, SLOT( toggleOsci( bool ) ) );
    toolsmenu->addAction( d->aosciconnect );
    d->aresultviewerconnect = new QAction( "Connect Result-Viewer", this );
    d->aresultviewerconnect->setCheckable( true );
    connect( d->aresultviewerconnect, SIGNAL( toggled( bool ) ), this, SLOT( toggleViewer( bool ) ) );
    toolsmenu->addAction( d->aresultviewerconnect );

    QTimer::singleShot( 1, this, SLOT( lateInit() ) );
}
EMEAWindow::~EMEAWindow() {
    qDebug() << "EMEAWindow::~EMEAWindow()";
}

void EMEAWindow::closeEvent( QCloseEvent* ev ) {
    //qDebug() << "EMEAWindow::closeEvent(" << ev << ")";

    // save window geometry
    QSettings s;
    s.setValue("mainWindowSize", size());
    
    d->stack->unfloatAllWidgets();
    ev->accept();
}

void EMEAWindow::lateInit() {
    //qDebug() << "EMEAWindow::lateInit()";
    QStringList arguments = QCoreApplication::arguments();

    /// Get the number of threads wanted/possible
    int threads = (QThreadPool::globalInstance()->maxThreadCount() - QThreadPool::globalInstance()->activeThreadCount()) / 3;
    for ( QStringList::iterator it=arguments.begin(); it!=arguments.end(); ++it )
        if ( *it == "-threads" )
            threads = ( ++it )->toInt();
    threads = qMax( 1, threads );

    d->base->setupChain( threads );
    d->base->d->updateChainConnections();

    
    /// React to the arguments given on the commandline
    // Load xml-files
    foreach( QString arg, arguments )
        if ( arg.endsWith( ".xml" ) )
            if ( QFile::exists( arg ) )
                d->base->loadXmlFromFile( arg );
    // Autostart the evaluation
    if ( arguments.contains( "-autostart" ) )
        QTimer::singleShot( 2000, d->base, SIGNAL( start() ) );
    else {
        d->aosciconnect->setChecked( true );
        d->aresultviewerconnect->setChecked( true );
    }
    // Autostop the evaluation
    if ( arguments.contains( "-autostop" ) ) {
        foreach ( NSCore::DataSetProducer* producer, d->base->d->producers ) {
            connect( producer, SIGNAL( finished() ), this, SLOT( delayStop() ) );
        }
    }
    //qDebug() << " lateInit() finished";

    // adjust width of first col
    //d->treeview->resizeColumnToContents(0);
    //d->treeview->setColumnWidth(0, d->treeview->columnWidth(0)*12/10);
    d->treeview->expandToDepth( 1 );
    d->treeview->setColumnWidth(0, d->treeview->size().width()/2);
}

void EMEAWindow::delayStop() {
    QTimer* timer = new QTimer( this );
    timer->setInterval( 10000 );
    timer->setSingleShot( true );
    connect( timer, SIGNAL( timeout() ), d->spectrumguicreator, SIGNAL( autoSave() ) );
    connect( timer, SIGNAL( timeout() ), this, SLOT( close() ) );
    timer->start();
}

void EMEAWindow::toggleOsci( bool n ) {
    //qDebug() << "EMEAWindow::toggleOsci(" << n << ")";
    d->base->d->toggleTool( d->osci->chainMember(), n );
}
void EMEAWindow::toggleViewer( bool n ) {
    //qDebug() << "EMEAWindow::toggleViewer(" << n << ")";
    d->base->d->toggleTool( d->viewer->chainMember(), n );
}

void EMEAWindow::addWidgetToStack( QWidget* w, bool scrollarea ) {
    //qDebug() << "EMEAWindow::addWidgetToStack(" << w << ")";
    if ( scrollarea ) {
        QScrollArea* tmp = new QScrollArea( this );
        tmp->setObjectName( w->objectName() );
        tmp->setWindowIcon( w->windowIcon() );
        tmp->setWidget( w );
        tmp->setWidgetResizable( true );
        d->stack->addWidget( tmp );
    } else
        d->stack->addWidget( w );
}


void EMEAWindow::loadXmlFromFile() {
    //qDebug() << "EMEAWindow::loadXmlFromFile()";
    QString path = QFileDialog::getOpenFileName( this, "File to load from", QString(), "xml-Files (*.xml)" );
    if ( ! path.isEmpty() )
        d->base->loadXmlFromFile( path );
}
void EMEAWindow::saveXmlToFile() {
    //qDebug() << "EMEAWindow::saveXmlToFile()";
    QString path = QFileDialog::getSaveFileName( this, "Save to file", QString(), "xml-Files (*.xml)" );
    if ( ! path.isEmpty() ) {
        QFile f( path );
        if ( f.open( QIODevice::WriteOnly | QIODevice::Text ) ) {
            QDomDocument doc( "eposxml" );
            QDomNode node = doc.createElement( "emea" );
            doc.appendChild( node );
            d->base->saveXml( node, doc );
            f.write( doc.toByteArray() );
            f.close();
        }
    }
}




class EMEAText::Private
{
    public:
        EMEABase* base;

        Private( QObject* p ) {
        }
};

EMEAText::EMEAText( EMEABase* base ) : QObject( 0 ), d( new EMEAText::Private( this ) ) {
    qDebug() << "EMEAText::EMEAText(" << base << ")";
    d->base = base;

    connect( d->base, SIGNAL( newSpectrum( NSSpectrum::AbstractSpectrumModel* ) ),
            this, SLOT( newSpectrum( NSSpectrum::AbstractSpectrumModel* ) ) );

    QTimer::singleShot( 1, this, SLOT( lateInit() ) );
}
EMEAText::~EMEAText() {
    qDebug() << "EMEAText::~EMEAText()";
}

void EMEAText::lateInit() {
    QStringList arguments = QCoreApplication::arguments();

    /// Get the number of threads wanted/possible
    int threads = (QThreadPool::globalInstance()->maxThreadCount() - QThreadPool::globalInstance()->activeThreadCount()) / 3;
    for ( QStringList::iterator it=arguments.begin(); it!=arguments.end(); ++it )
        if ( *it == "-threads" )
            threads = ( ++it )->toInt();
    threads = qMax( 1, threads );

    d->base->setupChain( threads );
    d->base->d->updateChainConnections();

    // Load xml-files
    foreach( QString arg, arguments )
        if ( arg.endsWith( ".xml" ) )
            if ( QFile::exists( arg ) )
                d->base->loadXmlFromFile( arg );

    QTimer::singleShot( 5000, d->base, SIGNAL( start() ) );

    foreach ( NSCore::DataSetProducer* producer, d->base->d->producers ) {
        connect( producer, SIGNAL( finished() ), this, SLOT( delayStop() ) );
    }
}


void EMEAText::delayStop() {
    qDebug() << "EMEAText::delayStop() Will stop the evaluation in a few seconds.";
    QTimer* timer = new QTimer( this );
    timer->setInterval( 10000 );
    timer->setSingleShot( true );
    connect( timer, SIGNAL( timeout() ), this, SIGNAL( autoSave() ) );
    connect( timer, SIGNAL( timeout() ), QCoreApplication::instance(), SLOT( quit() ) );
    timer->start();
}

void EMEAText::newSpectrum( NSSpectrum::AbstractSpectrumModel* spectrum ) {
    connect( this, SIGNAL( autoSave() ), spectrum, SLOT( autoSave() ) );
}


