/*
    Copyright 2008-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 3 of the License.

    This program is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file gpl-3.0.txt. If not, see
    <http://www.gnu.org/licenses/>.
*/
#ifndef EMEA_H
#define EMEA_H

#include <QtGui/QMainWindow>
#include <QtGui/QApplication>
#include <QtCore/QThreadPool>
#include <QtXml/QDomDocument>

#include <thread_template.h>
#include <spectrumgui.h>
#include <statisticslogging.h>

namespace NSCore {
    class DataSetProducer;
    class DataSetChainMember;
    class QTreeItem;
};

/**
 * @brief Base for the framework.
 */
class EMEABase : public QObject
{
    Q_OBJECT
    public:
        /// default constructor
        EMEABase( QObject* );
        /// destructor
        ~EMEABase();

        /**
         * @brief initialize the chain of evaluation
         *
         * This function is called at the right time during startup to
         * initialize the needed chain of producers and chain-members for the
         * given task.
         *
         * Your compiler will tell you that you have to implement this function.
         * Trust him.
         *
         * Use addToChainStart() and addToChain() to add elements to the chain
         * and use addWidgetToStack() to add widgets to the stacked gui.
         */
        virtual void setupChain( int threads ) =0;

        /**
         * @brief Add elements to the start of the chain.
         */
        void addToChainStart( NSCore::DataSetProducer* );
        /**
         * @{
         *
         * @brief Add elements to the chain
         *
         * Each call will create a new level in the chain. The elements of each
         * level are connected with all elements of the previous (thus
         * producing) level and with all elements of the next (consuming) level.
         */
        void addToChain( NSCore::DataSetChainMember*, bool addToGui=true );
        void addToChain( QList<NSCore::DataSetChainMember*>, bool addToGui=true );
        // @}
        void addToGui( NSCore::QTreeItem* );

        /**
         * @brief Add the threads to a list that gets stopped automaticly
         */
        void addThread( QThread* );
        /**
         * @brief Instantiate an object in its own thread and return the pointer
         *
         * This one makes multi-thread emea apps most easy. Simply create
         * the object via this template function and you get threads for free.
         *
         * You have to add the objects to the chain by hand.
         */
        template<class ThreadClass>
        ThreadClass* objectInNewThread( bool forceown =true ) {
            ThreadClass* ret = 0;
            QThreadPool* pool = QThreadPool::globalInstance();
            if ( pool->maxThreadCount() > pool->activeThreadCount() || forceown ) {
                NSCore::Thread_Template<ThreadClass>* t = new NSCore::Thread_Template<ThreadClass>( this );
                t->start();
                t->waitForStarted();
                pool->reserveThread();
                addThread( t );
                ret = t->item();
            } else {
                ret = new ThreadClass( this );
            }
            return ret;
        }

    signals:

        void newSpectrum( NSSpectrum::AbstractSpectrumModel* );

        void newTreeItem( NSCore::QTreeItem* );

        void newLoggingValues( QMap<QString, double> );

        void start();
        void stop();
        void abort();
        void reset();

    public slots:
        /**
         * @{
         * @brief Xml-stuff...
         */
        void loadXmlFromFile( QString );
        void loadXmlFromString( QString );
        void loadXml( QDomDocument );

        void saveXml( QDomNode, QDomDocument );
        // @}

        /**
         * @brief Status-functions
         */
        void checkStatus();

    private slots:
        void statisticsTimeout();

        void changeStatus( QString newstatus );

        void saveXmlToString( QString& );

        /**
         * @{
         * @brief handle the spectra
         */
        void newSpectrumCreated( NSSpectrum::AbstractSpectrumModel* );
        void spectrumDeleted( QObject* );
        void saveSpectrum( QString name, QByteArray& data );
        // @}

    public:
        class Private;
        Private* d;
};

/**
 * @brief Generic framework for various evaluation apps.
 *
 * EMEA stands for Epos Modularized Evaluation Applications.
 *
 * @todo Tutorial how to write emea-apps
 */
class EMEAWindow : public QMainWindow
{
    Q_OBJECT
    public:
        /// default constructor
        EMEAWindow( EMEABase*, QWidget* =0 );
        /// destructor
        ~EMEAWindow();

        /**
         * @brief Intercept the close event
         */
        void closeEvent( QCloseEvent* );

    protected slots:
        /**
         * @brief Add a widget to the stacked gui
         *
         * When the widget has an object name (set with setObjectName()) this
         * will be used, if it has an icon (set with setWindowIcon()) this will
         * be used too.
         *
         * @param w The widget to add.
         * @param scrollarea Whether the widget should be added inside a QScrollArea. Defaults to false.
         */
        void addWidgetToStack( QWidget* w, bool scrollarea = false );

    private slots:
        /**
         * @brief Do the real setup after all the constructors are run.
         *
         * C++ forbids to call the virtual setupChain() from the constructor of
         * EMEAWindow, so we schedule lateInit() first after entering the event
         * -loop.
         */
        void lateInit();

        /**
         * @brief Switch the connection to the osci
         */
        void toggleOsci( bool );

        /**
         * @brief Switch the connection to the result-viewer
         */
        void toggleViewer( bool );

        /**
         * @brief Delay the stop a bit to catch all datasets still in the queue.
         */
        void delayStop();

        /**
         * @{
         * @brief Xml-stuff...
         */
        void loadXmlFromFile();

        void saveXmlToFile();
        // @}

    private:
        class Private;
        Private* d;
};


class EMEAText : public QObject
{
    Q_OBJECT
    public:
        EMEAText( EMEABase* );
        ~EMEAText();

    public slots:

        void lateInit();

        void delayStop();

    signals:

        void autoSave();

    private slots:
        void newSpectrum( NSSpectrum::AbstractSpectrumModel* );

    private:
        class Private;
        Private* d;
};

template<class BaseClass> int mymain( int argc, char** argv )
{
    bool text = false;
    for ( int i=1; i<argc; i++ ) {
        if ( QString( argv[i] ) == "-text" ) {
            qDebug() << "Text-only requested";
            text = true;
        }
    }
    QCoreApplication::setOrganizationName("epos.fzd.de");
    QCoreApplication::setOrganizationDomain("epos.fzd.de");
    QCoreApplication::setApplicationName("emea");
    if ( text ) {

        QCoreApplication app( argc, argv );

        EMEABase* b = new BaseClass( &app );

        EMEAText* t = new EMEAText( b );

        int ret = app.exec();

        delete t;
        delete b;

        return ret;

    } else {

        QApplication app( argc, argv );

        EMEABase* b = new BaseClass( &app );

        EMEAWindow* w = new EMEAWindow( b );
        w->show();

        int ret = app.exec();

        delete w;
        delete b;

        return ret;
    }
}

#define EMEA_MAIN( baseclass ) \
    int main( int argc, char** argv )\
    { return mymain<baseclass>( argc, argv ); }

#endif // EMEA_H
