/*
    Copyright 2005-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "convolvedproducer.h"
#include "convolvedproducer.moc"

#include "random.h"
#include <QtCore/QDateTime>
#include <QtCore/QTimer>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>

#include <tiaction.h>
#include <tiparameter.h>

using namespace NSCore;

namespace NSCore {
    struct cache_t {
        int samples, bits;
        double noise;
        double g1sigma;
        double g2sigma;
        double g2fraction;
        double decay;
        double timingsigma;
        double *heights;
        double *heightsigmas;

        double *d;
    };
};

ConvolvedProducer::ConvolvedProducer( QObject* p )
    : DataSetProducer( "Convolved-Producer", p )
    , _togenerate( 0 )
    , _p( new TiParameter( "Parameters", QVariant::Invalid, this ) )
    , _dirty( true )
    , _cache( new cache_t )
{
    _cache->d = 0;
    _cache->heights = new double[ 2 ];
    _cache->heightsigmas = new double[ 2 ];
    srand( QDateTime::currentDateTime().toTime_t() );
    _ownthread = ( thread() != QCoreApplication::instance()->thread() );

    connect( _p, SIGNAL( guiValueChanged() ), this, SLOT( cacheDirty() ) );
    _p->addParameter( "Samples", 100 );
    _p->getParameter( "Samples" )->setRange( 0, 1024 );
    _p->addParameter( "Resolution", QVariant( 0 ) );
    _p->getParameter( "Resolution" )->setRange( 0, 64 );
    _p->getParameter( "Resolution" )->setToolTip( "Special value 0 means to use double-precision" );
    _p->addParameter( "Noise Level", 0.1 );
    _p->getParameter( "Noise Level" )->setRange( 0.0, 100.0 );
    _p->addParameter( "Timing Sigma", 0.01 );
    _p->getParameter( "Timing Sigma" )->setRange( 0.0, 100.0 );
    _p->getParameter( "Timing Sigma" )->setToolTip( "The sigma for the timing-distribution." );
    _p->addParameter( "Gauss Sigma", 2.5 );
    _p->getParameter( "Gauss Sigma" )->setRange( 0.0, 100.0 );
    _p->addParameter( "Gauss2 Sigma", 10.0 );
    _p->getParameter( "Gauss2 Sigma" )->setRange( 0.0, 100.0 );
    _p->addParameter( "Gauss2 Fraction", 0.1 );
    _p->getParameter( "Gauss2 Fraction" )->setRange( 0.0, 1.0 );
    _p->addParameter( "Decay", 20.0 );
    _p->getParameter( "Decay" )->setRange( 0.0, 100.0 );
    _p->addParameter( "Channel 1/Height", 1.1 );
    _p->addParameter( "Channel 1/Sigma", 0.2 );
    _p->addParameter( "Channel 2/Height", 0.5 );
    _p->addParameter( "Channel 2/Sigma", 0.2 );

    new TiAction( "Generate one event!", this, ":/icons/kde/cr22-action-player_play.png", this, SLOT( generateOneDataSet() ) );
    TiAction* action = new TiAction( "Generate N events!", this, ":/icons/kde/cr22-action-player_fwd.png", this, SLOT( generateNDataSets() ) );
    _number = new TiParameter( "Events to generate", 1, action );
}
ConvolvedProducer::~ConvolvedProducer() {
    delete[] _cache->heights;
    delete[] _cache->heightsigmas;
}

void ConvolvedProducer::start() {
    generateNDataSets();
}
void ConvolvedProducer::stop() {
    generateNDataSets( 0 );
}
void ConvolvedProducer::generateNDataSets() {
    generateNDataSets( _number->toInt() );
}
void ConvolvedProducer::generateNDataSets( int n ) {
    //qDebug() << "ConvolvedProducer::generateNDataSets(" << n << ")";
    if ( _togenerate == 0 )
        QTimer::singleShot( 100, this, SLOT( generateDataSet() ) );
    _togenerate += n;
}

QDomNode ConvolvedProducer::saveToXML( QDomDocument doc ) {
    QDomElement node = doc.createElement( "reader" );
    node.setAttribute( "type", "convolvedproducer" );
    return node;
}

bool ConvolvedProducer::restoreFromXML( QDomNode node ) {
    if ( node.nodeName() != "reader" || !node.isElement() || node.toElement().attribute( "type" ) != "convolvedproducer" )
        return false;
    return true;
}

void ConvolvedProducer::generateDataSet() {

    if ( QThread::currentThread() != thread() ) {
        QTimer::singleShot( 10, this, SLOT( generateDataSet() ) );
        return;
    }

    if ( _dirty ) {
        _cache->samples = _p->getParameter( "Samples" )->toInt();
        _cache->bits = _p->getParameter( "Resolution" )->toInt();
        _cache->noise = _p->getParameter( "Noise Level" )->toDouble();
        _cache->g1sigma = _p->getParameter( "Gauss Sigma" )->toDouble();
        _cache->g2fraction = _p->getParameter( "Gauss2 Fraction" )->toDouble();
        _cache->g2sigma = _p->getParameter( "Gauss2 Sigma" )->toDouble();
        _cache->decay = _p->getParameter( "Decay" )->toDouble();
        _cache->timingsigma = _p->getParameter( "Timing Sigma" )->toDouble();

        for ( int i=0; i<2; ++i ) {
            _cache->heights[ i ] = _p->getParameter( QString( "Channel %1/Height" ).arg( i+1 ) )->toDouble();
            _cache->heightsigmas[ i ] = _p->getParameter( QString( "Channel %1/Sigma" ).arg( i+1 ) )->toDouble();
        }

        if ( _cache->d != 0 ) {
            delete _cache->d;
            _cache->d = 0;
        }
        _cache->d = new double[ _cache->samples ];
        for ( int i=0; i<_cache->samples; i+=1 ) {
            _cache->d[ i ] = exp( - i / _cache->decay );
        }

        _dirty = false;
    }

    //qDebug() << "ConvolvedProducer::generateDataSet()";

    for ( int c = 512; c>0 && _togenerate > 0; --c ) {
        //qDebug() << _togenerate << "datasets to go";

        const int channels = 2;

        AcqisitionDataSet data( channels, _cache->samples );

        const double offset = 0;

        const double fullrange = int( pow( 2, _cache->bits ) );

        double pos = 0;

        //qDebug() << "Have to generate data for" << data.channels() << "channels.";

        for ( int i=0; i<data.channels(); ++i ) {

            // Move the first pulse a random amount and the second one relative to the first one
            if ( i==0 )
                pos = rand()*4.0/RAND_MAX - 2;
            else {
                double p = _cache->timingsigma * polarmethodrandom();
                pos += p;
                data.addResult( "horizontalposition", p );
            }

            double noise = 0;

            double impulseheight = polarmethodrandom() * _cache->heightsigmas[ i ] + _cache->heights[ i ];
            data.addResult( QString( "channel %1/given energy" ).arg( i ), impulseheight );

            // Cache to be faster...
            double *channel = data.getRawChannel( i );
            double s = data.samples();
            for ( int j=0; j<s; ++j )
                channel[ j ] = 0.0;

            for ( int j=0; j<s; ++j ) {
                noise = ( rand()*2.0/RAND_MAX - 1 ) * _cache->noise;
                double g = - 1/( 2* pow( _cache->g1sigma, 2 ) ) * exp( - pow( j - _cache->samples/2 - pos, 2 ) / (2* pow( _cache->g1sigma, 2 ) ) );
                g += - _cache->g2fraction/( 2* pow( _cache->g2sigma, 2 ) ) * exp( - pow( j - _cache->samples/2 - pos, 2) / (2* pow( _cache->g2sigma, 2 ) ) );
                for ( int d=0; j+d<s; ++d ) {
                    channel[ j+d ] += g * _cache->d[ d ];
                }
                channel[ j ] *= impulseheight;
                channel[ j ] += noise;
                if ( _cache->bits != 0 )
                    channel[ j ] = int( channel[ j ] * fullrange ) / fullrange;
            }

            data.addResult( QString( "channel %1/fullscale" ).arg( i ), 1.0 );
            data.addResult( QString( "channel %1/offset" ).arg( i ), 0.45 - offset );
        }
        data.updateMinMax();

        enqueue( data );

        --_togenerate;
    }

    if ( _togenerate > 0 ) {
        if ( _ownthread )
            QTimer::singleShot( 0, this, SLOT( generateDataSet() ) );
        else
            QTimer::singleShot( 10, this, SLOT( generateDataSet() ) );
    } else
        emit finished();
}


SingleConvolvedProducer::SingleConvolvedProducer( QObject* p )
    : DataSetProducer( "SingleConvolved-Producer", p )
    , _togenerate( 0 )
    , _p( new TiParameter( "Parameters", QVariant::Invalid, this ) )
    , _dirty( true )
    , _cache( new cache_t )
{
    _cache->d = 0;
    _cache->heights = new double[ 1 ];
    _cache->heightsigmas = new double[ 1 ];
    srand( QDateTime::currentDateTime().toTime_t() );
    _ownthread = ( thread() != QCoreApplication::instance()->thread() );

    connect( _p, SIGNAL( guiValueChanged() ), this, SLOT( cacheDirty() ) );
    _p->addParameter( "Samples", 100 );
    _p->getParameter( "Samples" )->setRange( 0, 1024 );
    _p->addParameter( "Resolution", QVariant( 0 ) );
    _p->getParameter( "Resolution" )->setRange( 0, 64 );
    _p->getParameter( "Resolution" )->setToolTip( "Special value 0 means to use double-precision" );
    _p->addParameter( "Noise Level", 0.1 );
    _p->getParameter( "Noise Level" )->setRange( 0.0, 100.0 );
    _p->addParameter( "Gauss Sigma", 2.5 );
    _p->getParameter( "Gauss Sigma" )->setRange( 0.0, 100.0 );
    _p->addParameter( "Gauss2 Sigma", 10.0 );
    _p->getParameter( "Gauss2 Sigma" )->setRange( 0.0, 100.0 );
    _p->addParameter( "Gauss2 Fraction", 0.1 );
    _p->getParameter( "Gauss2 Fraction" )->setRange( 0.0, 1.0 );
    _p->addParameter( "Decay", 20.0 );
    _p->getParameter( "Decay" )->setRange( 0.0, 100.0 );
    _p->addParameter( "Channel 1/Minimum", 0.3 );
    _p->addParameter( "Channel 1/Maximum", 0.9 );

    new TiAction( "Generate one event!", this, ":/icons/kde/cr22-action-player_play.png", this, SLOT( generateOneDataSet() ) );
    TiAction* action = new TiAction( "Generate N events!", this, ":/icons/kde/cr22-action-player_fwd.png", this, SLOT( generateNDataSets() ) );
    _number = new TiParameter( "Events to generate", 1, action );
}
SingleConvolvedProducer::~SingleConvolvedProducer() {
    delete[] _cache->heights;
    delete[] _cache->heightsigmas;
}

void SingleConvolvedProducer::start() {
    generateNDataSets();
}
void SingleConvolvedProducer::stop() {
    generateNDataSets( 0 );
}

void SingleConvolvedProducer::generateNDataSets() {
    generateNDataSets( _number->toInt() );
}
void SingleConvolvedProducer::generateNDataSets( int n ) {
    //qDebug() << "SingleConvolvedProducer::generateNDataSets(" << n << ")";
    if ( _togenerate == 0 )
        QTimer::singleShot( 100, this, SLOT( generateDataSet() ) );
    _togenerate += n;
}

QDomNode SingleConvolvedProducer::saveToXML( QDomDocument doc ) {
    QDomElement node = doc.createElement( "reader" );
    node.setAttribute( "type", "singleconvolvedproducer" );
    return node;
}

bool SingleConvolvedProducer::restoreFromXML( QDomNode node ) {
    if ( node.nodeName() != "reader" || !node.isElement() || node.toElement().attribute( "type" ) != "singleconvolvedproducer" )
        return false;
    return true;
}

void SingleConvolvedProducer::generateDataSet() {

    if ( QThread::currentThread() != thread() ) {
        QTimer::singleShot( 10, this, SLOT( generateDataSet() ) );
        return;
    }

    if ( _dirty ) {
        _cache->samples = _p->getParameter( "Samples" )->toInt();
        _cache->bits = _p->getParameter( "Resolution" )->toInt();
        _cache->noise = _p->getParameter( "Noise Level" )->toDouble();
        _cache->g1sigma = _p->getParameter( "Gauss Sigma" )->toDouble();
        _cache->g2fraction = _p->getParameter( "Gauss2 Fraction" )->toDouble();
        _cache->g2sigma = _p->getParameter( "Gauss2 Sigma" )->toDouble();
        _cache->decay = _p->getParameter( "Decay" )->toDouble();

        _cache->heights[ 0 ] = _p->getParameter( "Channel 1/Maximum" )->toDouble();
        _cache->heightsigmas[ 0 ] = _p->getParameter( "Channel 1/Minimum" )->toDouble();

        if ( _cache->d != 0 ) {
            delete _cache->d;
            _cache->d = 0;
        }
        _cache->d = new double[ _cache->samples ];
        for ( int i=0; i<_cache->samples; i+=1 ) {
            _cache->d[ i ] = exp( - i / _cache->decay );
        }

        _dirty = false;
    }

    //qDebug() << "SingleConvolvedProducer::generateDataSet()";

    for ( int c = 512; c>0 && _togenerate > 0; --c ) {
        //qDebug() << _togenerate << "datasets to go";

        const int channels = 1;

        AcqisitionDataSet data( channels, _cache->samples );

        const double offset = 0;

        const double fullrange = int( pow( 2, _cache->bits ) );

        double pos = 0;

        //qDebug() << "Have to generate data for" << data.channels() << "channels.";

        for ( int i=0; i<data.channels(); ++i ) {

            // Move the pulse a random amount
            pos = rand()*4.0/RAND_MAX - 2;
            data.addResult( "given time", pos );

            double noise = 0;

            double impulseheight = _cache->heightsigmas[0] + (_cache->heights[0]-_cache->heightsigmas[0]) * rand()*1.0/RAND_MAX;
            /*polarmethodrandom() * _cache->heightsigmas[ i ] + _cache->heights[ i ];*/
            data.addResult( QString( "channel %1/given energy" ).arg( i ), impulseheight );

            // Cache to be faster...
            double *channel = data.getRawChannel( i );
            double s = data.samples();
            for ( int j=0; j<s; ++j )
                channel[ j ] = 0.0;

            for ( int j=0; j<s; ++j ) {
                noise = ( rand()*2.0/RAND_MAX - 1 ) * _cache->noise;
                double g = - 1/( 2* pow( _cache->g1sigma, 2 ) ) * exp( - pow( j - _cache->samples/2 - pos, 2 ) / (2* pow( _cache->g1sigma, 2 ) ) );
                g += - _cache->g2fraction/( 2* pow( _cache->g2sigma, 2 ) ) * exp( - pow( j - _cache->samples/2 - pos, 2) / (2* pow( _cache->g2sigma, 2 ) ) );
                for ( int d=0; j+d<s; ++d ) {
                    channel[ j+d ] += g * _cache->d[ d ];
                }
                channel[ j ] *= impulseheight;
                channel[ j ] += noise;
                if ( _cache->bits != 0 )
                    channel[ j ] = int( channel[ j ] * fullrange ) / fullrange;
            }

            data.addResult( QString( "channel %1/fullscale" ).arg( i ), 1.0 );
            data.addResult( QString( "channel %1/offset" ).arg( i ), 0.45 - offset );
        }
        data.updateMinMax();

        enqueue( data );

        --_togenerate;
    }

    if ( _togenerate > 0 ) {
        if ( _ownthread )
            QTimer::singleShot( 0, this, SLOT( generateDataSet() ) );
        else
            QTimer::singleShot( 10, this, SLOT( generateDataSet() ) );
    } else
        emit finished();
}

