/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "datasetchain.h"
#include "datasetchain.moc"

#include <QtCore/QDebug>
#include <QtCore/QThread>
#include <QtCore/QTime>
#include <QtCore/QTimer>
#include <QtCore/QPointer>
#include <QtCore/QSemaphore>

#include <sys/prctl.h>

#include "statisticslogging.h"

namespace NSCore {

	class DataSetProducer::DataSetProducerPrivate {
		public:
			QReadWriteLock* consumerslock;
			QList< QPointer<DataSetChainMember> > consumers;
			DataSetProducerPrivate()
			{
				consumerslock = new QReadWriteLock(/* QReadWriteLock::Recursive*/ );
				counter = 0;
				dispatching = DispatchToAll;
				dispatchiterator = 0; //consumers.begin();

				logger = new StatisticsAndLogging();
			}
			~DataSetProducerPrivate() {
				delete consumerslock;
				delete logger;
			}

			int counter;

			int dispatching;
			//QList< QPointer<DataSetChainMember> >::Iterator dispatchiterator;
			int dispatchiterator;

			StatisticsAndLogging* logger;
	};

	DataSetProducer::DataSetProducer( QString n, QObject* p ) : QTreeItem( n, p ), _d( new DataSetProducerPrivate() ) {
		//qDebug() << "DataSetProducer::DataSetProducer() living in thread" << QThread::currentThread();
		QTimer* t = new QTimer( this );
		connect( t, SIGNAL( timeout() ), this, SLOT( countrate() ) );
		t->start( 1000 );
	}
	DataSetProducer::~DataSetProducer() {
		//qDebug() << "DataSetProducer::~DataSetProducer()";
		QWriteLocker( _d->consumerslock );
		while ( ! _d->consumers.isEmpty() )
			_d->consumers.removeFirst();
		delete _d;
	}
	void DataSetProducer::countrate() {
		_d->logger->addStatistics( "countrate", _d->counter );
		_d->counter = 0;
	}

	void DataSetProducer::registerConsumer( DataSetChainMember* consumer ) {
		//qDebug() << "DataSetProducer::registerConsumer(" << consumer << ")";
		if ( consumer == 0 )
			qWarning() << "Tried to register Consumer with address 0!";
		else {
			QWriteLocker( _d->consumerslock );
			if (!_d->consumers.contains(consumer)) {
				_d->consumers.append( QPointer<DataSetChainMember>( consumer ) );
				connect( consumer, SIGNAL( destroyed( NSCore::DataSetChainMember* ) ),
					this, SLOT( unregisterConsumer( NSCore::DataSetChainMember* ) ) );
			}
		}
	}
	void DataSetProducer::unregisterConsumer( DataSetChainMember* consumer ) {
		//qDebug() << "DataSetProducer::unregisterConsumer(" << consumer << ")";
		QWriteLocker( _d->consumerslock );
		foreach( QPointer<DataSetChainMember> pointer, _d->consumers ) {
			if ( pointer.data() == consumer )
				_d->consumers.removeAll( pointer );
		}
	}
	void DataSetProducer::unregisterAllConsumers() {
		while ( ! _d->consumers.isEmpty() )
			unregisterConsumer( _d->consumers.first().data() );
	}

	void DataSetProducer::setDispatching( Dispatching n ) {
		_d->dispatching = n;
		_d->dispatchiterator = 0;
	}

	void DataSetProducer::enqueue( AcqisitionDataSet data ) {
		//qDebug() << "DataSetProducer::enqueue(" << &data << ") this =" << this;
		if ( !( _d->consumers.isEmpty() ) ) {
			QReadLocker( _d->consumerslock );
			if ( _d->dispatching == DispatchEqual ) {
				if ( _d->dispatchiterator >= _d->consumers.size() ) {
					_d->dispatchiterator = 0;
				}
				_d->consumers.at( _d->dispatchiterator ).data()->addToQueue( data );
				_d->dispatchiterator += 1;
			}
			if ( _d->dispatching == DispatchToAll ) {
				for ( int i=0; i<_d->consumers.size(); ++i ) {
					_d->consumers.at( i ).data()->addToQueue( data );
				}
			}
		}
		_d->counter++;
	}

	int DataSetProducer::consumersSize() const {
		QReadLocker( _d->consumerslock );
		return _d->consumers.size();
	}
	bool DataSetProducer::consumersIsEmpty() const {
		QReadLocker( _d->consumerslock );
		return _d->consumers.isEmpty();
	}

	StatisticsAndLogging* DataSetProducer::logging() const { return _d->logger; }


	class ChainMemberThread : public QThread
	{
		public:
			ChainMemberThread( DataSetChainMember* p ) : QThread( p ), _queue( 1024 ), _run( true ), _member( p ) {
			}
			~ChainMemberThread() {
				//qDebug() << "ChainMemberThread::~ChainMemberThread()";
				_run = false;
				if ( ! wait( 5000 ) ) {
					qWarning( "ChainMemberThread::~ChainMemberThread() Have to kill this thread the hard way!" );
					terminate();
					wait( 5000 );
				}
			}

			void myexit() {
				_run = false;
				_queue.readyRead.wakeAll();
			}
			void run() {
				//qDebug() << "Will set the thread name to:" << qPrintable( QString( "%1-worker" ).arg( _member->objectName() ) ) << endl;
				prctl(PR_SET_NAME, qPrintable( QString( "%1-w" ).arg( _member->objectName().left(13) ) ) );
				QMutex _readyMutex;
				_run = true;
				//qDebug() << "ChainMemberThread::run() started. &mutex =" << &_readyMutex;
				_readyMutex.lock();
				while ( _run ) {
					if ( _queue.isEmpty() )
						_queue.readyRead.wait( & _readyMutex, 100 );
					while ( _run && ! _queue.isEmpty() ) {
						NSCore::AcqisitionDataSet data = _queue.first();
						if ( _member->dataArrived( data ) ) {
							_member->enqueue( data );
						}
						_queue.removeFirst();
					}
				}
				_readyMutex.unlock();
				//qDebug() << "ChainMemberThread::run() finished!";
			}

			AcqisitionDataSetRingBuffer _queue;
		private:
			volatile bool _run;
			DataSetChainMember* _member;
	};

	DataSetChainMember::DataSetChainMember( QString n, QObject* p )
		: DataSetProducer( n, p )
		, _thread( new ChainMemberThread( this ) )
	{
		//qDebug() << "DataSetChainMember::DataSetChainMember()";
		_thread->start();
	}
	DataSetChainMember::~DataSetChainMember() {
		//qDebug() << "DataSetChainMember::~DataSetChainMember()";
		_thread->myexit();
		emit destroyed( this );
	}

	void DataSetChainMember::addToQueue( NSCore::AcqisitionDataSet data ) {
		//qDebug() << "DataSetChainMember::addToQueue(" << &data << ")";
		//if ( &data == 0 ) return;
		_thread->_queue.enqueue( data );
	}

	void DataSetChainMember::abort() {
		_thread->_queue.dropAll();
	}
	bool DataSetChainMember::isRunning() {
		return !(_thread->_queue.isEmpty());
	}

};

