/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DATASETCHAIN_H
#define DATASETCHAIN_H

#include <list>

#include <datatypes.h>

#include <qtreeitem.h>

namespace NSCore {

	class DataSetProducer;
	class DataSetChainMember;

	class StatisticsAndLogging;

	/**
	 * @brief Baseclass for Producers
	 *
	 * Subclass to produce AcquisitionDataSets and feed them into a chain.
	 */
	class DataSetProducer : public QTreeItem
	{
		Q_OBJECT
		public:
			/**
			 * @brief empty constructor
			 */
			DataSetProducer( QString, QObject* =0 );
			/**
			 * @brief destructor
			 *
			 * Deregisters itself at the consumers.
			 */
			~DataSetProducer();

			/**
			 * @brief Return the number of registered consumers.
			 *
			 * This way a producer can decide not to do expensive conversions
			 * and similar stuff when there is no consumer to actually use the
			 * data.
			 */
			int consumersSize() const;
			/**
			 * @brief Return wether there are consumers or not.
			 *
			 * Even simplier then consumersSize().
			 */
			bool consumersIsEmpty() const;

			enum Dispatching { DispatchToAll, DispatchEqual };
			/**
			 * @todo Document this!
			 */
			void setDispatching( Dispatching );

			/**
			 * @brief The statistics-gathering object
			 */
			StatisticsAndLogging* logging() const;

			/**
			 * @brief Return whether this object is busy.
			 */
			virtual bool isRunning() =0;

		public slots:
			/**
			 * @brief Register a consumer
			 *
			 * This registers a DataSetChainMember as this objects consumer.
			 *
			 * The emitted data is sent to all consumers via the (internal)
			 * DataSetEvent through Qt's event system.
			 */
			void registerConsumer( NSCore::DataSetChainMember* );
			/**
			 * @brief Unregister a consumer
			 */
			void unregisterConsumer( NSCore::DataSetChainMember* );
			/**
			 * @brief Unregister all consumers
			 */
			void unregisterAllConsumers();

			/**
			 * @brief Start the automatic run
			 */
			virtual void start() =0;
			/**
			 * @brief Stop the automatic run
			 */
			virtual void stop() =0;
			/**
			 * @brief Stop the automatic run instantly and drop all remaining activities
			 */
			virtual void abort() = 0;
			/**
			 * @brief Reset all settings to prepare for a new automatic run
			 */
			virtual void reset() {};

		signals:
			/**
			 * @brief Signal the end of the producing
			 *
			 * If the producer was given a finite number of datasets to produce or a defined time
			 * has passed or the producer otherwise decides that it is finished, emit this signal.
			 */
			void finished();

		protected:
			/**
			 * @brief Add data to the queue
			 *
			 * The data is sent to all registered consumers.
			 */
			void enqueue( AcqisitionDataSet );

		private slots:
			/**
			 * @internal
			 * @brief Check the countrate once a second
			 */
			void countrate();

		private:
			/**
			 * Private class for the internals
			 * @internal
			 */
			class DataSetProducerPrivate;
			/**
			 * Internal pointer
			 * @internal
			 */
			DataSetProducerPrivate* _d;

			Q_DISABLE_COPY( DataSetProducer );
	};

	class ChainMemberThread;

	/**
	 * @brief Consume datasets, do something with them and give them to the next consumer
	 *
	 * Formerly there was producer and consumer and the chainmember to combine
	 * both. But consumers alone are stupid, even if things like the osci don't
	 * modify the data, they aren't always the last in the chain.
	 *
	 * So the DataSetChainMember is the base class for elements consuming
	 * datasets, doing something with them and reaching them further down the
	 * chain.
	 */
	class DataSetChainMember : public DataSetProducer
	{
		Q_OBJECT
		friend class ChainMemberThread;
		public:
			/// Defaultconstructor
			DataSetChainMember( QString, QObject* =0 );
			/// Deconstructor
			~DataSetChainMember();

			/**
			 * @brief Add new data to the incoming queue
			 */
			void addToQueue( NSCore::AcqisitionDataSet );

			void start() {}
			void stop() {}

			void abort();

			bool isRunning();

		signals:
			/**
			 * @brief Specialized signal to unregister at producers.
			 */
			void destroyed( NSCore::DataSetChainMember* );

		protected:
			/**
			 * @brief Data arrived!
			 *
			 * Subclass this function to handle the data. Return true when the
			 * data should be sent to the next in the chain. Otherwise it gets
			 * deleted.
			 */
			virtual bool dataArrived( NSCore::AcqisitionDataSet ) =0;

		private:
			ChainMemberThread* _thread;

			Q_DISABLE_COPY( DataSetChainMember );
	};

};

#endif // DATASETCHAIN_H

