/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "datatypes.h"

#include <cmath>
#include <vector>
#include <QtCore/QMutex>
#include <QtCore/QMutexLocker>
#include <QtCore/QSemaphore>
#include <QtCore/QReadWriteLock>
#include <QtCore/QWaitCondition>
#include <QtCore/QStringList>
#include <QtCore/QFile>
#include <QtCore/QDebug>

#include "qsemaphore64.h"

namespace NSCore {

	class MemoryPool
	{
		private:
			MemoryPool() {
				activeDataSets.release( int( pow( 2,14 ) ) );
			}
			MemoryPool( MemoryPool& ) {}
			MemoryPool( MemoryPool* ) {}
			~MemoryPool() {
				/*
				 * Will this ever get executed???
				 */
				while ( ! chunkPool.empty() ) {
					int size = chunkPool.begin()->first;
					qDebug() << " Deleting all elements of size" << size;
					std::multimap<int,double*>::const_iterator begin = chunkPool.lower_bound( size );
					std::multimap<int,double*>::const_iterator end = chunkPool.upper_bound( size );
					while ( begin != end ) {
						delete[] begin->second;
						begin++;
					}
					chunkPool.erase( size );
				}
			}

			QMutex lock;
			typedef std::multimap<int, double*> doublemap;
			typedef std::pair<int, double*> doublepair;
			doublemap chunkPool;
			doublemap minmaxPool;

			QSemaphore64 _poolSemaphore;

		public:
			static MemoryPool* the() {
				//static QMutex* memutex = new QMutex();
				//QMutexLocker locker( memutex );
				static MemoryPool* me = 0;
				if ( ! me ) {
					me = new MemoryPool();
					QFile* f = new QFile( "/proc/meminfo" );
					if ( f->open( QIODevice::ReadOnly ) ) {
						long long int globmem;
						//qDebug() << "Successfully opened /proc/meminfo";
						QString tmp = f->readAll();
						globmem = tmp.section( "\r", 0 ).section( " ", 1, 1, QString::SectionSkipEmpty ).toLongLong() * 1024 / 4;
						//qDebug() << " Global memory is" << globmem << "bytes, will reserve about a quarter of it...";
						f->close();
						me->_poolSemaphore.release( globmem );
						qDebug( "MemoryPool::the() Reserved %lli bytes for the buffer. That is %lli kB, %lli MB", globmem, globmem/1024, globmem/1024/1024 );
					} else {
						me->_poolSemaphore.release( int(pow(2, 16)) );
					}
					delete f;
				}
				return me;
			}

			QSemaphore64 activeDataSets;

			double* getMem( int size ) {
				//qDebug() << "MemoryPool::getMem(" << size << ")";
				_poolSemaphore.acquire( size * sizeof(double) );
				QMutexLocker l( &lock );
				double* tmp = 0;
				if ( chunkPool.count( size ) > 0 ) {
					doublemap::iterator it = chunkPool.find( size );
					tmp = it->second;
					chunkPool.erase( it );
					//qDebug() << "MemoryPool::getMem() getting a chunk out of the pool";
				} else {
					//qDebug() << "MemoryPool::getMem(" << size << ") have to allocate new space\n  chunkPool.size() is" << chunkPool.size() << "\n  activeChunks.available() is" << activeChunks.available();
					tmp = new double[ size ];
				}
				return tmp;
				//return (double*) malloc( size * sizeof(double) );
			}
			void freeMem( double* mem, int size ) {
				_poolSemaphore.release( size * sizeof(double) );
				QMutexLocker l( &lock );
				//qDebug() << "MemoryPool::freeMem(" << mem << ")";
				if ( chunkPool.size() > 10000 )
					delete[] mem;
				else
					chunkPool.insert( doublepair( size, mem ) );
				//free( mem );
			}
			double* getMinMax( int size ) {
				/*QMutexLocker l( &lock );
				double* tmp = 0;
				if (minmaxPool.count( size ) > 0 ) {
					doublemap::iterator it = minmaxPool.find( size );
					tmp = it->second;
					minmaxPool.erase( it );
				} else {
					tmp = new double[ size ];
				}
				return tmp;*/
				return new double[size];
			}
			void freeMinMax( double* mem, int size ) {
				/*QMutexLocker l( &lock );
				if ( minmaxPool.size() > 10000 )
					delete[] mem;
				else
					minmaxPool.insert( doublepair( size, mem ) );*/
				delete[] mem;
			}
	};



	class AcqisitionDataSet_Private : public QSharedData
	{
	private:
		bool empty;
	public:

		//mutable QReadWriteLock lock;
		mutable QMutex mutex;

		QHash<QString,double> results;

		void addresult( QString name, double value ) {
			name = cache.currentPrefix + '/' + name;
			while ( name.startsWith( '/' ) )
				name = name.remove( 0, 1 );
			while ( name.contains( "//" ) )
				name.replace( "//" /*QRegExp( "/{2,}" )*/, "/" );
			results[ name ] = value;
		}

		struct RealDataSet {
			// The data itself
			int channels;
			int samples;
			double* data;

			// caches
			double* min;
			double* max;

			// generations
			int generationNumber;
			QString generationPrefix;

			/// constructor
			RealDataSet( const int& _channels, const int& _samples, const int& _generationNumber, const QString& _generationPrefix )
				: channels( _channels )
				, samples( _samples )
				, generationNumber( _generationNumber )
				, generationPrefix( _generationPrefix )
			{
				data = MemoryPool::the()->getMem( channels*samples );
				min = new double[ channels ];
				max = new double[ channels ];
				for ( int i=0; i<channels; ++i ) {
					min[i] = INFINITY;
					max[i] = -INFINITY;
				}
			}
			~RealDataSet() {
				MemoryPool::the()->freeMem( data, channels * samples );
				delete[] min;
				delete[] max;
			}
			private:
				Q_DISABLE_COPY( RealDataSet );
		};
		std::vector<RealDataSet*> datasets;

		struct cache_t {
			int currentGeneration;
			QString currentPrefix;
		} cache;

		AcqisitionDataSet_Private( bool e=false ) : empty( e )// : mutex( QMutex::Recursive ) //lock( QReadWriteLock::Recursive )
		{
			cache.currentGeneration = -1;
			cache.currentPrefix = "";
			if ( ! empty )
				MemoryPool::the()->activeDataSets.acquire();
		}
		~AcqisitionDataSet_Private() {
			while ( datasets.size() > 0 ) {
				delete datasets.back();
				datasets.pop_back();
			}
			if ( ! empty )
				MemoryPool::the()->activeDataSets.release();
		}
		private:
			Q_DISABLE_COPY( AcqisitionDataSet_Private )
	};

	AcqisitionDataSet::AcqisitionDataSet()
		: d( new AcqisitionDataSet_Private( true ) )
	{
	}
	AcqisitionDataSet::AcqisitionDataSet( int channels, int samples )
		: d( new AcqisitionDataSet_Private )
	{
		newGeneration( "", channels, samples );
	}
	AcqisitionDataSet::AcqisitionDataSet( const AcqisitionDataSet& c )
	{
		//QWriteLocker lock( & c.d->lock );
		//QMutexLocker lock(  & c.d->mutex );
		d = c.d;
		//qDebug() << "AcqisitionDataSet::AcqisitionDataSet(" << &c << ") copy-constructor";
	}
	AcqisitionDataSet& AcqisitionDataSet::operator=( const AcqisitionDataSet& c ) {
		//QWriteLocker lock( & c.d->lock );
		//QMutexLocker lock(  & c.d->mutex );
		d = c.d;
		return *this;
	}
	AcqisitionDataSet::~AcqisitionDataSet() {
		//QWriteLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		//qDebug() << "AcqisitionDataSet::~AcqisitionDataSet()";
	}

	int AcqisitionDataSet::currentGenerationNumber() const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		return d->cache.currentGeneration;
	}
	int AcqisitionDataSet::parentGeneration( int gen ) const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		//qDebug() << "AcqisitionDataSet::parentGeneration(" << gen << ")";
		QString prefix = d->cache.currentPrefix;
		if ( gen != -1 ) {
			for ( uint i=0; i<d->datasets.size(); ++i ) {
				if ( d->datasets[i]->generationNumber == gen ) {
					prefix = d->datasets[i]->generationPrefix;
				}
			}
		}
		//qDebug() << " prefix of generation is" << prefix;
		if ( prefix.endsWith( '/' ) ) prefix.chop( 1 );
		prefix = prefix.section( '/', 0, -2 ) + '/';
		//qDebug() << " prefix of parent is" << prefix;
		for ( uint i=0; i<d->datasets.size(); ++i ) {
			if ( d->datasets[i]->generationPrefix == prefix ) {
				return d->datasets[i]->generationNumber;
			}
		}
		// return the root-generation by default. Correct would be to return -1 but returning 0
		// probably gives less errors.
		return 0;
	}
	void AcqisitionDataSet::debugPrint() const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		//qWarning( "debugPrint is not available anymore!" );
		foreach ( QString key, d->results.keys() ) {
			qDebug() << qPrintable( key ) << " : " << d->results[ key ];
		}
	}

	void AcqisitionDataSet::newGeneration( QString prefix, int channels, int samples ) {
		//QWriteLocker lock( d->lock );
		//qDebug() << "AcqisitionDataSet::newGeneration(" << prefix << "," << channels << "," << samples << ")";
		pushPrefix( prefix );
		QMutexLocker lock(  & d->mutex );
		//QWriteLocker lock( & d->lock );
		AcqisitionDataSet_Private::RealDataSet* tmp = new AcqisitionDataSet_Private::RealDataSet( channels, samples, d->datasets.size(), d->cache.currentPrefix );
		d->cache.currentGeneration = d->datasets.size();
		d->datasets.push_back( tmp );
	}

	void AcqisitionDataSet::pushPrefix( QString n ) {
		//QWriteLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		//qDebug() << "AcqisitionDataSet::pushPrefix(" << n << ")";
		d->cache.currentPrefix = QString( d->cache.currentPrefix + '/' + n + "/" ).replace( "//", "/" );
		//qDebug() << " new prefix is" << d->cache.currentPrefix;
	}
	void AcqisitionDataSet::popPrefix( QString n ) {
		//QWriteLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		//qDebug() << "AcqisitionDataSet::popPrefix(" << n << ")";
		if ( d->cache.currentPrefix.endsWith( QString( n + '/' ).replace( "//", "/" ) ) ) {
			d->cache.currentPrefix.chop( n.size() + 1 );
			int generation = -1;
			QString prefix = d->cache.currentPrefix;
			while ( generation == -1 && prefix.size() > 0 ) {
				foreach( AcqisitionDataSet_Private::RealDataSet* tmp, d->datasets ) {
					if ( tmp->generationPrefix == prefix )
						generation = tmp->generationNumber;
				}
				prefix = prefix.section( '/', 0, -2 );
			}
			d->cache.currentGeneration = generation;
		}
	}

	QString AcqisitionDataSet::getPrefix( int generation ) const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		if ( generation == -1 || generation == d->cache.currentGeneration )
			return d->cache.currentPrefix;

		QString ret;
		for ( uint i=0; ret=="" && i<d->datasets.size(); ++i ) {
			if ( d->datasets.at( i )->generationNumber == generation )
				ret = d->datasets.at( i )->generationPrefix;
		}
		return ret;
	}

	double AcqisitionDataSet::sample( int channel, int sample, int generation ) const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		if ( generation == -1 ) generation = d->cache.currentGeneration;
		if ( generation == -1 ) return 0.0;
		AcqisitionDataSet_Private::RealDataSet* tmp =  d->datasets[ generation ];
		double ret = 0;
		if ( sample >=0 && sample < tmp->samples )
			ret = tmp->data[ channel*tmp->samples + sample ];
		//qDebug() << "AcqisitionDataSet::sample(" << channel << "," << sample << "," << generation << ") returning" << ret;
		return ret;
	}
	void AcqisitionDataSet::setSample( int channel, int sample, double value, int generation ) {
		//QWriteLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		//qDebug() << "AcqisitionDataSet::setSample(" << channel << "," << sample << "," << value << "," << generation << ")";
		if ( generation == -1 ) generation = d->cache.currentGeneration;
		if ( generation == -1 ) return;
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		tmp->data[ channel*tmp->samples + sample ] = value;
		if ( value > tmp->max[ channel ] ) {
			tmp->max[ channel ] = value;
			addResult( QString( "channel %1/max" ).arg( channel ), value );
			addResult( QString( "channel %1/max at" ).arg( channel ), sample );
		}
		if ( value < tmp->min[ channel ] ) {
			tmp->min[ channel ] = value;
			addResult( QString( "channel %1/min" ).arg( channel ), value );
			addResult( QString( "channel %1/min at" ).arg( channel ), sample );
		}
	}
	int AcqisitionDataSet::samples( int generation ) const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		//if ( d->datasets.size() == 0 ) return 0;
		Q_ASSERT( d->datasets.size() > 0 );
		if ( generation == -1 ) generation = d->cache.currentGeneration;
		if ( generation == -1 ) return 0;
		//Q_ASSERT( d->datasets.size() > ( unsigned )generation );
		return d->datasets[ generation ]->samples;
	}
	int AcqisitionDataSet::channels( int generation ) const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		//if ( d->datasets.size() == 0 ) return 0;
		Q_ASSERT( d->datasets.size() > 0 );
		if ( generation == -1 ) generation = d->cache.currentGeneration;
		if ( generation == -1 ) return 0;
		//if ( ! ((signed)d->datasets.size() > generation) ) {
		//	qDebug() << "!" << d->datasets.size() << ">" << generation;
		//}
		Q_ASSERT( (signed)d->datasets.size() > generation );
		Q_ASSERT( generation > -1 );
		return d->datasets[ generation ]->channels;
	}

	double* AcqisitionDataSet::getRawChannel( int channel, int generation ) {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		if ( generation == -1 ) generation = d->cache.currentGeneration;
		if ( generation == -1 ) return NULL;
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		return &( tmp->data[ channel*tmp->samples ] );
	}
	double* AcqisitionDataSet::getRawData( int generation ) {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		if ( generation == -1 ) generation = d->cache.currentGeneration;
		if ( generation == -1 ) return NULL;
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		return tmp->data;
	}
	void AcqisitionDataSet::updateMinMax( int generation ) {
		int c = channels( generation );
		int s = samples( generation );
		QMutexLocker lock(  & d->mutex );
		if ( generation == -1 ) generation = d->cache.currentGeneration;
		if ( generation == -1 ) return;
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		double *ch = d->datasets[ generation ]->data;
		QString tmp2;
		//QWriteLocker lock( & d->lock );
		for ( int i=0; i<c; ++i ) {
			tmp->max[ i ] = -INFINITY;
			tmp->min[ i ] = INFINITY;
			int max_at =0, min_at =0;
			double value;
			for ( int j=0; j<s; ++j ) {
				value = ch[ s*i + j ];
				if ( value > tmp->max[ i ] ) {
					tmp->max[ i ] = value;
					max_at = j;
				}
				if ( value < tmp->min[ i ] ) {
					tmp->min[ i ] = value;
					min_at = j;
				}
			}
			tmp2 = QString( "channel %1/%2" ).arg( i );
			d->addresult( tmp2.arg( "max" ), tmp->max[ i ] );
			d->addresult( tmp2.arg( "max at" ), max_at );
			d->addresult( tmp2.arg( "min" ), tmp->min[ i ] );
			d->addresult( tmp2.arg( "min at" ), min_at );
		}
	}
	void AcqisitionDataSet::resetMinMaxCache( int generation ) {
		//QWriteLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		if ( generation == -1 ) generation = d->cache.currentGeneration;
		if ( generation == -1 ) return;
		lock.unlock();
		int c = channels( generation );
		lock.relock();
		AcqisitionDataSet_Private::RealDataSet* tmp = d->datasets[ generation ];
		for ( int i=0; i<c; ++i ) {
			tmp->max[ i ] = -INFINITY;
			tmp->min[ i ] = INFINITY;
		}
	}

	void AcqisitionDataSet::addResult( QString name, double value ) {
		//QWriteLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		d->addresult( name, value );
	}
	double AcqisitionDataSet::getResult( QString name ) const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		QString prefix = d->cache.currentPrefix;
		//qDebug() << "AcqisitionDataSet::getResult(" << name << ") prefix=" << prefix;
		while ( name.startsWith( "../" ) ) {
			name.remove( 0, 3 );
			prefix = prefix.section( '/', 0, -3 );
			//qDebug() << " new name" << name << " new prefix" << prefix;
		}
		//qDebug() << " will build the new name from" << prefix << "+ / +" << name;
		name = prefix + '/' + name;
		while ( name.contains( "//" ) )
			name.replace( "//" /*QRegExp( "/{2,}" )*/, "/" );
		while ( name.startsWith( '/' ) )
			name = name.remove( 0, 1 );
		if ( d->results.contains( name ) ) {
			return d->results[ name ];
		}
		//qDebug() << "AcqisitionDataSet::result(" << name << ")";
		//qWarning( " Will return the invalid result!\n Valid results are:" );
		//debugPrint();
		return NAN;
	}
	QStringList AcqisitionDataSet::getResultNames() const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		return d->results.keys();
	}
	QHash<QString,double> AcqisitionDataSet::getResultHash() const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		return d->results;
	}
	QMap<QString,double> AcqisitionDataSet::getResultMap() const {
		//QReadLocker lock( & d->lock );
		QMutexLocker lock(  & d->mutex );
		QMap<QString,double> ret;
		foreach ( QString key, d->results.keys() ) {
			ret[ key ] = d->results[ key ];
		}
		return ret;
	}


	struct AcqisitionDataSetRingBufferPrivate
	{
		int bufferSize;
		QVector<AcqisitionDataSet> buffer;
		int reader;
		int writer;
		QMutex* mutex;
		QSemaphore sem;
	};
	AcqisitionDataSetRingBuffer::AcqisitionDataSetRingBuffer( int buffersize )
		: _p( new AcqisitionDataSetRingBufferPrivate() )
	{
		_p->bufferSize = buffersize;
		_p->buffer.resize( buffersize );
		_p->reader = _p->writer = 0;
		_p->sem.release( buffersize-1 );
		_p->mutex = new QMutex();
	}

	AcqisitionDataSetRingBuffer::~AcqisitionDataSetRingBuffer() {
		delete _p->mutex;
		delete _p;
	}

	bool AcqisitionDataSetRingBuffer::isEmpty() const { return ( _p->reader == _p->writer ); }

	void AcqisitionDataSetRingBuffer::enqueue( AcqisitionDataSet& data ) {
		QMutexLocker lock( _p->mutex );
		_p->sem.acquire( 1 );
		_p->buffer[ _p->writer ] = data;
		Q_ASSERT( _p->writer + 1 != _p->reader );
		_p->writer = ( _p->writer + 1 ) % _p->bufferSize;
		readyRead.wakeAll();
	}

	AcqisitionDataSet& AcqisitionDataSetRingBuffer::first() {
		return _p->buffer[ _p->reader ];
	}

	void AcqisitionDataSetRingBuffer::removeFirst() {
		//QMutexLocker lock( &mutex );
		_p->buffer[ _p->reader ] = NSCore::AcqisitionDataSet();
		_p->reader = ( _p->reader + 1 ) % _p->bufferSize;
		_p->sem.release( 1 );
	}

	void AcqisitionDataSetRingBuffer::dropAll() {
		QMutexLocker lock( _p->mutex );
		while ( ! isEmpty() )
			removeFirst();
	}
};

