/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef CORE_INTTOFLOAT_H
#define CORE_INTTOFLOAT_H

#include <QtCore/QObject>
#include <math.h>

namespace NSCore {

	class TiParameter;

	/// Converts int-signals into float-signals and backwards
	/**
	 * IntToFloat converts int-signals to float-signals with a given precision.
	 *
	 * The precision means that that the int is multiplied with precision to
	 * form the float value. So an int of 8 times a precision of 1E-2 gives
	 * 0.08 for the result. The conversion backwards is also done by this class.
	 */
	class IntToFloat : public QObject {
		Q_OBJECT
		public:
			/**
			 * @brief Creates a new Integer-to-Double-converter with the given precision.
			 *
			 * @param precision should be given in the style 1E-3
			 * @param p the parent QObject
			 */
			IntToFloat( double precision, QObject* p )
				: QObject( p )
				, _precision( precision )
				, _parameter( 0 )
			{ }
			/**
			 * @brief Creates a new Integer-to-Double-converter with the given
			 * precision and connects to the given
			 * NSCore::TiParameter.
			 *
			 * @param precision should be given in the style 1E-3
			 * @param item the TiParameter to connect to
			 * @param p the parent QObject
			 */
			IntToFloat( double precision, NSCore::TiParameter* item, QObject* p );
		public slots:
			/**
			 * @brief convert int to double
			 *
			 * This function converts the given integer to a float by
			 * multiplying with the precision given at creation-time.
			 */
			void value( int n ) {
				emit valueChanged( double( n * _precision ) );
			}
			/**
			 * @brief convert double to int
			 *
			 * This function converts the given double to an int by dividing
			 * with the precision and rounding to int.
			 */
			void value( double n ) {
				emit valueChanged( int( round( n / _precision ) ) );
			}
		signals:
			/// Announce the new double value.
			void valueChanged( double );
			/// Announce the new int value.
			void valueChanged( int );
		private:
			double _precision;
			NSCore::TiParameter* _parameter;
	};

};

#endif

