/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef MATHEXPRESSION_P_H
#define MATHEXPRESSION_P_H

#include <tiparameter.h>
#include <datatypes.h>

namespace NSCore {

/**
 * @brief Simple result
 *
 * This returns a simple result.
 * The parameter is the name of the result to fetch.
 */
class MathExpression_Result : public MathExpression
{
	public:
		MathExpression_Result( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return QString( "result: %1" ).arg( _name ); }
		QDomNode saveToXML( QDomDocument doc ) {
			QDomElement node = doc.createElement( "result" );
			node.setAttribute( "name", _name );
			return node;
		}
		bool restoreFromXML( QDomNode node ) {
			if ( node.nodeName() == "result" && node.isElement() )
				_name = node.toElement().attribute( "name" );
			return false;
		}

		QVariant guiValue( int ) const { return _name; }
		void guiSetValue( QVariant n ) { _name = n.toString(); }
		Qt::ItemFlags guiFlags() const { return MathExpression::guiFlags() | Qt::ItemIsEditable; }

		QString _name;
};

/**
 * @brief Simply a constant
 */
class MathExpression_Constant : public MathExpression
{
	public:
		MathExpression_Constant( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

	private:
		QString debug() const { return QString( "constant: %1" ).arg( _value ); }
		QDomNode saveToXML( QDomDocument doc ) {
			QDomElement node = doc.createElement( "constant" );
			node.setAttribute( "value", _value );
			return node;
		}
		bool restoreFromXML( QDomNode node ) {
			if ( node.nodeName() == "constant" && node.isElement() )
				_value = node.toElement().attribute( "value" ).toDouble();
			return false;
		}

		QVariant guiValue( int ) const { return QString::number( _value ); }
		void guiSetValue( QVariant n ) { _value = n.toDouble(); }
		Qt::ItemFlags guiFlags() const { return MathExpression::guiFlags() | Qt::ItemIsEditable; }

		double _value;
};

/**
 * @brief Addition of results
 *
 * This adds the child results together. There are no parameters for this.
 */
class MathExpression_Add : public MathExpression
{
	public:
		MathExpression_Add( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "add"; }

		QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "add" ); }
		bool restoreFromXML( QDomNode node ) { return node.nodeName() == "add"; }
};

/**
 * @brief Multiplication of results
 *
 * This multiplies the child results together. There are no parameters for this.
 */
class MathExpression_Mul : public MathExpression
{
	public:
		MathExpression_Mul( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "mul"; }

		QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "mul" ); }
		bool restoreFromXML( QDomNode node ) { return node.nodeName() == "mul"; }
};

/**
 * @brief The minimum of a number of values
 */
class MathExpression_Min : public MathExpression
{
	public:
		MathExpression_Min( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "min"; }

		QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "min" ); }
		bool restoreFromXML( QDomNode node ) { return node.nodeName() == "min"; }
};
/**
 * @brief The maximum of a number of values
 */
class MathExpression_Max : public MathExpression
{
	public:
		MathExpression_Max( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "max"; }

		QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "max" ); }
		bool restoreFromXML( QDomNode node ) { return node.nodeName() == "max"; }
};


/**
 * @brief Negate the result
 *
 * This negates the childs result. If there are multiple childs, only the last
 * result will be used.
 */
class MathExpression_Neg : public MathExpression
{
	public:
		MathExpression_Neg( NSCore::QTreeItem* parent );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "neg"; }

		QString guiString() const { return "Negate ( * -1 )"; }

		QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "neg" ); }
		bool restoreFromXML( QDomNode node ) { return node.nodeName() == "neg"; }
};

/**
 * @brief Reduce the precision of a result
 *
 * This reduces the precision of the result so far. The given parameter is the precision which can be any number to which the result is trimmed to. For example 5E-2 will give you results with a precision of 0.05.
 *
 * If there are multiple childs, only the last result will be used.
 */
class MathExpression_Precision : public MathExpression
{
	public:
		MathExpression_Precision( NSCore::QTreeItem* p );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return QString( "precision: %1" ).arg( _precision ); }

		QDomNode saveToXML( QDomDocument doc ) {
			QDomElement node = doc.createElement( "precision" );
			node.setAttribute( "precision", _precision );
			return node;
		}
		bool restoreFromXML( QDomNode node ) {
			if ( node.nodeName() == "precision" && node.isElement() )
				_precision = node.toElement().attribute( "precision" ).toDouble();
			return false;
		}

		QVariant guiValue( int ) const { return QString::number( _precision ); }
		void guiSetValue( QVariant n ) { _precision = n.toDouble(); }
		Qt::ItemFlags guiFlags() const { return MathExpression::guiFlags() | Qt::ItemIsEditable; }

	private:
		double _precision;
};

/**
 * @brief Rather complicated SmallerMinusBigger
 *
 * This one compares a result on the first two channels to select the bigger and
 * the smaller one and then substractes a second kind of result of both
 * channels. Which channels result gets subtracted from the other is selected by
 * the first decision.
 */
class MathExpression_SmB : public MathExpression
{
	public:
		MathExpression_SmB( NSCore::QTreeItem* p );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return QString( "Smaller-Minus-Bigger: selection='%1' calculus='%2'" ).arg( _selection->toString() ).arg( _calculus->toString() ); }

		QString guiString() const { return "Smaller-Minus-Bigger"; }

		QString selection() const { return _selection->toString(); }
		QString calculus() const { return _calculus->toString(); }

		QDomNode saveToXML( QDomDocument doc ) {
			QDomElement node = doc.createElement( "smb" );
			node.setAttribute( "selection", _selection->toString() );
			node.setAttribute( "calculus", _calculus->toString() );
			return node;
		}
		bool restoreFromXML( QDomNode node ) {
			if ( node.nodeName() == "smb" && node.isElement() ) {
				QDomElement elem = node.toElement();
				_selection->setValue( elem.attribute( "selection" ) );
				_calculus->setValue( elem.attribute( "calculus" ) );
			}
			return false;
		}
	private:
		NSCore::TiParameter *_selection, *_calculus;
};

/**
 * @brief Greater-than-Zero
 */
class MathExpression_GtZ : public MathExpression
{
	public:
		MathExpression_GtZ( NSCore::QTreeItem* );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "Greater-than-Zero"; }
		QString guiString() const { return "Greater-than-Zero"; }

		QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "gtz" ); }
		bool restoreFromXML( QDomNode node ) { return node.nodeName() == "gtz"; }
};
/**
 * @brief Less-than-Zero
 */
class MathExpression_LtZ : public MathExpression
{
	public:
		MathExpression_LtZ( NSCore::QTreeItem* );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "Less-than-Zero"; }
		QString guiString() const { return "Less-than-Zero"; }

		QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "ltz" ); }
		bool restoreFromXML( QDomNode node ) { return node.nodeName() == "ltz"; }
};
/**
 * @brief Pass the value if its inside the window
 */
class MathExpression_Window : public MathExpression
{
	public:
		MathExpression_Window( NSCore::QTreeItem* );

		double value( NSCore::AcqisitionDataSet& data );

		QString debug() const { return "Window"; }
		QString guiString() const { return "Window"; }

		QDomNode saveToXML( QDomDocument doc ) {
			return doc.createElement( "window" );
		}
		bool restoreFromXML( QDomNode node ) {
			if ( node.nodeName() == "window" && node.isElement() ) {
				_min->setValue( node.toElement().attribute( "min" ) );
				_max->setValue( node.toElement().attribute( "max" ) );
			}
			return true;
		}
	private:
		TiParameter *_min, *_max;
};

/**
 */
class MathExpression_FuzzyWindow : public MathExpression
{
	public:
		MathExpression_FuzzyWindow( NSCore::QTreeItem* p )
			: MathExpression( "FuzzyWindow", true, p )
			, _center( new TiParameter( "center", 0.0, this ) )
			, _sigma( new TiParameter( "sigma", 0.5, this ) )
		{
		}

		double value( NSCore::AcqisitionDataSet& data ) {
			double center = _center->toDouble();
			double sigma = _sigma->toDouble();
			QList<MathExpression*> childs = childobjects<MathExpression>();
			if ( childs.size() > 0 ) {
				double tmp = childs[ 0 ]->value( data );
				// check whether the first childs value is good to...
				if ( rand()*1.0/RAND_MAX < exp( - pow( tmp-center,2 ) / ( 2* pow( sigma,2 ) ) ) ) {
					if ( childs.size() > 1 ) {
						// ...return the second childs value if there is a second child...
						return childs[ 1 ]->value( data );
					} else
						// ...otherwise return the first childs value.
						return tmp;
				} else
					// If the first childs value is bad, the third childs value will be returned if it exists...
					if ( childs.size() > 2 )
						return childs[ 2 ]->value( data );
			}
			/// ...otherwise we just return NAN
			return NAN;
		}

		QString debug() const { return "FuzzyWindow"; }
		QString guiString() const { return "FuzzyWindow"; }

		QDomNode saveToXML( QDomDocument doc ) {
			return doc.createElement( "fuzzywindow" );
		}
		bool restoreFromXML( QDomNode node ) {
			return ( node.nodeName() == "fuzzywindow" );
		}

	private:
		TiParameter *_center, *_sigma;
};

}; // NSCore

#endif // MATHEXPRESSION_P_H

