/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef QTREEITEM_H
#define QTREEITEM_H

#include <QtCore/QObject>
#include <QtCore/QStringList>
#include <QtCore/QDebug>
#include <QtCore/QVariant>

#include <QtXml/QtXml>

/**
 * @brief Essential things are here.
 *
 * Stuff in this namespace (and therefor in the libcore directory) needs only QtCore and QtXml as
 * base.
 */
namespace NSCore {

	class QTreeItemPointer;

	/**
	 * @brief Abstract class to arrange items in a tree.
	 *
	 * This class takes care of accessing the items and numbering them.
	 *
	 * <div style="text-decoration:line-through;">An QTreeItem can have a
	 * maximum number of childs set, so adding one to many might fail.</div>
	 * This is currently not implemented. If it is really needed it can come
	 * back but I don't see why.
	 *
	 * Managing childs is mainly done by QObject. But as we have the opportunity
	 * to add Childs which don't belong to this QObject, please add the foreign
	 * childs with addChild(...) and use childobjects() to retrieve the
	 * childrens instead of QObject::children().
	 *
	 * @todo Document the xml-style of actions!
	 */
	class QTreeItem : public QObject {
		Q_OBJECT

		/**
		 * @brief Tooltip to display in the gui ( if appropriate )
		 */
		Q_PROPERTY( QString toolTip READ toolTip WRITE setToolTip )

		/**
		 * @brief The value to be shown in guis
		 */
		Q_PROPERTY( QVariant guiValue READ guiValue WRITE guiSetValue )

		/**
		 * @brief Basicly the name to display in guis
		 *
		 * This actually uses guiIdentifier( int ) with a default of 0
		 * (=Qt::DisplayRole).
		 *
		 * If you want to access the icon and tooltip, use guiIdentifier()
		 * directly.
		 */
		Q_PROPERTY( QVariant guiIdentifier READ guiIdentifier )

		public:
			/**
			 * @brief Create a new QTreeItem
			 *
			 * Be aware that if the parent isn't 0 and is a QTreeItem it adds
			 * itself to the parents childs. So no further action has to be done.
			 *
			 * @param p The parent QTreeItem or QObject this one got inserted to.
			 */
			QTreeItem( QObject* p=0 ) __attribute__((deprecated))
				: QObject( p )
				, _deferredDelete( false )
				, _treeparent( 0 )
			{
				//qDebug() << "QTreeItem::QTreeItem("<< p <<")";
				if ( p!=0 ) {
					QTreeItem* parent = dynamic_cast<QTreeItem*>( p );
					if ( parent )
						parent->addChild( this );
				}
			}
			/**
			 * @brief Create a new named QTreeItem
			 *
			 * Be aware that if the parent isn't 0 and is a QTreeItem it adds
			 * itself to the parents childs. So no further action has to be done.
			 *
			 * @param n The name of this QTreeItem. Allows accessing this item 
			 * and childs by name / path.
			 * @param p The parent QTreeItem or QObject this one got inserted to.
			 */
			QTreeItem( QString n, QObject* p=0 )
				: QObject( p )
				, _deferredDelete( false )
				, _treeparent( 0 )
			{
				setObjectName( n );
				//qDebug() << "QTreeItem::QTreeItem("<< p <<")";
				if ( p!=0 ) {
					QTreeItem* parent = dynamic_cast<QTreeItem*>( p );
					if ( parent )
						parent->addChild( this );
				}
			}
			/**
			 * @brief Destructor
			 */
			virtual ~QTreeItem() {
				//qDebug() << "QTreeItem::~QTreeItem() [" << this << "]";
				emit aboutToDeleteMe( this );
			};

			/**
			 * @brief Set the objectname and check for duplicates in parents childs
			 *
			 * @note Please use this one instead of QObject::setObjectName()
			 * @note Haha, of course you will use this one as its the same name as
			 * QObject::setObjectName() :-)
			 */
			void setObjectName( const QString& );

			/// Return the number of childs.
			int childs() const { return childobjects<NSCore::QTreeItem>().size(); }

			/**
			 * @brief Return pointers to all the child-treeitems.
			 *
			 * This extends QObject::children() in that it also returns childs
			 * that are not children in the QObject-way.
			 *
			 * @note For usage on QTreeItem's tree-structure this has to be
			 * used instead of the QObject::children()!
			 * @todo Maybe just overwrite QObject:children? Like with setObjectName()?
			 */
			const QObjectList childobjects() const {
				QObjectList list = children();
				foreach( QObject* obj, _morechilds )
					list.append( obj );

				foreach( QObject* obj, list ) {
					QTreeItem* item = dynamic_cast<QTreeItem*>( obj );
					//if ( !item )
					//	qWarning() << "QTreeItem::childobjects() have a child here that is not a QTreeItem*!";
					if ( item && item->_deferredDelete ) {
						//qDebug() << "QTreeItem::childobjects() removing one item that is in deferred delete state!";
						list.removeAll( obj );
					}
				}

				return list;
			}

			/**
			 * @brief Return pointers to all child-treeitems of a certain type
			 *
			 * This extends childobjects() in that it only returns the addresses
			 * of the direct children that are of the given type.
			 */
			template<class T>
			QList<T*> childobjects() const {
				QList<T*> ret;
				foreach( QObject* obj, childobjects() ) {
					T* tmp = dynamic_cast<T*>( obj );
					if ( tmp )
						ret.append( tmp );
				}
				return ret;
			}

			/**
			 * @brief The names of the direct childs.
			 *
			 * This returns a QStringList with the names of the direct children.
			 */
			const QStringList childlist() const {
				QStringList res;
				foreach( NSCore::QTreeItem* child, childobjects<NSCore::QTreeItem>() ) {
					res << child->objectName();
				}
				return res;
			}
			template<class T>
			const QStringList childlist() const {
				QStringList res;
				foreach( T* child, childobjects<T>() ) {
					res << child->objectName();
				}
				return res;
			}

			/**
			 * @brief Gives access to the childs by their number.
			 *
			 * @param n the child you want to access
			 */
			QTreeItem* child( int n ) const {
				QList<QObject*> list = childobjects();
				if ( n<list.size() && dynamic_cast<QTreeItem*>( list[ n ] ) )
					return dynamic_cast<QTreeItem*>( list[ n ] );
				else
					return 0;
			}

			/**
			 * @brief Get the named direct child
			 *
			 * This function is non-recursive and only searches the direct
			 * childs for a complete hit of the name.
			 *
			 * @param n name of the child to find.
			 */
			QTreeItem* child( QString n ) const {
				foreach( QObject* child, childobjects() ) {
					QTreeItem* item = dynamic_cast<QTreeItem*>( child );
					if ( item != 0 && n == item->objectName() )
						return item;
				}
				return 0;
			}
			/**
			 * @brief Get the named direct child of the given type
			 *
			 * This functions extends child() and returns a casted pointer. If a
			 * child with the given name and type exists. Otherwise it returns 0.
			 *
			 * @param n name of the child to find
			 */
			template <class T>
			T* child( QString n ) const {
				foreach( T* child, childobjects<T>() ) {
					if ( child->objectName() == n )
						return child;
				}
				return 0;
			}

			/**
			 * @brief Get the named item recursivly
			 *
			 * This returns the item you select by path or 0 if no item is found.
			 *
			 * @param n the path to the item you search
			 */
			QTreeItem* getNamedChild( QString n ) const {
				//qDebug() << "AbstractNamedTreeItem::getNamedChild( QString" << n << ")";
				//qDebug() << " known childs are:" << childlist();
				//qDebug() << " own name is:" << objectName();
				QTreeItem* item = child( n.section( "/", 0,0 ) );
				if ( item != 0 ) {
					QString path = n.section( "/", 1,-1 );
					if ( path == "" ) {
						//qDebug() << "  path is empty, returning item(" << item << ")";
						return item;
					} else {
						//qDebug() << "  returning item->getNamedChild(" << path << ")";
						return item->getNamedChild( path );
					}
				}
				if ( n.isEmpty() ) {
					//qDebug() << " requested childname is empty, returning myself";
					return const_cast<QTreeItem*>( this );
				}
				return 0;
			}
			/**
			 * @brief Get the named item of the given type recursivly
			 *
			 * This returns the item you select by path and type or 0 if no valid item is found.
			 *
			 * @param n the path to the item you search
			 */
			template <class T>
			T* getNamedChild( QString n ) const {
				return dynamic_cast<T*>( getNamedChild( n ) );
			};

			/**
			 * @brief Returns the row-number of this item inside its parent.
			 */
			int row() const {
				if ( dynamic_cast<QTreeItem*>( parent() ) )
					return dynamic_cast<QTreeItem*>( parent() )->row( this );
				return -1;
			}
			/**
			 * @brief Returns the row-number of the given child.
			 *
			 * @param child the child that is searched for.
			 */
			int row( const QTreeItem* child ) const {
				int ret = -1;
				QList<QObject*> list = childobjects();
				for ( int i=0; i<list.size(); ++i )
					if ( list[ i ] == child ) ret = i;
				return ret;
			}

			/**
			 * @brief return the parent path
			 *
			 * This checks for all the parent QTreeItem and returns their
			 * names as a unix-like-path.
			 */
			QString path() const {
				QString ret="";
				QTreeItem* item = dynamic_cast<QTreeItem*>( parent() );
				if ( item )
					ret = item->path() + "/";
				return ret + objectName();
			}

			/**
			 * @brief GUI: return a/the value to display
			 *
			 * If you implement this, always remember to call the parent
			 * guiValue(int) if you can't return the right thing.
			 */
			virtual QVariant guiValue( int role =0 ) const {
				//qDebug() << "QTreeItem::guiValue(" << role << ") base-implementation called. returning an invalid QVariant.";
				return QVariant();
			}
			/**
			 * @brief GUI: set the value
			 *
			 * This function will be called by GUIs to change the value. You
			 * should implement it accordingly.
			 */
			virtual void guiSetValue( QVariant ) {}

			/**
			 * @brief GUI: return different types of data used to identify the
			 * treeitem
			 *
			 * Things like the name, icon, statustip, tooltip and so on will be
			 * accessed by this.
			 *
			 * @param role is actually a combination of Qt::ItemDataRole
			 */
			virtual QVariant guiIdentifier( /*Qt::ItemDataRole*/ int role=0 ) const;
			/**
			 * @brief GUI: the Flags for this item
			 */
			virtual Qt::ItemFlags guiFlags() const { return Qt::ItemIsEnabled; }

			/**
			 * @brief set the tooltip for this item
			 */
			void setToolTip( const QString& n ) { _tooltip = n; }

			/**
			 * @{
			 *
			 * @brief Xml functions
			 *
			 * Storing works as the old functions did.
			 *
			 * Restore calls restoreFromXML() and if that returns true (ie. the
			 * state of the item itself was successfully restored ),
			 * createFromXML() is called to give you a chance to create child
			 * elements as needed. If that too returns true, fromXml() is called
			 * on every child with every child-xml-element as argument.
			 */
			QDomNode toXml( QDomDocument );
			bool fromXml( QDomNode );
			// @}
			/**
			 * @brief recursive determination whether the save to xml is needed.
			 *
			 * @todo Should this be protected or private?
			 */
			bool saveToXmlNeeded();


			QTreeItem* treeParent() const { return _treeparent; }

		public slots:
			/**
			 * @brief deferred deletion
			 *
			 * Use this one instead of QObject::deleteLater()!!!
			 */
			void treeDeleteLater() {
				qDebug() << endl << "QTreeItem::treeDeleteLater()" << endl << endl;
				_deferredDelete = true;
				//clearPointers();
				setParent( 0 );
				deleteLater();
			}

			/**
			 * @brief Adds new childs to this item
			 *
			 */
			bool addChild( NSCore::QTreeItem* item );

			/// Debugoutput
			/**
			 * This prints a tree-view of this item and its childs to qDebug.
			 * Every subclass has to implement debug() to return the internal
			 * debug-data.
			 *
			 * Call it without an argument.
			 */
			void debugPrint( QString prefix="" ) const {
				if ( prefix == "" )
					qDebug() << "TreeItem::debugPrint()";
				qDebug( "%s", qPrintable( QString( "%1 +%2 (childs %3, name '%4')" ).arg( prefix ).arg( debug() ).arg( childs() ).arg( objectName() ) ) );
				foreach( QObject* obj, childobjects() ) {
					//qDebug() << " obj=" << obj;
					QTreeItem* item = 0;
					if ( ( item = dynamic_cast<QTreeItem*>( obj ) ) )
						item->debugPrint( prefix+" |" );
				}
			}

		signals:
			/**
			 */
			void newChildAdded( NSCore::QTreeItem* item );
			/**
			 */
			void aboutToDeleteMe( NSCore::QTreeItem* me );

			/**
			 * @brief The gui value changed.
			 *
			 * Emit this signal in implementations when the value shown in the gui has changed.
			 *
			 * This signal is automaticly connected to parents so when watching a tree, you only
			 * have to connect to the parents signal.
			 */
			void guiValueChanged();

		protected:
			/**
			 * @brief Custom debug text
			 *
			 * This function has to be implemented and has to return the
			 * internal debug-data that is to be printed...
			 */
			virtual QString debug() const =0;

			/**
			 * @brief Store the information of this element into a new QDomElement
			 *
			 * This function is only to save the state of this element, all the
			 * recursive stuff will be done by toXml().
			 *
			 * @param doc The QDomDocument to use for creation of the element
			 */
			virtual QDomNode saveToXML( QDomDocument doc ) =0;
			/**
			 * @brief Whether this item needs to be saved or not.
			 *
			 * If the item itself and all its childs don't need to be saved, it will be skipped.
			 *
			 * The default is to save.
			 */
			virtual bool needsToSaveToXml() { return true; }
			/**
			 * @brief Restore the state of this element from the given QDomNode
			 *
			 * You don't have to do the recursive stuff here, that is handled in
			 * fromXml() for you. Only the things done in saveToXML need/can be
			 * restored here.
			 *
			 * If you succeeded and want the childs to be restored too, return
			 * true.
			 *
			 * @param node The QDomNode to restore from
			 */
			virtual bool restoreFromXML( QDomNode node ) =0;
			/**
			 * @brief Create childs from XML
			 *
			 * On fromXml() this gets called after successfull return from
			 * restoreFromXML(). If you want to create childs of this item from
			 * the childs of the QDomNode given, you can do so.
			 *
			 * The return value defines whether fromXml will be called on the
			 * childs for the child-xml-items.
			 *
			 * The default implementation returns true, so the child items get a
			 * chance to restore their state too.
			 *
			 * @param node The QDomNode to restore from
			 */
			virtual bool createFromXML( QDomNode node ) { return true; }

			/**
			 * private because only qt's metaobject system needs to access it
			 */
			QString toolTip() const { return _tooltip; }
			QString _tooltip;

		private slots:
			void workOnXmlCache();

			void removeChild( QObject* );

		private:

			/// The additional "childs" not handled by qobjects system
			QObjectList _morechilds;
			/// Whether the object is in deferred deletion
			bool _deferredDelete;

			QList<QDomNode> _xmltoparse;
			bool _xmlcachereturn;

			QTreeItem* _treeparent;

			Q_DISABLE_COPY( QTreeItem )
	};

	/**
	 * @brief A pointer to QTreeItem
	 *
	 * Normally this just points to the item you point it to. But if you want automatic
	 * child-pointers pointing to the childs of the pointed item, you have to subclass this and
	 * implement newChildAdded() and itemDeleted().
	 */
	class QTreeItemPointer : public QTreeItem, public QPointer<QTreeItem>
	{
		Q_OBJECT
		public:
			/// Constructor
			QTreeItemPointer( QTreeItem* object, QObject* parent=0 );

			/**
			 * @{
			 *
			 * @brief Forwards to access the pointed-items functions.
			 */
			QVariant guiValue( int role ) const { return data()->guiValue( role ); }
			void guiSetValue( QVariant n ) { data()->guiSetValue( n ); }
			QVariant guiIdentifier( int role ) const { return data()->guiIdentifier( role ); }
			Qt::ItemFlags guiFlags() const { return data()->guiFlags(); }
			// @}

			QTreeItem* object() { return data(); }

		public slots:
			/**
			 * @brief A new child was added.
			 *
			 * This signals that a new child was added to the item pointed to.
			 * Can be used to create the new pointer to the new child.
			 */
			virtual void newChildAdded( NSCore::QTreeItem* ) {}
			/**
			 * @brief The item pointed to is deleted.
			 *
			 * Implement this functions if you want to be notified of the
			 * deletion of the item pointed to. Better not use the item anymore,
			 * with threads it can be that the item is already deleted.
			 */
			virtual void itemDeleted() {}

		private:
			QDomNode saveToXML( QDomDocument ) {
				return QDomNode();
			}
			bool restoreFromXML( QDomNode ) {
				return false;
			}
			QString debug() const { return QString( "Pointer to %1" ).arg( (long int)data() ); }
	};
};

#endif // ABSTRACTTREEITEM_H

