/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>
			Matthias Nagl <mnagl@uni-goettingen.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "qtreeitemmodel.h"
#include "qtreeitemmodel.moc"

#include <QtCore/QDebug>
#include <QtCore/QSize>
#include <QtCore/QTimer>
#include <QtCore/QThread>
#include <tiaction.h>

#include <modeltest.h>

namespace NSCore {

	class QTreeItemModel_Root : public QTreeItem
	{
		friend class QTreeItemModel;
		private:
			QTreeItemModel_Root( QObject* p ) : QTreeItem( "RootItem", p ) {
				//maximumChilds( 1 );
			}
			QString debug() const { return "QTreeItemModel_Root"; }
			QDomNode saveToXML( QDomDocument doc ) {
				return doc.createElement("root");
			}
			bool restoreFromXML( QDomNode ) { return false; }
	};

QTreeItemModel::QTreeItemModel( QObject* p )
	: QAbstractItemModel( p )
	, _roottreeitem( new QTreeItemModel_Root( this ) )
	, _rootitem( new QTreeItemNode( _roottreeitem, 0 ) )
{
	//new ModelTest( this, this );
	//qDebug() << "QTreeItemModel::QTreeItemModel(" << p << ") living in thread " << thread();
	connect( _rootitem, SIGNAL( itemDeleted( QTreeItemNode* ) ), this, SLOT( nodeDeleted( QTreeItemNode* ) ) );
	connect( _rootitem, SIGNAL( newChild( QTreeItemNode*, QTreeItemNode* ) ), this, SLOT( nodeAdded( QTreeItemNode*, QTreeItemNode* ) ) );
	connect( _rootitem, SIGNAL( newChildAboutToBeAdded( QTreeItemNode* ) ), this, SLOT( addingNode( QTreeItemNode* ) ) );
}
QTreeItemModel::~QTreeItemModel() {
	//qDebug() << "QTreeItemModel::~QTreeItemModel()";
	delete _rootitem;
	//delete _roottreeitem;
}

int QTreeItemModel::columnCount( const QModelIndex& parent ) const {
	//qDebug() << "QTreeItemModel::columnCount( parent=" << parent << ")";
	if ( _rootitem->isNull() )
		return 0;

	return 2;
}
int QTreeItemModel::rowCount( const QModelIndex& parent ) const {
	//qDebug() << "QTreeItemModel::rowCount( parent=" << parent << ")";
	if ( parent.column() > 0 )
		return 0;
	QTreeItemNode* p = itemFromIndex( parent );
	if ( !p )
		return 0;
	//qDebug() << "QTreeItemModel::rowCount( parent=" << parent << ") will return" << p->childobjects<QTreeItem>().size();
	return p->childobjects<QTreeItem>().size();
}

QModelIndex QTreeItemModel::index( int row, int col, const QModelIndex& parent ) const {
	if ( _rootitem->isNull() )
		return QModelIndex();
	//qDebug() << "QTreeItemModel::index( row=" << row << ", col=" << col << ", parent=" << parent << ")";

	QTreeItemNode* p = itemFromIndex( parent );
	if ( row >= 0 && row < p->childs() && col >= 0 && col < columnCount() )
		return createIndex( row, col, p->child( row ) );

	return QModelIndex();
}
QModelIndex QTreeItemModel::parent( const QModelIndex& index ) const {
	//qDebug() << "QTreeItemModel::parent( index=" << index << ")";

	QTreeItemNode* item = itemFromIndex( index );
	if ( !item )
		return QModelIndex();

	QTreeItemNode* parentItem = item->parentNode();
	if ( !parentItem )
		return QModelIndex();

	int row = parentItem->row();
	if ( row < 0 )
		return QModelIndex();
	return createIndex( row, 0, parentItem );
}

QVariant QTreeItemModel::data( const QModelIndex& index, int role ) const {
	if ( ! index.isValid() )
		return QVariant();

	QVariant ret;
	NSCore::QTreeItemNode* tmp = static_cast<NSCore::QTreeItemNode*>( index.internalPointer() );
	if ( tmp != 0 && tmp->object() != 0 ) {
		if (index.column() == 0 || (role == Qt::DecorationRole && qVariantCanConvert<QDomDocument>(tmp->object()->guiValue(Qt::DisplayRole)))) {
			ret = tmp->object()->guiIdentifier( role );
		}
		else if ( role == QTreeItemModel::MouseActionStateRole ) {
		    ret = tmp->retrieveMouseActionState();
		}
		else {
			if ( index.column() == 1 && (role == Qt::DisplayRole || role == Qt::EditRole) ) {
				//qDebug() << "Returing the guiValue of" << tmp->object();
				ret = tmp->object()->guiValue( role );
				//qDebug() << " its" << ret;
			}
		}
	}
	return ret;
}

bool QTreeItemModel::setData( const QModelIndex& index, const QVariant& value, int role ) {
	NSCore::QTreeItemNode* tmp = static_cast<NSCore::QTreeItemNode*>( index.internalPointer() );
	if ( index.column() == 1  && role == Qt::EditRole && tmp->object()->guiValue( Qt::EditRole ).type() != QVariant::Invalid ) {
		tmp->guiSetValue( value );
		emit dataChanged(index, index);
		return true;
	}
	else if (  role == QTreeItemModel::MouseActionStateRole ) {
	    tmp->cacheMouseActionState( value );
	    emit dataChanged(index, index);
	    return true;
	}
	return false;
}

QSize QTreeItemModel::span (const QModelIndex & index) const {
    if (index.isValid() && index.column() == 0) {
	qDebug() << "span called";
	NSCore::QTreeItemNode* tmp = static_cast<NSCore::QTreeItemNode*>( index.internalPointer() );
	if (qVariantCanConvert<QDomDocument>(tmp->object()->guiValue(Qt::DisplayRole)))
	    return QSize(2,1);
    }
    return QSize(1,1);
}

Qt::ItemFlags QTreeItemModel::flags( const QModelIndex& index ) const {
	Qt::ItemFlags ret;
	NSCore::QTreeItemNode* tmp = static_cast<NSCore::QTreeItemNode*>( index.internalPointer() );
	if ( tmp && tmp->object() )
		ret = tmp->object()->guiFlags();
	if ( index.column() != 1 )
		ret &= ~Qt::ItemIsEditable;
	return ret;
}

QVariant QTreeItemModel::headerData( int section, Qt::Orientation orientation, int role ) const {
	if ( orientation == Qt::Horizontal && role == Qt::DisplayRole ) {
		if ( section == 0 )
			return "Name";
		if ( section == 1 )
			return "Value";
		if ( section == 2 )
			return "Actions";
	}
	return QVariant();
}

void QTreeItemModel::addToRootItem( NSCore::QTreeItem* item ) {
	//qDebug() << "QTreeItemModel::addToRootItem(" << item << ")";
	_roottreeitem->addChild( item );
}

NSCore::QTreeItem* QTreeItemModel::rootItem() const {
	return _roottreeitem;
}


QTreeItemNode* QTreeItemModel::itemFromIndex( const QModelIndex& index ) const {
	if ( index.isValid() )
		return static_cast<QTreeItemNode*>( index.internalPointer() );
	else
		return _rootitem;
}
QModelIndex QTreeItemModel::indexFromItem( const QTreeItemNode* item ) const {
	if ( item == _rootitem )
		return createIndex( 0, 0, _rootitem );
	if ( item == 0 )
		return QModelIndex();

	QModelIndex parentindex = indexFromItem( item->parentNode() );
	return index( item->row(), 0, parentindex );
}

void QTreeItemModel::nodeDeleted( QTreeItemNode* item ) {
	//qDebug() << "QTreeItemModel::nodeDeleted(" << item << ")";
	if ( thread() == QThread::currentThread() ) {
		int row = indexFromItem( item ).row();
		QModelIndex parentindex = indexFromItem( item ).parent();
		beginRemoveRows( parentindex, row, row );
		//delete item; /// Can't remove the item directly, this might be inside an event-handler.
		item->setParent( 0 ); /// If we don't reparent the object, it might still get accessed before its finally deleted.
		item->deleteLater(); /// Schedule the object for deletion on the next run of the event-loop.
		endRemoveRows();
	} else {
		//qDebug() << "QTreeItemModel::nodeDeleted(" << item << ") will schedule the deletion for later";
		_nodesToDelete.prepend( item );
		QTimer::singleShot( 20, this, SLOT( checkForChanges() ) );
	}
}
void QTreeItemModel::addingNode( QTreeItemNode* childsparent ) {
	//qDebug() << "QTreeItemModel::addingNode(" << childsparent << ")";
	QModelIndex parentindex( indexFromItem( childsparent ) );
	//qDebug() << " index is" << parentindex;
	int row = childsparent->childs();
	//qDebug() << " row is" << row;
	beginInsertRows( parentindex, row, row );
}
void QTreeItemModel::nodeAdded( QTreeItemNode* newChild, QTreeItemNode* childsparent ) {
	//qDebug() << "QTreeItemModel::nodeAdded(" << newChild << "," << childsparent << ")";
	endInsertRows();
}

void QTreeItemModel::checkForChanges() {
	while ( ! _nodesToDelete.isEmpty() )
		nodeDeleted( _nodesToDelete.takeFirst() );
}


QTreeItemNode::QTreeItemNode( QTreeItem* item, QTreeItemNode* p )
	: QTreeItemPointer( item, p )
	, _deferredDeletion( false )
{
	//qDebug() << "QTreeItemNode::QTreeItemNode(" << item << "," << p << ") me (" << this << ") is living in thread " << thread();
	if ( item ) {
		//setObjectName( item->objectName() );
		//qDebug() << " item has" << item->childs() << "childs:" << item->childlist();
		foreach( QString n, item->childlist() ) {
			newChildAdded( item->child( n ) );
		}
	}

	startTimer( 100 );
}
QTreeItemNode::~QTreeItemNode() {
	//qDebug() << "QTreeItemNode::~QTreeItemNode()" << this;
}

void QTreeItemNode::itemDeleted() {
	//qDebug() << "QTreeItemNode::itemDeleted() [" << this << "] item is" << object();
	//if ( object() != 0 )
	//	qDebug() << " item has childs?" << object()->childs();
	//qDebug() << " own childs() is" << childs();
	if ( thread() != QThread::currentThread() || childs() > 0 ) {
		//qDebug() << " scheduling deferred deletion";
		_deferredDeletion = true;
	} else {
		Q_ASSERT( thread() == QThread::currentThread() );
		Q_ASSERT( this != 0 );
		Q_ASSERT( childs() == 0 );
		if ( this != 0 ) {
			//qDebug() << " will emit itemDeleted(" << this << ")";
			_deferredDeletion = false;
			emit itemDeleted( this );
		}
	}
}
void QTreeItemNode::newChildAdded( QTreeItem* item ) {
	if ( item == 0 )
		return;
	//qDebug() << "QTreeItemNode::newChildAdded(" << item << ") me =" << this;
	if ( thread() != QThread::currentThread() ) {
		//qWarning( "QTreeItemNode::newChildAdded() Will buffer the new event" );
		_bufferedNewItems.push_back( item );
	} else {
		if ( item == 0 )
			return;
		emit newChildAboutToBeAdded( this );
		//qDebug() << "QTreeItemNode::newChildAdded(" << item << ")";
		QTreeItemNode* tmp = new QTreeItemNode( item, this );
		connect( tmp, SIGNAL( itemDeleted( QTreeItemNode* ) ), this, SIGNAL( itemDeleted( QTreeItemNode* ) ) );
		connect( tmp, SIGNAL( newChild( QTreeItemNode*, QTreeItemNode* ) ), this, SIGNAL( newChild( QTreeItemNode*, QTreeItemNode* ) ) );
		connect( tmp, SIGNAL( newChildAboutToBeAdded( QTreeItemNode* ) ), this, SIGNAL( newChildAboutToBeAdded( QTreeItemNode* ) ) );
		emit newChild( tmp, this );
	}
}


QTreeItemNode* QTreeItemNode::parentNode() const {
	if ( !QTreeItem::parent() ) {
		//qWarning() << "QTreeItemNode::parentNode() Got no parent anymore";
		return 0;
	}
	return dynamic_cast<QTreeItemNode*>( QTreeItem::parent() );
}

QTreeItemNode* QTreeItemNode::child( int n ) const {
	return static_cast<QTreeItemNode*>( QTreeItem::child( n ) );
}

void QTreeItemNode::cacheMouseActionState(const QVariant &value) {
    mouseactionstate = value;
}

QVariant QTreeItemNode::retrieveMouseActionState() {
    return mouseactionstate;
}

QString QTreeItemNode::debug() const {
	if ( data() )
		return QString( "Pointer to: %1" ).arg( data()->objectName() );
	return "Pointer to 0!";
}

void QTreeItemNode::timerEvent( QTimerEvent* ) {
	//qDebug() << "QTreeItemNode::timerEvent( QTimerEvent* )";
	while ( !_bufferedNewItems.isEmpty() )
		newChildAdded( _bufferedNewItems.takeFirst() );

	if ( _deferredDeletion ) {
		//qDebug() << "QTreeItemNode::timerEvent() processing deferred deletion";
		_deferredDeletion = false;
		itemDeleted();
	}
}

};
