/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>
			Matthias Nagl <mnagl@uni-goettingen.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef QTREEITEMMODEL_H
#define QTREEITEMMODEL_H

#include <QtCore/QAbstractItemModel>
#include <QtCore/QList>
#include <qtreeitem.h>

namespace NSCore {

	class QTreeItemNode;
	class QTreeItemModel_Root;

	/**
	 * @brief Provides a standard model for trees of NSCore::QTreeItem
	 *
	 * The classes in this module combine the Model-View-Pattern used in Qt with
	 * the coolness of trees of QTreeItem.
	 *
	 * With this class it is possible to access a tree of
	 * NSCore::QTreeItem as a simple model of Qt's Model-View-Pattern.
	 * So all the different views can be used.
	 *
	 * Usage inside a qobject is like the following:
	 * @code
	 * TreeItemModel* model = new TreeItemModel( this );
	 * model->setRootPath( rootitem ); // Any QTreeItem will do...
	 * QTreeWidget* tw = new QTreeWidget( this );
	 * tw->setModel( model );
	 * @endcode
	 *
	 * @todo Support drag-and-drop to rearrange items.
	 */
	class QTreeItemModel : public QAbstractItemModel
	{
		Q_OBJECT
		public:
			/// Default constructor
			QTreeItemModel( QObject* );
			/// destructor
			~QTreeItemModel();

			/// implements QAbstractItemModel::columnCount()
			int columnCount( const QModelIndex& parent = QModelIndex() ) const;
			/// implements QAbstractItemModel::rowCount()
			int rowCount( const QModelIndex& parent = QModelIndex() ) const;

			/// implements QAbstractItemModel::index()
			QModelIndex index( int row, int col, const QModelIndex& parent = QModelIndex() ) const;
			/// implements QAbstractItemModel::parent()
			QModelIndex parent( const QModelIndex& index ) const;

			/// implements QAbstractItemModel::data()
			QVariant data( const QModelIndex& index, int role = Qt::DisplayRole ) const;
			/// implements QAbstractItemModel::setData()
			bool setData( const QModelIndex& index, const QVariant& value, int role = Qt::EditRole );
			/// implements QAbstractItemModel::headerData()
			QVariant headerData( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const;

			/// implements QAbstractItemModel::flags()
			Qt::ItemFlags flags( const QModelIndex& ) const;

			QTreeItem* rootItem() const;// { return _roottreeitem; }

			static const int MouseActionStateRole = 67;

			QSize span (const QModelIndex & index) const;

		public slots:
			/**
			 * @brief Add an item to the root.
			 *
			 * The added item can be any QTreeItem, even one in the middle of a
			 * tree.
			 */
			void addToRootItem( NSCore::QTreeItem* );

		private slots:
			/// A node was deleted
			void nodeDeleted( QTreeItemNode* );
			/// A new node was added
			void nodeAdded( QTreeItemNode* newchild, QTreeItemNode* childsparent );
			/// A new node is about to be added
			void addingNode( QTreeItemNode* childsparent );

			void checkForChanges();

		private:
			QTreeItemNode* itemFromIndex( const QModelIndex& ) const;
			QModelIndex indexFromItem( const QTreeItemNode* ) const;

			QTreeItemModel_Root* _roottreeitem;
			QTreeItemNode* _rootitem;
			QList<QTreeItemNode*> _nodesToDelete;
	};

	/**
	 * @internal
	 * @brief Internal class to sort tree-items for QTreeItemModel.
	 */
	class QTreeItemNode : public QTreeItemPointer
	{
		Q_OBJECT
		public:
			QTreeItemNode( QTreeItem*, QTreeItemNode* );
			~QTreeItemNode();

			void itemDeleted();

			void newChildAdded( QTreeItem* );

			QTreeItemNode* parentNode() const;

			QTreeItemNode* child( int n ) const;

			void cacheMouseActionState(const QVariant &value);
			QVariant retrieveMouseActionState();

		signals:
			/**
			 * @brief A child was deleted
			 *
			 * The QTreeItemNode* is either this item itself ( if its the one
			 * that was deleted ) or one of its childs.
			 */
			void itemDeleted( QTreeItemNode* );
			/**
			 * @brief A new child was added
			 *
			 * @param child The new item that was added
			 * @param parent The parent item the new child was added to
			 */
			void newChild( QTreeItemNode* child, QTreeItemNode* parent );
			void newChildAboutToBeAdded( QTreeItemNode* parent );

		private:
			QString debug() const;
			void timerEvent( QTimerEvent* );
			QList<QTreeItem*> _bufferedNewItems;
			bool _deferredDeletion;
			QVariant mouseactionstate;
	};

};


#endif // NAMEDTREEMODEL_H

