/*
    Copyright 2009-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
            Matthias Nagl <mnagl@uni-goettingen.de>
            Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef STATISTICALPRODUCER_H
#define STATISTICALPRODUCER_H

#include <QtCore/QObject>
#include "datasetchain.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

namespace NSCore {

    class TiParameter;

    /**
     * @brief Produces pulses that look like real scintillator/PMT pulses by
     * simulating statistical processes in scintillators.
     *
     * This class can (and will) be used to determine the mathematical
     * resolution of algorithms or to simulate and test other parts of the
     * system.
     *
     * It produces pulses that look like real sctintillator/PMT signals by
     * simulating the electron generation and amplification induced by photons
     * from a scintillator in PMTs. Several options can be changed:
     * <ul>
     * <li>The number of electrons created by photons for up to three components
     * (this number should be set to the number of photons that hit the PMT's
     * cathode multiplied by the PMT's PHR.)
     * <li>The decay time for up to three components
     * <li>The anode pulse width of the PMT that broadens the single electron peaks
     * </ul>
     */
    class StatisticalProducer : public DataSetProducer
    {
        Q_OBJECT
        public:
            /// Simple constructor
            StatisticalProducer( QObject* =0 );
            /// Simple destructor
            ~StatisticalProducer();

            void start();
            void stop();
            void abort() { stop(); }

            bool isRunning() { return (_togenerate!=0); }

        public slots:

            /**
             * @brief Generate a number of datasets
             */
            void generateNDataSets( int );

        private slots:
            /**
             * @brief The real working function
             */
            void generateDataSet();

            /**
             * @brief mark the internal parameter-cache as dirty.
             * @internal
             */
            void cacheDirty() { _dirty = true; }

            void generateOneDataSet() { generateNDataSets( 1 ); }
            void generateNDataSets(); // { generateNDataSets( _number->toInt() ); }

        private:
            QString debug() const { return "StatisticalProducer"; }

            QDomNode saveToXML( QDomDocument doc );
            bool restoreFromXML( QDomNode );

            /**
             * @brief Number of datasets to generate
             *
             * This internal variable is decremented ( by generateDataSet() )
             * every time a data set is generated and incremented ( by
             * generateNDataSets( int ) ) when more data sets are to be
             * generated.
             */
            int _togenerate;
            /**
             * @brief whether this is running in the main-thread or not
             *
             * Basicly this caches the information whether this object lives in
             * a child-thread or in the main ( gui- )thread. The rescheduling of
             * generateDataSet depends on that. If this information wasn't
             * cached, QThread::currentThread() and this->thread() would be
             * called every time a data set is generated...
             */
            bool _ownthread;

            /// Configuration elements
            TiParameter* _p;

            /// Number of datasets to create on "Generate N events"
            TiParameter* _number;

            gsl_rng * rand;

            /**
             * @{
             * @brief Caching of parameters for speed
             */
            bool _dirty;
            struct cache_t;
            cache_t* _cache;
            // @}
    };

};

#endif // STATISTICALPRODUCER_H

