/*
    Copyright 2009-2011 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Matthias Nagl <mnagl@uni-goettingen.de>
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "tabulatedstatistics.h"
#include "tabulatedstatistics.moc"

#include <QtCore/QDateTime>
#include <QtCore/QTimer>
#include <QtCore/QCoreApplication>
#include <QtCore/QThread>
#include <math.h>

#include <tiaction.h>
#include <tiparameter.h>

using namespace NSCore;


struct TabulatedStatistics::cache_t {
    int samples;
    double samplerate;
    double noise;
    double gain;
    double risetime;
    double tts;
    int dynodes;
    double hoffset;
    bool hoffsetjitter;
    double eff;

    std::vector< std::vector<double> > photontimes;
};

TabulatedStatistics::TabulatedStatistics( QObject* p )
    : DataSetProducer( "Tabulated Statistical Producer", p )
    , _togenerate( 0 )
    , _p( new TiParameter( "Parameters", QVariant::Invalid, this ) )
    , _dirty( true )
    , _cache( new cache_t )
{
    _ownthread = ( thread() != QCoreApplication::instance()->thread() );

    connect( _p, SIGNAL( guiValueChanged() ), this, SLOT( cacheDirty() ) );
    _p->addParameter( "Samples", 400 );
    _p->getParameter( "Samples" )->setRange( 0, pow(2,16) );
    _p->addParameter( "Sampling Rate", 4.0 );
    _p->getParameter( "Sampling Rate" )->setToolTip( "Sampling rate in giga-samples per second." );
    _p->addParameter( "Horizontal Offset", 20.0 );
    _p->getParameter( "Horizontal Offset" )->setToolTip( "Horizontal offset between the signal's onset and the first sample in nano seconds." );
    _p->addParameter( "Jitter Horizontal Offset", true );
    _p->getParameter( "Jitter Horizontal Offset" )->setToolTip( "Jitter simulated signal onset between two samples (if disabled all generated signals begin at the configured horizontal offset). Important for time determination studies." );
    _p->addParameter( "PMT/Gain", 2.5E6 );
    _p->getParameter( "PMT/Gain" )->setRange( 1.0, 1.0E10 );
    _p->getParameter( "PMT/Gain" )->setToolTip( "Gain of the PMT." );
    _p->addParameter( "PMT/Rise Time", 0.7 );
    _p->getParameter( "PMT/Rise Time" )->setRange( 0.0, 100.0 );
    _p->getParameter( "PMT/Rise Time" )->setToolTip( "10%/90% risetime in nano seconds." );
    _p->addParameter( "PMT/Dynode Stages", 8 );
    _p->getParameter( "PMT/Dynode Stages" )->setRange( 1, 20 );
    _p->getParameter( "PMT/Dynode Stages" )->setToolTip( "Number of dynode stages of the used PMT." );
    _p->addParameter( "PMT/Transit Time Spread", 0.37 );
    _p->getParameter( "PMT/Transit Time Spread" )->setRange( 0.0, 100.0 );
    _p->getParameter( "PMT/Transit Time Spread" )->setToolTip( "FWHM of the transit time distribution in nano seconds" );
    _p->addParameter( "PMT/Conversion Efficiency", 15.0 );
    _p->getParameter( "PMT/Conversion Efficiency" )->setRange( 0.0, 100.0 );
    _p->getParameter( "PMT/Conversion Efficiency" )->setToolTip( "Overall efficiency (in %) for converting one photon of the into a photo-electron (quantum efficiency · collection efficiency)." );

    _p->addParameter( "Scintillator/Photon File", QUrl() );


    new TiAction( "Generate one event!", this, ":/icons/kde/cr22-action-player_play.png", this, SLOT( generateOneDataSet() ) );
    TiAction* action = new TiAction( "Generate N events!", this, ":/icons/kde/cr22-action-player_fwd.png", this, SLOT( generateNDataSets() ) );
    _number = new TiParameter( "Events to generate", 1, action );

    rand = gsl_rng_alloc (gsl_rng_taus);
    gsl_rng_set(rand, QDateTime::currentDateTime().toTime_t());
}
TabulatedStatistics::~TabulatedStatistics() {
    gsl_rng_free(rand);
}

void TabulatedStatistics::start() {
    generateNDataSets();
}
void TabulatedStatistics::stop() {
    _togenerate = 0;
}
void TabulatedStatistics::generateNDataSets() {
    generateNDataSets( _number->toInt() );
}
void TabulatedStatistics::generateNDataSets( int n ) {
    //qDebug() << "TabulatedStatistics::generateNDataSets(" << n << ")";
    if ( _togenerate < 1 && n > 0 )
        QTimer::singleShot( 100, this, SLOT( generateDataSet() ) );
    _togenerate += n;
}

QDomNode TabulatedStatistics::saveToXML( QDomDocument doc ) {
    QDomElement node = doc.createElement( "reader" );
    node.setAttribute( "type", "statisticalproducer" );
    return node;
}

bool TabulatedStatistics::restoreFromXML( QDomNode node ) {
    if ( node.nodeName() != "reader" || !node.isElement() || node.toElement().attribute( "type" ) != "statisticalproducer" )
        return false;
    return true;
}

void TabulatedStatistics::generateDataSet() {

    if ( QThread::currentThread() != thread() ) {
        QTimer::singleShot( 10, this, SLOT( generateDataSet() ) );
        return;
    }

    if ( _dirty ) { // All parameters defined in the unit of samples instead of nano seconds here!!
        _cache->samples = _p->getParameter( "Samples" )->toInt();
        _cache->samplerate = _p->getParameter( "Sampling Rate" )->toDouble();
        _cache->gain = _p->getParameter( "PMT/Gain" )->toDouble();
        _cache->risetime = _p->getParameter( "PMT/Rise Time" )->toDouble(); // unit: ns
        _cache->tts = _p->getParameter( "PMT/Transit Time Spread" )->toDouble() / sqrt(8.0*M_LN2) * _cache->samplerate; // saved as sigma value (unit: samples)
        _cache->dynodes = _p->getParameter( "PMT/Dynode Stages" )->toInt();
        _cache->hoffset = _p->getParameter( "Horizontal Offset" )->toDouble() * _cache->samplerate; // unit: samples
        _cache->hoffsetjitter = _p->getParameter( "Jitter Horizontal Offset" )->toBool();
        _cache->eff = _p->getParameter( "PMT/Conversion Efficiency" )->toDouble() / 100.0;

        qDebug() << "Path is" << _p->getParameter( "Scintillator/Photon File" )->toString();

        QFile* f = new QFile( _p->getParameter( "Scintillator/Photon File" )->toString() );
        if ( f->open( QIODevice::ReadOnly | QIODevice::Text ) ) {
            qDebug() << "Will read the file of photon times now.";
            _cache->photontimes.clear();
            std::vector<double> currenttimes;
            QString line;
            while ( ! f->atEnd() ) {
                line = f->readLine();
                //qDebug() << "Parsing line '" << line << "'";

                if ( line.startsWith("# ") ) {
                } else {
                    if ( line.size() < 2 ) {
                        if ( currenttimes.size() > 0 ) {
                            _cache->photontimes.push_back( currenttimes );
                            currenttimes.clear();
                        }
                    } else {
                        currenttimes.push_back( line.toDouble() );
                    }
                }
            }
            if ( currenttimes.size() > 0 ) {
                _cache->photontimes.push_back( currenttimes );
            }

            qDebug() << "Finished reading the photontimes. Got " << _cache->photontimes.size();
            for ( uint i=0; i < _cache->photontimes.size(); ++i ) {
                qDebug() << "\t" << i << _cache->photontimes[i].size();
            }

            f->close();
        } else {
            qWarning( " ! Couldn't open input file!" );
            emit finished();
        }
        delete f;

        _dirty = false;
    }

    // sigma = 0.5927954880410816 * risetime!
    // sigma = fwhm / sqrt(8*ln(2))

    //qDebug() << "TabulatedStatistics::generateDataSet()";

    const int channels = 1;

    AcqisitionDataSet data( channels, _cache->samples );

    // calculate number of samples for computed for each single-electron-signal (4*risetime in negative and 10*risetime in positive time direction)
    const int numsamples = int(ceil(_cache->risetime * _cache->samplerate * 14.0))+2;
    const int samplesbeforeetime = int(ceil(_cache->risetime * _cache->samplerate * 4.0));

    // pre-compute parameters for the model function
    // evaluation with parameters:
    // a * exp( as*x*x ) + b * exp( bc + bl*x + bs*x*x ) + c * exp( cc + cl*x + cs*x*x )
    const double a = -0.30220021385127144 / _cache->risetime; // scaling factor
    const double as = -1.910443060465301 / _cache->risetime / _cache->risetime / _cache->samplerate / _cache->samplerate; // factor for x²
    const double b = -0.06674516227667812 / _cache->risetime; // scaling factor
    const double bc = -0.6833657773594571; // constant factor
    const double bl = 0.5305496751209839 / _cache->risetime / _cache->samplerate; // factor for x
    const double bs = -0.10297668067994235 / _cache->risetime / _cache->risetime / _cache->samplerate / _cache->samplerate; // factor for x²
    const double c = -0.11218881998806554 / _cache->risetime; // scaling factor
    const double cc = -0.55931178280345; // constant factor
    const double cl = 1.2199010997601758 / _cache->risetime / _cache->samplerate; // factor for x
    const double cs = -0.6651740312607746 / _cache->risetime / _cache->risetime / _cache->samplerate / _cache->samplerate; // factor for x²

    // calculate gain of each dynode
    const double dynodegain = pow(_cache->gain, 1.0 / _cache->dynodes);

    // compute measured voltage from charge-normalized amplitude
    const double amplitude2volt = 1.602E-19 * 50.0 * 1.0E9;

    //qDebug() << "Have to generate data for" << data.channels() << "channels.";

    for ( int i=0; i<data.channels(); ++i ) {
        //qDebug() << "Generating data for channel" << i;

        // Cache to be faster...
        double *channel = data.getRawChannel( i );

        // reset memory
        for (int m=0; m<_cache->samples; m++)
            channel[m] = 0.0;

        // create random jitter from the interval [0,1[ if enabled
        // horizontal 0-position of the complete signal
        // ## hpos in samples ##
        const double hpos = _cache->hoffset + (_cache->hoffsetjitter ? gsl_ran_flat(rand, 0.0, 1.0) : 0.0);
        data.addResult( QString( "channel %1/horizontalposition" ).arg( i ), hpos );

        // choose one of the given time-distributions of photons
        std::vector<double>& electrons = _cache->photontimes[ gsl_rng_uniform_int(rand, _cache->photontimes.size() ) ];

        // compute number of cathode electrons for this signal
        //const int numel = gsl_ran_poisson(rand, _cache->gammaenergy / _cache->gammareference * electrons.size() * _cache->eff );
        const int numel = electrons.size() * _cache->eff;
        //qDebug() << " Resulting number of electrons:" << numel << "from" << electrons.size() << "available";

        // loop over the electrons of this signal                       ___j___
        for (int j=0; j<numel; j++) {

            // determine time of this photon by adding exponential decay of the scintillation states and transit time spread from the PMT
            // ## etime in samples ##
            // ## modulo-division to protect against element-access errors
            double etime = hpos + electrons[ gsl_rng_uniform_int(rand, electrons.size()) ] * _cache->samplerate + gsl_ran_gaussian(rand, _cache->tts);

            // determine the random amplification modificator for this photon
            // note that the sum of poisson distributed random variables X_s = X_1 + X_2 is poisson distributed itself with λ_s = λ_1 + λ_2
            double currentgain = 1.0;
            for (int m=0; m<_cache->dynodes; m++)
                currentgain = gsl_ran_poisson(rand, dynodegain * currentgain);

            // add gaussian to data
            const int firstsample = int(floor(etime))-samplesbeforeetime;
            for (int m=0; m<numsamples; m++) {
                const int csample = firstsample+m;
                if (csample >= 0 && csample < _cache->samples) {
                    // Compute single electron signal as a sum of three gaussians (approximation of real-world signals).
                    // The Integral of the sum of gaussians is normalized. It is rescaled in the next step.
                    const double x = double(csample)-etime;
                    const double xsquared = x*x;
                    const double gsample = a * exp( as*xsquared ) + b * exp( bc + bl*x + bs*xsquared ) + c * exp( cc + cl*x + cs*xsquared );
                    // The normalized signal's amplitude (volts) * 50 Ohm * elementary charge equals the current in A.
                    // Additionally it must be divided by the sample interval and multiplied by the gain
                    channel[csample] += gsample * amplitude2volt * currentgain;
                }
            }
        }
    }
    data.updateMinMax();

    // workaround
    data.addResult( "channel 0/fullscale", qMin( -data.getResult( "channel 0/min" )*1.2, 2.0 ) );
    data.addResult( "channel 0/offset",  -data.getResult( "channel 0/min" )*0.5 );

    enqueue( data );

    --_togenerate;
    if ( _togenerate > 0 ) {
        if ( _ownthread )
            QTimer::singleShot( 0, this, SLOT( generateDataSet() ) );
        else
            QTimer::singleShot( 10, this, SLOT( generateDataSet() ) );
    } else
        emit finished();
}

// vim: et sw=4
