
#ifndef QTREEITEMTEST
#define QTREEITEMTEST

#include <QtTest/QtTest>

#include <datasetchain.h>

#include <thread_template.h>

using namespace NSCore;

class Producer : public NSCore::DataSetProducer
{
    Q_OBJECT
    public:
        Producer() : NSCore::DataSetProducer( "Producer" ), toProduce( 0 ) {}
        int toProduce;
        QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "bla" ); }
        bool restoreFromXML( QDomNode node ) { return node.nodeName() == "bla"; }
        QString debug() const { return "Producer"; }

        void stop() { toProduce = 0; }
        void abort() {}
        void reset() {}

        bool isRunning() { return (toProduce!=0); }

        void start() {
            while ( toProduce > 0 ) {
                int samples = 100;
                NSCore::AcqisitionDataSet data( 2, samples );
                for ( int i=0; i<data.channels(); ++i ) {
                    double* ch = data.getRawChannel( i );
                    for ( int j=0; j<samples; ++j )
                        ch[ j ] = rand()*2.0/RAND_MAX-1;
                }
                data.updateMinMax();
                enqueue( data );
                toProduce--;
            }
            QTimer::singleShot( 1000, this, SIGNAL( finished() ) );
        }
};

class Consumer : public NSCore::DataSetChainMember
{
    bool dataArrived( NSCore::AcqisitionDataSet data ) {
        //qDebug() << "Consumer::dataArrived(" << &data << ")";
        for ( int i=0; i<data.channels(); ++i ) {
            double val=0;
            double* ch = data.getRawChannel( i );
            int s = data.samples();
            for ( int j=0; j<s; ++j )
                val += ch[ j ];
            val /= s;
            //qDebug() << " Channel" << i << ": Average =" << val;
            data.addResult( QString( "channel %1/average" ).arg( i ), val );
            data.addResult( QString( "channel %1/samples processed" ).arg( i ), s );
        }
        //data.debugPrint();
        //qDebug() << " done.";
        return true;
    }
    public:
        Consumer() : NSCore::DataSetChainMember( "Consumer" ) {}

        QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "bla" ); }
        bool restoreFromXML( QDomNode node ) { return node.nodeName() == "bla"; }
        QString debug() const { return "Consumer"; }
};

class ChainMember : public NSCore::DataSetChainMember
{
    public:
        ChainMember() : NSCore::DataSetChainMember( "ChainMember" ) {}

        bool dataArrived( NSCore::AcqisitionDataSet data ) {
            //qDebug() << "ChainMember::dataArrived(" << &data << ")";
            int oldgeneration = data.currentGenerationNumber();
            data.newGeneration( "averaged", data.channels(), data.samples() );
            for ( int i=0; i<data.channels(); ++i ) {
                double* newch = data.getRawChannel( i );
                double* oldch = data.getRawChannel( i, oldgeneration );
                double avg[ 3 ];
                avg[ 0 ] = 0.0;
                avg[ 1 ] = 0.0;
                avg[ 2 ] = oldch[ 0 ];
                int s = data.samples();
                for ( int j=0; j<s; ++j ) {
                    avg[ 0 ] = avg[ 1 ];
                    avg[ 1 ] = avg[ 2 ];
                    avg[ 2 ] = oldch[ j+1 ];
                    double tmp = ( avg[ 0 ] + avg[ 1 ] + avg[ 2 ] ) / 3.0;
                    newch[ j ] = tmp;
                }
                data.updateMinMax();
                double val=0;
                for ( int j=0; j<s; ++j )
                    val += newch[ j ];
                val /= s;
                //qDebug() << " Channel" << i << ": Average =" << val;
                data.addResult( QString( "channel %1/average" ).arg( i ), val );
            }
            data.popPrefix( "averaged" );
            return true;
        }
        QString debug() const { return "ChainMember"; }
        QDomNode saveToXML( QDomDocument ) { return QDomNode(); }
        bool restoreFromXML( QDomNode node ) { return node.nodeName() == "bla"; }
};

#define MYBENCHMARK( CODE ) \
    QTime t; \
    t.start(); \
    CODE \
    double time = t.elapsed() - 1000; \
    qDebug() << toproduce / time * 1000 << "counts/s"; \
    time = toproduce / time * 1000;

class DataSetChainTest : public QObject
{
    Q_OBJECT
    private:
        NSCore::Thread_Template<Producer>* t1;
        Producer *producer;
        NSCore::Thread_Template<ChainMember>* t21, *t22;
        ChainMember *chainmember1, *chainmember2;
        NSCore::Thread_Template<Consumer> *t3, *t4;
        Consumer *consumer1, *consumer2;
        double avg;
        int c;
    private slots:
        void initTestCase() {
            t1 = new NSCore::Thread_Template<Producer>( this );
            t1->start();
            t1->waitForStarted();
            producer = t1->item(); //new Producer();
            QVERIFY( connect( producer, SIGNAL( finished() ), QThread::currentThread(), SLOT( quit() ) ) );
            t21 = new NSCore::Thread_Template<ChainMember>( this );
            t21->start();
            t21->waitForStarted();
            chainmember1 = t21->item(); //new ChainMember();
            t22 = new NSCore::Thread_Template<ChainMember>( this );
            t22->start();
            t22->waitForStarted();
            chainmember2 = t22->item(); //new ChainMember();
            t3 = new NSCore::Thread_Template<Consumer>( this );
            t3->start();
            t3->waitForStarted();
            consumer1 = t3->item(); //new Consumer();
            t4 = new NSCore::Thread_Template<Consumer>( this );
            t4->start();
            t4->waitForStarted();
            consumer2 = t4->item(); //new Consumer();
        }
        void cleanupTestCase() {
            //qDebug() << "Deleting consumer2";
            delete t4;
            QVERIFY( chainmember1->consumersIsEmpty() );
            //qDebug() << "Deleting consumer1";
            delete t3;
            //qDebug() << "Deleting chainmember";
            delete t21;
            delete t22;
            QVERIFY( producer->consumersIsEmpty() );
            //qDebug() << "Deleting producer";
            delete t1;
        }
        void simplesetup() {
            producer->registerConsumer( consumer1 );
            QVERIFY( ! producer->consumersIsEmpty() );
            avg = 0.0;
            c = 0;
        }
        void simplebenchmark_data() {
            QTest::addColumn<int>( "toproduce" );
            //QTest::newRow( "single" ) << 1;
            //QTest::newRow( "hundred" ) << 100;
            QTest::newRow( "thousand-1" ) << 1000;
            //QTest::newRow( "thousand-2" ) << 1000;
            QTest::newRow( "5-thousand" ) << 5000;
            QTest::newRow( "many-1.1  " ) << 10000;
            //QTest::newRow( "many-1.2  " ) << 10000;
            //QTest::newRow( "many-1.3  " ) << 10000;
            //QTest::newRow( "many-2.1  " ) << 100000;
            //QTest::newRow( "many-2.2  " ) << 100000;
        }
        void simplebenchmark() {
            QFETCH( int, toproduce );
            MYBENCHMARK(
                producer->toProduce = toproduce;
                QTimer::singleShot( 1, producer, SLOT( start() ) );
                QCoreApplication::exec();
                )
            avg += time;
            c++;
            QVERIFY( producer->toProduce == 0 );
        }
        void simplesetdown() {
            producer->unregisterAllConsumers();
            QVERIFY( producer->consumersIsEmpty() );
            avg /= c;
            qDebug() << "Average countrate was" << avg << "counts/s.";
        }
        void complexsetup() {
            avg = 0.0;
            c = 0;
            //producer->setDispatching( NSCore::DataSetProducer::DispatchEqual );
            producer->registerConsumer( consumer1 );
            //consumer1->registerConsumer( chainmember );
            producer->registerConsumer( chainmember1 );
            producer->registerConsumer( chainmember2 );
            chainmember1->registerConsumer( consumer2 );
            chainmember2->registerConsumer( consumer2 );
            QVERIFY( ! producer->consumersIsEmpty() );
            //QVERIFY( ! consumer1->consumersIsEmpty() );
            QVERIFY( ! chainmember1->consumersIsEmpty() );
            QVERIFY( ! chainmember2->consumersIsEmpty() );
            QVERIFY( consumer2->consumersIsEmpty() );
        }
        void complexbenchmark_data() {
            QTest::addColumn<int>( "toproduce" );
            QTest::newRow( "thousand-1" ) << 1000;
            //QTest::newRow( "thousand-2" ) << 1000;
            //QTest::newRow( "5-thousand" ) << 5000;
            QTest::newRow( "many-1    " ) << 10000;
            //QTest::newRow( "many-2.1  " ) << 100000;
            //QTest::newRow( "many-2.2  " ) << 100000;
        }
        void complexbenchmark() {
            QFETCH( int, toproduce );
            MYBENCHMARK(
                producer->toProduce = toproduce;
                QTimer::singleShot( 1, producer, SLOT( start() ) );
                QCoreApplication::exec();
                )
            avg += time;
            c++;
            QVERIFY( producer->toProduce == 0 );
        }
        void complexsetdown() {
            producer->unregisterAllConsumers();
            consumer1->unregisterAllConsumers();
            consumer2->unregisterAllConsumers();
            chainmember1->unregisterAllConsumers();
            chainmember2->unregisterAllConsumers();
            QVERIFY( producer->consumersIsEmpty() );
            QVERIFY( consumer1->consumersIsEmpty() );
            QVERIFY( chainmember1->consumersIsEmpty() );
            QVERIFY( chainmember2->consumersIsEmpty() );
            QVERIFY( consumer2->consumersIsEmpty() );
            avg /= c;
            qDebug() << "Average countrate was" << avg << "counts/s.";
        }
};

#endif
