
#ifndef QTREEITEMTEST
#define QTREEITEMTEST

#include <QtTest/QtTest>
#include <cmath>

#include <mathexpression.h>

using namespace NSCore;

class MathExpressionTest : public QObject
{
	Q_OBJECT
	private:
		MathTreeRoot* tree;
		AcqisitionDataSet* dataset;
	private slots:
		void initTestCase() {
			tree = new MathTreeRoot( "Math", 0 );
			QCOMPARE( tree->childs(), 1 );
			dataset = new AcqisitionDataSet( 1, 15 );
			double* ch = dataset->getRawChannel( 0 );
			for ( int i=0; i<15; ++i ) {
				ch[ i ] = fabs( -i+5.0 ) - 5.0;
			}
			dataset->updateMinMax();
			dataset->debugPrint();
			QCOMPARE( dataset->getResultNames().size(), 4 );
		}
		void cleanupTestCase() {
			QPointer<MathTreeRoot> tmp = tree;
			delete tree;
			QVERIFY( tmp == 0 );
		}
		void emptyMath() {
			QVERIFY( isnan( tree->value( *dataset ) ) );
		}

		void mathBenchmark_data() {
			QTest::addColumn<QString>( "xml" );
			QTest::addColumn<double>( "result" );

			QTest::newRow( "constant" ) << "<path name=\"Math\"><constant value=\"2.0\" /></path>" << 2.0;
			QTest::newRow( "result" ) << "<path name=\"Math\"><result name=\"channel 0/min\" /></path>" << -5.0 ;
			QTest::newRow( "add" ) << "<path name=\"Math\"><add><result name=\"channel 0/min\" /><constant value=\"1\" /></add></path>" << -4.0 ;
			QTest::newRow( "multiply" ) << "<path name=\"Math\"><mul><result name=\"channel 0/min\" /><constant value=\"2\" /></mul></path>" << -10.0 ;
			QTest::newRow( "minimum" ) << "<path name=\"Math\"><min><result name=\"channel 0/min\" /><result name=\"channel 0/max\" /></min></path>" << -5.0 ;
			QTest::newRow( "maximum" ) << "<path name=\"Math\"><max><result name=\"channel 0/min\" /><result name=\"channel 0/max\" /></max></path>" << 4.0 ;
			QTest::newRow( "negation" ) << "<path name=\"Math\"><neg><result name=\"channel 0/min\" /></neg></path>" << 5.0 ;
			QTest::newRow( "precision0.1" ) << QString( "<path name=\"Math\"><precision precision=\"0.1\"><constant value=\"%1\" /></precision></path>" ).arg( M_PI ) << 3.1;
			QTest::newRow( "precision0.001" ) << QString( "<path name=\"Math\"><precision precision=\"0.001\"><constant value=\"%1\" /></precision></path>" ).arg( M_PI ) << 3.141;
			QTest::newRow( "precision0.00001" ) << QString( "<path name=\"Math\"><precision precision=\"0.00001\"><constant value=\"%1\" /></precision></path>" ).arg( M_PI) << 3.14158;
			//QTest::newRow( "precision0.0000001" ) << QString( "<path name=\"Math\"><precision precision=\"0.0000001\"><constant value=\"%1\" /></precision></path>" ).arg( M_PI ) << 3.141592654;
			QWARN( "SmB isn't checked at the moment" );
			QString tmp = "<constant value=\"1\" /><constant value=\"2\" />";
			QTest::newRow( "ltz-true"  ) << "<path name=\"Math\"><ltz><result name=\"channel 0/min\" />" + tmp + "</ltz></path>" << 1.0;
			QTest::newRow( "ltz-false" ) << "<path name=\"Math\"><ltz><result name=\"channel 0/max\" />" + tmp + "</ltz></path>" << 2.0;
			QTest::newRow( "gtz-true"  ) << "<path name=\"Math\"><gtz><result name=\"channel 0/max\" />" + tmp + "</gtz></path>" << 1.0;
			QTest::newRow( "gtz-false" ) << "<path name=\"Math\"><gtz><result name=\"channel 0/min\" />" + tmp + "</gtz></path>" << 2.0;
			QTest::newRow( "window-low" ) << "<path name=\"Math\"><window><parameter name=\"minimum\"\
				type=\"double\" value=\"-3\" /><parameter name=\"maximum\" type=\"double\"\
				value=\"-1\" /><result name=\"channel 0/min\" />"+tmp + "</window>" << 2.0;
			QTest::newRow( "window-high" ) << "<path name=\"Math\"><window><parameter name=\"minimum\"\
				type=\"double\" value=\"-3\" /><parameter name=\"maximum\" type=\"double\"\
				value=\"-1\" /><result name=\"channel 0/max\" />"+tmp + "</window>" << 2.0;
			QTest::newRow( "window-fits" ) << "<path name=\"Math\"><window><parameter name=\"minimum\"\
				type=\"double\" value=\"-10\" /><parameter name=\"maximum\" type=\"double\"\
				value=\"0\" /><result name=\"channel 0/min\" />"+tmp + "</window>" << 1.0;
			QTest::newRow( "fuzzy-low" ) << "<path name=\"Math\"><fuzzywindow><parameter name=\"center\"\
				type=\"double\" value=\"0\" /><parameter name=\"sigma\" type=\"double\"\
				value=\"0.1\" /><result name=\"channel 0/min\" />"+tmp + "</fuzzywindow>" << 2.0;
			QTest::newRow( "fuzzy-high" ) << "<path name=\"Math\"><fuzzywindow><parameter name=\"center\"\
				type=\"double\" value=\"-10\" /><parameter name=\"sigma\" type=\"double\"\
				value=\"0.1\" /><result name=\"channel 0/max\" />"+tmp + "</fuzzywindow>" << 2.0;
			QTest::newRow( "fuzzy-fits" ) << "<path name=\"Math\"><fuzzywindow><parameter name=\"center\"\
				type=\"double\" value=\"-5\" /><parameter name=\"sigma\" type=\"double\"\
				value=\"0.1\" /><result name=\"channel 0/min\" />"+tmp + "</fuzzywindow>" << 1.0;
		}
		void mathBenchmark() {
			QFETCH( QString, xml );
			QFETCH( double, result );
			QDomDocument doc( "bla" );
			doc.setContent( xml );
			tree->fromXml( doc.documentElement() );
			QCOMPARE( tree->childs(), 2 );
			QBENCHMARK {
				QCOMPARE( tree->value( *dataset ), double( result ) );
			}
		}
};

#endif
