/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef TIENUM_H
#define TIENUM_H

#include <qtreeitem.h>

#include <QtXml/QDomNode>
#include <QtCore/QUrl>

namespace NSCore {

	/**
	 * @brief Enum in a named tree structur
	 */
	class TiEnum : public QTreeItem
	{
		Q_OBJECT

		Q_FLAGS( GuiHint )
		/**
		 * @brief Hints for the GUI to display the parameter the right way.
		 *
		 * Several hints can be given via this property.
		 *
		 * Read with guiHint(), set with setGuiHint().
		 */
		Q_PROPERTY( GuiHint guiHint READ guiHint WRITE setGuiHint )
		
		/**
		 * @brief The value of this parameter
		 *
		 * Probably the most important property...
		 */
		//Q_PROPERTY( QVariant value READ value WRITE setValue RESET reset )

		/**
		 * @brief The type as QString
		 */
		//Q_PROPERTY( QString typeName READ typeName )


		public:
			/**
			 * @brief Hint-types for the GUI
			 */
			enum GuiHint {
				/** This parameter should be read only. For example for control-messages. */
				ReadAllowed = 1,
				/** This parameter is write-only... */
				WriteAllowed = 2,
				/** The parameter is a normal read-and-write-parameter. */
				ReadWrite = 3,
				/** The gui should provide a reset-button. */
				ProvideReset = 4,
				/** This parameter is a counter. It is read-only and should be resetted. */
				Counter = 5,
				/** Url: Get a filename for read. */
				UrlRead = 1,
				/** Url: Get a filename for write. */
				UrlWrite = 2,
				/** Url: Get a directory name. */
				UrlDir = 8
			};

			/**
			 * @brief Creates a new named parameter with value v
			 *
			 * @param n the name of this parameter
			 * @param v the value of this parameter, the type is determined
			 * from this
			 * @param p the parent TiParameter
			 */
			TiEnum( QString n, QStringList, QString, QTreeItem* p );
			/**
			 * @brief Creates a new named parameter with type t
			 *
			 * @param n the name of this parameter
			 * @param t the type of this parameter
			 * @param p the parent TiParameter
			 */
			TiEnum( QString n, QStringList, int, QTreeItem* p );

			/**
			 * @brief Copy a TiParameter and possibly its childs
			 * into another tree
			 *
			 * Childs are only copied if they are also TiParameters...
			 *
			 * @param item The item that is to be copied
			 * @param p The parent of the new item
			 */
			//TiEnum( TiParameter& item, QTreeItem* p );

			/// destructor
			~TiEnum();

			/// Returns the type.
			//QVariant::Type type() const { return _value.type(); }
			/// Returns the type as string.
			//QString typeName() const { return QVariant::typeToName( _value.type() ); }

			/// Returns the value.
			//QVariant value() const { return _value; }

			/// Returns the guiHint
			int guiHint() const { return _guihint; }
			/// Set the guiHint
			void setGuiHint( int n );

			/// Converts this to a QVariant
			//operator QVariant() { return value(); }

			/**
			 * @{
			 *
			 * @brief Simplify the access
			 */
			int toInt() const { return _current; }
			bool toBool() const { return _current; }
			QString toString() const { return _list[ _current ]; }
			// @}

			/**
			 * @brief access child parameters
			 *
			 * This function returns Parameters by path ( if they exist ).
			 *
			 * @param path The path to the parameter
			 */
			//TiParameter* getParameter( QString path );

			/**
			 * @brief adds a new Parameter as child
			 *
			 * @param path The path to the new item.
			 * @param value The initial value of the new item.
			 * @param tooltip A string to be shown as tooltip in GUIs.
			 */
			//void addParameter( QString path, QVariant value, QString tooltip=0 );
			/**
			 * @brief adds a new Parameter as child
			 *
			 * @param path The path to the new item.
			 * @param type The initial type of the new item.
			 * @param tooltip A string to be shown as tooltip in GUIs.
			 */
			//void addParameter( QString path, QVariant::Type type, QString tooltip=0 );

			QVariant guiValue( int role ) const;
			void guiSetValue( QVariant v ) { setValue( v ); }
			Qt::ItemFlags guiFlags() const;
			QVariant guiIdentifier( int role ) const;


		public slots:
			/**
			 * @brief Set the value
			 *
			 * Sets the value of the option.
			 * If the type doesn't match, nothing is done.
			 * If the value exceeds min/max, the respective value is set.
			 *
			 * This slot is protected against recursions with signals/slots.
			 */
			void setValue( QVariant );
			/**
			 * @{
			 *
			 * @brief Set the value
			 *
			 * This function is for convenience and behaves like setValue(QVariant)
			 */
			void setValue( int );
			void setValue( bool );
			void setValue( QString );
			// @}

			/**
			 * @brief Reset the value to the default
			 */
			void reset();

		private slots:
			void unlockUpdate();

		private:
			QString debug() const {
				QString tmp = QString( "(%1 from %2)" ).arg( toString() ).arg( _list.join(",") );
				if ( _list.isEmpty() )
					tmp="";
				return QString( "%1%2" ).arg( objectName() ).arg( tmp );
			}

			bool needsToSaveToXml();
			QDomNode saveToXML( QDomDocument );
			bool restoreFromXML( QDomNode );

			/**
			 * @brief masked copy-constructor
			 *
			 * has to be private since the new parent needs to be known
			 */
			/*TiEnum( const TiParameter& )
				: QTreeItem( "", 0 )
				{}*/
			Q_DISABLE_COPY( TiEnum );

			QStringList _list;
			int _current;

			bool _inupdate;
			int _guihint;
	};

};

#endif // TIENUM_H

