/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "tiparameter.h"
#include "tiparameter.moc"

#include <QtCore/QDebug>
#include <QtCore/QUrl>

#include "tiaction.h"

namespace NSCore {

	TiParameter::TiParameter( QString n, QVariant v, QTreeItem* p )
		: QTreeItem( n, p )
		, _value( v )
		, _min( QVariant::Invalid )
		, _max( QVariant::Invalid )
		, _inupdate( false )
		, _guihint( 3 )
	{ }
	TiParameter::TiParameter( QString n, QVariant::Type t, QTreeItem* p )
		: QTreeItem( n, p )
		, _value( QVariant( t ) )
		, _min( QVariant::Invalid )
		, _max( QVariant::Invalid )
		, _inupdate( false )
		, _guihint( 3 )
	{ }

	TiParameter::TiParameter( TiParameter& item, QTreeItem* p )
		: QTreeItem( item.objectName(), p )
	{
		foreach( QString s, item.childlist() ) {
			TiParameter* tmp = dynamic_cast<TiParameter*>( item.child( s ) );
			if ( tmp != 0 )
				new TiParameter( *tmp, this );
		}
	}
	TiParameter::~TiParameter() {
		//qDebug() << "TiParameter::~TiParameter()" << path() << " " << objectName();
	}

	void TiParameter::setValue( QVariant v ) {
		//qDebug() << "TiParameter::setValue(" << v << ")";
		if ( _value.type() == QVariant::Url )
			v = QUrl( v.toString() );
		//qDebug() << " _inupdate=" << _inupdate << " canConvert()" << v.canConvert( _value.type() );

		if ( !_inupdate && v.canConvert( _value.type() ) )
		{
			if ( _min.type() != QVariant::Invalid && _max.type() != QVariant::Invalid ) {
				if (
					( _value.type()==QVariant::Double && v.toDouble() >= _min.toDouble() && v.toDouble() <= _max.toDouble() )
					||
					( _value.type()==QVariant::Int && v.toInt() >= _min.toInt() && v.toInt() <= _max.toInt() )
					||
					( _value.type() != QVariant::Double && _value.type() != QVariant::Int )
				)
				{
					//qDebug() << " Boundaries are checked and okay";
				} else {
					// Missed boundaries
					qWarning() << "TiParameter::setValue(" << v << ")";
					qWarning() << QString( " Boundaries missed! [%1,%2]" ).arg( _min.toString() ).arg( _max.toString() );
					return;
				}
			}
			//else
			//	qDebug() << " No boundaries set";

			_inupdate = true; // Begin with blocking the update for other calls

			v.convert( _value.type() );

			_value = v;

			emit guiValueChanged();

			_inupdate = false; // Release the lock
			//QTimer::singleShot( 20, this, SLOT( unlockUpdate() ) );
		}
		//else
		//	qWarning() << "Parameter::setValue( QVariant ) Either in update or wrong type (" << _value.type() << "|" << v.type() << "), _inupdate =" << _inupdate;
	}

	void TiParameter::setValue( int n ) { setValue( QVariant( n ) ); }
	void TiParameter::setValue( bool n ) { setValue( QVariant( n ) ); }
	void TiParameter::setValue( double n ) { setValue( QVariant( n ) ); }
	void TiParameter::setValue( QString n ) { setValue( QVariant( n ) ); }
	void TiParameter::setValue( QUrl n ) { setValue( QVariant( n ) ); }

	void TiParameter::setMin( QVariant v ) { _min = v; }
	void TiParameter::setMax( QVariant v ) { _max = v; }

	void TiParameter::reset() {
		setValue( QVariant( _value.type() ) );
	}

	void TiParameter::unlockUpdate() {
		_inupdate = false;
	}

	TiParameter* TiParameter::getParameter( QString path ) {
		//qDebug() << "TiParameter::getParameter(" << path << ")";
		QTreeItem* item = getNamedChild( path );
		TiParameter* parameter = dynamic_cast<TiParameter*>( item );
		if ( parameter ) {
			//qDebug() << " found a valid Parameter" << parameter;
			return parameter;
		}
		//qWarning() << "TiParameter::getParameter(" << path << ")\n  Will return 0!";
		return 0;
	}

	void TiParameter::addParameter( QString path, QVariant value, QString tooltip ) {
		//qDebug() << "TiParameter::addParameter(" << path << "," << value << ")";
		QString parentpath = path.section( "/", 0, -2 );
		QTreeItem* p = getNamedChild( parentpath );
		if ( parentpath != "" && p == 0 ) {
			QString firstpath = parentpath.section( "/", 0, 0 );
			TiParameter* item = dynamic_cast<TiParameter*>( getParameter( firstpath ) );
			if ( !item )
				addParameter( firstpath, QVariant::Invalid );
			item = dynamic_cast<TiParameter*>( getParameter( firstpath ) );
			if ( item )
				item->addParameter( path.section( "/", 1, -1 ), value );
			else
				qWarning() << "! Warning: couldn't create child!";
		} else {
			if ( parentpath == "" )
				p = this;
			TiParameter* tmp = new TiParameter( path.section( "/", -1, -1 ), value, p );
			tmp->setToolTip( tooltip );
		}
	}
	void TiParameter::addParameter( QString path, QVariant::Type type, QString tooltip ) {
		//qDebug() << "TiParameter::addParameter(" << path << "," << type << ")";
		addParameter( path, QVariant( type ), tooltip );
	}

	void TiParameter::setGuiHint( int n ) {
		if ( _guihint & TiParameter::ProvideReset && ! n & TiParameter::ProvideReset ) {
			delete getNamedChild( "Reset" );
		}
		_guihint = n;
		if ( _guihint & TiParameter::ProvideReset ) {
			TiAction* action = new TiAction( "Reset", this );
			connect( action, SIGNAL( activated() ), this, SLOT( reset() ) );
		}
	}

	Qt::ItemFlags TiParameter::guiFlags() const {
		Qt::ItemFlags ret = QTreeItem::guiFlags()| Qt::ItemIsSelectable;
		if ( _value.type() != QVariant::Invalid )
			ret |= Qt::ItemIsEditable ;
		return ret;
	}
	/*QStringList TiParameter::guiActions() const {
		QStringList ret = QTreeItem::guiActions();
		if ( guiHint() & TiParameter::ProvideReset )
			ret << "Reset";
		if ( _value.type() == QVariant::Url )
			ret << "Sel. File";
		return ret;
	}
	void TiParameter::guiExecAction( QString action, GuiServer_Interface* gui ) {
		qDebug() << "TiParameter::guiExecAction(" << action << "," << gui <<")";
		if ( action == "Reset" ) {
			setValue( 0 );
		} if ( action == "Sel. File" ) {
			QString file;
			//qDebug() << " guiHint is" << _guihint << " &UrlRead =" << int( _guihint&UrlRead ) << " &UrlWrite =" << int( _guihint&UrlWrite ) << " &UrlDir =" << int( _guihint&UrlDir );
			if ( _guihint == 3 )
				_guihint = UrlWrite;
			if ( _guihint & UrlRead )
				file = gui->getFileName( "Select a file to open for reading", "", _urlfilter, GuiServer_Interface::Read );
			if (  _guihint & UrlWrite )
				file = gui->getFileName( "Select a file to open for writing", "", _urlfilter, GuiServer_Interface::Write );
			if (  _guihint & UrlDir )
				file = gui->getFileName( "Select a dir for reading/writing", "", "", GuiServer_Interface::Dir );
			if ( ! file.isNull() )
				setValue( file );
		} else
			QTreeItem::guiExecAction( action, gui );
	}*/
	QVariant TiParameter::guiIdentifier( int role ) const {
		switch ( role ) {
			case Qt::ToolTipRole:
				return _tooltip;
				break;
			default:
				return QTreeItem::guiIdentifier( role );
		}
		return QVariant();
	}

	bool TiParameter::needsToSaveToXml() {
		if ( type() == QVariant::Url )
			return true;
		if ( ! ( _guihint & WriteAllowed ) )
			return false;
		if ( type() == QVariant::Invalid )
			return false;
		return true;
	}
	QDomNode TiParameter::saveToXML( QDomDocument doc ) {
		QDomElement node = doc.createElement( "parameter" );
		node.setAttribute( "name", objectName() );
		node.setAttribute( "type", QVariant::typeToName( type() ) );
		if ( type() != QVariant::Invalid )
			node.setAttribute( "value", toString() );
		return node;
	}
	bool TiParameter::restoreFromXML( QDomNode node ) {
		if ( node.nodeName() == "parameter" ) {
			QDomElement elem = node.toElement();
			QString name, type, value;

			name = elem.attribute( "name" );
			type = elem.attribute( "type" );
			value = elem.attribute( "value" );

			if ( name == objectName() && type == QVariant::typeToName( this->type() ) ) {
				setValue( value );
				return true;
			}
		}
		return false;
	}

};

