/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef TIPARAMETER_H
#define TIPARAMETER_H

#include <qtreeitem.h>

#include <QtXml/QDomNode>
#include <QtCore/QUrl>

namespace NSCore {

	/**
	 * @brief Parameter in a named tree structur
	 *
	 * This class manages Parameters in an QTreeItem structure.
	 *
	 * @note If a class lets you acces the configuration via
	 * TiParameter or TiParameterPointer, you should
	 * <b>never</b> delete one of the childs as this will result in very strange
	 * error!
	 */
	class TiParameter : public QTreeItem
	{
		Q_OBJECT

		Q_FLAGS( GuiHint )
		/**
		 * @brief Hints for the GUI to display the parameter the right way.
		 *
		 * Several hints can be given via this property.
		 *
		 * Read with guiHint(), set with setGuiHint().
		 */
		Q_PROPERTY( GuiHint guiHint READ guiHint WRITE setGuiHint )
		
		/**
		 * @brief The value of this parameter
		 *
		 * Probably the most important property...
		 */
		Q_PROPERTY( QVariant value READ value WRITE setValue RESET reset )
		/**
		 * @brief The minimum value.
		 */
		Q_PROPERTY( QVariant minimum READ min WRITE setMin )
		/**
		 * @brief The maximum value.
		 */
		Q_PROPERTY( QVariant maximum READ max WRITE setMax )

		/**
		 * @brief The type as QString
		 */
		Q_PROPERTY( QString typeName READ typeName )

		/**
		 * @brief The filter for Urls
		 */
		Q_PROPERTY( QString urlFilter READ urlFilter WRITE setUrlFilter )

		public:
			/**
			 * @brief Hint-types for the GUI
			 */
			enum GuiHint {
				/** This parameter should be read only. For example for control-messages. */
				ReadAllowed = 1,
				/** This parameter is write-only... */
				WriteAllowed = 2,
				/** The parameter is a normal read-and-write-parameter. */
				ReadWrite = 3,
				/** The gui should provide a reset-button. */
				ProvideReset = 4,
				/** This parameter is a counter. It is read-only and should be resetted. */
				Counter = 5,
				/** Url: Get a filename for read. */
				UrlRead = 1,
				/** Url: Get a filename for write. */
				UrlWrite = 2,
				/** Url: Get a directory name. */
				UrlDir = 8
			};

			/**
			 * @brief Creates a new named parameter with value v
			 *
			 * @param n the name of this parameter
			 * @param v the value of this parameter, the type is determined
			 * from this
			 * @param p the parent TiParameter
			 */
			TiParameter( QString n, QVariant v, QTreeItem* p );
			/**
			 * @brief Creates a new named parameter with type t
			 *
			 * @param n the name of this parameter
			 * @param t the type of this parameter
			 * @param p the parent TiParameter
			 */
			TiParameter( QString n, QVariant::Type t, QTreeItem* p );

			/**
			 * @brief Copy a TiParameter and possibly its childs
			 * into another tree
			 *
			 * Childs are only copied if they are also TiParameters...
			 *
			 * @param item The item that is to be copied
			 * @param p The parent of the new item
			 */
			TiParameter( TiParameter& item, QTreeItem* p );

			/// destructor
			~TiParameter();

			/// Returns the type.
			QVariant::Type type() const { return _value.type(); }
			/// Returns the type as string.
			QString typeName() const { return QVariant::typeToName( _value.type() ); }

			/// Returns the value.
			QVariant value() const { return _value; }

			/// Returns the minimum allowed.
			QVariant min() const { return _min; }
			/// Returns the maximum allowed.
			QVariant max() const { return _max; }

			/// Returns the guiHint
			int guiHint() const { return _guihint; }
			/// Set the guiHint
			void setGuiHint( int n );

			/// Converts this to a QVariant
			operator QVariant() { return value(); }

			/**
			 * @{
			 *
			 * @brief Simplify the access
			 */
			int toInt() const { return _value.toInt(); }
			bool toBool() const { return _value.toBool(); }
			double toDouble() const { return _value.toDouble(); }
			QString toString() const {
				if ( _value.type() == QVariant::Url )
					return _value.toUrl().toString();
				return _value.toString();
			}
			QUrl toUrl() const { return _value.toUrl(); }
			// @}

			/**
			 * @brief access child parameters
			 *
			 * This function returns Parameters by path ( if they exist ).
			 *
			 * @param path The path to the parameter
			 */
			TiParameter* getParameter( QString path );

			/**
			 * @brief adds a new Parameter as child
			 *
			 * @param path The path to the new item.
			 * @param value The initial value of the new item.
			 * @param tooltip A string to be shown as tooltip in GUIs.
			 */
			void addParameter( QString path, QVariant value, QString tooltip=QString() );
			/**
			 * @brief adds a new Parameter as child
			 *
			 * @param path The path to the new item.
			 * @param type The initial type of the new item.
			 * @param tooltip A string to be shown as tooltip in GUIs.
			 */
			void addParameter( QString path, QVariant::Type type, QString tooltip=QString() );

			QVariant guiValue( int ) const { return value(); }
			void guiSetValue( QVariant v ) { setValue( v ); }
			Qt::ItemFlags guiFlags() const;
			QVariant guiIdentifier( int role ) const;

			void setUrlFilter( const QString& filter ) { _urlfilter = filter; }
			QString urlFilter() const { return _urlfilter; }

		public slots:
			/**
			 * @brief Set the value
			 *
			 * Sets the value of the option.
			 * If the type doesn't match, nothing is done.
			 * If the value exceeds min/max, the respective value is set.
			 *
			 * This slot is protected against recursions with signals/slots.
			 */
			void setValue( QVariant );
			/**
			 * @{
			 *
			 * @brief Set the value
			 *
			 * This function is for convenience and behaves like setValue(QVariant)
			 */
			void setValue( int );
			void setValue( bool );
			void setValue( double );
			void setValue( QString );
			void setValue( QUrl );
			// @}

			/// Sets the minimum allowed.
			void setMin( QVariant );
			/// Sets the maximum allowed.
			void setMax( QVariant );
			/// Sets min and max.
			void setRange( QVariant min, QVariant max ) { setMin( min ); setMax( max ); }

			/**
			 * @brief Reset the value to the default
			 */
			void reset();

		private slots:
			void unlockUpdate();

		private:
			QString debug() const {
				QString tmp = QString( "(%1,%2)" ).arg( toString() ).arg( QMetaType::typeName( _value.userType() ) );
				if ( tmp == "(,)" )
					tmp="";
				return QString( "%1%2" ).arg( objectName() ).arg( tmp );
			}

			bool needsToSaveToXml();
			QDomNode saveToXML( QDomDocument );
			bool restoreFromXML( QDomNode );

			/**
			 * @brief masked copy-constructor
			 *
			 * has to be private since the new parent needs to be known
			 */
			TiParameter( const TiParameter& )
				: QTreeItem( "", 0 )
				{}

			QVariant _value;
			QVariant _min;
			QVariant _max;
			bool _inupdate;
			int _guihint;
			QString _urlfilter;
	};


	/**
	 * @brief A protected pointer to a TiParameter
	 *
	 * @note If a class lets you acces the configuration via
	 * TiParameter or TiParameterPointer, you should
	 * <b>never</b> delete one of the childs as this will result in very strange
	 * error!
	 */
	class TiParameterPointer : public QPointer<TiParameter>
	{
		public:
			/// Construct a TiParameterPointer from a pointer to a TiParameter
			TiParameterPointer( TiParameter* p ) : QPointer<TiParameter>( p )
			{ }

			/**
			 * @brief subscription operator
			 *
			 * This gives you access to the named childs like an array.
			 */
			TiParameter* operator[]( QString path ) {
				if ( data() && dynamic_cast<TiParameter*>( data() ) )
					return dynamic_cast<TiParameter*>( data() )->getParameter( path );
				return 0;
			}

	};
};

#endif // TIPARAMETER_H

