/*
    Copyright 2009-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/

#include "abstractfilereader.h"
#include "abstractfilereader.moc"

#include "qtiocompressor.h"

#include <tiparameter.h>
#include <tiaction.h>

using namespace NSDiskIO;

AbstractFileReader::AbstractFileReader( QString n, QObject* p )
    : NSCore::DataSetProducer( n, p )
    , _config( new NSCore::TiParameter( "Parameters", QVariant::Invalid, this ) )
    , _counter( 0 )
    , _file( 0 )
    , _compressor( 0 )
    , _timer( new QTimer( this ) )
{
    qDebug() << "AbstractFileReader::AbstractFileReader(" << n << "," << p << ")";
    new NSCore::TiAction( "Read One Datasets", this, ":/icons/kde/cr22-action-player_play.png", this, SLOT( readDataSet() ) );
    new NSCore::TiAction( "Read Ten Datasets", this, ":/icons/kde/cr22-action-player_fwd.png", this, SLOT( readTenDataSets() ) );
    new NSCore::TiAction( "Read All Datasets", this, ":/icons/kde/cr22-action-player_end.png", this, SLOT( readAllDataSets() ) );
    new NSCore::TiAction( "Stop reading", this, ":/icons/kde/cr22-action-player_stop.png", this, SLOT( stopReading() ) );
    new NSCore::TiAction( "Reset file to the beginning", this, ":/icons/kde/cr22-action-player_start.png", this, SLOT( resetFile() ) );

    _config->addParameter( "File to read from", QVariant::Url );
    _config->getParameter( "File to read from" )->setGuiHint( NSCore::TiParameter::UrlRead );
    connect( _config, SIGNAL( guiValueChanged() ), this, SLOT( optionsChanged() ) );

    _timer->setInterval( 0 );
    connect( _timer, SIGNAL( timeout() ), this, SLOT( workOnCounter() ) );
    _timer->setSingleShot( true );
}
AbstractFileReader::~AbstractFileReader() {
    qDebug() << "AbstractFileReader::~AbstractFileReader()";
    delete _file;
}

QDomNode AbstractFileReader::saveToXML( QDomDocument doc ) {
    return doc.createElement( "reader" );
}
bool AbstractFileReader::restoreFromXML( QDomNode node ) {
    return node.nodeName() == "reader";
}

void AbstractFileReader::workOnCounter() {
    //qDebug() << "AbstractFileReader::workOnCounter() _counter =" << _counter << " _file =" << _file;

    if ( _file && _counter != 0 ) {

        for ( uint c=256*4; c>0 && _counter != 0; c-- ) {
            NSCore::AcqisitionDataSet data = readSingleDataSet();
            if ( data.channels() == 0 && data.samples() == 0 ) {
                readNDataSets( 0 );
            } else enqueue( data );
            float percent = _file->pos() * 1.0 / _file->size();
            emit progress(percent);
            emit progressPercent(percent*100);
            emit progressPromille(percent*1000);
            QMutexLocker lock( &_counter_lock );
            if ( _counter > 0 ) _counter -= 1;
        }

        _timer->start();
    }
}
void AbstractFileReader::optionsChanged() {
    QString file = _config->getParameter( "File to read from" )->toString();
    if ( !_file || file != _file->fileName() ) {
        openFile( file );
    }
    configChanged();
}

void AbstractFileReader::readNDataSets( int n ) {
    qDebug() << "AbstractFileReader::readNDataSets(" << n << ")";
    QMutexLocker lock( &_counter_lock );
    if ( n > 0 && _counter > -1 ) {
        _counter += n;
    }
    if ( n <= 0 ) {
        _counter = n;
    }
    if ( _counter != 0 )
        _timer->start();
    else
        emit finished();
}
void AbstractFileReader::openFile( QString filename ) {
    qDebug() << "AbstractFileReader::openFile(" << filename << ")";
    if (_compressor) {
        fileAboutToBeClosed( _compressor );
        delete _compressor;
        _compressor = 0;
    } else
        fileAboutToBeClosed( _file );
    delete _file;
    _file = 0;
    emit progress(0.0);
    emit progressPercent(0);
    emit progressPromille(0);
    if ( QFile::exists( filename ) ) {
        _file = new QFile( filename );
        if ( _file->open( QIODevice::ReadOnly ) ) {
            qDebug() << " open successful";
            if ( filename.endsWith( ".gz" ) ) {
                _compressor = new QtIOCompressor( _file );
                _compressor->setStreamFormat( QtIOCompressor::GzipFormat );
                _compressor->open( QIODevice::ReadOnly );
                //qDebug() << " opened compressed file";
                fileOpened( _compressor );
            } else {
                if ( filename.endsWith( ".z" ) ) {
                    _compressor = new QtIOCompressor( _file );
                    _compressor->setStreamFormat( QtIOCompressor::ZlibFormat );
                    _compressor->open( QIODevice::ReadOnly );
                    //qDebug() << " opened compressed file";
                    fileOpened( _compressor );
                } else
                    fileOpened( _file );
            }
            //qDebug() << "  ready.";
        } else {
            qWarning() << "Something went wrong...";
            delete _file;
            _file = 0;
        }
    }
}
void AbstractFileReader::resetFile() {
    if ( _file ) {
        QString filename = _file->fileName();
        openFile( "" );
        openFile( filename );
    }
}

QStringList AbstractFileReader::compressions() const {
    QStringList ret;
    ret << "z";
    if (QtIOCompressor::isGzipSupported())
        ret << "gz";
    return ret;
}

