/*
    Copyright 2009-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef ABSTRACTFILEREADER_H
#define ABSTRACTFILEREADER_H

#include <datasetchain.h>

namespace NSCore
{
    class TiParameter;
} // namespace NSCore

class QTimer;
class QtIOCompressor;

/**
 * @brief Disk-IO-things should be in this namespace
 *
 * @todo Tutorial on how to write readers and writers
 */
namespace NSDiskIO
{

    /**
     * @brief Abstract base class for reading data from a file
     *
     * This class provides the basic stuff to read AcqisitionDataSet from files.
     */
    class AbstractFileReader : public NSCore::DataSetProducer
    {
        Q_OBJECT
        public:
            /// destructor
            ~AbstractFileReader();

            void start() { readAllDataSets(); }
            void stop() { stopReading(); }
            void abort() { stop(); }
            void reset() { resetFile(); }

            bool isRunning() { return (_counter!=0); }

        public slots:
            /**
             * @brief Schedule a number of dataset for reading
             *
             * This adds the given number to the internal number of datasets scheduled for reading.
             *
             * There are two exceptions:
             * <ul>
             * <li>"Adding" 0 makes the reading stop by setting the counter to 0.
             * <li>"Adding" -1 makes the reader read till the end of the file.
             * </ul>
             */
            void readNDataSets( int );

            /**
             * @{
             * @brief Convinience functions to readNDataSets
             */
            void readDataSet() { readNDataSets( 1 ); }
            void readTenDataSets() { readNDataSets( 10 ); }
            void readAllDataSets() { readNDataSets( -1 ); }

            void stopReading() { readNDataSets( 0 ); }
            // @}

            /**
             * @brief Open a new file
             *
             * This closes the old file and opens the new given file.
             * If the filepath is empty, it only closes the old file.
             */
            void openFile( QString filepath );

            /**
             * @brief Close and reopen the current file.
             *
             * This resets the file stream back to the beginning by closing the file and reopening it.
             */
            void resetFile();

        signals:
            /**
             * @brief Report the current position in the file as progress.
             *
             * Range is from 0.0 to 1.0.
             */
            void progress(double);
            /**
             * @brief Report the current position in the file as progress.
             *
             * Range is percentage from 0 to 100.
             */
            void progressPercent(int);
            /**
             * @brief Report the current position in the file as progress.
             *
             * Range is in tenth of percentage from 0 to 1000.
             */
            void progressPromille(int);

        protected:
            /**
             * @brief Create a new reader
             *
             * This is protected additionally to the pure virtual functions so you don't try to
             * instantiate an abstract reader directly.
             */
            AbstractFileReader( QString, QObject* );

            /**
             * @brief The real work
             *
             * In this function you have to do all the real work to produce an AcqisitionDataSet
             * from the data in the file.
             *
             * If the end of the file is reached, return a NSCore::AcqisitionDataSet( 0,0 ), an
             * empty dataset, and the reading will stop.
             */
            virtual NSCore::AcqisitionDataSet readSingleDataSet() =0;

            /**
             * @brief A new file was opened
             *
             * A new file was opened for reading, do what needs to be done to successfully read
             * datasets from the file
             */
            virtual void fileOpened( QIODevice* ) =0;
            /**
             * @brief The current file is about to be closed
             *
             * Do what is needed before the file-handle is closed.
             */
            virtual void fileAboutToBeClosed( QIODevice* ) =0;

            /**
             * @brief Pointer to the configuration items
             */
            NSCore::TiParameter* _config;

            /** @{
             * @brief xml-handling
             */
            QDomNode saveToXML( QDomDocument );
            bool restoreFromXML( QDomNode );
            // @}

            /**
              @brief return the endings for supported compressions
              */
            QStringList compressions() const;

            /**
              @brief The configuration has changed.
              */
            virtual void configChanged() {}

        private slots:
            void workOnCounter();
            void optionsChanged();

        private:
            QMutex _counter_lock;
            int _counter;

            QFile* _file;
            QtIOCompressor* _compressor;

            QTimer* _timer;
    };

} // namespace NSDiskIO

#endif // ABSTRACTFILEREADER_H
