/*
    Copyright 2009-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/

#include "abstractfilewriter.h"
#include "abstractfilewriter.moc"

#include "qtiocompressor.h"

#include <tiparameter.h>

#include <QtCore/QDebug>

using namespace NSDiskIO;

AbstractFileWriter::AbstractFileWriter( QString n, QObject* p )
    : NSCore::DataSetChainMember( n, p )
    , _config( new NSCore::TiParameter( "options", QVariant::Invalid, this ) )
    , _file( 0 )
    , _filemutex( new QMutex )
    , _counter( new NSCore::TiParameter( "Written datasets", QVariant::Int, this ) )
    , _compressor( 0 )
    , _finishemitted( false )
{
    qDebug() << "AbstractFileWriter::AbstractFileWriter(" << n << "," << p << ")";
    _config->addParameter( "File to write to", QVariant::Url );
    _config->getParameter( "File to write to" )->setGuiHint( NSCore::TiParameter::UrlWrite );
    _config->getParameter( "File to write to" )->setToolTip( "If the filename contains MTIME or DATE these will be replaced with the yyyymmdd-hhmmss of the time the file is created." );
    _config->addParameter( "Send stop after N events", QVariant::Int );
    _stopnumber = _config->getParameter( "Send stop after N events" );
    _stopnumber->setValue( -1 );
    _stopnumber->setToolTip( "Signal the connected\
            producers to stop 'measurement' after the given number of events is written to\
            the file. If the value is below one, no signal is sent." );
    connect( _config, SIGNAL( guiValueChanged() ), this, SLOT( optionsChanged() ) );
    _counter->setGuiHint( NSCore::TiParameter::Counter );
}
AbstractFileWriter::~AbstractFileWriter() {
    qDebug() << "AbstractFileWriter::~AbstractFileWriter()";
    _filemutex->lock();
    delete _compressor;
    delete _file;
    delete _filemutex;
}

QDomNode AbstractFileWriter::saveToXML( QDomDocument doc ) {
    return doc.createElement( "writer" );
}
bool AbstractFileWriter::restoreFromXML( QDomNode node ) {
    if ( node.isElement() && node.nodeName() == "writer" ) return true;
    return false;
}

void AbstractFileWriter::optionsChanged() {
    //qDebug() << "AbstractFileWriter::optionsChanged()";
    QString filename = _config->getParameter( "File to write to" )->toString();
    if ( !_file || (filename != _file->fileName()) )
        filenameChanged( filename );
    configChanged();
}

void AbstractFileWriter::filenameChanged( QString filename ) {
    qDebug() << "AbstractFileWriter::filenameChanged(" << filename << ")";
    if ( _file ) {
        QMutexLocker lock(_filemutex);
        if ( _compressor ) {
            fileAboutToBeClosed( _compressor );
            delete _compressor;
            _compressor = 0;
        } else
            fileAboutToBeClosed( _file );
        delete _file;
        _file = 0;
        _counter->reset();
        _finishemitted = false;
    }
    if ( filename != "" ) {
        QMutexLocker lock(_filemutex);
        QString currentdate = QDateTime::currentDateTime().toString( "yyyyMMdd-hhmmss" );
        _file = new QFile( filename.replace("MTIME", currentdate).replace("DATE", currentdate) );
        if ( _file->open( QIODevice::WriteOnly ) ) {
            qDebug() << " Opened new file";
            if ( filename.endsWith( ".gz" ) ) {
                _compressor = new QtIOCompressor( _file, 3 );
                _compressor->setStreamFormat( QtIOCompressor::GzipFormat );
                _compressor->open( QIODevice::WriteOnly );
                fileOpened( _compressor );
            } else {
                if ( filename.endsWith( ".z" ) ) {
                    _compressor = new QtIOCompressor( _file );
                    _compressor->setStreamFormat( QtIOCompressor::ZlibFormat );
                    _compressor->open( QIODevice::WriteOnly );
                    fileOpened( _compressor );
                } else
                    fileOpened( _file );
            }
        } else {
            delete _file;
            _file = 0;
        }
    }
}

bool AbstractFileWriter::dataArrived( NSCore::AcqisitionDataSet data ) {
    //qDebug() << "AbstractFileWriter::dataArrived(" << data.samples() << ")";
    QMutexLocker lock(_filemutex);
    if ( _file != 0 ) {
        if ( _stopnumber->toInt() > 0
                && _counter->toInt() >= _stopnumber->toInt()
                && !_finishemitted ) {
            _finishemitted = true;
            emit finished();
        }
        if ( (_stopnumber->toInt() < 1 ) || (_stopnumber->toInt() > _counter->toInt() ) ) {
            //qDebug() << " will write the data to disk";
            writeSingleDataset( data );
            _counter->setValue( _counter->toInt() + 1 );
        }
    }
    return true;
}

QStringList AbstractFileWriter::compressions() const {
    QStringList ret;
    ret << "z";
    if (QtIOCompressor::isGzipSupported())
        ret << "gz";
    return ret;
}

