/*
    Copyright 2009-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ABSTRACTFILEWRITER_H
#define ABSTRACTFILEWRITER_H

#include <datasetchain.h>

class QtIOCompressor;

namespace NSCore
{
    class TiParameter;
} // namespace NSCore

namespace NSDiskIO {

    /**
     * @brief Abstract base class to write datasets to a file
     *
     * This class provides the basic framework to write AcqisitionDataSet to files
     */
    class AbstractFileWriter : public NSCore::DataSetChainMember
    {
        Q_OBJECT
        public:
            /// destructor
            ~AbstractFileWriter();

        protected:

            /**
             * @brief Create a new writer.
             *
             * This is protected additionally to the pure virtual functions so you don't try to
             * instantiate an abstract writer directly.
             */
            AbstractFileWriter( QString, QObject* );

            /**
             * @brief the parent for options
             */
            NSCore::TiParameter* _config;

            /**
             * @brief Write one single dataset to the file
             */
            virtual void writeSingleDataset( NSCore::AcqisitionDataSet& ) =0;

            /**
             * @brief A new file was opened
             *
             * Do what you need to write datasets to it.
             */
            virtual void fileOpened( QIODevice* ) =0;
            /**
             * @brief The given file is about to be closed
             *
             * Do what is needed to be done before the closing.
             */
            virtual void fileAboutToBeClosed( QIODevice* ) =0;

            /**
             * @brief The configuration changed.
             *
             * Implement this to get noticed when the configuration changed.
             * This saves you from implementing your own slots.
             */
            virtual void configChanged() {}

            /**
              @brief return the endings for supported compressions
              */
            QStringList compressions() const;

            /**
             * @{
             * @brief Implement parent functions.
             */
            QDomNode saveToXML( QDomDocument );
            bool restoreFromXML( QDomNode );

            bool dataArrived( NSCore::AcqisitionDataSet );
            // @}

        private slots:
            void filenameChanged( QString );
            void optionsChanged();

        private:
            /// the file object to us
            QFile* _file;
            QMutex* _filemutex;

            /// the counter for the written datasets
            NSCore::TiParameter* _counter;

            NSCore::TiParameter* _stopnumber;

            QtIOCompressor* _compressor;

            bool _finishemitted;
    };

} // namespace NSDiskIO
#endif // ABSTRACTFILEWRITER_H
