/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/

#include "csvforprinting.h"

#include <mathexpression.h>
#include <tiparameter.h>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_interp.h>

namespace NSDiskIO {

    CSVWriter::CSVWriter( QObject* p ) : AbstractFileWriter( "CSVWriter", p ) {
        _config->getParameter( "File to write to" )->setUrlFilter( "*.csv" );
        _timealign = new NSCore::MathTreeRoot( "Align data to", this );
        QDomDocument doc("xml");
        doc.setContent( QString( "<path name=\"Align data to\"><add><result name=\"horizontalposition\" /><result name=\"triggerdelay\" /></add></path>" ) );
        _timealign->fromXml( doc.documentElement() );
        _timeinterval = new NSCore::MathTreeRoot( "Time interval to use", this );
        doc.setContent( QString( "<path name=\"Time interval to use\"><result name=\"samplinginterval\" /></path>" ) );
        _timeinterval->fromXml( doc.documentElement() );
        _config->addParameter( "Generations to save", "0" );
        _config->getParameter( "Generations to save" )->setToolTip( "Comma-separated list of generations to save to the file. Data format will be: time, <list of all channels for each given generation>" );
        //debugPrint();
    }
    void CSVWriter::fileOpened( QIODevice* f ) {
        _stream.setDevice( f );
    }
    void CSVWriter::fileAboutToBeClosed( QIODevice* ) {
        _stream.setDevice( 0 );
    }
    void CSVWriter::writeSingleDataset( NSCore::AcqisitionDataSet& data ) {
        QList<int> generations;
        foreach( QString s, _config->getParameter( "Generations to save" )->toString().split( ",", QString::SkipEmptyParts ) )
            generations.append( s.toInt() );

        double samplinginterval = _timeinterval->value( data );
        if ( isnan( samplinginterval ) || samplinginterval == 0 )
            samplinginterval = 1;

        double horizontalposition = _timealign->value( data );
        if ( isnan( horizontalposition ) )
            horizontalposition = 0;

        for ( int i=0; i<data.samples(); ++i ) {
            _stream << horizontalposition + i*samplinginterval;
            foreach( int gen, generations )
                for ( int j=0; j<data.channels( gen ); ++j ) {
                    //double* ch = data.getRawChannel( j, gen );
                    if ( i < data.samples( gen ) )
                        _stream << ", " << data.getRawChannel( j, gen )[ i ];
                    else _stream << ", " << 0;
                }
            _stream << endl;
        }

        _stream << endl << endl;
    }



    CombinedCSV::CombinedCSV( QObject* p )
        : AbstractFileWriter( "CombinedCSV", p )
        , _combine_counter( 0 )
    {
        _config->getParameter( "File to write to" )->setUrlFilter( "*.csv" );
        _config->addParameter( "Datasets to average over", 1 );
        _config->getParameter( "Datasets to average over" )->setRange( 1, int( 1E8 ) );
        _config->addParameter( "Generation", QVariant::Int );
    }

    void CombinedCSV::fileOpened( QIODevice* f ) {
        _stream.setDevice( f );
    }
    void CombinedCSV::fileAboutToBeClosed( QIODevice* ) {
        _stream.setDevice( 0 );
    }

    void CombinedCSV::writeSingleDataset( NSCore::AcqisitionDataSet& data ) {
        int generation = _config->getParameter( "Generation" )->toInt();

        int samples = data.samples( generation );
        for ( int c=0; c<data.channels( generation ); ++c ) {
            double* ch = data.getRawChannel( c, generation );
            while ( _save[ c ].size() < samples )
                _save[ c ].append( 0.0 );
            for ( int i=0; i<samples; ++i ) {
                _save[ c ][ i ] += ch[ i ];
            }
        }
        _combine_counter++;

        if ( _combine_counter >= _config->getParameter( "Datasets to average over" )->toInt() ) {
            qDebug() << " Will save" << _combine_counter << "averaged datasets to disk.";

            for ( int i=0; i<samples; ++i ) {

                _stream << i;
                foreach( int c, _save.keys() ) {
                    _stream << " " << _save[ c ][ i ] / _combine_counter;
                }
                _stream << endl;
            }
            _stream << endl << endl;

            _save.clear();
            _combine_counter = 0;
        }

    }



    MergeCSV::MergeCSV( QObject* p ) : AbstractFileWriter( "MergeCSV", p ) {
        cache.samples = 0;
        cache.xa = 0;
        cache.interp = 0;
        cache.firstposition = NAN;
        cache.averagecounter = 0;
        cache.acc = gsl_interp_accel_alloc();
        _config->getParameter( "File to write to" )->setUrlFilter( "*.csv" );
        _config->addParameter( "Oversampling factor", 4 );
        _config->getParameter( "Oversampling factor" )->setRange( 1, int(1E6) );
        _config->addParameter( "Datasets to average over", 10 );
        _config->getParameter( "Datasets to average over" )->setRange( -1, int( 1E6 ) );
        _config->addParameter( "Write all intermediate averages", false );
        _timealign = new NSCore::MathTreeRoot( "Align data to", this );
    }
    MergeCSV::~MergeCSV() {
        fileAboutToBeClosed( _stream.device() );
        delete cache.xa;
        gsl_interp_free( cache.interp );
        gsl_interp_accel_free( cache.acc );
        while ( cache.channeldata.size() > 0 )
            delete[] cache.channeldata.takeFirst();
    }

    void MergeCSV::fileOpened( QIODevice* f ) {
        _stream.setDevice(f);
        cache.firstposition = NAN;
    }
    void MergeCSV::fileAboutToBeClosed( QIODevice* ) {
        dumpChannelDataToStream();
        _stream.flush();
        _stream.setDevice( 0 );
    }

    void MergeCSV:: configChanged() {
        _overfactor = _config->getParameter( "Oversampling factor" )->toInt();
        _intermediateaverages = _config->getParameter( "Write all intermediate averages" )->toBool();
        _average = _config->getParameter( "Datasets to average over" )->toInt();
    }

    void MergeCSV::writeSingleDataset( NSCore::AcqisitionDataSet& data ) {
        int samples = data.samples();
        int channels = data.channels();

        if ( cache.samples != samples ) {
            delete[] cache.xa;
            cache.xa = new double[samples];
            for ( int j=0; j<samples; ++j )
                cache.xa[j] = j;
            gsl_interp_free( cache.interp );
            cache.interp = gsl_interp_alloc( gsl_interp_cspline, samples );
            while ( cache.channeldata.size() > 0 )
                delete[] cache.channeldata.takeFirst();

            cache.samples = samples;
        }
        while ( cache.channeldata.size() < channels ) {
            double* tmp = new double[samples*_overfactor];
            for ( int i=0; i<samples*_overfactor; ++i)
                tmp[i] = 0.0;
            cache.channeldata.append( tmp );
        }

        if ( isnan(cache.firstposition) ) {
            cache.firstposition = _timealign->value( data );
            if ( isnan(cache.firstposition) )
                cache.firstposition = 0;
        }
        double position = _timealign->value( data );
        if ( isnan(position) ) position = 0;
        double toffset = position - cache.firstposition;
        //qDebug() << "MergeCSV::writeSingleDataset() firstpos =" << cache.firstposition
        //    << " my position =" << position
        //    << " toffset = " << toffset;
        for ( int i=0; i<channels; ++i ) {
            double *ch = data.getRawChannel( i );
            gsl_interp_accel_reset( cache.acc );
            gsl_interp_init( cache.interp, cache.xa, ch, samples );
            for ( int j=abs(toffset); j<samples-abs(toffset); ++j ) {
                for ( int m=0; m<_overfactor; ++m ) {
                    double y = j+m*1.0/_overfactor + toffset;
                    cache.channeldata[i][j*_overfactor+m] += gsl_interp_eval( cache.interp, cache.xa, ch, y, cache.acc );
                }
            }
        }

        cache.averagecounter++;

        if ( _intermediateaverages || (_average > 0 && cache.averagecounter >= _average) ) {
            dumpChannelDataToStream();
        }
        if ( _average > 0 && cache.averagecounter >= _average ) {
            while ( cache.channeldata.size() > 0 ) {
                delete[] cache.channeldata.takeFirst();
            }
            cache.averagecounter = 0;
        }
    }
    void MergeCSV::dumpChannelDataToStream() {
        if (cache.averagecounter == 0)
            return;
        for ( int j=0; j<cache.samples*_overfactor; ++j ) {
            _stream << j*1.0/_overfactor - cache.firstposition;
            for ( int i=0; i<cache.channeldata.size(); ++i ) {
                _stream << " " << cache.channeldata[i][j]/cache.averagecounter;
            }
            _stream << endl;
        }
        _stream << endl << endl;
    }

}


// vim: et
