/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef CSVFORPRINTING_H
#define CSVFORPRINTING_H

#include "abstractfilewriter.h"
#include <gsl/gsl_interp.h>

namespace NSCore {
    class MathTreeRoot;
};

//typedef struct gsl_interp gsl_interp;

namespace NSDiskIO {

    /**
     * @brief write CSV-files for gnuplot
     *
     * This writes CSV files which gives the posibility to load and show the
     * waveforms in gnuplot.
     *
     * Columns are: sample, ch1, ch2,... <br>
     * The blocks are separated by two newlines, just as gnuplot likes it.
     */
    class CSVWriter : public AbstractFileWriter {
        public:
            CSVWriter( QObject* =0 );

            void writeSingleDataset( NSCore::AcqisitionDataSet& data );

            void fileOpened( QIODevice* );
            void fileAboutToBeClosed( QIODevice* );

            QString debug() const { return "CSVWriter"; }

        private:
            NSCore::MathTreeRoot *_timealign;
            NSCore::MathTreeRoot *_timeinterval;

            QTextStream _stream;
    };

    /**
     * @brief Combined CSV-Files for Origin, etc.
     *
     * Seems as if this one is superseeded by the MergeCSV-writer...
     */
    class CombinedCSV : public AbstractFileWriter {
        public:
            CombinedCSV( QObject* =0 );

            void writeSingleDataset( NSCore::AcqisitionDataSet& data );

            void fileOpened( QIODevice* );
            void fileAboutToBeClosed( QIODevice* );

            QString debug() const { return "CombinedCSV"; }

        private:
            QMap< int,QList<double> > _save;
            int _combine_counter;

            QTextStream _stream;
    };

    /**
     * @brief Merge pulses. Align them by time and average them with sub-sampling applied...
     */
    class MergeCSV : public AbstractFileWriter
    {
        public:
            MergeCSV( QObject* =0 );
            ~MergeCSV();

            void writeSingleDataset( NSCore::AcqisitionDataSet& data );

            void fileOpened( QIODevice* );
            void fileAboutToBeClosed( QIODevice* );

            QString debug() const { return "MergeCSV"; }

            void configChanged();

        private:
            // @{
            NSCore::MathTreeRoot* _timealign;
            /// Factor of oversampling
            int _overfactor;
            /// Write all intermediate averages to the file?
            bool _intermediateaverages;
            /// How many datasets will be averaged over
            int _average;
            // @}
            //QMap<double, double>

            struct cache_t
            {
                int samples;
                double *xa;

                gsl_interp* interp;
                gsl_interp_accel* acc;

                QList<double*> channeldata;

                int averagecounter;
                double firstposition;
            } cache;

            void dumpChannelDataToStream();

            QTextStream _stream;
    };
};

#endif // CSVFORPRINTING_H
// vim: et

