/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/

#include "eposbin.h"

#include "qtiocompressor.h"

#include <tiparameter.h>

namespace NSDiskIO {

	EPOSBinWriter::EPOSBinWriter( QObject* p )
		: AbstractFileWriter( "EPOSBinWriter", p )
		, _strip( false )
	{
		//qDebug() << "EPOSBinWriter::EPOSBinWriter(" << p << ")";
		_config->addParameter( "Strip sampled data", QVariant::Bool, "If activated only the results will be saved to disk, the samples of the data will be stripped." );
		_config->getParameter( "Strip sampled data" )->setValue( false );
		QString types = "EPOS binary file (*.eposbinfile)\n%1All files (*.*)";
		if ( QtIOCompressor::isGzipSupported() ) {
			types.arg( "Compressed EPOS binary file (*.eposbinfile.gz)\n" );
		} else {
			types.arg( "" );
		}
		_config->getParameter( "File to write to" )->setUrlFilter( types );
	}
	EPOSBinWriter::~EPOSBinWriter() {
	}

	void EPOSBinWriter::configChanged() {
		_strip = _config->getParameter( "Strip sampled data" )->toBool();
	}

	void EPOSBinWriter::fileOpened( QIODevice* f ) {
		//qDebug() << "EPOSBinWriter::fileOpened(" << f << ")";
		_stream.setDevice( f );
	}
	void EPOSBinWriter::fileAboutToBeClosed( QIODevice* ) {
		_stream.setDevice( 0 );
	}

	void EPOSBinWriter::writeSingleDataset( NSCore::AcqisitionDataSet& data ) {
		//qDebug() << "EPOSBinWriter::writeSingleDataset(" << &data << ")";
		//if ( !_file )
		//	return;

		_stream.writeRawData( "\%EPOSBIN4\%", 10 );
		_stream << data.channels()
		         << ( _strip?0:data.samples() )
		         ;
		if ( ! _strip )
			for ( int i=0; i<data.channels(); ++i ) {
				double* ch = data.getRawChannel( i );
				for ( int j=0; j<data.samples(); ++j )
					_stream << ch[ j ];
			}
		QStringList keys = data.getResultNames();
		_stream << keys.size();
		foreach ( QString k, keys ) {
			_stream << k.toUtf8() << data.getResult( k );
		}
	}


	void version3RecursiveRead( QDataStream& stream, const int childs, QString prefix, NSCore::AcqisitionDataSet& data ) {
		for ( int i=0; i<childs; ++i ) {
			QByteArray name;
			QVariant value;
			int newchilds;
			stream >> name >> value >> newchilds;
			if ( name == "data" ) name = "";
			QString n = prefix + "/" + QString::fromUtf8( name );
			data.addResult( n, value.toDouble() );
			version3RecursiveRead( stream, newchilds, n, data );
		}
	}
	EPOSBinReader::EPOSBinReader( QObject* p )
		: AbstractFileReader( "EPOSBinReader", p )
		, _headertoskip( 0 )
		, _file( 0 )
	{
		QString types = "EPOS binary file (*.eposbinfile)\n";
		foreach ( QString ending, compressions() )
			types += QString( "Compressed EPOS binary file (*.eposbinfile.%1)\n" ).arg(ending);
		types += "All files (*.*)";
		_config->getParameter( "File to read from" )->setUrlFilter( types );
	}
	EPOSBinReader::~EPOSBinReader() {
	}
	void EPOSBinReader::fileOpened( QIODevice* f ) {
		_file = f;
		_stream.setDevice( _file );
		_headertoskip = 0;
		_fileversion = 0;
		qDebug() << _file->peek( 20 );
		if ( _file != 0 && _file->peek( 1 ) == "\%" ) {
			QByteArray tmp = _file->read( 1 );
			QByteArray r;

			do {
				r = _file->read( 1 );
				tmp.append( r );
			} while ( r != "\%" );

			_headertoskip = tmp.size();

			//if ( tmp == "\%EPOSBIN2\%" )
			//	qDebug() << "EPOSBinReader::fileOpened() seems to be a version 2 file.";
			if ( tmp == "\%EPOSBIN3\%" ) {
				qDebug() << "EPOSBinReader::fileOpened() found a version 3 file.";
				_fileversion = 1;
			}
			if ( tmp == "\%EPOSBIN4\%" ) {
				qDebug() << "EPOSBinReader::fileOpened() found a version 4 file.";
				_fileversion = 2;
				_headertoskip = 0;
				_file->seek( 0 );
			}
		}
	}
	void EPOSBinReader::fileAboutToBeClosed( QIODevice* ) {
		_stream.setDevice( 0 );
		_file = 0;
	}
	NSCore::AcqisitionDataSet EPOSBinReader::readSingleDataSet() {
		//qDebug() << "EPOSBinReader::readSingleDataSet()";
		if ( _file && ! _stream.atEnd() ) {

			//qDebug() << " fileversion is" << _fileversion;

			if ( _fileversion == 0 ) {

				double samplinginterval, triggerdelay, horizontalposition, timestamp;
				int channels, samples;
				_stream >> samplinginterval >> triggerdelay >> horizontalposition >> timestamp >> channels >> samples;

				NSCore::AcqisitionDataSet data( channels, samples );
				data.addResult( "samplinginterval", samplinginterval );
				data.addResult( "triggerdelay", triggerdelay );
				data.addResult( "horizontalposition", horizontalposition );
				data.addResult( "timestamp", timestamp );

				double fullscale, offset;
				for ( int i=0; i<channels; ++i ) {
					_stream >> fullscale >> offset;
					data.addResult( QString( "channel %1/fullscale" ).arg( i ), fullscale );
					data.addResult( QString( "channel %1/offset" ).arg( i ), offset );

					double* ch = data.getRawChannel( i );
					for ( int j=0; j<samples; ++j ) {
						_stream >> ch[ j ];
					}
				}
				data.updateMinMax();

				return data;

			} else if ( _fileversion == 1 ) {

				try {
					int channels, samples;
					_stream >> channels >> samples;
					NSCore::AcqisitionDataSet data( channels, samples );
					version3RecursiveRead( _stream, 1, "", data );

					for ( int i=0; i<data.channels(); ++i ) {
						double* ch = data.getRawChannel( i );
						for ( int j=0; j<data.samples(); ++j ) {
							_stream >> ch[ j ];
						}
					}
					data.updateMinMax();
					//data.debugPrint();
					return data;
				}
				catch( std::exception& e ) {
					qDebug() << "EPOSBinReader::readSingleDataSet() caught exception" << e.what();
					readNDataSets( 0 );
				}

			} else if ( _fileversion == 2 ) {
				//qDebug() << " Reading a version 4 dataset";
				try {
					while ( _file->peek( 1 ) != "\%" ) _file->read( 1 );
					QByteArray tmp = _file->read( 10 );
					if ( tmp != "\%EPOSBIN4\%" ) qFatal( "Not a version 4 file!" );
					int channels, samples;
					_stream >> channels >> samples;
					NSCore::AcqisitionDataSet data( channels, samples );
					for ( int i=0; i<channels; ++i ) {
						double* ch = data.getRawChannel( i );
						for ( int j=0; j<samples; ++j ) {
							_stream >> ch[j];
						}
					}
					if ( _file->bytesAvailable() == 0 ) {
						throw std::exception();
					}
					int keysize;
					_stream >> keysize;
					for ( int i=0; i<keysize && _file->bytesAvailable()!=0; ++i ) {
						QByteArray name;
						double value;
						_stream >> name >> value;
						data.addResult( QString::fromUtf8( name ), value );
					}
					//data.debugPrint();
					return data;
				}
				catch( std::exception& e ) {
					qDebug() << "EPOSBinReader::readSingleDataSet() caught exception" << e.what();
					readNDataSets( 0 );
				}
			}
		}
		return NSCore::AcqisitionDataSet( 0,0 );
	}

};

