/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/

#include "pragreader.h"
#include "pragreader.moc"

#include <QtCore/QDebug>
#include <QtCore/QCoreApplication>
#include <QtCore/QBuffer>

#include <tiparameter.h>

#include "qtiocompressor.h"

namespace NSDiskIO {

    PragReader::PragReader( QObject* p )
        : AbstractFileReader( "PragReader", p )
    {
        _config->addParameter( "samples", 300 );
        _config->addParameter( "channels", 2 );
        _config->addParameter( "offset", 32 );
        configChanged();
        tdata.reserve( 2048 );
    }
    PragReader::~PragReader() {
    }

    void PragReader::configChanged() {
        _samples = _config->getParameter( "samples" )->toInt();
        _channels = _config->getParameter( "channels" )->toInt();
        _offset = _config->getParameter( "offset" )->toInt();
    }

    void PragReader::fileOpened( QIODevice* f ) {
        _stream.setDevice( f );
    }
    void PragReader::fileAboutToBeClosed( QIODevice* ) {
        _stream.setDevice( 0 );
    }

    NSCore::AcqisitionDataSet PragReader::readSingleDataSet() {
        if ( ! _stream.atEnd() ) {
            NSCore::AcqisitionDataSet data( _channels, _samples );
            //qDebug() << " _channels=" << _channels << " _samples=" << _samples;

            qint8 tmp;
            double* ch[ _channels ];
            for ( int j=0; j<_channels; ++j ) ch[ j ] = data.getRawChannel( j );
            for ( int i=0; i<_samples; ++i ) {
                for ( int j=0; j<_channels; ++j ) {
                    //qDebug() << "  i =" << i << " j =" << j;
                    _stream >> tmp;
                    ch[ j ][ i ] = ( tmp-127.0 )/256.0;
                }
            }
            for ( int i=0; i<32*_channels; ++i ) {
                _stream >> tmp;
            }

            data.updateMinMax();
            data.addResult( "samples", _samples );
            for ( int i=0; i<2; ++i ) {
                data.addResult( QString( "channel %1/fullscale" ).arg( i ), qMax( fabs( data.getResult( QString( "channel %1/max" ).arg( i ) ) ), fabs( data.getResult( QString( "channel %1/min" ).arg( i ) ) ) ) * 2 );
                data.addResult( QString( "channel %1/offset" ).arg( i ), 0.0 );
            }

            return data;
        }
        return NSCore::AcqisitionDataSet( 0,0 );
    }


    PragDopplerReader::PragDopplerReader( QObject* p )
        : AbstractFileReader( "Prag-Doppler Reader", p )
    {
    }
    PragDopplerReader::~PragDopplerReader() {
    }

    void PragDopplerReader::fileOpened( QIODevice* device ) {
        _stream.setDevice( device );
        _stream.setByteOrder( QDataStream::LittleEndian );
        _stream >> _samples;
        //qDebug() << "PragDopplerReader::fileOpened() _samples =" << _samples;
        _channels = 2;
        /* Throw away the second and third number, these just define the total number of datasets in
         * the file. Useless, we just read till the end:-) */
        qint32 tmp;
        _stream >> tmp;
        _stream >> tmp;
    }
    void PragDopplerReader::fileAboutToBeClosed( QIODevice* device ) {
    }
    NSCore::AcqisitionDataSet PragDopplerReader::readSingleDataSet() {
        int max = pow( 2, 16 );
        if ( ! _stream.atEnd() ) {
            NSCore::AcqisitionDataSet data( _channels, _samples );

            double* ch = data.getRawData();
            for ( int i=0; i<_samples; ++i ) {
                qint16 in;
                for ( int j=0; j<_channels; ++j ) {
                    _stream >> in;
                    ch[ j*_samples + i ] = in * 1.0 / max;
                }
            }

            data.updateMinMax();
            data.addResult( "samples", _samples );
            for ( int i=0; i<2; ++i ) {
                data.addResult( QString( "channel %1/fullscale" ).arg( i ), qMax( fabs( data.getResult( QString( "channel %1/max" ).arg( i ) ) ), fabs( data.getResult( QString( "channel %1/min" ).arg( i ) ) ) ) * 2 );
                data.addResult( QString( "channel %1/offset" ).arg( i ), 0.0 );
            }
            //data.debugPrint();

            return data;
        }
        return NSCore::AcqisitionDataSet();
    }

};

