/*
    Copyright 2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef DOPPLERGSL_H
#define DOPPLERGSL_H

#include <plugins_interface.h>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_fit.h>
#include <gsl/gsl_multimin.h>

class GSLExponentialFit : public NSEvaluation::Analyzer
{
	private:
		double* x;
		double* y;
		int samples;

	public:
		GSLExponentialFit( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ), x( 0 ), y( 0 ), samples( 0 ) {
		}
		~GSLExponentialFit() {
			delete[] x;
			delete[] y;
		}

		void analyze( NSCore::AcqisitionDataSet& data ) {
			if ( samples != data.samples() ) {
				delete[] x;
				delete[] y;
				samples = data.samples();
				x = new double[ samples ];
				for ( int j=0; j<samples; ++j )
					x[ j ] = j;
				y = new double[ samples ];
			}

			for ( int i=0; i<data.channels(); ++i ) {
				int max_at = data.getResult( QString( "channel %1/max at" ).arg( i ) );

				if ( max_at + 130 < samples ) {
					double* ch = data.getRawChannel( i );
					double mean = 0;
					for ( int j=0; j<20; ++j )
						mean += ch[ j ];
					mean /= 20;

					for ( int j=0; j<data.samples(); ++j ) {
						y[ j ] = qMin( 0., log( ch[ j ] - mean ) );
					}

					double c0, c1, cov00, cov01, cov11, chisq;
					gsl_fit_linear( &( x[ max_at+30 ] ), 1, &( y[ max_at+30 ] ), 1, samples-30-max_at, &c0, &c1, &cov00, &cov01, &cov11, &chisq );
					double energy, energy_error;
					gsl_fit_linear_est( -30, c0, c1, cov00, cov01, cov11, &energy, &energy_error );
					data.addResult( QString( "channel %1/Zeroline" ).arg( i ), mean );
					data.addResult( QString( "channel %1/Energy" ).arg( i ), exp( energy ) );
					data.addResult( QString( "channel %1/Energy Error" ).arg( i ), energy_error );
					data.addResult( QString( "channel %1/ChiSq" ).arg( i ), chisq );
					data.addResult( QString( "channel %1/Time Constant" ).arg( i ), c1 );
				}

			}
		}

		QString type() const { return "GSLExponentialFit"; }
};


class GSLMultiExpFit : public NSEvaluation::Analyzer
{
	private:
		double *x, *y;
		int samples;

		gsl_multimin_fminimizer* minimizer;

		struct parameters {
			int samples;
			double* ch;
		};

		static double myfunction( const gsl_vector* v, void* params ) {
			parameters* p = ( parameters* )params;
			double bg = gsl_vector_get( v, 0 );
			double A = gsl_vector_get( v, 1 );
			double t1 = gsl_vector_get( v, 2 );
			double tau1 = gsl_vector_get( v, 3 );

			double result = 0;
			for ( int i=0; i<p->samples; ++i ) {
				double tmp = bg;
				if ( i >= t1 )
					tmp += A * exp( - ( i-t1 )/tau1 );
				result += fabs( p->ch[ i ] - tmp );
			}
			return result;
		}
	public:
		GSLMultiExpFit( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ), x( 0 ), y( 0 ), samples( 0 ) {
		}

		void analyze( NSCore::AcqisitionDataSet& data ) {
			if ( samples != data.samples() ) {
				delete[] x;
				delete[] y;
				samples = data.samples();
				x = new double[ samples ];
				for ( int j=0; j<samples; ++j )
					x[ j ] = j;
				y = new double[ samples ];
			}

			for ( int i=0; i< data.channels(); ++i ) {
				double *ch = data.getRawChannel( i );
				for ( int j=0; j<samples; ++j )
					y[ j ] = qMin( 0., log( ch[ j ] ) );

				minimizer = gsl_multimin_fminimizer_alloc( gsl_multimin_fminimizer_nmsimplex, 4 );
				gsl_multimin_function minex_func;
				minex_func.n = 4;
				minex_func.f = &myfunction;
				parameters p;
				p.samples = samples;
				p.ch = y;
				minex_func.params = &p;

				gsl_vector *x, *ss;
				x = gsl_vector_alloc( 4 );
				gsl_vector_set( x, 0, 0.1 );
				gsl_vector_set( x, 1, 0.3 );
				gsl_vector_set( x, 2, 100 );
				gsl_vector_set( x, 3, 0.1 );
				ss = gsl_vector_alloc( 4 );
				gsl_vector_set( x, 0, 0.1 );
				gsl_vector_set( x, 1, 0.1 );
				gsl_vector_set( x, 2,  10 );
				gsl_vector_set( x, 3, 0.1 );

				gsl_multimin_fminimizer_set( minimizer, &minex_func, x, ss );

				int status, iter;
				double size;

				do {
					iter++;
					status = gsl_multimin_fminimizer_iterate( minimizer );
					if ( status )
						break;
					size = gsl_multimin_fminimizer_size( minimizer );
					status = gsl_multimin_test_size( size, 1e-4 );
					if ( status == GSL_SUCCESS )
						qDebug( " Found a minimum of value %g", size );
				}
				while ( status == GSL_CONTINUE && iter < 100 );

				gsl_multimin_fminimizer_free( minimizer );
			}
		}


		QString type() const { return "GSLMultiExpFit"; }
};

#endif // DOPPLERGSL_H

