/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "plugins_factory.h"

#include <math.h>
#include <QtCore/QLibrary>
#include <QtCore/QCoreApplication>
#include <QtCore/QDir>
#include <QtCore/QMutex>
#include <QtCore/QMutexLocker>

#include "config.h"

namespace NSEvaluation {

	PluginFactory::PluginFactory() : mutex(  new QMutex( QMutex::Recursive ) ) {
		//qDebug() << "PluginFactory::PluginFactory()";
	}

	struct pluginlib_t {
		QString libname;
		QString libpath;
		QStringList symbols;
	};

	QList<pluginlib_t> searchplugins( QDir dir ) {
		//qDebug() << "searchplugins(" << dir.absolutePath() << ")";
		QList<pluginlib_t> ret;
		QStringList libs = dir.entryList( QStringList() << "*.eposplugin" );
		foreach( QString lib, libs ) {
			//qDebug() << " trying" << lib;
			pluginlib_t plugin;
			QFile f( dir.absoluteFilePath( lib ) );
			if ( f.open( QIODevice::ReadOnly ) ) {
				QString tmp = f.readAll();
				plugin.symbols = tmp.split( "\n", QString::SkipEmptyParts );
				//qDebug() << " Symbols found" << plugin.symbols;
				for ( QStringList::iterator it=plugin.symbols.begin(); it!=plugin.symbols.end(); ++it )
					( *it ) += "_start";
				//qDebug() << " used symbols will be" << plugin.symbols;
				f.close();
			}
			lib.chop( 11 );
			plugin.libpath = dir.absoluteFilePath( lib );
			plugin.libname = lib.section( "/", -1, -1 );
			//qDebug() << "Found lib '" << lib << "' in '" << dir.absolutePath() << "'";
			ret.append( plugin );
		}
		foreach( QString child, dir.entryList( QStringList() << "lib*", QDir::Dirs | QDir::NoDotAndDotDot ) ) {
			ret += searchplugins( dir.absoluteFilePath( child ) );
		}
		return ret;
	}

	PluginFactory* PluginFactory::the() {
		static PluginFactory* factory = 0;
		static QMutex themutex( QMutex::Recursive );
		QMutexLocker lock( &themutex );
		if ( ! factory ) {
			factory = new PluginFactory();
			QMutexLocker lock( factory->mutex );
			QList<pluginlib_t> libraries = searchplugins( QDir() );
			qDebug() << "Plugindir is " << PLUGINSDIR;
			libraries += searchplugins( QDir( PLUGINSDIR ) );

			qDebug() << " * PluginFactory loading plugins" << libraries.size();

			QStringList loaded_libs;

			foreach( pluginlib_t plugin, libraries ) {
				if ( ! loaded_libs.contains( plugin.libname ) ) {
					QLibrary* lib = new QLibrary( plugin.libpath );
					lib->load();
					if ( lib->isLoaded() ) {
						loaded_libs.append( plugin.libname );
						if ( plugin.symbols.isEmpty() )
							plugin.symbols.append( "start" );
						foreach( QString symbol, plugin.symbols ) {
							typedef void ( *MyPrototype )();
							MyPrototype func = ( MyPrototype )lib->resolve( symbol.toStdString().c_str() );
							if ( func )
								func();
							else
								qWarning() << QString( "Symbol %1 not found in lib %2" ).arg( symbol ).arg( plugin.libname );
						}
					} else
						qWarning() << QString( " * Couldn't load lib '%1' (%2)! *\n * Error was '%3' *" ).arg( plugin.libname ).arg( plugin.libpath ).arg( lib->errorString() );
				}
			}
			qDebug() << endl << " List of plugins usable:" << factory->plugins() << endl;
		}
		return factory;
	}

	void PluginFactory::registerClient( PluginFactory_Client* client ) {
		QMutexLocker lock( mutex );
		//qDebug() << "PluginFactory::registerClient(" << client << ") this =" << this;
		_factories.push_back( client );
	}

	Plugin* PluginFactory::newPlugin( QString s, NSCore::QTreeItem* p ) {
		QMutexLocker lock( mutex );
		//qDebug() << "PluginFactory::newPlugin(" << s << ") this =" << this;
		//qDebug() << " List is" << plugins();
		for ( QList<PluginFactory_Client*>::Iterator it = _factories.begin(); it != _factories.end(); ++it ) {
			if ( ( *it )->canCreate().contains( s ) )
				return ( *it )->create( s, p );
		}
		return 0;
	}

	QStringList PluginFactory::plugins() const {
		QMutexLocker lock( mutex );
		QStringList tmp;
		foreach( PluginFactory_Client* client, _factories ) {
			tmp += client->canCreate();
		}
		return tmp;
	}

	QString PluginFactory::type( QString plugin ) const {
		//qDebug() << endl << "PluginFactory::type(" << plugin << ")";
		QMutexLocker lock( mutex );
		foreach( PluginFactory_Client* client, _factories ) {
			if ( client->canCreate().contains( plugin ) ) {
				//qDebug() << " Checking" << client << endl << " has plugins" << client->canCreate();
				return client->type( plugin );
			}
		}
		return "";
	}

	QString PluginFactory::help( QString plugin ) const {
		QMutexLocker lock( mutex );
		foreach( PluginFactory_Client* client, _factories ) {
			if ( client->canCreate().contains( plugin ) )
				return client->help( plugin );
		}
		return "";
	}
	QStringList PluginFactory::categories( QString plugin ) const {
		QMutexLocker lock( mutex );
		foreach( PluginFactory_Client* client, _factories ) {
			if ( client->canCreate().contains( plugin ) )
				return client->categories( plugin );
		}
		return QStringList();
	}

	QList<QDomNode> PluginFactory::getPluginsList( QDomDocument doc ) const {
		QMutexLocker lock( mutex );
		QList<QDomNode> list;
		foreach ( QString plugin, PluginFactory::the()->plugins() ) {
			QDomElement item = doc.createElement( "item" );
			QDomElement title = doc.createElement( "title" );
			item.appendChild( title );
			title.appendChild( doc.createTextNode( plugin ) );
			QDomElement help = doc.createElement( "help" );
			item.appendChild( help );
			help.appendChild( doc.createTextNode( this->help( plugin ) ) );
			QDomElement categories = doc.createElement( "categories" );
			item.appendChild( categories );
			QStringList catlist = this->categories( plugin );
			if ( ! catlist.contains( type( plugin ) ) )
				catlist += type( plugin );
			foreach ( QString cat, catlist ) {
				QDomElement catelem = doc.createElement( "category" );
				catelem.appendChild( doc.createTextNode( cat ) );
				categories.appendChild( catelem );
			}
			list << item;
		}
		return list;
	}
};

