/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "chartview.h"
#include "chartview.moc"

#include <QtCore/QDebug>
#include <QtCore/QModelIndex>
#include <QtGui/QItemSelection>
#include <QtGui/QItemSelectionModel>
#include <QtGui/QScrollBar>
#include <QtGui/QPainter>
#include <QtGui/QMenu>
#include <QtGui/QAction>
#include <QtGui/QContextMenuEvent>
#include <QtGui/QTextDocument>

#include <cmath>

double distance( const QPointF& p1, const QPointF p2 ) {
	return sqrt( pow( p2.x()-p1.x(), 2 ) + pow( p2.y()-p1.y(), 2 ) );
}

QList<double> tics( double min, double max, double inc ) {
	QList<double> ret;
	for ( double tic=(min-fmod(min,inc)); tic<=max; tic+=inc )
		ret << tic;
	return ret;
}

ChartView::ChartView( QWidget* p )
	: QAbstractItemView( p )
	, _title( "(no title)" )
	, _mouseInside( false )
	, _range_drag( 0 )
{
	//qDebug() << "ChartView::ChartView( QWidget*" << p << ")";

	/*colors
		<< QColor(255,0,0)
		<< QColor(0,0,255)
		<< QColor(0,255,0)
		<< QColor(227,167,0)
		<< QColor(0,255,255)
		<< QColor(255,0,255)
		<< QColor(192,36,0);*/
	/*colors
		<< QColor(64, 0, 0 ) //	Very dark red
		<< QColor(0, 64, 0 ) //	Very dark green
		<< QColor(64, 64, 0 ) //	Very dark yellow
		<< QColor(0, 64, 64 ); //	Very dark cyan*/

	if ( palette().color(QPalette::Base).valueF() > 0.3 )
		colors
			<< QColor(128, 0, 0 ) //	Dark red
			<< QColor(0, 128, 0 ) //	Dark green
			<< QColor(0, 0, 128 ) //	Dark blue
			<< QColor(128, 128, 0 ) //	Dark yellow
			<< QColor(0, 128, 128 ) //	Dark cyan
			<< QColor(128, 0, 128 ) //	Dark magenta
			<< QColor(192, 88, 0 ); //	Dark orange

	colors
		<< QColor(192, 0, 0 ) //	Mid red
		<< QColor(0, 192, 0 ) //	Mid green
		<< QColor(0, 0, 192 ) //	Mid blue
		<< QColor(192, 192, 0 ) //	Mid yellow
		<< QColor(0, 192, 192 ) //	Mid cyan
		<< QColor(192, 0, 192 ); //	Mid magenta

	colors
		<< QColor(255, 0, 0 ) //	Red
		<< QColor(0, 255, 0 ) //	Green
		<< QColor(0, 0, 255 ) //	Blue
		<< QColor(255, 255, 0 ) //	Yellow
		<< QColor(0, 255, 255 ) //	Cyan
		<< QColor(255, 0, 255 ) //	Magenta
		<< QColor(255, 128, 0 ); //	Orange

	if ( palette().color(QPalette::Base).valueF() < 0.7 )
		colors
			<< QColor(192, 255, 192 ) //	Light green
			<< QColor(255, 192, 192 ) //	Light red
			<< QColor(192, 192, 255 ) //	Light blue
			<< QColor(255, 255, 192 ) //	Light yellow
			<< QColor(192, 255, 255 ) //	Light cyan
			<< QColor(255, 192, 255 ) //	Light magenta
			<< QColor(255, 168, 88 )  //	Light orange
			<< QColor(255, 220, 168 ); //	Very light orange

	_zoomfactors << 1.0 << 2.0 << 3.0 << 4.0 << 8.0 << 16.0;
	_currentzoom = _zoomfactors.begin();
	_scrollfraction = 0.0;

	horizontalScrollBar()->setRange( 0,0 );
	verticalScrollBar()->setRange( 0,0 );

	QAction* toggleFWHM = new QAction( "Toggle FWHM-line", this );
	toggleFWHM->setCheckable( true );
	connect( toggleFWHM, SIGNAL( triggered( bool ) ), this, SLOT( setFWHMLine( bool ) ) );
	addAction( toggleFWHM );

	QAction* toggleLog = new QAction( "Toggle logarithmic scale", this );
	toggleLog->setCheckable( true );
	connect( toggleLog, SIGNAL( triggered( bool ) ), this, SLOT( setLogarithmicScale( bool ) ) );
	addAction( toggleLog );

	_action_zoomin = new QAction( QIcon( ":/icons/kde/cr22-action-viewmag+.png" ), "Zoom In", this );
	connect( _action_zoomin, SIGNAL( triggered() ), this, SLOT( increaseZoom() ) );
	addAction( _action_zoomin );
	_action_zoomout = new QAction( QIcon( ":/icons/kde/cr22-action-viewmag-.png" ), "Zoom Out", this );
	connect( _action_zoomout, SIGNAL( triggered() ), this, SLOT( decreaseZoom() ) );
	addAction( _action_zoomout );

	QAction* addRange = new QAction( "Add Range Marker", this );
	connect( addRange, SIGNAL( triggered() ), this, SLOT( addRangeMarker() ) );
	addAction( addRange );

	//addAreaMarker();
	//qDebug() << "ChartView::ChartView() finished";
}
ChartView::~ChartView() {
	//qDebug( "ChartView::~ChartView()" );
	/// Should be unnecessary because the ranges are QObjects and childs of this widget:
	while ( ! _ranges.isEmpty() )
		delete _ranges.takeFirst();
}

QRect ChartView::visualRect( const QModelIndex& ) const {
	//qDebug( "ChartView::visualRect( const QModelIndex& )" );
	return QRect();
}
void ChartView::scrollTo( const QModelIndex&, ScrollHint ) {
	qDebug( "ChartView::scrollTo( const QModelIndex&, ScrollHint )" );
}
QModelIndex ChartView::indexAt( const QPoint& point ) const {
	//qDebug( "ChartView::indexAt( const QPoint& (%i,%i) )", point.x(), point.y() );
	return QModelIndex();
}
void ChartView::setModel( QAbstractItemModel* model ) {
	QAbstractItemView::setModel( model );
	connect( model, SIGNAL( rowsRemoved( const QModelIndex&, int, int ) ), this, SLOT( rowsRemoved( const QModelIndex&, int, int ) ) );
	connect( model, SIGNAL( columnsRemoved( const QModelIndex&, int, int ) ), this, SLOT( columnsRemoved( const QModelIndex&, int, int ) ) );
}

void ChartView::dataChanged( const QModelIndex& topLeft, const QModelIndex& bottomright ) {
	if ( model() == 0 ) {
		return;
	}
	//qDebug( "ChartView::dataChanged( const QModelIndex&, const QModelIndex& )" );
	//qDebug() << topLeft << endl << bottomright;
	QAbstractItemView::dataChanged( topLeft, bottomright );

	QModelIndex index;
	double time, count;
	int rowstart = 0;
	int rowend = model()->rowCount( rootIndex() )-1;
	if ( /*! _cache.dirty &&*/ topLeft.isValid() && bottomright.isValid() ) {
		if ( topLeft.row() == bottomright.row() ) // That probably means a single value was edited,
			_cache.dirty = true;                    // so we better recalculate min and max...
		else {                                    //
			rowstart = topLeft.row();         // Else its just a new value and we check wether
			rowend = bottomright.row();       // this one could be new min/max.
		}

	}
	if ( _cache.dirty ) {
		_cache.reset();
		int b = _cache.borderwidth;
		_cache.visualRect = viewport()->rect().adjusted( b, b, -b, -b );
		// Zoom:
		horizontalScrollBar()->setRange( 0, (_currentzoom!=_zoomfactors.begin()) * _cache.visualRect.width() );

		_rect_zoomin = QRect( 0,0, 30,30 );
		_rect_zoomin.moveTop( _cache.visualRect.top() + _cache.borderwidth );
		_rect_zoomin.moveRight( _cache.visualRect.right() - _cache.borderwidth );
		_rect_zoomout = QRect( 0,0, 30,30 );
		_rect_zoomout.moveTop( _cache.visualRect.top() + _cache.borderwidth );
		_rect_zoomout.moveRight( _rect_zoomin.left() - _cache.borderwidth );
	}
	for ( int row=rowstart; row<=rowend; ++row ) {
		// Key
		time = model()->headerData( row, Qt::Vertical ).toDouble();
		_cache.data_xmin = qMin( time, _cache.data_xmin );
		_cache.data_xmax = qMax( time, _cache.data_xmax );
		// Value
		for ( int col=0; col<model()->columnCount(); ++col ) {
			index = model()->index( row, col, rootIndex() );
			count = data( index );
			_cache.data_ymin = qMin( count, _cache.data_ymin );
			_cache.data_ymax = qMax( count, _cache.data_ymax );
		}

		//qDebug( "values are : (%e|%e)", time, count );
		//qDebug( "xmin = %e, xmax = %e, ymin = %e, ymax = %e", xmin, xmax, ymin, ymax );
	}
	//qDebug( "xmin = %e, xmax = %e, ymin = %e, ymax = %e", _cache.data_xmin, _cache.data_xmax, _cache.data_ymin, _cache.data_ymax );

	_cache.normalizeModel.reset();
	_cache.normalizeModel.translate( -_cache.data_xmin/_cache.data_xrange(), -_cache.data_ymin/_cache.data_yrange() );
	_cache.normalizeModel.scale( 1/_cache.data_xrange(), 1/_cache.data_yrange() );

	_cache.dirty = false;

	//qDebug( "ChartView::dataChanged() finished." );
	if ( ! _inPaintEvent )
		viewport()->update();
}
void ChartView::rowsInserted( const QModelIndex&, int, int ) {
	qDebug( "ChartView::rowsInserted( const QModelIndex&, int, int )" );
}
void ChartView::rowsRemoved( const QModelIndex&, int, int ) {
	//qDebug() << "ChartView::rowsRemoved(...)";
	_cache.dirty = true;
	dataChanged();
}
void ChartView::columnsRemoved( const QModelIndex&, int, int ) {
	//qDebug() << "ChartView::columnsRemoved(...)";
	_cache.dirty = true;
	dataChanged();
}

bool ChartView::edit( const QModelIndex&, EditTrigger, QEvent* ) {
	//qDebug( "ChartView::edit( const QModelIndex&, EditTrigger, QEvent* )" );
	return false;
}
QModelIndex ChartView::moveCursor( QAbstractItemView::CursorAction, Qt::KeyboardModifiers ) {
	//qDebug( "ChartView::moveCursor( QAbstractItemView::CursorAction, Qt::KeyboardModifiers )" );
	return QModelIndex();
}

int ChartView::horizontalOffset() const {
	//qDebug( "ChartView::horizontalOffset()" );
	return horizontalScrollBar()->value();
}
int ChartView::verticalOffset() const {
	//qDebug( "ChartView::verticalOffset()" );
	return verticalScrollBar()->value();
}

bool ChartView::isIndexHidden( const QModelIndex& ) const {
	qDebug( "ChartView::isIndexHidden( const QModelIndex& )" );
	return false;
}

void ChartView::setSelection( const QRect&, QItemSelectionModel::SelectionFlags ) {
	//qDebug( "ChartView::setSelection( const QRect&, QItemSelectionModel::SelectionFlags )" );
}

void ChartView::paintEvent( QPaintEvent* ) {
	//qDebug( "ChartView::paintEvent( QPaintEvent* )" );

	if ( ! model() || model()->rowCount() == 0 || model()->columnCount() == 0 ) {
		return;
	}
	_inPaintEvent = true;
	if ( _cache.dirty )
		dataChanged( QModelIndex(), QModelIndex() );
	_inPaintEvent = false;

	if ( _cache.data_yrange() == 0 || _cache.data_xrange() == 0 ) {
		return;
	}

	QPainter painter( viewport() );
	// Paint the border
	//painter.setPen( QColor( 200, 200, 200 ) );
	//painter.drawRect( _cache.visualRect );

	painter.save();
	painter.translate( _cache.visualRect.bottomLeft() );

	QTransform ct;
	ct.translate( _scrollfraction * _cache.visualRect.width() * (*_currentzoom -1), 0 );
	ct.scale( _cache.visualRect.width() * (*_currentzoom), -_cache.visualRect.height() );
	ct = _cache.normalizeModel * ct;

	// Calculate X- and Y-Zero-Lines and grid
	QVector<QLineF> gridlines;
	double inc = 1E-6;
	double x0 = qMin( _cache.data_ymax, qMax( _cache.data_ymin, 0.0 ) );
	while ( fabs( (QPointF(0,0)*ct).x() - (QPointF(inc,0)*ct).x() ) < 20 )
		inc *= 10;
	QList<double> xtics = tics( _cache.data_xmin, _cache.data_xmax, inc );
	QList<double> mxtics = tics( _cache.data_xmin, _cache.data_xmax, inc/2 );
	QPolygonF xpoints;
	foreach ( double x, xtics )
		xpoints << QPointF( x, x0 );
	foreach ( double x, mxtics )
		gridlines << QLineF( QPointF( x, _cache.data_ymin ), QPointF( x, _cache.data_ymax ) );
	inc = 1E-6;
	double y0 = qMin( _cache.data_xmax, qMax( _cache.data_xmin, 0.0 ) );
	while ( fabs( (QPointF(0,0)*ct).y() - (QPointF(0,inc)*ct).y() ) < 20 )
		inc *= 10;
	QList<double> ytics;
	QList<double> mytics;
	if ( _cache.logscale ) {
		double inc = 1;
		for ( double y = 1; y < exp(_cache.data_ymax); y+=inc ) {
			if ( fmod( y, inc*10 ) == 0 ) {
				inc *= 10;
				ytics << log(y);
			}
			mytics << log(y);
		}
	} else {
		ytics = tics( _cache.data_ymin, _cache.data_ymax, inc );
		mytics = tics( _cache.data_ymin, _cache.data_ymax, inc/2 );
	}
	QPolygonF ypoints;
	foreach ( double y, ytics )
		ypoints << QPointF( y0, y );
	foreach ( double y, mytics )
		gridlines << QLineF( QPointF( _cache.data_xmin, y ), QPointF( _cache.data_xmax,y ) );
	// FWHM
	QLineF fwhm;
	if ( _cache.fwhmline ) {
		double y = _cache.data_ymax / 2.0;
		if ( _cache.logscale ) {
			y = log( exp( _cache.data_ymax ) / 2.0 );
		}
		//fwhm.setLine( _cache.data_xmin, y, _cache.data_xmax, y );
		fwhm = QLineF( _cache.data_xmin, y, _cache.data_xmax, y );
	}

	// Draw Grid
	painter.setPen( palette().color(QPalette::Button) );
	foreach ( QLineF line, gridlines ) {
		painter.drawLine( line*ct );
	}
	// Draw Axis
	painter.setPen( palette().color(QPalette::Text) );
	foreach ( QPointF p, xpoints ) {
		painter.drawLine( p*ct + QPointF(0,3), p*ct + QPointF(0,-3) );
		painter.drawText( p*ct, QString::number( p.x() ) );
	}
	painter.drawLine( QPointF(_cache.data_xmin,x0) * ct, QPointF(_cache.data_xmax,x0) * ct );
	foreach ( QPointF p, ypoints ) {
		painter.drawLine( p*ct + QPointF(3,0), p*ct + QPointF(-3,0) );
		if ( _cache.logscale )
			painter.drawText( p*ct+QPointF(4,(painter.fontMetrics().height()/4.0)), QString::number( exp(p.y()) ) );
		else
			painter.drawText( p*ct+QPointF(4,(painter.fontMetrics().height()/4.0)), QString::number( p.y() ) );
	}
	painter.drawLine( QPointF(y0,_cache.data_ymin) * ct, QPointF(y0,_cache.data_ymax) * ct );
	// Draw Ranges
	double oldsize = 0;
	foreach ( Range* range, _ranges ) {
		//qDebug() << "Drawing range" << range->min << "," << range->max;
		QColor bg = palette().color( QPalette::Highlight );
		if ( &(range->min) != _range_drag && &(range->max) != _range_drag )
			painter.setPen( Qt::NoPen );
		else
			painter.setPen( bg );
		bg.setAlpha( 100 );
		painter.setBrush( bg );
		QRectF r( QPointF( range->min, _cache.data_ymin ) * ct,
				QPointF( range->max, _cache.data_ymax ) * ct );
		painter.drawRect( r );
	}
	painter.setPen( palette().color( QPalette::HighlightedText ) );
	foreach ( Range* range, _ranges ) {
		//painter.drawText( r, Qt::AlignCenter|Qt::TextWordWrap, range->label );
		painter.save();
		QRectF r( QPointF( range->min, _cache.data_ymin ) * ct,
				QPointF( range->max, _cache.data_ymax ) * ct );
		painter.translate( r.bottomLeft() );
		painter.translate( 0, -oldsize );
		QTextDocument text( range->label );
		text.adjustSize();
		//qDebug() << "text.size" << text.size() << text.pageSize() << " r" << r;
		//qDebug() << painter.transform().map( QPointF(0.0, 0.0) );
		text.drawContents( &painter );
		oldsize -= text.size().height();
		painter.restore();
	}
	// Draw FWHM
	if ( ! fwhm.isNull() ) {
		painter.setPen( palette().color(QPalette::Highlight) );
		painter.drawLine( fwhm * ct );
	}

	painter.setRenderHint( QPainter::Antialiasing );
	// Draw Data
	for (int col=0; col<model()->columnCount(); ++col) {
		QPolygonF points;
		for (int row=0; row<model()->rowCount(); ++row) {
			points << QPointF( model()->headerData( row, Qt::Vertical ).toDouble(), data( model()->index( row, col ) ) ) * ct;
		}
		painter.setPen( colors[col%colors.size()] );
		painter.drawPolyline( points );
	}

	// Back to native coordinates
	painter.restore();

	// Draw icons for zoom-in and zoom-out
	if ( _mouseInside ) {
		QColor blend = palette().color(QPalette::Button);
		blend.setAlphaF( 0.75 );
		painter.fillRect( _rect_zoomin, blend );
		painter.fillRect( _rect_zoomout, blend );
		_action_zoomout->icon().paint( &painter, _rect_zoomout.toRect(), Qt::AlignCenter,
			(_currentzoom==_zoomfactors.begin())?QIcon::Disabled:QIcon::Normal );
		_action_zoomin->icon().paint( &painter, _rect_zoomin.toRect(), Qt::AlignCenter,
			(*_currentzoom==_zoomfactors.last())?QIcon::Disabled:QIcon::Normal );
	}

	// Draw Alphablending towards the border left and right
	QRect r( _cache.visualRect.right(), 0, _cache.borderwidth, viewport()->rect().height() );
	QLinearGradient grad( r.left()-2, 0, r.right(), 0 );
	QColor blend = palette().color(QPalette::Base);
	blend.setAlphaF(1.0);
	grad.setColorAt( 1, blend );
	QColor opaque = palette().color(QPalette::Base);
	opaque.setAlphaF(0.0);
	grad.setColorAt( 0, opaque );
	painter.fillRect( r, QBrush( grad ) );
	r = QRect( 0, 0, _cache.visualRect.left(), viewport()->rect().height() );
	grad.setStart( r.topRight() );
	grad.setFinalStop( r.topLeft() );
	painter.fillRect( r, QBrush( grad ) );
	//qDebug( "ChartView::paintEvent() finished." );
}

void ChartView::resizeEvent( QResizeEvent* ) {
	_cache.dirty = true;
}
void ChartView::scrollContentsBy( int dx, int dy ) {
	//qDebug() << "ChartView::scrollContentsBy(" << dx << "," << dy << ")";
	_scrollfraction += dx*1.0 / (horizontalScrollBar()->maximum());
	_scrollfraction = qMax( -1.0, qMin( 0.0, _scrollfraction ) );
	_cache.dirty = true;
	viewport()->update();
}

QRegion ChartView::visualRegionForSelection( const QItemSelection& ) const {
	qDebug( "ChartView::visualRegionForSelection( const QItemSelection& )" );
	return QRegion();
}

void ChartView::contextMenuEvent( QContextMenuEvent* ev ) {
	qDebug() << "ChartView::contextMenuEvent(" << ev << ")";
	QMenu::exec( actions(), ev->globalPos() );
}

void ChartView::keyPressEvent( QKeyEvent* ev ) {
	QAbstractItemView::keyPressEvent( ev );
	if ( ! ev->isAccepted() ) {
		//qDebug() << "ChartView::keyPressEvent(" << ev << ")";
		if ( ev->matches( QKeySequence::ZoomIn ) ) {
			increaseZoom();
			ev->accept();
		}
		if ( ev->matches( QKeySequence::ZoomOut ) ) {
			decreaseZoom();
			ev->accept();
		}
	}
}

void ChartView::increaseZoom() {
	//qDebug() << "ChartView::increaseZoom()";
	_currentzoom++;
	if ( _currentzoom == _zoomfactors.end() )
		_currentzoom--;
	//qDebug() << " New zoom is:" << *_currentzoom;
	_cache.dirty = true;
	viewport()->update();
}
void ChartView::decreaseZoom() {
	//qDebug() << "ChartView::decreaseZoom()";
	if ( _currentzoom != _zoomfactors.begin() )
		_currentzoom--;
	//qDebug() << " New zoom is:" << *_currentzoom;
	_cache.dirty = true;
	viewport()->update();
}
void ChartView::enterEvent( QEvent* ) {
	//qDebug() << "ChartView::enterEvent()";
	_mouseInside = true;
	viewport()->update();
}
void ChartView::leaveEvent( QEvent* ) {
	//qDebug() << "ChartView::leaveEvent()";
	_mouseInside = false;
	viewport()->update();
}

double ChartView::data( const QModelIndex& index ) {
	double ret = model()->data( index, Qt::DisplayRole ).toDouble();
	if ( _cache.logscale ) {
		if ( fabs(ret) > 0 )
			ret = log( fabs(ret) );
		else
			ret = 0.0;
	}
	return ret;
}

Range* ChartView::addRangeMarker() {
	qDebug() << "ChartView::addRangeMarker()";
	double quarter = _cache.data_xrange() / 4;
	_ranges.append( new Range( _cache.data_xmin + quarter, _cache.data_xmin + 3*quarter, this ) );
	emit rangeAdded( _ranges.last() );
	return _ranges.last();
}

QPointF ChartView::toLocalPos( const QPointF &pos ) const {
	QTransform inv = _cache.normalizeModel.inverted();
	QPointF p( (pos.x() - _cache.visualRect.left() - _scrollfraction *
				_cache.visualRect.width() * (*_currentzoom -1) ) / (_cache.visualRect.width() * (*_currentzoom)),
			pos.y() / _cache.visualRect.height() );
	p = p * inv;
	return p;
}
void ChartView::mousePressEvent( QMouseEvent* ev ) {
	//qDebug() << "ChartView::mousePressEvent(" << ev << ")";
	if ( ev->buttons() & Qt::LeftButton ) {
		QPointF pos = toLocalPos( ev->pos() );
		for ( int i=0; i<_ranges.size(); ++i ) {
			if ( _range_drag == 0 ) {
				if ( (ev->modifiers() & Qt::ShiftModifier) && (abs( pos.x() - _ranges[i]->min ) / _cache.data_xrange() < 0.05) )
					_range_drag = &(_ranges[i]->min);
				if ( (ev->modifiers() & Qt::ControlModifier) && (abs( pos.x() - _ranges[i]->max ) / _cache.data_xrange() < 0.05) )
					_range_drag = &(_ranges[i]->max);
			}
		}
		if ( _range_drag != 0 ) {
			//qDebug() << " starting a drag" << _range_drag;
			ev->accept();
		}
	} else {
		QAbstractItemView::mousePressEvent( ev );
	}
}
void ChartView::mouseMoveEvent( QMouseEvent* ev ) {
	//qDebug() << "ChartView::mouseMoveEvent(" << ev << ")";
	if ( _range_drag != 0 ) {
		//qDebug() << " _range_drag(" << _range_drag << ") old" << *_range_drag;
		(*_range_drag) = toLocalPos( ev->pos() ).x();
		//qDebug() << " _range_drag(" << _range_drag << ") new" << *_range_drag;
		viewport()->update();
		ev->accept();
	} else
		QAbstractItemView::mouseMoveEvent( ev );
}
void ChartView::mouseReleaseEvent( QMouseEvent* ev ) {
	//qDebug() << "ChartView::mouseReleaseEvent(" << ev << ")";
	if ( _range_drag != 0 ) {
		//qDebug() << "Gotta finish a drag";
		_range_drag = 0;
		foreach ( Range* r, _ranges ) {
			if ( r->max < r->min ) {
				double tmp = r->max;
				r->max = r->min;
				r->min = tmp;
			}
			emit r->rangeChanged();
		}
		viewport()->update();
		ev->accept();
		return;
	}
	if ( ev->button() == Qt::LeftButton ) {
		if ( _rect_zoomin.contains( ev->pos() ) ) {
			_action_zoomin->trigger();
			ev->accept();
		}
		if ( _rect_zoomout.contains( ev->pos() ) ) {
			_action_zoomout->trigger();
			ev->accept();
		}
	}
	if ( ! ev->isAccepted() ) {
		QAbstractItemView::mouseReleaseEvent( ev );
	}
}

