/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef CHARTVIEW_H
#define CHARTVIEW_H

#include <QtGui/QAbstractItemView>

#include <cmath>

class QAction;

class Range : public QObject
{
	Q_OBJECT
	friend class ChartView;
	public:
		Range( double r1, double r2, QObject* p=0 ) : QObject(p), min( r1 ), max( r2 ) {
		}

		double min, max;

		QString label;

	signals:
		void minChanged( double );
		void maxChanged( double );

		void rangeChanged();

	private:
		Q_DISABLE_COPY( Range );
};

class ChartView : public QAbstractItemView
{
	Q_OBJECT
	Q_PROPERTY( bool FWHMLine READ FWHMLine WRITE setFWHMLine )
	Q_PROPERTY( bool LogarithmicScale READ logarithmicScale WRITE setLogarithmicScale )
	Q_PROPERTY( double zoomFactor READ zoomFactor )
	public:
		ChartView( QWidget* =0 );
		~ChartView();

		QRect visualRect( const QModelIndex& ) const;
		void scrollTo( const QModelIndex&, ScrollHint =EnsureVisible );
		QModelIndex indexAt( const QPoint& ) const;
		void setModel( QAbstractItemModel* );

		bool logarithmicScale() const { return _cache.logscale; }
		bool FWHMLine() const { return _cache.fwhmline; }

		double zoomFactor() const { return *_currentzoom; }

		int rangeCount() const { return _ranges.size(); }
		Range* range( int i ) { return _ranges[i]; }

	public slots:
		void setLogarithmicScale( bool n ) {
			_cache.logscale = n;
			_cache.dirty = true;
			dataChanged( QModelIndex(), QModelIndex() );
		}
		void setFWHMLine( bool n ) { _cache.fwhmline = n; viewport()->update(); }
		void increaseZoom();
		void decreaseZoom();

		Range* addRangeMarker();


	signals:
		void rangeAdded( Range* );

	protected slots:
		void dataChanged( const QModelIndex& = QModelIndex(), const QModelIndex& = QModelIndex() );
		void rowsInserted( const QModelIndex&, int, int );

	protected:
		bool edit( const QModelIndex&, EditTrigger, QEvent* );
		QModelIndex moveCursor( QAbstractItemView::CursorAction, Qt::KeyboardModifiers );

		int horizontalOffset() const;
		int verticalOffset() const;

		bool isIndexHidden( const QModelIndex& ) const;

		void setSelection( const QRect&, QItemSelectionModel::SelectionFlags );

		void paintEvent( QPaintEvent* );
		void resizeEvent( QResizeEvent* );
		void scrollContentsBy( int, int );

		QRegion visualRegionForSelection( const QItemSelection& ) const;

		void contextMenuEvent( QContextMenuEvent* );

		void keyPressEvent( QKeyEvent* );

		void enterEvent( QEvent* );
		void leaveEvent( QEvent* );

		void mousePressEvent( QMouseEvent* );
		void mouseMoveEvent( QMouseEvent* );
		void mouseReleaseEvent( QMouseEvent* );

	private slots:
		void rowsRemoved( const QModelIndex&, int, int );
		void columnsRemoved( const QModelIndex&, int, int );

	private:
		QString _title;
		bool _inPaintEvent;

		/// Access the data and automaticly scale/log
		double data( const QModelIndex& );

		struct cache_t
		{
			/// The y-Axis is to be scaled logarithmic.
			bool logscale;
			/// Display a fwhm-line
			bool fwhmline;

			/// This cache is dirty and has to be re-set.
			bool dirty;
			/// Min and Max of the values in the model.
			double data_xmin, data_xmax, data_ymin, data_ymax;

			/// Visual rect in the viewport. Will be surrounded by the border.
			QRectF visualRect;
			/// border
			double borderwidth;

			/// Transform the coordinates of the model to normalized values between 0 and 1
			QTransform normalizeModel;

			cache_t()
				: logscale( false )
				, fwhmline( false )
				, dirty( true )
				, borderwidth( 15 )
			{
				reset();
			}
			void reset() {
				data_xmin = INFINITY;
				data_xmax = -INFINITY;
				data_ymin = INFINITY;
				data_ymax = -INFINITY;
			}

			double data_xrange() const { return data_xmax - data_xmin; }
			double data_yrange() const { return data_ymax - data_ymin; }

		} _cache;

		/// Zooming and scrolling.
		QList<double> _zoomfactors;
		QList<double>::Iterator _currentzoom;
		double _scrollfraction;
		QAction *_action_zoomin, *_action_zoomout;
		QRectF _rect_zoomin, _rect_zoomout;
		bool _mouseInside;

		/// Colors
		QList<QColor> colors;

		QPointF toLocalPos( const QPointF& ) const;

		QList<Range*> _ranges;
		double *_range_drag;
};

#endif // CHARTVIEW_H

