/*
    Copyright 2005-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "osciwidget_v3.h"
#include "osciwidget_v3.moc"

#include "osciwidget_v3_private.h"
#include "osciwidget_v3_private.moc"

#include <math.h>
#include <QtCore/QMutex>
#include <QtCore/QMutexLocker>
#include <QtGui/QPainter>
#include <QtGui/QVBoxLayout>
#include <QtGui/QInputDialog>
#include <QtGui/QAction>
#include <QtGui/QMenu>
#include <QtGui/QContextMenuEvent>
#include <QtCore/QTimer>

Q_DECLARE_METATYPE( NSCore::AcqisitionDataSet )

OsciWidgetV3_ChainMember::OsciWidgetV3_ChainMember( QObject* p )
	: NSCore::DataSetChainMember( "OsciConsumer", p )
{
	_slowdown.start();
	qRegisterMetaType<NSCore::AcqisitionDataSet>();
}
bool OsciWidgetV3_ChainMember::dataArrived( NSCore::AcqisitionDataSet d ) {
	//qDebug() << "OsciWidgetV3_ChainMember::dataArrived(" << d << ")";
	if ( _slowdown.elapsed() > 200 ) {
		emit data( d );
		_slowdown.restart();
	}
	return true;
}

OsciWidgetV3::OsciWidgetV3( QWidget* p )
	: QWidget( p )
	, _layout( new QVBoxLayout( this ) )
	, _options( new OsciWidgetV3_Options( this ) )
	, _chainmember( new OsciWidgetV3_ChainMember( this ) )
{
	connect( _chainmember, SIGNAL( data( NSCore::AcqisitionDataSet ) ), this, SLOT( dataAvailable( NSCore::AcqisitionDataSet ) ) );

	setMinimumSize( 100,100 );
	_layout->addWidget( _options, 1 );
	connect( _options->spinboxChannels, SIGNAL( valueChanged( int ) ), this, SLOT( setChannels( int ) ) );
	setChannels( _options->spinboxChannels->value() );

	setObjectName( "Oscilloscope" );
	setWindowIcon( QIcon( ":/icons/cr48-action-osci.png" ) );
}
OsciWidgetV3::~OsciWidgetV3() {
}

void OsciWidgetV3::dataAvailable( NSCore::AcqisitionDataSet data ) {
	//qDebug() << "OsciWidgetV3::dataAvailable( data )";
	for ( int i=0; i<_oscis.size(); ++i ) {
		_oscis[ i ]->data( data );
	}
}

void OsciWidgetV3::setChannels( int n ) {
	while ( _oscis.size() > n ) {
		delete _oscis.takeLast();
	}
	while ( _oscis.size() < n ) {
		_oscis.push_back( new OsciWidgetV3_Osci( this ) );
		_layout->addWidget( _oscis.last(), 255 );
		_oscis.last()->setChannel( _oscis.size()-1 );
	}
}

void OsciWidgetV3::forceMinMaxUsage( bool n ) {
	foreach ( OsciWidgetV3_Osci* osci, _oscis ) {
		osci->forceMinMaxUsage( n );
	}
}

NSCore::DataSetChainMember* OsciWidgetV3::chainMember() { return _chainmember; }



OsciWidgetV3_Osci::OsciWidgetV3_Osci( QWidget* p )
	: QWidget( p )
	, fullscale( 0 )
	, offset( 0 )
	, _channel( 0 )
	, _generation( 0 )
	, _mutex( new QMutex() )
	, _forceminmax( false )
{
	QAction* channelselect = new QAction( "Select Channel...", this );
	connect( channelselect, SIGNAL( triggered() ), this, SLOT( enterChannel() ) );
	addAction( channelselect );
	QAction* generationselect = new QAction( "Select Generation...", this );
	connect( generationselect, SIGNAL( triggered() ), this, SLOT( enterGeneration() ) );
	addAction( generationselect );

	x_actions = new QMenu( "Horizontal Data", this );
	addAction( x_actions->menuAction() );
	y_actions = new QMenu( "Vertical Data", this );
	addAction( y_actions->menuAction() );

	QAction* unselect = new QAction( "Unselect all Results", this );
	connect( unselect, SIGNAL( triggered() ), this, SLOT( uncheckAllResults() ) );
	addAction( unselect );

	QAction* forceminmax = new QAction( "Force usage of Min/Max for display", this );
	forceminmax->setCheckable( true );
	connect( forceminmax, SIGNAL( triggered( bool ) ), this, SLOT( forceMinMaxUsage( bool ) ) );
	addAction( forceminmax );

	setBackgroundRole( QPalette::Base );
	setAutoFillBackground( true );
}
OsciWidgetV3_Osci::~OsciWidgetV3_Osci() {
	delete _mutex;
}

void OsciWidgetV3_Osci::data( NSCore::AcqisitionDataSet data ) {
	if ( ! data.channels() > _channel )
		_channel = data.channels() - 1;

	QMutexLocker locker( _mutex );

	QString chstring = QString( "channel %1" ).arg( _channel );

	int generation = _generation;
	QString prefix = data.getPrefix( generation );
	//if ( data.generations() <= generation )
	//	generation = channel->currentGeneration();

	if ( !_forceminmax && data.getResult( prefix + chstring + "/fullscale" ) != NAN ) {
		fullscale = data.getResult( prefix + chstring + "/fullscale" );
		offset    = data.getResult( prefix + chstring + "/offset" );
	} else {
		double a = data.getResult( prefix + chstring + "/max" );
		double b = data.getResult( prefix + chstring + "/min" );
		//qDebug() << "max =" << a << " min =" << b;
		fullscale = a - b;
		offset = fullscale / -2.0 - b;
	}
	//qDebug() << "Fullscale =" << fullscale << " With Offset =" << offset;

	_data.clear();
	double* in = data.getRawChannel( _channel, generation );
	for ( int i=0; i<data.samples( generation ); i++ )
		_data.push_back( in[i] );

	// Get the list of results
	QStringList newresults = data.getResultNames();
	newresults = newresults.filter( chstring );
	//newresults.removeDuplicates();
	//qDebug() << " New list of available results:" << newresults;
	foreach ( QString result, newresults.filter(" at") ) {
		if ( ! _actions_x.contains( result ) ) {
			QAction* tmp = new QAction( result, x_actions );
			tmp->setCheckable( true );
			if ( result.contains("min") || result.contains("max") )
				tmp->setChecked( true );
			x_actions->addAction( tmp );
			_actions_x[result] = tmp;
		}
	}
	foreach ( QString result, newresults ) {
		if ( ! result.contains(" at") && ! _actions_y.contains( result ) ) {
			QAction* tmp = new QAction( result, y_actions );
			tmp->setCheckable( true );
			if ( result.contains("min") || result.contains("max") )
				tmp->setChecked( true );
			y_actions->addAction( tmp );
			_actions_y[result] = tmp;
		}
	}

	// Get the result-values that are to be displayed.
	_addons_x.clear();
	_addons_y.clear();
	foreach( QString name, _actions_x.keys() ) {
		if ( _actions_x[ name ]->isChecked() ) {
			double tmp = data.getResult( name );
			if ( tmp != NAN )
				_addons_x[ name ] = tmp;
		} else
			_addons_x.remove( name );
	}
	foreach( QString name, _actions_y.keys() ) {
		if ( _actions_y[ name ]->isChecked() ) {
			double tmp = data.getResult( name );
			if ( tmp != NAN )
				_addons_y[ name ] = tmp;
		} else
			_addons_y.remove( name );
	}

	update();
}

void OsciWidgetV3_Osci::paintEvent( QPaintEvent* ) {
	QMutexLocker locker( _mutex );
	QPainter p( this );
	if (fullscale == 0) {
		p.drawRect( rect().adjusted( 0,0, -1, -1 ) );
		return;
	}

	p.save();

	double dx = width() / double( _data.size() );
	double dy = - height() / fullscale;

	p.translate( 0, height()/2 + dy*offset );

	p.setPen( palette().color(QPalette::Button) );
	// Draw yticks
	double step = fullscale / 100.;
	while ( fabs(step * dy) < 30 )
		step *= 2;
	double min = - fullscale/2 - offset;
	min -= fmod(min, step);
	double max = fullscale/2 - offset;
	max -= fmod(max, step);
	//qDebug() << "fullscale =" << fullscale << " offset =" << offset << " min =" << min << " max =" << max << " step =" << step;
	for ( double y=min; y<max; y += step ) {
		p.drawLine(0, y*dy, width(), y*dy);
		p.drawText(5, y*dy - 1, QString("%1").arg( y ) );
	}

	// Draw xticks
	step = 10;
	while ( fabs(step * dx) < 50 )
		step *= 2;
	for ( double x=step; x<_data.size(); x+=step ) {
		p.drawLine(x*dx, -fullscale*dy, x*dx, fullscale*dy);
		p.save();
		p.translate(x*dx + 10, min*dy - 2);
		p.rotate(-90);
		p.drawText(0, 0, QString("%1").arg(x));
		p.restore();
	}

	p.setPen( palette().color(QPalette::Text) );
	// Draw zeroline
	p.drawLine(0, 0, width(), 0);

	// Draw data
	p.save();
	p.setRenderHints( QPainter::Antialiasing );
	if ( _data.size() > 0 ) {
		QPointF oldpoint( 0, _data[ 0 ] * dy );
		QPointF newpoint;
		for ( int i=1; i<_data.size(); ++i ) {
			newpoint = QPointF(  i*dx, _data[ i ] * dy );
			p.drawLine( oldpoint, newpoint );
			oldpoint = newpoint;
		}
	}
	p.restore();

	p.setPen( palette().color(QPalette::Highlight) );
	// Draw selected results
	{
		QMapIterator<QString,double> it( _addons_x );
		int i=0;
		while ( it.hasNext() ) {
			it.next();
			if ( it.value() != 0 ) {
				int x = int( dx * it.value() );
				p.drawLine( x, int( -fullscale*dy ), x, int( fullscale*dy ) );
				QString text( "%1:%2" );
				text = text.arg( it.key() ).arg( it.value() );
				QColor fill = palette().color( QPalette::Base );
				fill.setAlphaF( 0.75 );
				QRect rect = p.fontMetrics().boundingRect(text).translated( x+10, 30 + i*20 ).adjusted( -1, -1, 1, 1);
				p.fillRect( rect, fill );
				p.drawText( rect, text );
				++i;
			}
		}
	}
	{
		QMapIterator<QString,double> it( _addons_y );
		while ( it.hasNext() ) {
			it.next();
			if ( it.value() != 0 ) {
				int y = int( dy * it.value() );
				p.drawLine( 0, y, width(), y );
				QString text( "%1: %2" );
				text = text.arg( it.key() ).arg( it.value() );
				QColor fill = palette().color( QPalette::Base );
				fill.setAlphaF( 0.75 );
				QRect rect = p.fontMetrics().boundingRect(text).translated( 10, y-10 ).adjusted( -1, -1, 1, 1);
				p.fillRect( rect, fill );
				p.drawText( rect, text );
			}
		}
	}

	p.restore();
	p.setPen( palette().color(QPalette::Text) );
	p.drawRect( rect().adjusted( 0,0, -1, -1 ) );
}

void OsciWidgetV3_Osci::setChannel( int n ) {
	while ( ! _actions_x.isEmpty() ) delete _actions_x.take( _actions_x.keys().first() );
	while ( ! _actions_y.isEmpty() ) delete _actions_y.take( _actions_y.keys().first() );
	_channel = n;
}

void OsciWidgetV3_Osci::contextMenuEvent( QContextMenuEvent* ev ) {
	QMenu::exec( actions(), ev->globalPos() );
}

void OsciWidgetV3_Osci::enterChannel() {
	setChannel( QInputDialog::getInteger( this,
			"Select the new channel to show",
			"Channel to show:",
			_channel, 0 ) );
}
void OsciWidgetV3_Osci::enterGeneration() {
	_generation = QInputDialog::getInteger( this,
			"Select the new generation to show",
			"Generation to show:",
			_generation, 0 );
}

void OsciWidgetV3_Osci::uncheckAllResults() {
	foreach ( QString key, _actions_x.keys() )
		_actions_x[key]->setChecked( false );
	foreach ( QString key, _actions_y.keys() )
		_actions_y[key]->setChecked( false );
}

void OsciWidgetV3_Osci::forceMinMaxUsage( bool n ) {
	_forceminmax = n;
}


OsciWidgetV3_Options::OsciWidgetV3_Options( QWidget* p )
	: QFrame( p )
{
	setupUi( this );
}

