/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>
			Matthias Nagl <mnagl@uni-goettingen.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "qtreeitemmodeldelegate.h"
#include "qtreeitemmodeldelegate.moc"

#include <QtCore/QDebug>
#include <QtCore/QModelIndex>
#include <QtCore/QEvent>
#include <QtCore/QUrl>
#include <QtGui/QMouseEvent>
#include <QtGui/QPainter>
#include <QtGui/QItemEditorCreatorBase>
#include <QtGui/QDoubleSpinBox>
#include <QtGui/QComboBox>
#include <QtGui/QApplication>
#include <QtGui/QInputDialog>
#include <QtGui/QMessageBox>
#include <QtGui/QPushButton>
#include <QtGui/QToolButton>
#include <QStyleOption>

#include <QtXml/QtXml>

#include "fancylistview.h"
#include "libcore/qtreeitemmodel.h"
#include "qurlwidget.h"
#include "qsciencespinbox.h"

#include <tiaction.h>

class DoubleCreator : public QItemEditorCreatorBase
{
	public:
		QWidget* createWidget( QWidget* p ) const {
			return new QScienceSpinBox( p );
		}
		QByteArray valuePropertyName() const { return "value"; }
};

QTreeItemModelDelegate::QTreeItemModelDelegate( QObject* p ) : QStyledItemDelegate( p ) {
	QItemEditorFactory* factory = new QItemEditorFactory();
	factory->registerEditor( QVariant::Double, new DoubleCreator() );
	setItemEditorFactory( factory );
}

QStyleOptionButton QTreeItemModelDelegate::initStyleOptionButton(const QStyleOptionViewItem& option, const QModelIndex &index) const {
	QStyleOptionViewItemV4 opt(option);
	QStyleOptionButton buttonstyle;
	buttonstyle.rect = opt.rect;
	buttonstyle.palette = opt.palette;
	if (index.data(NSCore::QTreeItemModel::MouseActionStateRole).toBool())
		buttonstyle.state = QStyle::State_Sunken | QStyle::State_Enabled | QStyle::State_Active;
	else
		buttonstyle.state = QStyle::State_Enabled;
	QDomDocument doc = qVariantValue<QDomDocument>( index.data( Qt::DisplayRole ) );
	QDomNodeList actions = doc.elementsByTagName( "action" );
	/*if ( actions.size() > 0 )
		for ( int i=0; i<actions.size(); ++i )
			buttonstyle.text.append(actions.at( i ).toElement().attribute( "name" ));*/
	if ( actions.size() > 0 ) {
	    buttonstyle.text = actions.at(0).toElement().attribute( "name" );
	    if (actions.at(0).toElement().hasAttribute("icon"))
		buttonstyle.icon = QIcon(actions.at(0).toElement().attribute("icon"));
	}
	return buttonstyle;
}

QStyleOptionToolButton QTreeItemModelDelegate::initStyleOptionToolButton(const QStyleOptionViewItem& option, const QModelIndex &index) const {
	QStyleOptionViewItemV4 opt(option);
	QStyleOptionToolButton buttonstyle;
	buttonstyle.rect = opt.rect;
	buttonstyle.palette = opt.palette;
	buttonstyle.arrowType = Qt::NoArrow;
	//buttonstyle.activeSubControls = QStyle::SC_ToolButton;
	//buttonstyle.subControls = QStyle::SC_ToolButton;
	buttonstyle.features = QStyleOptionToolButton::None;
	if (index.data(NSCore::QTreeItemModel::MouseActionStateRole).toBool())
		buttonstyle.state = QStyle::State_Enabled | QStyle::State_Sunken;
	else
		buttonstyle.state = QStyle::State_Enabled | QStyle::State_Raised;
	buttonstyle.text = "Open File";
	buttonstyle.icon = QIcon(":/icons/kde/cr22-action-fileopen.png");
	int buttonsize = QApplication::style()->pixelMetric(QStyle::PM_ButtonIconSize, &option);
	buttonstyle.iconSize = QSize(buttonsize, buttonsize);
	return buttonstyle;
}

void QTreeItemModelDelegate::initStyleOption(QStyleOptionViewItem *option, const QModelIndex &index) const {
	QStyledItemDelegate::initStyleOption(option, index);
	QStyleOptionViewItemV4 *v4 = qstyleoption_cast<QStyleOptionViewItemV4 *>(option);

	QVariant value = index.data(Qt::DecorationRole);
	if (value.isValid() && !value.isNull() && value.type() == QVariant::String && v4) {
		v4->icon = QIcon(qvariant_cast<QString>(value));
		QIcon::Mode mode;
		if (!(option->state & QStyle::State_Enabled))
		    mode = QIcon::Disabled;
		else if (option->state & QStyle::State_Selected)
		    mode = QIcon::Selected;
		else
		    mode = QIcon::Normal;
		QIcon::State state = option->state & QStyle::State_Open ? QIcon::On : QIcon::Off;
		v4->decorationSize = v4->icon.actualSize(option->decorationSize, mode, state);
	}
	value = index.data(Qt::DisplayRole);
	if (value.isValid() && value.type() == QVariant::Url && v4 && index.column() == 0)
		v4->text = value.toUrl().toString();
}

QSize QTreeItemModelDelegate::sizeHint( const QStyleOptionViewItem& option, const QModelIndex& index ) const {
	Q_ASSERT(index.isValid());

	if (qVariantCanConvert<QDomDocument>( index.data( Qt::DisplayRole ) )) {
		QStyleOptionButton opt = initStyleOptionButton(option, index);
		QPushButton test(opt.icon, opt.text);
		return test.sizeHint();
	}
	else if (index.data(Qt::DisplayRole).type() == QVariant::Url) {
		QStyleOptionToolButton opt = initStyleOptionToolButton(option, index);
		QToolButton test;
		test.setText(opt.text);
		test.setIcon(opt.icon);
		QSize size(QStyledItemDelegate::sizeHint( option, index ));
		size.setWidth(size.width() + QApplication::style()->combinedLayoutSpacing(QSizePolicy::ToolButton, QSizePolicy::LineEdit, Qt::Horizontal) + test.sizeHint().width());
		size.setHeight(qMax(size.height(), test.sizeHint().height()));
		return size;
	}
	else if (index.data(Qt::EditRole).type() == QVariant::Double || index.data(Qt::EditRole).type() == QVariant::Int) {
		if (!spinboxsizebuffer.isValid()) {
			QScienceSpinBox test;
			test.setValue(index.data(Qt::EditRole).toDouble());
			spinboxsizebuffer = QSize(test.sizeHint());
		}
		return spinboxsizebuffer;
	}
	else
		return QStyledItemDelegate::sizeHint( option, index ).expandedTo(QSize(0,25));
}

void QTreeItemModelDelegate::paint( QPainter* p, const QStyleOptionViewItem& option, const QModelIndex& index ) const {
	Q_ASSERT(index.isValid());

	if (qVariantCanConvert<QDomDocument>( index.data( Qt::DisplayRole ) )) {
		QStyleOptionButton tmp = initStyleOptionButton(option, index);
		QApplication::style()->drawControl(QStyle::CE_ItemViewItem, &option, p);
		QApplication::style()->drawControl(QStyle::CE_PushButton, &tmp, p);
	}
	else if (index.data(Qt::DisplayRole).type() == QVariant::Url) {
		QStyleOptionToolButton tmp = initStyleOptionToolButton(option, index);
		QToolButton test;
		test.setText(tmp.text);
		test.setIcon(tmp.icon);
		QApplication::style()->drawControl(QStyle::CE_ItemViewItem, &option, p);
		int offset = test.sizeHint().width();
		tmp.rect.setWidth(offset);
		QApplication::style()->drawComplexControl(QStyle::CC_ToolButton, &tmp, p);
		offset += QApplication::style()->combinedLayoutSpacing(QSizePolicy::ToolButton, QSizePolicy::LineEdit, Qt::Horizontal);
		QStyleOptionViewItemV4 viewopt(option);
		initStyleOption(&viewopt, index);
		viewopt.rect.setX(viewopt.rect.x()+offset);
		viewopt.backgroundBrush = QBrush(Qt::transparent);
		QApplication::style()->drawControl(QStyle::CE_ItemViewItem, &viewopt, p);
	}
	else
		QStyledItemDelegate::paint( p, option, index );
}

QWidget* QTreeItemModelDelegate::createEditor( QWidget* p, const QStyleOptionViewItem& option, const QModelIndex& index ) const {
	if (qVariantCanConvert<QDomDocument>(index.data(Qt::EditRole)))
		return 0;
	else if (index.data(Qt::EditRole).type() == QVariant::StringList)
		return new QComboBox( p );
	else if (index.data(Qt::DisplayRole).type() == QVariant::Url) {
		QUrlWidget *w = new QUrlWidget(p);
		w->setIcon(QIcon(":/icons/kde/cr22-action-fileopen.png"));
		return w;
	}
	else
		return QStyledItemDelegate::createEditor( p, option, index );
}
void QTreeItemModelDelegate::setEditorData( QWidget* e, const QModelIndex& index ) const {
	if (index.data(Qt::EditRole).type() == QVariant::StringList) {
		QComboBox* combo = dynamic_cast<QComboBox*>(e);
		if (combo) {
			QStringList list = index.data(Qt::EditRole).toStringList();
			combo->addItems(list);
			combo->setCurrentIndex(list.indexOf(index.data(Qt::DisplayRole).toString()));
		}
		return;
	}
	if ( index.data(Qt::DisplayRole).type() == QVariant::Url ) {
		QUrlWidget *w = dynamic_cast<QUrlWidget*>(e);
		if (w)
			w->setText(index.data(Qt::EditRole).toString());
		return;
	}
	QStyledItemDelegate::setEditorData( e, index );
}
void QTreeItemModelDelegate::setModelData( QWidget* e, QAbstractItemModel* model, const QModelIndex& index ) const {
	if (index.data(Qt::EditRole).type() == QVariant::StringList) {
		QComboBox* combo = dynamic_cast<QComboBox*>(e);
		if (combo)
			model->setData(index, combo->currentIndex(), Qt::EditRole);
		return;
	}
	if ( index.data(Qt::DisplayRole).type() == QVariant::Url ) {
		QUrlWidget *w = dynamic_cast<QUrlWidget*>( e );
		if (w)
			model->setData(index, w->text(), Qt::EditRole);
		return;
	}
	QStyledItemDelegate::setModelData( e, model, index );
}

bool QTreeItemModelDelegate::editorEvent( QEvent* ev, QAbstractItemModel* model, const QStyleOptionViewItem& option, const QModelIndex& index ) {
	QVariant data = index.data( Qt::EditRole );

	// use base class event handler if field ist not of type DOM or URL
	if ( !index.isValid() || ( index.isValid()
				&& !qVariantCanConvert<QDomDocument>( data )
			//	&& data.type() != QVariant::Url
			//	&& data.type() != QVariant::StringList // not implemented
				)
			)
		return QStyledItemDelegate::editorEvent( ev, model, option, index );

	// all following code handles clicks on DOM, URL, or StringList fields

	QMouseEvent* mev = dynamic_cast<QMouseEvent*>( ev );

	// deactivate DOM-button if mouse cursor is moved out of the button's area while the mousebutton ist pressed
	if (mev && ev->type() == QEvent::MouseMove) {
	    if (option.rect.contains(mev->pos()))
		model->setData(index, false, NSCore::QTreeItemModel::MouseActionStateRole);
	}

	// don't handle events that are not from the mouse or no single click or release events
	if ( !mev || (ev->type() != QEvent::MouseButtonRelease && ev->type() != QEvent::MouseButtonPress) )
		return false;

/*	if ( data.type() == QVariant::Url && mev->type() == QEvent::MouseButtonRelease ) {
		qDebug() << "Handling editing of the url...";
		QRectF rect = option.rect;
		int s = sizeHint( option, index ).height();
		rect.setWidth( s );
		//qDebug() << mev->pos() << rect.contains( mev->pos() );
		if ( rect.contains( mev->pos() ) ) {
			qDebug() << " Show filedialog here...";
			return false;
		} else {
			QStyleOptionViewItem opt = option;
			opt.rect.adjust( 20, 0, 0, 0 );
			return QStyledItemDelegate::editorEvent( ev, model, opt, index );
		}
	}*/

	else if ( qVariantCanConvert<QDomDocument>( data ) && mev->type() == QEvent::MouseButtonPress ) {
	    model->setData(index, true, NSCore::QTreeItemModel::MouseActionStateRole);
	}
	else if ( qVariantCanConvert<QDomDocument>( data ) && mev->type() == QEvent::MouseButtonRelease ) {
		QDomDocument doc = qVariantValue<QDomDocument>( data );
		
		model->setData(index, false, NSCore::QTreeItemModel::MouseActionStateRole);
		
		QDomNodeList actions = doc.elementsByTagName( "action" );

		if ( actions.size() > 0 ) {
			//qDebug() << "TreeItemModelDelegate::editorEvent() with an action!";

			float w = option.rect.width();
			w = w/actions.size();
			QRectF rect = option.rect;
			rect.setWidth( w );

			for ( int i=0; i<actions.size(); ++i ) {
				if ( rect.intersects( QRect( mev->pos(), mev->pos() ) ) ) {
					qDebug() << QString( "Selected action (%1) is \"%2\"" ).arg( i ).arg( actions.at( i ).toElement().attribute( "name" ) );

					QDomElement action = actions.at( i ).toElement();
					qDebug() << " action" << action.attribute( "name" ) << "with" << action.childNodes().size() << "childs.";

					QDomDocument doc( "epos-action" );
					QDomElement call = doc.createElement( "action" );
					call.setAttribute( "name", action.attribute( "name" ) );

					QDomElement argument = action.firstChildElement( "argument" );
					while ( ! argument.isNull() ) {
						qDebug() << " argument type" << argument.attribute( "type" );

						QDomElement arg = doc.createElement( "argument" );
						QDomElement value = doc.createElement( "value" );

						QString title = argument.firstChildElement( "title").firstChild().nodeValue();
						QString label = argument.firstChildElement( "label").firstChild().nodeValue();
						QString defval = argument.firstChildElement( "default" ).firstChild().nodeValue();

						if ( argument.attribute( "type" ) == "string" ) {
							QString tmp = QInputDialog::getText( 0,
									title,
									label,
									QLineEdit::Normal,
									defval );
							if ( tmp.isEmpty() ) return false;
							arg.setAttribute( "type", "string" );
							QDomText text = doc.createTextNode( tmp );
							value.appendChild( text );
						}
						if ( argument.attribute( "type" ) == "double" ) {
							double tmp = QInputDialog::getDouble( 0,
									title,
									label,
									defval.toInt(),
									argument.firstChildElement( "range" ).attribute( "min", "0" ).toInt(),
									argument.firstChildElement( "range" ).attribute( "max", "1000" ).toInt()
									);
							arg.setAttribute( "type", "double" );
							value.appendChild( doc.createTextNode( QString::number( tmp ) ) );
						}
						if ( argument.attribute( "type" ) == "int" ) {
							int tmp = QInputDialog::getInteger( 0,
									title,
									label,
									defval.toInt(),
									argument.firstChildElement( "range" ).attribute( "min", "0" ).toInt(),
									argument.firstChildElement( "range" ).attribute( "max", "1000" ).toInt()
									);
							arg.setAttribute( "type", "int" );
							value.appendChild( doc.createTextNode( QString::number( tmp ) ) );
						}
						if ( argument.attribute( "type" ) == "bool" ) {
							QMessageBox::StandardButton btn = QMessageBox::question( 0,
									title,
									label,
									QMessageBox::Ok | QMessageBox::Cancel
									);
							arg.setAttribute( "type", "bool" );
							value.appendChild( doc.createTextNode( (btn&QMessageBox::Ok)?"true":"false" ) );
						}
						if ( argument.attribute( "type" ) == "listitem" ) {
							qDebug() << " ListItem wanted...";
							FancyListDialog dialog( title, label, 0 );
							QDomElement elem = argument.firstChildElement( "range" ).firstChildElement( "item" );
							while ( ! elem.isNull() ) {
								dialog.addXmlItem( elem );
								elem = elem.nextSiblingElement( "item" );
							}
							if ( dialog.exec() == QDialog::Accepted ) {
								arg.setAttribute( "type", "listitem" );
								value.appendChild( doc.createTextNode( dialog.currentItem ) );
							}
						}
						if ( argument.attribute( "type" ) == "stringlistitem" ) {
							qDebug() << "StringListItem wanted...";
							QStringList values;
							QDomElement elem = argument.firstChildElement( "range" ).firstChildElement( "item" );
							while ( ! elem.isNull() ) {
								values << elem.text();
								elem = elem.nextSiblingElement( "item" );
							}
							bool ok;
							QString ret = QInputDialog::getItem( 0, title, label, values, 0, false, &ok );
							if ( ok ) {
								arg.setAttribute( "type", "stringlistitem" );
								value.appendChild( doc.createTextNode( ret ) );
							}
						}
						arg.appendChild( value );
						call.appendChild( arg );
						argument = argument.nextSiblingElement( "argument" );
					}

					doc.appendChild( call );
					model->setData( index, qVariantFromValue<QDomDocument>( doc ) );
				}
				rect.translate( w, 0 );
			}
		}
		return true;
	}
	return false;
}

