/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "xyspecview.h"
#include "xyspecview.moc"

#include <cmath>
#include <QtCore/QDebug>
#include <QtGui/QPainter>
#include <QtGui/QHBoxLayout>
#include <QtGui/QLabel>
#include <QtGui/QAction>
#include <QtGui/QMenu>
#include <QtGui/QContextMenuEvent>

struct XYSpecView_Private {
	double zmax;
	std::vector< std::vector<double> > data;
	int sizex, sizey;

	void redoCacheFromModel( QAbstractItemModel* model ) {
		if ( model->columnCount() != sizex || model->rowCount() != sizey ) {
			//data.clear();
			sizex = model->columnCount();
			sizey = model->rowCount();
			//data = new double[ sizex ][ sizey ];
			data.resize( sizex );
			for ( int i=0; i<sizex; ++i )
				data[ i ].resize( sizey );
		}
		zmax = 0;
		for ( int x=0; x<sizex; ++x )
			for ( int y=0; y<sizey; ++y ) {
				data[ x ][ y ] = model->data( model->index( y, x ) ).toDouble();
				if ( logarithmicZ && data[ x ][ y ] > 1 )
					data[ x ][ y ] = log( data[ x ][ y ] );
				zmax = qMax( data[ x ][ y ], zmax );
			}
	}

	QColor color( double z ) {
		z = (z/zmax)*(colorsInUse.size()-1);
		for (int i=0; i<colorsInUse.size()-1; ++i) {
			if (z>=i && z<=i+1) {
				QColor c1 = colorsInUse[i];
				QColor c2 = colorsInUse[i+1];
				z = z-i;
				return QColor( c1.red()*(1-z)+c2.red()*z, c1.green()*(1-z)+c2.green()*z, c1.blue()*(1-z)+c2.blue()*z );
			}
		}
		return QColor(100,100,100);
	}

	QMap<QString, QList<QColor> > colors;
	QList<QColor> colorsInUse;
	bool logarithmicZ;
};

XYSpecView::XYSpecView( QWidget* p ) : QAbstractItemView( p ), d( new XYSpecView_Private() ) {
	qDebug() << "XYSpecView::XYSpecView(" << p << ")";
	d->zmax = 0;
	d->sizex = d->sizey = 0;
	d->logarithmicZ = false;

	QAction* toggleLog = new QAction( "Toggle logarithmic Z scale", this );
	toggleLog->setCheckable( true );
	connect( toggleLog, SIGNAL( triggered( bool ) ), this, SLOT( setLogarithmicZ( bool ) ) );
	addAction( toggleLog );

	d->colors["White-Red"] = QList<QColor>() << QColor(255,255,255,255) << QColor(255,0,0);
	d->colors["Black-Blue-Orange"] = QList<QColor>() << QColor(0,0,0) << QColor(0,0,255) << QColor(255,200,0);
	d->colors["Black-Blue-Red-Yellow"] = QList<QColor>() << QColor(0,0,0) << QColor(0,0,255) << QColor(255,0,0) << QColor(255,255,0);
	d->colors["Fire"] = QList<QColor>() << QColor(100,0,0) << QColor(255,0,0) << QColor(255,255,0);
	d->colors["Rainbow"] = QList<QColor>() << QColor(0,0,0,255);
	int steps = 6;
	for ( float i=0; i<steps; i+=1.0 ) d->colors["Rainbow"].append( QColor::fromHsvF( 1.0-i/steps, 1,1 ) );

	QMenu* colorMenu = new QMenu( "Colors", this );
	connect( colorMenu, SIGNAL( triggered( QAction* ) ), this, SLOT( setColor( QAction* ) ) );
	foreach ( QString key, d->colors.keys() ) {
		QPixmap pix( 22, 22 );
		QPainter p( &pix );
		QLinearGradient grad( QPointF(0,22), QPointF(0,0) );
		int s = d->colors[key].size();
		for ( int i=0; i<s; ++i ) {
			grad.setColorAt( i*1.0/(s-1), d->colors[key][i] );
		}
		p.fillRect( pix.rect(), grad );
		QAction* tmp = new QAction( QIcon( pix ), key, this );
		colorMenu->addAction( tmp );
		if ( d->colorsInUse.size() == 0 ) {
			d->colorsInUse = d->colors[ key ];
		}
	}
	addAction( colorMenu->menuAction() );
}
XYSpecView::~XYSpecView() {
	qDebug() << "XYSpecView::~XYSpecView()";
}

void XYSpecView::setLogarithmicZ( bool n ) {
	d->logarithmicZ = n;
	d->redoCacheFromModel( model() );
	update();
}
void XYSpecView::setColor( QAction* a ) {
	d->colorsInUse = d->colors[ a->text() ];
	update();
}

QRect XYSpecView::visualRect( const QModelIndex& ) const {
	return QRect();
}
void XYSpecView::scrollTo( const QModelIndex&, ScrollHint ) {
}
QModelIndex XYSpecView::indexAt( const QPoint& ) const {
	return QModelIndex();
}
void XYSpecView::setModel( QAbstractItemModel* model ) {
	qDebug() << "XYSpecView::setModel(" << model << ")";
	QAbstractItemView::setModel( model );
	d->redoCacheFromModel( model );
	update();
}

QModelIndex XYSpecView::moveCursor( QAbstractItemView::CursorAction, Qt::KeyboardModifiers ) {
	return QModelIndex();
}

int XYSpecView::horizontalOffset() const {
	return 0;
}
int XYSpecView::verticalOffset() const {
	return 0;
}

bool XYSpecView::isIndexHidden( const QModelIndex& ) const {
	return false;
}

void XYSpecView::setSelection( const QRect&, QItemSelectionModel::SelectionFlags ) {
}

QRegion XYSpecView::visualRegionForSelection( const QItemSelection& ) const {
	return QRegion();
}

void XYSpecView::dataChanged( const QModelIndex& _topLeft, const QModelIndex& _bottomRight ) {
	//qDebug() << "XYSpecView::dataChanged(" << _topLeft << "," << _bottomRight << ")";
	for ( int x = _topLeft.column(); x <= _bottomRight.column(); ++x )
		for ( int y = _topLeft.row(); y <= _bottomRight.row(); ++y ) {
			d->data[ x ][ y ] = model()->data( model()->index( y, x ) ).toDouble();
			if ( d->logarithmicZ && d->data[ x ][ y ] > 0 )
				d->data[ x ][ y ] = log( d->data[ x ][ y ] );
			//qDebug() << " saving data (" << x << "," << y << ") =" << d->data[ x ][ y ];
			d->zmax = qMax( d->data[ x ][ y ], d->zmax );
		}
	update();
}
void XYSpecView::paintEvent( QPaintEvent* ev ) {
	//qDebug() << "XYSpecView::paintEvent(" << ev << ")";
	if ( model() == 0 )
		return;

	QPainter p( viewport() );

	p.setRenderHints( QPainter::Antialiasing );

	int xskip = 0;
	double xstep = 0;
	while ( xstep < 1 ) {
		xskip++;
		xstep = 1.0 * viewport()->width() / ( d->data.size() / xskip );
	}
	int yskip = 0;
	double ystep = 0;
	while ( ystep < 1 ) {
		yskip++;
		ystep = 1.0 * viewport()->height() / ( d->data[ 0 ].size() / yskip );
	}
	//qDebug() << "X:" << viewport()->width() << d->data.size() << xskip << xstep;
	//qDebug() << "Y:" << viewport()->height() << d->data[ 0 ].size() << yskip << ystep;

	for ( uint i=0; i<( d->data.size() / xskip ); i++ ) {
		for ( uint j=0; j<( d->data[ i ].size() / yskip ); j++ ) {
			double z = d->data[ i*xskip ][ j*yskip ];
			if ( z != 0 ) {
				p.fillRect( QRectF( i*xstep, j*ystep, xstep, ystep ), d->color( z ) );
			}
		}
	}
}

void XYSpecView::contextMenuEvent( QContextMenuEvent* ev ) {
	QMenu::exec( actions(), ev->globalPos() );
	ev->accept();
}

