/*
    Copyright 2009-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/

#include "netclient.h"
#include "netclient.moc"

#include <QtCore/QStringList>
#include <QtCore/QTime>
#include <QtCore/QCoreApplication>
#include <QtCore/QDebug>

#include <tiparameter.h>
#include <tiaction.h>

using namespace NSNetwork;

class UpdateBlocker
{
	volatile bool* _block;
	public:
		UpdateBlocker( volatile bool* block ) : _block( block ) {
			*_block = true;
		}
		~UpdateBlocker() {
			*_block = false;
		}
};
AbstractNetClient::AbstractNetClient( QIODevice* dev, QObject* p )
	: QObject( p )
	, device( dev )
	, _blockupdates( false )
{
	connect( device, SIGNAL( readyRead() ), this, SLOT( readData() ) );
}
AbstractNetClient::~AbstractNetClient() {
}


void AbstractNetClient::readData() {
	//qDebug() << "AbstractNetClient::readData() _blockupdates =" << _blockupdates;
	int bytesavailable = device->bytesAvailable();
	while ( ((bytesavailable = device->bytesAvailable()) > 0) && device->peek( bytesavailable ).contains('\x3') ) {
		QString line = QString::fromUtf8( device->read( device->peek( bytesavailable ).indexOf('\x3')+1 )).remove(QRegExp("[\x3\n\r]+$"));
		//qDebug() << " line read:" << line;
		if ( line.startsWith( "UPDATE \"" ) ) {
			QString name = line.section( "\"", 1, 1 );
			QString value = line.section( "\"", 3, -2 );
			values[ name ] = value;
			if ( ! _blockupdates ) {
				updateVariable( name, value );
				_blockupdates = false;
			}
		} else if ( line.startsWith( "UPDATE VARIABLES \"" ) ) {
			QString tmp = line.section( " ", 2, -1 );
			while ( tmp.contains( QRegExp("[\x3\n\r]$") ) )
				tmp.chop( 1 );
			if ( tmp != "\"\"" ) {
				values[ "VARIABLES" ] = tmp;
				updateVariable( "VARIABLES", tmp );
			}
		} else if ( line.startsWith( "UPDATE TYPE \"" ) ) {
			//qDebug() << " getting type from:" << line;
			values[ QString("TYPE:%1").arg(line.section("\"", 1, 1)) ] = line.section( "\"", 3, -2 );
		} else {
			answers.append( line );
		}
	}
}

QStringList AbstractNetClient::allvariables() {
	//qDebug() << "AbstractNetClient::allvariables()";
	UpdateBlocker lock( &_blockupdates );
	values.remove( "VARIABLES" );
	device->write( "GET VARIABLES\x3" );
	/*
	   Wait maximum 10 seconds to send the request and another 10 seconds to receive an answer.
	*/
	device->waitForBytesWritten( 10000 );
	device->waitForReadyRead( 10000 );
	return values.take( "VARIABLES" ).remove(QRegExp("^\"")).remove(QRegExp("\"+$")).split("\",\"");
}

QString AbstractNetClient::gettype( const QString& name ) {
	QString tname = QString("TYPE:%1").arg( name );
	UpdateBlocker lock( & _blockupdates );
	//qDebug() << "AbstractNetClient::gettype(" << name << ")";
	device->write( QString( "GET TYPE \"%1\"\x3" ).arg( name ).toUtf8() );
	/*
	   Wait maximum 10 seconds to send the request and another 10 seconds to receive an answer.
	*/
	device->waitForBytesWritten( 5000 );
	device->waitForReadyRead( 10000 );
	QString ret = values.take( tname );
	if ( ret == "" ) ret = "invalid";
	//qDebug() << "AbstractNetClient::gettype() will return" << ret;
	return ret;
}

QString AbstractNetClient::getvariable( const QString& name ) {
	values.remove( name );
	UpdateBlocker lock( & _blockupdates );
	//qDebug() << "AbstractNetClient::getvariable(" << name << ")";
	device->write( QString( "GET \"%1\"\x3" ).arg( name ).toUtf8() );
	/*
	   Wait maximum 10 seconds to send the request and another 10 seconds to receive an answer.
	*/
	device->waitForBytesWritten( 10000 );
	device->waitForReadyRead( 10000 );
	return values.take( name );
}

QString AbstractNetClient::setvariable( const QString& name, const QString& value ) {
	values.remove( name );
	UpdateBlocker lock( & _blockupdates );
	//qDebug() << "AbstractNetClient::setvariable(" << name << "," << value << ")";
	device->write( QString( "SET \"%1\" \"%2\"\x3" ).arg( name ).arg( value ).toUtf8() );
	/*
	   Wait maximum 10 seconds to send the request and another 10 seconds to receive an answer.
	*/
	device->waitForBytesWritten( 10000 );
	device->waitForReadyRead( 10000 );
	return values.take( name );
}

void AbstractNetClient::subscribe( const QString& name ) {
	//qDebug() << "AbstractNetClient::subscribe(" << name << ")";
	device->write( QString( "SUBSCRIBE \"%1\"\x3" ).arg( name ).toUtf8() );
}
void AbstractNetClient::unsubscribe( const QString& name ) {
	device->write( QString( "UNSUBSCRIBE \"%1\"\x3" ).arg( name ).toUtf8() );
	device->waitForBytesWritten( 5000 );
	device->waitForReadyRead( 5000 );
	if ( !answers.removeOne( "OK" ) )
		qWarning() << "AbstractNetClient::unsubscribe() didn't receive OK";
}


QTreeItemNetClient::QTreeItemNetClient( QIODevice* dev, QObject* p )
	: AbstractNetClient( dev, p )
	, tree( 0 )
{
	if ( 0 == QMetaType::type( "QDomDocument" ) ) {
		qRegisterMetaType<QDomDocument>();
	}
	qRegisterMetaTypeStreamOperators<QDomDocument>( "QDomDocument" );
	//qDebug() << "type of QDomDocument:" << QMetaType::type( "QDomDocument" );
	//qDebug() << "QTreeItemNetClient::QTreeItemNetClient(" << dev << "," << p << ")";
	tree = new NSCore::TiParameter( "", QVariant::Invalid, 0 );
	connect( this, SIGNAL(internalUpdateTree(const QStringList&)),
			this, SLOT(updateTree(const QStringList&)), Qt::QueuedConnection );
	subscribe( "VARIABLES" );
}
QTreeItemNetClient::~QTreeItemNetClient() {
	//qDebug() << "QTreeItemNetClient::~QTreeItemNetClient()";
	delete tree;
}

void QTreeItemNetClient::updateVariable( const QString& n, const QString& value ) {
	//qDebug() << "QTreeItemNetClient::updateVariable(" << qPrintable(n) << "," << qPrintable(value) << ")";
	if ( n == "VARIABLES" ) {
		QString tmp = value;
		emit internalUpdateTree( tmp.remove(QRegExp("^\"")).remove(QRegExp("\"$")).split("\",\"") );
		return;
	}
	QString name = n;
	if ( name.startsWith( "/" ) ) name.remove( 0, 1 );
	if ( tree->getNamedChild( name ) ) {
		_blockupdates = true;
		tree->getNamedChild( name )->guiSetValue( value );
		_blockupdates = false;
	}
}

void QTreeItemNetClient::valueChanged() {
	//qDebug() << "QTreeItemNetClient::valueChanged() sender =" << sender();
	if ( _blockupdates )
		return;

	NSCore::TiParameter* item = qobject_cast<NSCore::TiParameter*>( sender() );
	if ( item )
		setvariable( item->path().replace( tree->objectName(), "" ), item->toString() );
}

void QTreeItemNetClient::actionExecuted( QDomDocument doc ) {
	//qDebug() << "QTreeItemNetClient::actionExecuted()";
	NSCore::TiAction* action = qobject_cast<NSCore::TiAction*>( sender() );
	if ( ! action )
		return;
	setvariable( action->path().remove( tree->objectName() ), doc.toString().remove( "\n" ) );
}

void QTreeItemNetClient::updateTree( const QStringList& list ) {
	//qDebug() << "QTreeItemNetClient::updateTree(" << qPrintable( list.join(",") ) << ")";
	QStringList remotepaths = list;
	remotepaths.replaceInStrings(QRegExp("^/"), "");
	remotepaths.removeAll(QString(""));
	//qDebug() << " new remotepaths=" << remotepaths;
	//qDebug() << " old remotepaths=" << _oldremotepaths;
	foreach ( QString path, _oldremotepaths ) {
		if ( ! remotepaths.contains(path) ) {
			tree->getNamedChild( path )->deleteLater();
		}
	}
	_oldremotepaths = remotepaths;
	QStringList toSubscribe;
	while ( ! remotepaths.isEmpty() ) {
		QString name = remotepaths.takeFirst();
		//qDebug() << name;
		//qDebug() << tree->getNamedChild( name );
		if ( ! tree->getNamedChild( name ) ) {
			//qDebug() << " trying the parent" << name.section( "/", 0, -2 );
			//qDebug() << tree->getNamedChild( name.section( "/", 0, -2 ) );
			if ( tree->getNamedChild( name.section( "/", 0, -2 ) ) ) {
				QString typeName = gettype( name );
				int type = QMetaType::type( typeName.toUtf8() );
				//qDebug() << name << " type:" << typeName << type;
				if ( type < QMetaType::User ) {
					//qDebug() << "Creating new item '" << name.section( "/", -1 ) << "'!";
					NSCore::TiParameter* tmp = new NSCore::TiParameter( name.section( "/", -1 ), QVariant( type, (void*) 0 ), tree->getNamedChild( name.section( "/", 0, -2 ) ) );
					tmp->guiSetValue( getvariable( name ) );
					//subscribe( name );
					toSubscribe.append( name );
					connect( tree->getNamedChild( name ), SIGNAL( guiValueChanged() ), this, SLOT( valueChanged() ) );
				} else if ( "QDomDocument" == typeName ) {
					QString tmp = getvariable( name );
					//qDebug() << " action xml:" << tmp;
					QDomDocument doc( "epos-actions" );
					doc.setContent( tmp );
					connect( new NSCore::TiAction( doc, tree->getNamedChild( name.section( "/", 0, -2 ) ) ),
							SIGNAL( activated( QDomDocument ) ),
							this, SLOT( actionExecuted( QDomDocument ) ) );
				}
			} else
				remotepaths.append( name );
		}
	}
	while ( ! toSubscribe.isEmpty() )
		subscribe( toSubscribe.takeFirst() );
}

