/*
    Copyright 2009-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/

#include "netprotocol.h"
#include "netprotocol.moc"

#include <QtCore/QDebug>
#include <QtCore/QStringList>

#include <qtreeitem.h>
#include <tiaction.h>

using namespace NSNetwork;

AbstractNetProtocol::AbstractNetProtocol( QIODevice* dev, QObject* p )
    : QObject( p )
    , device( dev )
    , _status( "Unknown" )
{
    //qDebug() << "AbstractNetProtocol::AbstractNetProtocol(" << dev << "," << p << ")";
    connect( dev, SIGNAL( readyRead() ), this, SLOT( canReadData() ) );
    startTimer( 500 );
}

void AbstractNetProtocol::canReadData() {
    //qDebug() << "AbstractNetProtocol::canReadData()";
    int bytesavailable = 0;
    while ( ( ( bytesavailable = device->bytesAvailable() ) > 0 )
            && device->peek( bytesavailable ).contains( "\x3" ) )
    {
        _updateSent = false;
        QString line = QString::fromUtf8( device->read(device->peek( bytesavailable).indexOf('\x3')+1 ) );
        //qDebug() << "AbstractNetProtocol::canReadData() have a command to read:" << qPrintable(line);
        if ( line.startsWith( "GET SUBSCRIPTIONS" ) ) {
            sendUpdate( "SUBSCRIPTIONS", subscriptions.uniqueKeys() );
        }
        if ( line.startsWith( "GET VARIABLES" ) ) {
            sendUpdate( "VARIABLES", getvariables() );
        }
        if ( line.startsWith( "GET SPECTRA" ) ) {
            sendUpdate( "SPECTRA", _spectra );
        }
        if ( line.startsWith( "GET \"SPECTRUM/" ) ) {
            QString spectrum = line.section( "\"", 1, 1 ).section( "/", 1, -1 );
            QByteArray data;
            emit getSpectrum( spectrum, data );
            //data.replace('\n', "<br />" );
            //data = data.toBase64();
            sendUpdate( QString( "\"SPECTRUM/%1\"" ).arg( spectrum ), data.data() );
            continue;
        }
        if ( line.startsWith( "GET XML" ) ) {
            sendUpdate( "XML", getxml() );
        }
        if ( line.startsWith( "GET STATUS" ) ) {
            sendUpdate( "STATUS", _status );
        }
        if ( line.startsWith( "GET TYPE" ) ) {
            QString variable = line.section( "\"", 1, 1 );
            //qDebug() << " Getting type for" << variable << "which is" << gettype( variable );
            device->write( QString( "UPDATE TYPE \"%1\" \"%2\"\x3" ).arg( variable ).arg( gettype( variable ) ).toUtf8() );
        }
        if ( line.startsWith( "GET \"" ) ) {
            QString variable = line.section( "\"", 1, 1 );
            sendUpdate( "\"" + variable + "\"", getvariable( variable ) );
        }
        if ( line.startsWith( "SET \"" ) ) {
            QString variable = line.section( "\"", 1, 1 );
            QString value = line.section( "\"", 3, -2 );
            setvariable( variable, value );
            sendUpdate( "\"" + variable + "\"", getvariable( variable ) );
        }
        if ( line.startsWith( "SET XML \"" ) ) {
            QString xml = line.section( "\"", 1, -2);
            setxml(xml);
            sendUpdate( "XML", getxml() );
        }
        if ( line.startsWith( "SET STATUS \"" ) ) {
            QString status = line.section("\"", 1, -2);
            emit setStatus(status);
            if ( ! _updateSent )
                sendUpdate( "STATUS", _status );
        }

        if ( line.startsWith( "SUBSCRIBE \"" ) ) {
            QString variable = line.section( "\"", 1, 1 );
            if ( ! subscriptions.contains( variable ) )
                subscriptions.insert( variable, QString() );
            //qDebug() << " Subscribing to" << variable;
            if ( variable == "VARIABLES" ) {
                //qDebug() << " Subscribing to variables-list";
                updateVariablesList();
                continue;
            }
            if ( variable == "SPECTRA" ) {
                sendUpdate( "SPECTRA", _spectra );
                continue;
            }
            sendUpdate( QString( "\"%1\"" ).arg( variable ), getvariable( variable ) );
        }
        if ( line == "UNSUBSCRIBE ALL\x3" ) {
            subscriptions.clear();
            device->write( "OK\x3" );
        }
        if ( line.startsWith( "UNSUBSCRIBE \"" ) ) {
            QString variable = line.section( "\"", 1, 1 );
            while ( subscriptions.contains( variable ) )
                subscriptions.take( variable );
            //subscriptions.removeAll( line.section( "\"", 1, 1 ) );
            device->write( "OK\x3" );
        }
        //qDebug() << line;
    }
}

void AbstractNetProtocol::sendUpdate( const QString& variable, const QString& value ) {
    device->write( QString( "UPDATE %1 \"%2\"\x3" ).arg( variable ).arg( value ).toUtf8() );
    _updateSent = true;
}
void AbstractNetProtocol::sendUpdate( const QString& variable, const QStringList& values ) {
    sendUpdate( variable, values.join("\",\"") );
}

void AbstractNetProtocol::updateVariable( const QString& variable, const QString& value ) {
    //qDebug() << "AbstractNetProtocol::updateVariable(" << variable << "," << value << ")";
    //qDebug() << " subscriptions is" << subscriptions;
    if ( !variable.isEmpty() && (subscriptions.contains( variable ) || variable == "STATUS") ) {
        //qDebug() << "AbstractNetProtocol::updateVariable(" << variable << "," << value << ")";
        //qDebug() << " will send the update.";
        QString var = QString( "\"%1\"" ).arg( variable );
        if (
                variable == "SPECTRA" ||
                variable == "STATUS" ||
                variable == "VARIABLES" ||
                variable == "SUBSCRIPTIONS"
                ) {
            //qDebug() << "Special variable" << variable;
            var = variable;
        }
        sendUpdate( var, value );
    }
}
void AbstractNetProtocol::updateVariablesList() {
    //qDebug() << "AbstractNetProtocol::updateVariablesList()";
    sendUpdate( "VARIABLES", getvariables() );
}
void AbstractNetProtocol::updateStatus( const QString& status ) {
    _status = status;
    updateVariable( "STATUS", _status );
}
void AbstractNetProtocol::setxml( QString newxml ) {
    emit loadXml( newxml );
}
QString AbstractNetProtocol::getxml() {
    QString ret;
    emit saveXml( ret );
    return ret;
}

void AbstractNetProtocol::addToSpectrumList( const QString& spec ) {
    if ( ! _spectra.contains(spec) )
        _spectra.append(spec);
    updateVariable( "SPECTRA", _spectra.join("\",\"") );
}
void AbstractNetProtocol::removeFromSpectrumList( const QString& spec ) {
    _spectra.removeAll(spec);
    updateVariable( "SPECTRA", _spectra.join("\",\"") );
}

void AbstractNetProtocol::timerEvent( QTimerEvent* ) {
    foreach ( QString variable, subscriptions.uniqueKeys() ) {
        if ( !variable.isEmpty() ) {
            QString value = getvariable( variable );
            if ( value != subscriptions[ variable ] ) {
                sendUpdate( QString( "\"%1\"" ).arg( variable ), value );
                subscriptions[ variable ] = value;
            }
        }
    }
}


QTreeItemNetProtocol::QTreeItemNetProtocol( NSCore::QTreeItem* t, QIODevice* io )
    : AbstractNetProtocol( io, io )
    , tree( t )
    , _blockupdates( false )
{
    if ( 0 == QMetaType::type( "QDomDocument" ) ) {
        qRegisterMetaType<QDomDocument>();
    }
    qRegisterMetaTypeStreamOperators<QDomDocument>( "QDomDocument" );
    //qDebug() << "QTreeItemNetProtocol::QTreeItemNetProtocol(" << t << "," << io << ")";
    QStringList names = recursiveVariableNames( tree ).replaceInStrings( QRegExp( "/{2,}" ), "/" );
    //qDebug() << names;
    //tree->debugPrint();
    foreach ( QString n, names ) {
        //qDebug() << n;
        NSCore::QTreeItem* item = tree->getNamedChild(
                n.remove(QRegExp("^/")).remove(QRegExp(QString("^%1").arg(tree->objectName()))).remove(QRegExp("^/")) );
        //qDebug() << n;
        if ( item )
            connectNewChilds( item );
        else {
            qDebug() << "item:" << item;
            //qDebug() << "item->metaObject():" << item->metaObject();
            //qDebug() << "item->metaObject()->indexOfSignal( \"guiValueChanged()\" ):" << item->metaObject()->indexOfSignal( "guiValueChanged()" );
        }
    }
    //qDebug() << "QTreeItemNetProtocol::QTreeItemNetProtocol() finished";
}
QTreeItemNetProtocol::~QTreeItemNetProtocol() {
    //qDebug() << "QTreeItemNetProtocol::~QTreeItemNetProtocol()";
}

QStringList QTreeItemNetProtocol::recursiveVariableNames( NSCore::QTreeItem* item ) {
    QStringList ret;
    ret.append( "" );
    foreach ( NSCore::QTreeItem* child, item->childobjects<NSCore::QTreeItem>() ) {
        ret << recursiveVariableNames( child );
    }
    for ( QStringList::Iterator it=ret.begin(); it!=ret.end(); ++it ) {
        (*it) = QString( "/%1%2" ).arg( item->objectName() ).arg( *it );
    }
    return ret;
}
QStringList QTreeItemNetProtocol::getvariables() {
    //qDebug() << "QTreeItemNetProtocol::getvariables()";
    return recursiveVariableNames( tree ).replaceInStrings( "//", "/" )/*.replaceInStrings( QRegExp( "/$" ), "" )*/;
}
QString QTreeItemNetProtocol::gettype( const QString& n ) {
    NSCore::QTreeItem* item = getItem( n );
    if ( item && item->guiValue().type() != QVariant::Invalid ) {
        //qDebug() << " Will return" << int( item->guiValue().userType() ) << "as" << QMetaType::typeName( item->guiValue().userType() );
        return QMetaType::typeName( item->guiValue().userType() );
    }
    return "invalid";
}
QString QTreeItemNetProtocol::getvariable( const QString& n ) {
    NSCore::QTreeItem* item = getItem( QString(n).replace(QRegExp("^/"), "") );
    if ( item ) {
        if ( item->guiValue().type() != QVariant::UserType )
            return item->guiValue().toString();
        else if ( item->guiValue().userType() == QMetaType::type( "QDomDocument" ) ) {
            QDomDocument doc = qVariantValue<QDomDocument>( item->guiValue() );
            return doc.toString();
        }
    }
    return "";
}
void QTreeItemNetProtocol::setvariable( const QString& n, const QString& value ) {
    NSCore::QTreeItem* item = getItem( n );
    if ( item ) {
        _blockupdates = true;
        item->guiSetValue( QVariant( value ) );
        _blockupdates = false;
    }
}

void QTreeItemNetProtocol::valueChanged() {
    //qDebug() << "QTreeItemNetProtocol::valueChanged() sender =" << sender();
    NSCore::QTreeItem* item = qobject_cast<NSCore::QTreeItem*>( sender() );
    //qDebug() << " is treeitem" << item;
    //qDebug() << " with path" << item->path();
    //qDebug() << " _blockupdates =" << _blockupdates;
    if ( item && ! _blockupdates ) {
        //qDebug() << "QTreeItemNetProtocol::valueChanged() sender =" << sender();
        updateVariable( item->path(), item->guiValue().toString() );
    }
}

NSCore::QTreeItem* QTreeItemNetProtocol::getItem( const QString& n ) {
    QString name = n;
    name = name.replace( QRegExp( "/{2,}" ), "/" ).replace( QRegExp( "^/" ), "" )/*.replace( QRegExp( "/$" ), "" )*/;
    //qDebug() << "QTreeItemNetProtocol::getItem(" << name << ")";
    if ( tree->objectName() != "" && name.startsWith( tree->objectName() ) )
        name = name.section( "/", 1 );
    if ( name == "" )
        return tree;
    NSCore::QTreeItem* item = tree->getNamedChild( name );
    return item;
}

void QTreeItemNetProtocol::connectNewChilds( NSCore::QTreeItem* item ) {
    //qDebug() << "QTreeItemNetProtocol::connectNewChilds(" << item << ")";
    connect( item, SIGNAL( newChildAdded( NSCore::QTreeItem* ) ),
            this, SLOT( connectNewChilds( NSCore::QTreeItem* ) ) );
    connect( item, SIGNAL( newChildAdded( NSCore::QTreeItem* ) ),
            this, SLOT( updateVariablesList() ), Qt::QueuedConnection );
    connect( item, SIGNAL( destroyed() ),
            this, SLOT( updateVariablesList() ), Qt::QueuedConnection );
}

