/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "networkchain.h"
#include "networkchain.moc"

#include "bonjourserviceregister.h"
#include "bonjourservicebrowser.h"
#include "bonjourserviceresolver.h"

using namespace NSNetwork;

NetworkDataReceiver::NetworkDataReceiver( QObject* p )
    : NSCore::DataSetProducer( "NetworkDataReceiver", p )
    , _socket( new QUdpSocket( this ) )
    , _buffer( new QByteArray() )
    , _config( new NSCore::TiParameter( "options", QVariant::Invalid, this ) )
    , _service( 0 )
{
    connect( _config, SIGNAL( guiValueChanged() ), this, SLOT( configChanged() ) );

    _config->addParameter( "Publish in network group", "(default)" );

    _socket->bind();
    _socket->open( QIODevice::ReadOnly );

    configChanged();
    start();
}
NetworkDataReceiver::~NetworkDataReceiver() {
}
void NetworkDataReceiver::configChanged() {
    delete _service;
    _socket->close();
    _socket->bind();
    _socket->open( QIODevice::ReadOnly );
    _service = new BonjourServiceRegister( this );
    BonjourRecord record( "EPOS raw-data receiver", "_x-raw._udp", "" );
    record.txt["network group"] = _config->getParameter( "Publish in network group" )->toString().toLocal8Bit();
    _service->registerService( record, _socket->localPort());
}
void NetworkDataReceiver::start() {
    connect( _socket, SIGNAL( readyRead() ), this, SLOT( readData() ) );
}
void NetworkDataReceiver::stop() {
    disconnect( _socket, SIGNAL( readyRead() ), this, SLOT( readData() ) );
    _buffer->clear();
}
void NetworkDataReceiver::readData() {
    //qDebug() << "NetworkDataReceiver::readData()";
    //qDebug() << " bytes available =" << _socket->bytesAvailable();

    _buffer->append( _socket->readAll() );
    //qDebug() << " _buffer.size() is now" << _buffer->size();

    while ( _buffer->contains( "#BEGIN" ) && _buffer->contains( "END#" ) ) {
        //qDebug() << " buffer contains start and stop markers";
        while ( !_buffer->startsWith( "#B" ) )
            _buffer->remove( 0, 1 );
        _buffer->remove( 0, 6 );

        QDataStream stream( _buffer, QIODevice::ReadOnly );
        int samples, channels;
        stream >> channels >> samples;
        NSCore::AcqisitionDataSet data( channels, samples );
        int s = 0;
        stream >> s;
        for ( int i=0; i<s; ++i ) {
            QString name;
            double value;
            stream >> name >> value;
            data.addResult( name, value );
        }
        double* ch = data.getRawData();
        for ( int i=0; i<data.channels()*data.samples(); ++i )
            stream >> ch[ i ];

        _buffer->remove( 0, _buffer->indexOf( "END#" ) + 4 );

        if ( s < 1 )
            data.updateMinMax();

        enqueue( data );
    }
}



NetworkDataSender::NetworkDataSender( QObject* p )
    : NSCore::DataSetChainMember( "NetworkDataSender", p )
    , _config( new NSCore::TiParameter( "options", QVariant::Invalid, this ) )
    , currentbalancingindex( 0 )
    , _resolver( new BonjourServiceResolver( this ) )
{
    connect( _config, SIGNAL( guiValueChanged() ), this, SLOT( configChanged() ) );

    QString host = "localhost";

    connect( _resolver, SIGNAL( bonjourRecordResolved( BonjourRecord ) ),
            this, SLOT( recordResolved( BonjourRecord ) ) );
    BonjourServiceBrowser* browser = new BonjourServiceBrowser( this );
    connect( browser, SIGNAL(currentBonjourRecordsChanged(const QList<BonjourRecord>&)),
            this, SLOT(currentRecordsChanged(const QList<BonjourRecord>&)));
    browser->browseForServiceType( "_x-raw._udp" );

    _config->addParameter( "Group to send data to", "(default)" );
    //_config->addParameter( "Balance the sending load?", false );
    //_config->getParameter( "Balance the sending load?" )->setToolTip( "<qt>When balancing is activated, each dataset will be sent to only one of the available receivers. Otherwise all datasets are sent to all receivers.</qt>" );
    _config->addParameter( "Send samples", true );
    _config->addParameter( "Send results", true );

    configChanged();
}
NetworkDataSender::~NetworkDataSender() {
    foreach ( QString k, receivers.keys() ) {
        while ( ! receivers[k].isEmpty() ) {
            delete receivers[k].takeFirst();
        }
    }
}

void NetworkDataSender::configChanged() {
    //qDebug() << "NetworkDataSender::configChanged()";
    currentgroup = &( receivers[ _config->getParameter("Group to send data to")->toString() ] );
    _sendSamples = _config->getParameter( "Send samples" )->toBool();
    _sendResults = _config->getParameter( "Send results" )->toBool();
}
void NetworkDataSender::currentRecordsChanged( const QList<BonjourRecord>& list ) {
    //qDebug() << __func__ << "(" << list.size() << ")";
    foreach ( const BonjourRecord& record, list ) {
        //qDebug() << record.serviceName << record.registeredType << record.replyDomain << record.txt;
        _resolver->resolveBonjourRecord( record );
    }
    currentgroup = 0;
    foreach ( QString k, receivers.keys() ) {
        while ( ! receivers[k].isEmpty() ) {
            delete receivers[k].takeFirst();
        }
    }
}
void NetworkDataSender::recordResolved( BonjourRecord record ) {
    //qDebug() << __func__ << "(" << record.host.hostName() << ":" << record.port << ")";
    //qDebug() << " txt:" << record.txt;
    //qDebug() << record.txt["network group"];
    bool exists=false;
    foreach ( Receiver* recv, receivers[ record.txt["network group"] ] )
        if ( recv->record == record )
            exists = true;
    if ( ! exists )
        receivers[ record.txt["network group"] ].append( new Receiver( record ) );
    configChanged();
}

bool NetworkDataSender::dataArrived( NSCore::AcqisitionDataSet data ) {
    //qDebug() << "NetworkDataSender::dataArrived( data )";

    if ( currentgroup == 0 || currentgroup->size() < 1 ) return true;

    QByteArray buffer;

    QDataStream stream( &buffer, QIODevice::WriteOnly );

    int samples = 0;
    if ( _sendSamples )
        samples = data.samples();

    stream << data.channels() << samples;
    if ( _sendResults ) {
        QStringList keys = data.getResultNames();
        stream << keys.size();
        foreach ( QString k, keys ) {
            stream << k << data.getResult( k );
        }
    } else
        stream << 0;

    double* ch = data.getRawData();
    for ( int i=0; i<data.channels()*samples; ++i )
        stream << ch[ i ];

    buffer.prepend( "#BEGIN" );
    buffer.append( "END#" );

    //qDebug() << " will write" << buffer.size();

    //qDebug() << " receiver group has" << currentgroup->size() << "entries";
    if ( currentbalancingindex >= currentgroup->size() ) currentbalancingindex = 0;
    //qDebug() << " will send data to index" << currentbalancingindex;
    if ( currentgroup->at(currentbalancingindex)->socket->isWritable() ) {
        //qDebug() << currentgroup->at(currentbalancingindex)->socket->state();
        while ( buffer.size() > 0 ) {
            const int buffersizetosend = 512;
            currentgroup->at(currentbalancingindex)->socket->write( buffer.left( buffersizetosend ) );
            buffer.remove( 0, buffersizetosend );
        }
        currentbalancingindex++;

    }

    return true;
}

