/*
    Copyright 2005-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef NETWORKCHAIN_H
#define NETWORKCHAIN_H

#include <datasetchain.h>
#include <tiparameter.h>

#include <QtNetwork/QUdpSocket>
#include <bonjourrecord.h>

class QUdpSocket;

/**
 * @brief Networking stuff combined
 *
 * This namespace contains all the stuff needed for networking.
 */
namespace NSNetwork {

    class BonjourServiceRegister;
    class BonjourServiceResolver;

    /**
     * @brief Receive datasets over network.
     *
     * The datasets should be sent by NetworkDataSender which determines its receivers via Zeroconf.
     * Which means that this NetworkReceiver is publishing itself via zeroconf.
     *
     * You can set groups of receivers with the gui-config-option "Publish in network group" which
     * becomes "network group" in the txt field of zeroconf.
     */
    class NetworkDataReceiver : public NSCore::DataSetProducer
    {
        Q_OBJECT
        public:
            NetworkDataReceiver( QObject* =0 );
            ~NetworkDataReceiver();

            void start();
            void stop();
            void abort() { stop(); }

            bool isRunning() { return true; }

        private slots:
            void readData();

            void configChanged();

        private:

            QString debug() const { return "NetworkReceiver"; }

            QDomNode saveToXML( QDomDocument ) { return QDomNode(); }
            bool restoreFromXML( QDomNode ) { return false; }

            QUdpSocket* _socket;

            QByteArray* _buffer;

            NSCore::TiParameter* _config;

            BonjourServiceRegister *_service;
    };

    /**
     * @brief Send datasets across the net
     */
    class NetworkDataSender : public NSCore::DataSetChainMember
    {
        Q_OBJECT
        public:
            NetworkDataSender( QObject* =0 );
            ~NetworkDataSender();

        private slots:
            void configChanged();

            void currentRecordsChanged( const QList<BonjourRecord> &list );
            void recordResolved( BonjourRecord );

        private:

            QString debug() const { return "NetworkDataSender"; }

            QDomNode saveToXML( QDomDocument ) { return QDomNode(); }
            bool restoreFromXML( QDomNode ) { return false; }

            bool dataArrived( NSCore::AcqisitionDataSet );

            NSCore::TiParameter* _config;

            struct Receiver
            {
                BonjourRecord record;
                QUdpSocket* socket;

                Receiver( BonjourRecord r ) : record( r ), socket( new QUdpSocket() ) {
                    socket->connectToHost( record.host.hostName(), record.port, QIODevice::WriteOnly );
                }
                ~Receiver() {
                    delete socket;
                }
            };
            typedef QList<Receiver*> ReceiverList;

            QMap< QString,ReceiverList> receivers;
            ReceiverList* currentgroup;
            int currentbalancingindex;

            bool _sendSamples, _sendResults;

            BonjourServiceResolver* _resolver;
    };

}; // NSNetwork

#endif // NETWORKCHAIN_H
