
#ifndef IOBUFFER_H
#define IOBUFFER_H

#include <QtCore/QIODevice>
#include <QtCore/QByteArray>

class IOBuffer: public QIODevice
{
	public:
		void addToIn( QByteArray b ) {
			in.append( b );
			emit readyRead();
		}
		QByteArray readOut() {
			return out;
		}
		QByteArray readIn() {
			return in;
		}
		qint64 cleanBuffers() {
			qint64 s = in.size();
			in.clear();
			out.clear();
			return s;
		}

		bool isSequential() const { return true; }

		bool canReadLine() const {
			return in.contains("\n") || QIODevice::canReadLine();
		}
		qint64 bytesAvailable() const {
			return in.size() + QIODevice::bytesAvailable();
		}
		qint64 readData( char* d, qint64 s ) {
			//qDebug() << "IOBuffer::readData(" << &d << "," << s << ")";
			qint64 size = qMin( qint64(in.size()), s );
			memcpy( d, in.data(), size );
			in.remove( 0, size );
			return size;
		}

		qint64 writeData( const char* d, qint64 s ) {
			qint64 oldsize = out.size();
			out.append( d );
			return out.size() - oldsize;
		}

	private:
		QByteArray in, out;
};

class DebuggingPipe;

class PipeDevice: public QIODevice
{
	Q_OBJECT
	friend class DebuggingPipe;
	public:

		bool isSequential() const { return true; }

		qint64 writeData( const char* d, qint64 s ) {
			//qDebug() << "PipeDevice[" << objectName() << "]::writeData(" << d << "," << s << ")";
			qint64 oldsize = out->size();
			out->append( d );
			qint64 ret = out->size() - oldsize;
			emit bytesWritten( ret );
			return ret;
		}
		qint64 readData( char* d, qint64 s ) {
			//qDebug() << "PipeDevice[" << objectName() << "]::readData( char*," << s << ")";
			qint64 size = qMin( qint64(in->size()), s );
			memcpy( d, in->data(), size );
			in->remove( 0, size );
			return size;
		}
		qint64 bytesAvailable() const {
			//qDebug() << "PipeDevice[" << objectName() << "]::bytesAvailable() will return" << in->size() + QIODevice::bytesAvailable();
			return in->size() + QIODevice::bytesAvailable();
		}
		bool canReadLine() const {
			return in->contains( "\n" ) || QIODevice::canReadLine();
		}
	public slots:
		void dataArrived() {
			//qDebug() << "PipeDevice[" << objectName() << "]::dataArrived() have " << bytesAvailable() << "bytes";
			emit readyRead();
		}
	private:
		PipeDevice( QByteArray* i, QByteArray* o, QObject* p )
			: QIODevice( p )
			, in( i )
			, out( o )
		{
		}

		QByteArray *in, *out;
};

class DebuggingPipe: public QObject
{
	//Q_OBJECT
	public:
		DebuggingPipe( QObject* p =0 ) : QObject( p ) {
			b1 = new QByteArray();
			b2 = new QByteArray();
			one = new PipeDevice( b1, b2, this );
			one->open( QIODevice::ReadWrite );
			two = new PipeDevice( b2, b1, this );
			two->open( QIODevice::ReadWrite );
			connect( one, SIGNAL( bytesWritten( qint64 ) ), two, SLOT( dataArrived() ) );
			connect( two, SIGNAL( bytesWritten( qint64 ) ), one, SLOT( dataArrived() ) );
			one->setObjectName("one");
			two->setObjectName("two");
		}

		QIODevice* first() const { return one; }
		QIODevice* second() const { return two; }

	private:
		PipeDevice *one, *two;
		QByteArray *b1, *b2;
};

#endif
