/*
    Copyright 2005-2008 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "lowpass.h"

#include <cmath>

LowPass::LowPass( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	_config->addParameter( "Filtering Order", 2 );
	_config->addParameter( "Cutoff Frequency", 0.2 );
	_config[ "Cutoff Frequency" ]->setToolTip( "Filter frequency with respect to the samplingrate. Real frequency is cutoff * samplingrate." );
	_config[ "Cutoff Frequency" ]->setRange( 0.0, 0.5 );
}
void LowPass::configChanged() {
	_cache.ns = _config[ "Filtering Order" ]->toInt();
	_cache.freq = _config[ "Cutoff Frequency" ]->toDouble();

	while ( _cache.coef.size() > 0 )
		_cache.coef.pop_back();
	_cache.coef = lpdes( _cache.freq, 1, _cache.ns );

	_cache.samples = 0;
}

void LowPass::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	if ( data.samples() != _cache.samples ) {
		qDebug() << " Redoing the cache...";

		_cache.samples = data.samples();

		if ( _cache.buffers.size() > 0 ) {
			_cache.buffers.pop_back();
			while ( _cache.buffers.size() > 1 ) {
				delete[] _cache.buffers.at( _cache.buffers.size() - 1 );
				_cache.buffers.pop_back();
			}
			_cache.buffers.pop_back();
		}
		//qDebug() << _cache.buffers.size();
		_cache.buffers.push_back( 0 );
		while ( signed( _cache.buffers.size() ) < _cache.ns )
			_cache.buffers.push_back( new double[ data.samples() ] );
		_cache.buffers.push_back( 0 );
		//qDebug() << _cache.buffers.size();

		//qDebug() << "  Cache done.";
	}

	for ( int i=0; i<data.channels(); ++i ) {
		_cache.buffers.at( 0 ) = data.getRawChannel( i, oldgeneration );
		_cache.buffers.at( _cache.buffers.size()-1 ) = data.getRawChannel( i );

		for ( uint k=0; k<_cache.coef.size(); ++k ) {

			double* in = _cache.buffers.at( k );
			double* out = _cache.buffers.at( k+1 );

			for ( int j=0; j<3; ++j )
				_cache.v[ j ] = in[ 0 ];

			out[ 0 ] = _cache.coef.at( k ).a * weighted_v( _cache.v ) - _cache.coef.at( k ).b * in[ 0 ] - _cache.coef.at( k ).c * in[ 0 ];
			out[ 1 ] = _cache.coef.at( k ).a * weighted_v( _cache.v ) - _cache.coef.at( k ).b * out[ 0 ] - _cache.coef.at( k ).c * in[ 0 ];
			for ( int j=2; j<data.samples()-1; ++j ) {
				out[ j ] = _cache.coef.at( k ).a * weighted_v( _cache.v ) - _cache.coef.at( k ).b * out[ j-1 ] - _cache.coef.at( k ).c * out[ j-2 ];
				_cache.push_v( in[ j+1 ] );
			}
			out[ data.samples()-1 ] = _cache.coef.at( k ).a * weighted_v( _cache.v ) - _cache.coef.at( k ).b * out[ data.samples()-2 ] - _cache.coef.at( k ).c * out[ data.samples()-3 ];

		}
	}

	data.updateMinMax();
}

std::vector<triple> LowPass::lpdes( double fc, double t, int ns ) {
	std::vector<triple> ret;
	double wcp = tan( fc * t * M_PI );
	for ( int i=1; i<=ns; ++i ) {
		double cs = cos( ( 2 * ( i+ns ) - 1 ) * M_PI/( 4*ns ) );
		double x = 1 / ( 1 + wcp*wcp - 2 * wcp * cs );
		ret.push_back( triple( wcp*wcp*x, 2 * ( wcp*wcp - 1 ) * x, ( 1 + wcp*wcp + 2 * wcp * cs ) * x ) );
	}
	return ret;
}

HighPass::HighPass( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	_config->addParameter( "Filtering Order", 2 );
	_config->addParameter( "Cutoff Frequency", 0.2 );
	_config[ "Cutoff Frequency" ]->setToolTip( "Filter frequency with respect to the samplingrate. Real frequency is cutoff * samplingrate." );
	_config[ "Cutoff Frequency" ]->setRange( 0.0, 0.5 );
}
void HighPass::configChanged() {
	_cache.ns = _config[ "Filtering Order" ]->toInt();
	_cache.freq = _config[ "Cutoff Frequency" ]->toDouble();

	while ( _cache.coef.size() > 0 )
		_cache.coef.pop_back();
	_cache.coef = hpdes( _cache.freq, 1, _cache.ns );

	_cache.samples = 0;
}

void HighPass::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	if ( data.samples() != _cache.samples ) {
		qDebug() << " Redoing the cache...";

		_cache.samples = data.samples();

		if ( _cache.buffers.size() > 0 ) {
			_cache.buffers.pop_back();
			while ( _cache.buffers.size() > 1 ) {
				delete[] _cache.buffers.at( _cache.buffers.size() - 1 );
				_cache.buffers.pop_back();
			}
			_cache.buffers.pop_back();
		}
		//qDebug() << _cache.buffers.size();
		_cache.buffers.push_back( 0 );
		while ( signed( _cache.buffers.size() ) < _cache.ns )
			_cache.buffers.push_back( new double[ data.samples() ] );
		_cache.buffers.push_back( 0 );
		//qDebug() << _cache.buffers.size();

		//qDebug() << "  Cache done.";
	}

	for ( int i=0; i<data.channels(); ++i ) {
		_cache.buffers.at( 0 ) = data.getRawChannel( i, oldgeneration );
		_cache.buffers.at( _cache.buffers.size()-1 ) = data.getRawChannel( i );

		for ( uint k=0; k<_cache.coef.size(); ++k ) {

			double* in = _cache.buffers.at( k );
			double* out = _cache.buffers.at( k+1 );

			for ( int j=0; j<3; ++j )
				_cache.v[ j ] = in[ 0 ];

			out[ 0 ] = _cache.coef.at( k ).a * weighted_v( _cache.v ) - _cache.coef.at( k ).b * in[ 0 ] - _cache.coef.at( k ).c * in[ 0 ];
			out[ 1 ] = _cache.coef.at( k ).a * weighted_v( _cache.v ) - _cache.coef.at( k ).b * out[ 0 ] - _cache.coef.at( k ).c * in[ 0 ];
			for ( int j=2; j<data.samples()-1; ++j ) {
				out[ j ] = _cache.coef.at( k ).a * weighted_v( _cache.v ) - _cache.coef.at( k ).b * out[ j-1 ] - _cache.coef.at( k ).c * out[ j-2 ];
				_cache.push_v( in[ j+1 ] );
			}
			out[ data.samples()-1 ] = _cache.coef.at( k ).a * weighted_v( _cache.v ) - _cache.coef.at( k ).b * out[ data.samples()-2 ] - _cache.coef.at( k ).c * out[ data.samples()-3 ];

		}
	}

	data.updateMinMax();
}

std::vector<triple> HighPass::hpdes( double fc, double t, int ns ) {
	std::vector<triple> ret;
	double wcp = tan( fc * t * M_PI );
	for ( int i=1; i<=ns; ++i ) {
		double cs = cos( ( 2 * ( i+ns ) - 1 ) * M_PI/( 4*ns ) );
		double x = 1 / ( 1 + wcp*wcp - 2 * wcp * cs );
		ret.push_back( triple( x, 2 * ( wcp*wcp - 1 ) * x, ( 1 + wcp*wcp + 2 * wcp * cs ) * x ) );
	}
	return ret;
}

