/*
    Copyright 2005-2011 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "analyzers.h"
#include <math.h>

InterpolatedMinimumAnalyzerLagrange::InterpolatedMinimumAnalyzerLagrange( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {
	//_config->addParameter( "last minimum", 0.0 );
	//_config->addParameter( "last minimum at", QVariant( 0 ) );
}

void InterpolatedMinimumAnalyzerLagrange::analyze( NSCore::AcqisitionDataSet& data ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double min = 1000;
		int min_at = 0;
		double* ch = data.getRawChannel( i );
		// Suche nach minimum
		for ( int j=5; j<data.samples()-4; ++j ) {
			if ( ch[ j ] < min ) {
				min = ch[ j ];
				min_at = j;
			}
		}
		/// minimum nach 3er Ordnung Lagrange
		double interpolated_min = min_interpolated3( min_at, ch );
		double interpolated_min_at = p( interpolated_min, min_at-1, ch );
		data.addResult( QString( "channel %1/interpolated minimum (quadric) at" ).arg( i ), interpolated_min );
		data.addResult( QString( "channel %1/interpolated minimum (quadric)" ).arg( i ), interpolated_min_at );
		data.addResult( QString( "channel %1/interpolated minimum at" ).arg( i ), interpolated_min );
		data.addResult( QString( "channel %1/interpolated minimum" ).arg( i ), interpolated_min_at );
	}
}

double InterpolatedMinimumAnalyzerLagrange::L( double x, int k, int x0 ) const {
	double zaehler = 1;
	for ( int i=0; i<3; ++i )
		zaehler = zaehler * pow( x-( x0+i ), i!=k );
	double nenner = 1;
	for ( int i=0; i<3; ++i )
		nenner = nenner * pow( x0+k - ( x0+i ), i!=k );
	return zaehler/nenner;
}
double InterpolatedMinimumAnalyzerLagrange::p( double x, int x0, double* ch ) const {
	double sum=0;
	for ( int i=0; i<3; ++i )
		sum += L( x, i, x0 ) * ch[ x0 + i ];
	return sum;
}

double InterpolatedMinimumAnalyzerLagrange::min_interpolated3( int x_min, double* ch ) const {
	double a = ch[ x_min-1 ] / 2;
	double b = - ch[ x_min ];
	double c = ch[ x_min+1 ] / 2;
	double A = a*( x_min + x_min+1 ) + b*( x_min-1 + x_min+1 ) + c*( x_min-1 + x_min );
	double B = a + b + c;
	return A/( 2*B );
}



ThresholdTriggerAnalyzer::ThresholdTriggerAnalyzer( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ), PolynomInterpolate( 3 ) {
	_config->addParameter( "threshold", (_threshold = -0.1) );
	_config[ "threshold" ]->setMin( -5.0 );
	_config[ "threshold" ]->setMax( 5.0 );
}

void ThresholdTriggerAnalyzer::configChanged() {
	_threshold = _config->getParameter( "threshold" )->toDouble();
}

void ThresholdTriggerAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {
	int samples = data.samples();
	for ( int i=0; i<data.channels(); ++i ) {
		double* ch = data.getRawChannel( i );
		int j=0;
		while ( ch[ j ] < _threshold && j < samples )
			++j;
		while ( ch[ j ] > _threshold && j < samples )
			++j;

		//qDebug() << i << " Found first smaller sample at" << j;

		if ( j+10 < samples ) {

			setData( ch );
			setMiddle( j );

			coord a = coord( j-1, ch[ j-1 ] );
			coord b = coord( j-0.5, polint( j-0.5 ) );
			coord c = coord( j, ch[ j ] );

			//qDebug() << " a=(" << ax << "," << ay << ") b=(" << bx << "," << by << ") c=(" << cx << "," << cy << ")";

			double error = fabs( b.y - _threshold );
			int wc=0;
			while ( error > 0.001 && wc < 100 ) {
				wc++;
				//qDebug() << " a=(" << ax << "," << ay << ") b=(" << bx << "," << by << ") c=(" << cx << "," << cy << ") error =" << error;
				if ( b.y > _threshold )
					a = b;
				else
					c = b;

				b.x = a.x + ( c.x-a.x )/2;
				b.y = polint( b.x );

				error = fabs( b.y - _threshold );
			}
			if ( wc >= 100 )
				qDebug() << "Watchdog-Counter reached maximum...";

			//qDebug() << " a=(" << a.x << "," << a.y << ") b=(" << b.x << "," << b.y << ") c=(" << c.x << "," << c.y << ") error =" << error;

			data.addResult( QString( "channel %1/interpolated threshold at" ).arg( i ), b.x );
			data.addResult( QString( "channel %1/interpolated threshold" ).arg( i ), b.y );
		}

	}
}


CFlinAnalyzer::CFlinAnalyzer( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {
	_config->addParameter( "fraction point", 0.4 );
	_config[ "fraction point" ]->setMin( 0.0 );
	_config[ "fraction point" ]->setMax( 1.0 );
}
CFlinAnalyzer::~CFlinAnalyzer() {
}
void CFlinAnalyzer::configChanged() {
	_fraction_point = _config->getParameter( "fraction point" )->toDouble();
}

void CFlinAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double _minimum = data.getResult( QString( "channel %1/interpolated minimum" ).arg( i ) );
		double _minimum_at = data.getResult( QString( "channel %1/interpolated minimum at" ).arg( i ) );
		if ( _minimum == NAN || _minimum_at == NAN ) {
			qWarning() << "No interpolated minimum present on the dataset.";
			return;
		}

		if ( _minimum_at > 10 && _minimum_at < data.samples() ) {
			double* ch = data.getRawChannel( i );
			double zeroline = data.getResult( QString( "channel %1/zeroline" ).arg( i ) );
			if ( zeroline == NAN ) {
				zeroline = 0;
				for ( int j=1; j<_minimum_at-10; ++j )
					zeroline += ch[ j ];
				zeroline /= _minimum_at-11;
			}

			double _fraction = zeroline - ( zeroline - _minimum ) * _fraction_point;

			// Find the first sample to exceed _fraction
			bool _updown = false;
			if ( ch[ 2 ] > _fraction )
				_updown = true;

			int _fraction_at_int = 0;
			for ( int j=int( _minimum_at )-10; j<int( _minimum_at ) && _fraction_at_int==0; ++j ) {
				if ( _updown && ch[ j ] < _fraction )
					_fraction_at_int = j;
				if ( !_updown && ch[ j ] > _fraction )
					_fraction_at_int = j;
			}
			double y0 = ch[ _fraction_at_int ];
			double ym1 = ch[ _fraction_at_int - 1 ];

			double _fraction_at = ( _fraction - ym1 )/( y0 - ym1 ) + _fraction_at_int - 1;
			data.addResult( QString( "channel %1/interpolated fraction point at" ).arg( i ), _fraction_at );
			data.debugPrint();
			qDebug() << "interpolated fraction point at" << _fraction_at;
		}

	}
}

ZeroLineAnalyzer::ZeroLineAnalyzer( NSCore::QTreeItem* p )
	:NSEvaluation::Analyzer( p )
{
	_config->addParameter( "work on minimum", true );
	_config->getParameter( "work on minimum" )->setToolTip( "Wether to search for a minimum or a maximum" );
	_config->addParameter( "skip samples before the extremum", 20 );
	_config->getParameter( "skip samples before the extremum" )->setToolTip( "Stop the determination of the zeroline some samples before the extremum. So the rising slope isn't included in the zeroline." );
	_config->addParameter( "skip samples at the beginning", 4 );
	_config->getParameter( "skip samples at the beginning" )->setToolTip( "The number of samples skipped at the beginning of the dataset. Sometimes there is noise from the digitizer on it..." );
}
void ZeroLineAnalyzer::configChanged() {
	_negative = _config->getParameter( "work on minimum" )->toBool();
	_samplesbefore = _config->getParameter( "skip samples before the extremum" )->toInt();
	_skipbegin = _config->getParameter( "skip samples at the beginning" )->toInt();
}
void ZeroLineAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {

	int c = data.channels();
	for ( int i=0; i<c; ++i ) {
		double _extremum_at = data.getResult( QString( "channel %1/interpolated minimum at" ).arg( i ) );
		if ( !_negative )
			_extremum_at = data.getResult( QString( "channel %1/interpolated maximum at" ).arg( i ) );
		if ( _extremum_at != NAN && _extremum_at-_samplesbefore > _skipbegin ) {
			double* ch = data.getRawChannel( i );
			double zeroline = 0;
			double zerodeviation = 0;
			for ( int j=_skipbegin; j<_extremum_at-_samplesbefore; ++j )
				zeroline += ch[ j ];
			zeroline /= ( _extremum_at - _samplesbefore - _skipbegin );
			for ( int j=_skipbegin; j<_extremum_at-_samplesbefore; ++j )
				zerodeviation += pow( ch[ j ] - zeroline, 2 );
			zerodeviation = sqrt( zerodeviation / ( _extremum_at - _samplesbefore - _skipbegin ) );

			data.addResult( QString( "channel %1/zeroline" ).arg( i ), zeroline );
			data.addResult( QString( "channel %1/zeroline deviation" ).arg( i ), zerodeviation );
		}/* else {
			qWarning() << "ZeroLineAnalyzer::analyze() didn't find an extremum to work with";
			data.debugPrint();
		}*/
	}
}

OldCFAnalyzer::OldCFAnalyzer( NSCore::QTreeItem* p )
	: NSEvaluation::Analyzer( p )
{
	_config->addParameter( "fraction point", 0.2 );
	_config[ "fraction point" ]->setMin( 0.0 );
	_config[ "fraction point" ]->setMax( 1.0 );
	_config->addParameter( "interpolation order", 3 );
	_config[ "interpolation order" ]->setMin( 1 );
	_config[ "interpolation order" ]->setMax( 5 );
	_config->addParameter( "cf on minimum", true );
	_config->addParameter( "cf before extremum", true );
	_config->addParameter( "precision", 6 );
	_config[ "precision" ]->setMin( 1 );
	_config[ "precision" ]->setMax( 12 );
	_config[ "precision" ]->setToolTip( "The precision of the fraction search in terms of significant numbers." );
}
OldCFAnalyzer::~OldCFAnalyzer() {
}
void OldCFAnalyzer::configChanged() {
	_fraction_point = _config->getParameter( "fraction point" )->toDouble();
	setOrder( _config->getParameter( "interpolation order" )->toInt() );
	_negative = _config->getParameter( "cf on minimum" )->toBool();
	_before = _config->getParameter( "cf before extremum" )->toBool();
	precision = _config[ "precision" ]->toInt();
}

void OldCFAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {

	int c = data.channels();
	for ( int i=0; i<c; ++i ) {
		double _extremum = data.getResult( QString( "channel %1/interpolated minimum" ).arg( i ) );
		double _extremum_at = data.getResult( QString( "channel %1/interpolated minimum at" ).arg( i ) );
		if ( !_negative ) {
			_extremum = data.getResult( QString( "channel %1/interpolated maximum" ).arg( i ) );
			_extremum_at = data.getResult( QString( "channel %1/interpolated maximum at" ).arg( i ) );
		}
		if ( _extremum == NAN || _extremum_at == NAN ) {
			qDebug() << " no extremum found, skipping";
			return;
		}
		int samples = data.samples();

		if ( _extremum_at > 10 && _extremum_at < samples ) {

			double zeroline = 0;
			double zerodeviation = 0;

			double* ch = data.getRawChannel( i );

			if ( data.getResult( QString( "channel %1/zeroline" ).arg( i ) ) ) {
				zeroline = data.getResult( QString( "channel %1/zeroline" ).arg( i ) );
				zerodeviation = data.getResult( QString( "channel %1/zeroline deviation" ).arg( i ) );
			} else {

				double zerocount = fmin( 70, _extremum_at-20 );
				for ( int j=1; j<zerocount; ++j )
					zeroline += ch[ j ];
				zeroline /= zerocount;
				for ( int j=1; j<zerocount; ++j )
					zerodeviation += pow( ch[ j ] - zeroline, 2 );
				zerodeviation = sqrt( zerodeviation / ( zerocount-1.0 ) );

				data.addResult( QString( "channel %1/zeroline" ).arg( i ), zeroline );
				data.addResult( QString( "channel %1/zeroline deviation" ).arg( i ), zerodeviation );
			}

			double _fraction = zeroline - ( zeroline - _extremum ) * _fraction_point;

			int _fraction_at_int = 0;

			for ( int j=int( _extremum_at ); j>0 && j<samples && _fraction_at_int==0; ) {
				if (
					( ch[ j ] > _fraction && _negative )
					||
					( ch[ j ] < _fraction && !_negative )
				)
					_fraction_at_int = j;
				if ( _before )
					--j;
				else
					++j;
			}

			//qDebug() << " _fraction_at_int =" << _fraction_at_int;

			if ( _fraction_at_int > 0 ) {
				setMiddle( _fraction_at_int );
				setData( ch );

				points[ 0 ] = coord( _fraction_at_int-1, polint( _fraction_at_int-1 ) );
				points[ 1 ] = coord( _fraction_at_int, polint( _fraction_at_int ) );
				points[ 2 ] = coord( _fraction_at_int+1, polint( _fraction_at_int+1 ) );

				int watch = 0;
				while ( fabs( points[ 1 ].y - _fraction ) > powf( 10, -precision ) && points[ 0 ].x != points[ 2 ].x && watch < 10000 ) {

					if ( points[ 1 ].y < _fraction && points[ 0 ].y > _fraction )
						points[ 2 ] = points[ 1 ];
					else
						points[ 0 ] = points[ 1 ];
					double newx = points[ 0 ].x + ( points[ 2 ].x - points[ 0 ].x )/2;
					points[ 1 ] = coord( newx, polint( newx ) );

					watch++;
				}

				if ( watch < 10000 ) {
					data.addResult( QString( "channel %1/old interpolated fraction point at" ).arg( i ), points[ 1 ].x );
					data.addResult( QString( "channel %1/old interpolated %2 fraction point at" ).arg( i ).arg( _fraction_point ), points[ 1 ].x );
					double error = fabs( ( points[ 1 ].y - _fraction ) /* / _fraction */ );
					/*if ( error > 2 ) {
						qDebug() << "\tCF-error =" << error;
						qDebug() << "\t\t_fraction =" << _fraction << " points[1].y =" << points[ 1 ].y;
						data.debugPrint();
					}*/
					data.addResult( QString( "channel %1/old cf error" ).arg( i ), error );
				} else
					qDebug() << " Watchdog stopped the interpolation, 10000 iterations seems not enough...";
			}
		}
	}
}

#if 0
CFSplineAnalyzer::CFSplineAnalyzer( NSCore::QTreeItem* p )
	: NSEvaluation::Analyzer( p )
{
	_config->addParameter( "BackupMinimumAnalyzer", QVariant::Invalid );
	_minimumanalyzer = new InterpolatedMinimumAnalyzer( _config->getParameter( "BackupMinimumAnalyzer" ) );

	_config->addParameter( "fraction point", 0.2 );
	_config[ "fraction point" ]->setMin( 0.0 );
	_config[ "fraction point" ]->setMax( 1.0 );
}
CFSplineAnalyzer::~CFSplineAnalyzer() {
	delete _minimumanalyzer;
}

void CFSplineAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {
	double _fraction_point = _config->getParameter( "fraction point" )->toDouble();
	for ( int i=0; i<data.channels(); ++i ) {
		NSCore::ChannelDataSet* ch = data.channel( i );
		QVariant* _min_parameter = ch->getResult( "interpolated minimum" );
		QVariant* _min_at_parameter = ch->getResult( "interpolated minimum at" );
		if ( _min_parameter == 0 || _min_at_parameter == 0 ) {
			qDebug() << " running own Analyzer";
			_minimumanalyzer->analyze( data );
			_min_parameter = ch->getResult( "interpolated minimum" );
			_min_at_parameter = ch->getResult( "interpolated minimum at" );
		}
		double _minimum = ( ( _min_parameter )?_min_parameter->toDouble():0 );
		double _minimum_at = ( ( _min_at_parameter )?_min_at_parameter->toDouble():0 );

		if ( _minimum_at > 10 && _minimum_at < ch->samples() ) {
			double zeroline=0;
			for ( int i=1; i<_minimum_at-10; ++i )
				zeroline += ch->sample( i );
			zeroline /= _minimum_at-11;

			double _fraction = zeroline - ( zeroline - _minimum ) * _fraction_point;

			// Find the first sample to exceed _fraction
			bool _updown = false;
			if ( ch->sample( 2 ) > _fraction )
				_updown = true;

			int _fraction_at_int = 0;
			for ( int i=/*int( _minimum_at )-*/10; i<int( _minimum_at ) && _fraction_at_int==0; ++i ) {
				if ( _updown && ch->sample( i ) < _fraction )
					_fraction_at_int = i;
				if ( !_updown && ch->sample( i ) > _fraction )
					_fraction_at_int = i;
			}

			//setMiddle( _fraction_at_int );
			setData( ch->sampledata(), ch->samples() );
			setDerivates( 1e40, 1e40 );

			double x0,x1,x2, y0,y1,y2;
			x0 = _fraction_at_int-1;
			y0 = splint( _fraction_at_int-1 );
			x1 = _fraction_at_int;
			y1 = splint( _fraction_at_int );
			x2 = _fraction_at_int+1;
			y2 = splint( _fraction_at_int+1 );

			while ( fabs( y1 - _fraction ) > 1E-4 && x0 != x2 ) {
				if ( y1 < _fraction && y0 > _fraction ) {
					x2 = x1;
					y2 = y1;
				} else {
					x0 = x1;
					y0 = y1;
				}
				double newx = x0 + ( x2 - x0 )/2.0;
				x1 = newx;
				y1 = splint( newx );
			}

			//qDebug() << "trying to add a result (int. fraction point at =" << x1 << ")";
			//qDebug() << " to channel" << ch;
			ch->addResult( "interpolated fraction point at", x1 );
			ch->addResult( "interpolated fraction point", y1 );
			ch->addResult( "zeroline", zeroline );
			double error = fabs( ( y1 - _fraction ) /* / _fraction */ );
			if ( error > 2 ) {
				qDebug() << "\tCF-error =" << error;
				qDebug() << "\t\t_fraction =" << _fraction << " points[1].y =" << y1;
				ch->debugPrint();
			}
			ch->addResult( "error", error );
		}
	}
}
//#define AVERAGE 0
double CFSplineAnalyzer::y( int n ) {
#ifdef AVERAGE
	double ret = 0;
	int m = 0;
	if ( n>0 ) {
		ret += _y[ n-1 ];
		m++;
	}
	ret += 2*_y[ n ];
	m+=2;
	if ( n<_n ) {
		ret += _y[ n+1 ];
		m++;
	}
	return ret / m;
#else
	return _y[ n ];
#endif
}
#endif

IntegralAnalyzer::IntegralAnalyzer( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {
	_begin = new NSCore::MathTreeRoot( "Begin", _config );
	_begin->setSingleChild( true );
	_end = new NSCore::MathTreeRoot( "End", _config );
	_end->setSingleChild( true );

	index = 0;
	QString tmp = p->objectName();
	//qDebug() << "IntegralAnalyzer::IntegralAnalyzer() tmp =" << tmp;
	if ( tmp.contains( "-" ) ) {
		//qDebug() << "IntegralAnalyzer::IntegralAnalyzer() tmp.section()" << tmp.section( "-", -1 );
		index = tmp.section( "-", -1 ).toInt() + 1;
	}

	//qDebug() << "IntegralAnalyzer::IntegralAnalyzer() index is" << index;
}
void IntegralAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {
	int begin = _begin->value( data );
	if ( begin<0 || begin>data.samples() ) begin = 0;
	int end = _end->value( data );
	if ( end<begin || end>data.samples() ) end = data.samples();
	//qDebug() << "IntegralAnalyzer::analyze() will integrate from" << begin << "to" << end;
	for ( int i=0; i<data.channels(); ++i ) {
		double* in = data.getRawChannel( i );
		double integral = 0.0;
		for ( int j=begin; j<end; ++j ) {
			integral += in[ j ];
		}
		if ( index == 0 )
			data.addResult( QString( "channel %1/integral" ).arg( i ), integral );
		data.addResult( QString( "channel %1/integral %2" ).arg( i ).arg( index ), integral );
	}
	//data.debugPrint();
}
PerChannelIntegral::PerChannelIntegral( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {
	_config->addParameter( "Skip invalid datasets", false );
	_config->getParameter( "Skip invalid datasets" )->setToolTip( "<dl><dt>true<dd>Sip datasets when the before/after range doesn't completely fit into the available samples. Prevents pulses on the edge from getting a (wrong) energy <dt>false<dd>Calculate the integral regardless whether the full interval fits into the available samples or not. Use this if you want before/after to be -1.");
	_config->addParameter( "Work on minimum", true );
	_config->addParameter( "Samples before the extremum", 10 );
	_config->getParameter( "Samples before the extremum" )->setRange( -1, 2048 );
	_config->getParameter( "Samples before the extremum" )->setToolTip( "-1 means to start at the beginning. This only works if \"Skip invalid datasets\" is set to false." );
	_config->addParameter( "Samples after the extremum", 50 );
	_config->getParameter( "Samples after the extremum" )->setRange( -1, 2048 );
	_config->getParameter( "Samples after the extremum" )->setToolTip( "-1 means to stop at the end of the datasets. This only works if \"Skip invalid datasets\" is set to false." );
}
void PerChannelIntegral::configChanged() {
	skipinvalid = _config->getParameter( "Skip invalid datasets" )->toBool();
	minimum = _config->getParameter( "Work on minimum" )->toBool();
	before = _config->getParameter( "Samples before the extremum" )->toInt();
	if ( before == -1 ) before = 4096;
	after = _config->getParameter( "Samples after the extremum" )->toInt();
	if ( after == -1 ) after = 4096;
}
void PerChannelIntegral::analyze( NSCore::AcqisitionDataSet& data ) {
	int c = data.channels();
	int s = data.samples();
	int extrema[ c ];
	for ( int i=0; i<c; ++i )
		if ( minimum )
			extrema[ i ] = data.getResult( QString( "channel %1/min at" ).arg( i ) );
		else
			extrema[ i ] = data.getResult( QString( "channel %1/max at" ).arg( i ) );
	if ( skipinvalid )
		for ( int i=0; i<c; ++i ) {
			if ( extrema[ i ]-before < 0 || extrema[ i ]+after > s )
				return;
		}
	for ( int i=0; i<c; ++i ) {
		int start = qMax( extrema[ i ]-before, 0 );
		int end = qMin( extrema[ i ]+after, s );
		double integral = 0.0;
		double* d = data.getRawChannel( i );
		for ( int j=start; j<end; ++j )
			integral += d[ j ];
		data.addResult( QString( "channel %1/integral" ).arg( i ), integral );
	}
}

OldDynamicConstantFraction::OldDynamicConstantFraction( NSCore::QTreeItem* p )
	: NSEvaluation::Analyzer( p )
	, PolynomInterpolate( 3 )
{
	_config->addParameter( "weight", 0.01 );
	_tiweight = _config->getParameter( "weight" );
	_config->addParameter( "Skip invalid datasets", false );
	_config->getParameter( "Skip invalid datasets" )->setToolTip( "Skip datasets/channels that don't contain a valid integral beforehand. Otherwise the integral is calculated by this plugin if it is not already there." );
}
void OldDynamicConstantFraction::configChanged() {
	_weight = _tiweight->toDouble();
	_skipinvalid = _config->getParameter( "Skip invalid datasets" )->toBool();
}
void OldDynamicConstantFraction::analyze( NSCore::AcqisitionDataSet& data ) {
	//qDebug() << "OldDynamicConstantFraction::analyze()";
	int samples = data.samples();
	int channels = data.channels();
	for ( int i=0; i<channels; ++i ) {
		double* ch = data.getRawChannel( i );

		double integral = data.getResult( QString( "channel %1/integral" ).arg( i ) );
		if ( isnan( integral ) && _skipinvalid )
			break;
		if ( isnan( integral ) ) {
			integral = 0.0;
			for ( int j=0; j<samples; ++j ) integral += ch[ j ];
		}

		double fraction = integral * _weight;
		//qDebug() << " integral is" << integral << " fraction is" << fraction;

		/*double y0 = NAN, y1 = NAN, x = NAN;
		if ( fraction > data.getResult( QString( "channel %1/min" ).arg( i ) ) ) {
			for ( int j=data.getResult( QString( "channel %1/min at" ).arg( i ) ); j>1; --j ) {
				if ( ch[j] <= fraction && ch[j-1] > fraction ) {
					//qDebug() << "  found the nearest samples at" << j;
					y0 = ch[ j-1 ];
					y1 = ch[ j ];
					x = j;
				}
			}
		}
		if ( ! isnan(x) ) {
			x += ( fraction - y1 ) / ( y0 - y1 );
			data.addResult( QString( "channel %1/dynamic fraction point" ).arg( i ), fraction );
			data.addResult( QString( "channel %1/dynamic fraction point at" ).arg( i ), x );
		}*/
		if ( fraction > data.getResult( QString( "channel %1/min" ).arg( i ) ) ) {
			int fraction_at_int = 0;
			for ( int j=data.getResult( QString( "channel %1/min at" ).arg( i ) ); j>0 && fraction_at_int == 0; --j )
				if ( ch[j] > fraction )
					fraction_at_int = j;
			//qDebug() << " integer_position of the fraction is" << fraction_at_int;
			if ( fraction_at_int != 0 ) {
				setMiddle( fraction_at_int );
				setData( ch );
				points[ 0 ] = coord( fraction_at_int-1, polint( fraction_at_int-1 ) );
				points[ 1 ] = coord( fraction_at_int, polint( fraction_at_int ) );
				points[ 2 ] = coord( fraction_at_int+1, polint( fraction_at_int+1 ) );

				int watch = 0;
				while ( fabs( points[ 1 ].y - fraction ) > powf( 10, -6 ) && points[ 0 ].x != points[ 2 ].x && watch < 10000 ) {

					if ( points[ 1 ].y < fraction && points[ 0 ].y > fraction )
						points[ 2 ] = points[ 1 ];
					else
						points[ 0 ] = points[ 1 ];
					double newx = points[ 0 ].x + ( points[ 2 ].x - points[ 0 ].x )/2;
					points[ 1 ] = coord( newx, polint( newx ) );

					watch++;
				}
				if ( watch < 10000 ) {
					data.addResult( QString( "channel %1/old dynamic fraction point" ).arg( i ), fraction );
					data.addResult( QString( "channel %1/old dynamic fraction point at" ).arg( i ), points[1].x );
				} else
					qDebug() << " Watchdog stopped the interpolation, 10000 iterations seems not enough...";
			}
		}
	}
}

NewResultAnalyzer::NewResultAnalyzer( NSCore::QTreeItem* p ) : NSEvaluation::Analyzer( p ) {
	_config->addParameter( "New Results Name", QVariant::String );
	_math = new NSCore::MathTreeRoot( "Math", _config );
}
void NewResultAnalyzer::analyze( NSCore::AcqisitionDataSet& data ) {
	QString name = _config[ "New Results Name" ]->toString();
	double value = _math->value( data );
	if ( name.isEmpty() || isnan( value ) )
		return;
	data.addResult( name, value );
}
