/*
    Copyright 2009-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef ANALYZER_EXTREMA_H
#define ANALYZER_EXTREMA_H

#include <plugins_interface.h>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_interp.h>
#include <gsl/gsl_min.h>

struct arguments
{
	gsl_interp* interp;
	double* xa;
	double* ya;
};

double interpol( double x, void* params ) {
	struct arguments * args = static_cast<arguments*>( params );
	return gsl_interp_eval( args->interp, args->xa, args->ya, x, 0 );
}

class GSLExtrema : public NSEvaluation::Analyzer
{

	protected:
		gsl_interp* interp;
		double *_xa, *_ya;

		int order;

		gsl_min_fminimizer* minimizer;
		gsl_function F;

		double prec_interval;
		int iter_max;

		bool do_minimum, do_maximum;
		QList<int> skipChannels;

	public:
		GSLExtrema( NSCore::QTreeItem* p )
			: NSEvaluation::Analyzer( p )
			, interp( 0 )
			, _xa( 0 )
			, _ya( 0 )
			, minimizer( 0 )
		{
			//qDebug() << "GSLExtrema::GSLExtrema(" << p << ")";

			prec_interval = pow( 10, -6 );
			iter_max = pow( 10, 7 );

			F.params = 0;

			_config->addParameter( "interpolation order", 4 );
			_config->getParameter( "interpolation order" )->setRange( 1, 10 );
			_config->addParameter( "interpolate the minimum", true );
			_config->addParameter( "interpolate the maximum", true );
			_config->addParameter( "skip channels", "" );

			gsl_set_error_handler_off();
		}
		~GSLExtrema() {
			if ( interp )
				gsl_interp_free( interp );
			if ( minimizer )
				gsl_min_fminimizer_free( minimizer );
			qDebug() << "GSLExtrema::~GSLExtrema()";
		}

		void configChanged() {
			//qDebug() << "GSLExtrema::configChanged()";
			if ( _config->getParameter( "interpolate the minimum" ) != 0 )
				do_minimum = _config->getParameter( "interpolate the minimum" )->toBool();
			if ( _config->getParameter( "interpolate the maximum" ) != 0 )
				do_maximum = _config->getParameter( "interpolate the maximum" )->toBool();
			order = _config->getParameter( "interpolation order" )->toInt();

			QList<int> tmpSkipChannels;
			QString s = _config->getParameter( "skip channels" )->toString();
			foreach( QString part, s.split( ",", QString::SkipEmptyParts ) ) {
				int i = part.toInt();
				tmpSkipChannels << i;
			}
			skipChannels = tmpSkipChannels;

			if ( interp != 0 ) delete interp;
			interp = gsl_interp_alloc( gsl_interp_polynomial, order+1 );
			//qDebug() << "GSLExtrema::configChanged() order=" << order << " interp=" << interp;

			if (_xa != 0 ) delete _xa;
			if (_ya != 0 ) delete _ya;
			_xa = new double[ order+1 ];
			_ya = new double[ order+1 ];

			if ( minimizer != 0 ) delete minimizer;
			minimizer = gsl_min_fminimizer_alloc( gsl_min_fminimizer_goldensection );

			F.function = interpol;
			struct arguments* args = new arguments;
			args->interp = interp;
			args->xa = _xa;
			args->ya = _ya;
			if ( F.params != 0 )
				delete static_cast<struct arguments*>( F.params );
			F.params = args;
		}

		int do_the_work( double* xa, double* ya, int s) {
			gsl_interp_init( interp, xa, ya, order+1 );

			//qDebug() << GSL_FN_EVAL(&F,s-1) << GSL_FN_EVAL(&F,s) << GSL_FN_EVAL(&F,s+1);
			int s0 = s-(order/2);
			int s1 = s+(order/2);
			if ( ya[order/2] == ya[order/2+1] ) {
				s1 += 1;
			}
			//qDebug() << " s0 =" << s0 << " s1 =" << s1;
			if ( GSL_EINVAL != gsl_min_fminimizer_set( minimizer, &F, s, s0, s1 ) ) {
				int j=0;
				for ( ; j<iter_max && gsl_min_test_interval(
							gsl_min_fminimizer_x_lower(minimizer),
							gsl_min_fminimizer_x_upper(minimizer), prec_interval, prec_interval
							) != GSL_SUCCESS; ++j )
					gsl_min_fminimizer_iterate( minimizer );
				return j;
			}/* else
				qDebug() << "gsl_min_fminimizer_set returned" << GSL_EINVAL;*/
			return iter_max;
		}
		void analyze( NSCore::AcqisitionDataSet& data ) {
			int channels = data.channels();
			int samples = data.samples();

			for ( int i=0; i<channels; ++i ) {
				if ( skipChannels.size() > 0 && skipChannels.contains( i ) ) {
					//qDebug() << "skipChannels.size() =" << skipChannels.size() << " and has" << i << "in it?" << skipChannels.contains(i);
					break;
				}
				double* ch = data.getRawChannel( i );

				if ( do_minimum ) {
					int s = data.getResult( QString( "channel %1/min at" ).arg( i ) );
					if ( s < order || s > samples - order ) break;
					for ( int j=0; j<order+1; ++j ) {
						_xa[j] = s - int((order+1)/2) + j;
						_ya[j] = ch[ s - int((order+1)/2) + j ];
					}

					if ( do_the_work( _xa, _ya, s ) >= iter_max )
						if ( GlobalLimiter::the()->show( "GSLExtrema/IterationCount" ) )
							qDebug() << "GSLExtrema: Seems as if" << iter_max << "iterations is not enough to determine the minimum!";
					//qDebug() << "minimum of" << gsl_min_fminimizer_f_minimum(minimizer) << "at" << gsl_min_fminimizer_x_minimum(minimizer);
					data.addResult( QString( "channel %1/interpolated minimum" ).arg(i),
							gsl_min_fminimizer_f_minimum(minimizer) );
					//data.addResult( QString( "channel %1/gsl:interpolated minimum" ).arg(i),
					//		gsl_min_fminimizer_f_minimum(minimizer) );
					data.addResult( QString( "channel %1/interpolated minimum at" ).arg(i),
							gsl_min_fminimizer_x_minimum(minimizer) );
					//data.addResult( QString( "channel %1/gsl:interpolated minimum at" ).arg(i),
					//		gsl_min_fminimizer_x_minimum(minimizer) );
				}
				if ( do_maximum ) {
					int s = data.getResult( QString( "channel %1/max at" ).arg( i ) );
					if ( s < 4 || s > samples - 4 ) break;
					for ( int j=0; j<order+1; ++j ) {
						_xa[j] = s - int((order+1)/2) + j;
						_ya[j] = - ch[ s - int((order+1)/2) + j ];
					}

					if ( do_the_work( _xa, _ya, s ) >= iter_max )
						if ( GlobalLimiter::the()->show( "GSLExtrema/IterationCount" ) )
							qDebug() << "GSLExtrema: Seems as if" << iter_max << "iterations is not enough to determine the maximum!";
					//qDebug() << "maximum of" << gsl_min_fminimizer_f_minimum(minimizer) << "at" << gsl_min_fminimizer_x_minimum(minimizer);
					data.addResult( QString( "channel %1/interpolated maximum" ).arg(i),
							- gsl_min_fminimizer_f_minimum(minimizer) );
					//data.addResult( QString( "channel %1/gsl:interpolated maximum" ).arg(i),
					//		- gsl_min_fminimizer_f_minimum(minimizer) );
					data.addResult( QString( "channel %1/interpolated maximum at" ).arg(i),
							gsl_min_fminimizer_x_minimum(minimizer) );
					//data.addResult( QString( "channel %1/gsl:interpolated maximum at" ).arg(i),
					//		gsl_min_fminimizer_x_minimum(minimizer) );
				}

			}
		}


		QString type() const { return "GSLExtrema"; }
};

class InterpolatedMinimumAnalyzer : public GSLExtrema
{
	public:
		InterpolatedMinimumAnalyzer( NSCore::QTreeItem* p ) : GSLExtrema( p ) {
			delete _config->getParameter( "interpolate the minimum" );
			delete _config->getParameter( "interpolate the maximum" );
			do_minimum = true;
			do_maximum = false;
		}
		QString type() const { return "InterpolatedMinimumAnalyzer"; }
};
class InterpolatedMaximumAnalyzer : public GSLExtrema
{
	public:
		InterpolatedMaximumAnalyzer( NSCore::QTreeItem* p ) : GSLExtrema( p ) {
			delete _config->getParameter( "interpolate the minimum" );
			delete _config->getParameter( "interpolate the maximum" );
			do_maximum = true;
			do_minimum = false;
		}
		QString type() const { return "InterpolatedMaximumAnalyzer"; }
};

#endif // ANALYZER_EXTREMA_H
