/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef ANALYZERS_GSL_CF_H
#define ANALYZERS_GSL_CF_H

#include "plugins_interface.h"

#include "globallimiter.h"

#include <gsl/gsl_errno.h>
#include <gsl/gsl_interp.h>
#include <gsl/gsl_min.h>

struct cf_arguments
{
	gsl_interp* interp;
	double* xa;
	double* ya;
	double cf;
};

double minimize_cf( double x, void* params ) {
	struct cf_arguments* args = static_cast<cf_arguments*>( params );
	double ret = fabs( gsl_interp_eval( args->interp, args->xa, args->ya, x, 0 ) - args->cf );
	//qDebug() << "minimize_cf( x=" << x << ", params ) returning" << ret;
	return ret;
}


class GenericCFAnalyzer : public NSEvaluation::Analyzer
{
	private:
		gsl_interp* interp;
		double *_xa, *_ya;

		gsl_min_fminimizer* minimizer;
		gsl_function F;

	protected:
		int order;

		double prec_interval;
		int iter_max;

		bool _negative;
		bool _before;

	public:
		GenericCFAnalyzer( NSCore::QTreeItem* p )
			: NSEvaluation::Analyzer( p )
			, interp( 0 )
			, _xa( 0 )
			, _ya( 0 )
			, minimizer( 0 )
		{
			prec_interval = pow( 10, -8 );
			iter_max = pow( 10, 7 );

			F.params = 0;

			_config->addParameter( "interpolation order", 4 );
			_config->getParameter( "interpolation order" )->setRange( 1, 10 );
			_config->addParameter( "cf on minimum", true );
			_config->addParameter( "cf before extremum", true );

			gsl_set_error_handler_off();
		}
		~GenericCFAnalyzer() {
			if ( interp )
				gsl_interp_free( interp );
			if ( minimizer )
				gsl_min_fminimizer_free( minimizer );
		}

		void configChanged() {
			//qDebug() << "GSL_CFAnalyzer::configChanged()";
			order = _config->getParameter( "interpolation order" )->toInt();
			if ( _config->getParameter( "cf on minimum" ) )
				_negative = _config->getParameter( "cf on minimum" )->toBool();
			if ( _config->getParameter( "cf before extremum" ) )
				_before = _config->getParameter( "cf before extremum" )->toBool();

			//qDebug() << "order:" << order << _negative << _before;

			if ( interp != 0 ) delete interp;
			interp = gsl_interp_alloc( (order==1?gsl_interp_linear:gsl_interp_polynomial), order+1 );

			if (_xa != 0 ) delete _xa;
			if (_ya != 0 ) delete _ya;
			_xa = new double[ order+1 ];
			_ya = new double[ order+1 ];

			if ( minimizer != 0 ) delete minimizer;
			minimizer = gsl_min_fminimizer_alloc( gsl_min_fminimizer_goldensection );

			F.function = minimize_cf;
			struct cf_arguments* args = new cf_arguments;
			args->interp = interp;
			args->xa = _xa;
			args->ya = _ya;
			if ( F.params != 0 )
				delete static_cast<struct cf_arguments*>( F.params );
			F.params = args;
		}

		virtual double getFractionLevel( NSCore::AcqisitionDataSet& data, int channel ) =0;
		virtual void addResults( NSCore::AcqisitionDataSet& data, int channel, gsl_min_fminimizer* minimizer ) =0;

		void analyze( NSCore::AcqisitionDataSet& data ) {
			//qDebug() << "GenericCFAnalyzer::analyze()";
			int channels = data.channels();
			int samples = data.samples();
			int inc = 1;
			if ( _before )
				inc = -1;

			for ( int i=0; i<channels; ++i ) {
				double* ch = data.getRawChannel( i );

				int s = data.getResult( QString( "channel %1/min at" ).arg(i) );
				if ( !_negative )
					s = data.getResult( QString( "channel %1/max at" ).arg(i) );
				double cf = getFractionLevel( data, i );
				if ( isnan( cf ) && GlobalLimiter::the()->show( "GenericCFAnalyzer/ValidCFPoint" ) )
					qWarning() << "GenericCFAnalyzer::analyze() Didn't get a valid cf-point.";
				if ( isnan( cf ) || isnan( s ) || s < 0 || s >= samples )
					break;

				static_cast<cf_arguments*>( F.params )->cf = cf;

				if ( _negative ) {
					//qDebug( " s = %i, inc = %i, samples = %i", s, inc, samples );
					while ( s > order && s < samples-order && ch[s] < cf )
						s+=inc;
					//qDebug( " s = %i, inc = %i, samples = %i", s, inc, samples );
					if ( fabs( ch[s] - cf ) > fabs( ch[s-inc] - cf ) )
						s-=inc;
					//qDebug( " s = %i, inc = %i, samples = %i", s, inc, samples );
				} else {
					while ( ch[s] > cf && s > order && s < samples-order )
						s+=inc;
					if ( fabs( ch[s] - cf ) > fabs( ch[s-inc] - cf ) )
						s-=inc;
				}
				//qDebug() << s << ":" << ch[s] << " " << s+1 << ":" << ch[s+1];
				if ( s < order || s > samples - order ) {
					qWarning() << "GenericCFAnalyzer::analyze() s has a strange value:" << s;
					break;
				}
				for ( int j=0; j<order+1; ++j ) {
					_xa[j] = s - int((order+1)/2) + j;
					_ya[j] = ch[ s - int((order+1)/2) + j ];
					//qDebug() << " _ya[" << s-int((order+1)/2) + j << "] =" << _ya[j];
				}

				gsl_interp_init( interp, _xa, _ya, order+1 );

				int s0 = s-(order/2);
				int s1 = s+(order/2);
				if ( _ya[order/2] == _ya[order/2+1] )
					s1 += 1;

				if ( GSL_EINVAL != gsl_min_fminimizer_set( minimizer, &F, s, s0, s1 ) ) {
					int j=0;
					for ( ; j<iter_max && gsl_min_test_interval(
								gsl_min_fminimizer_x_lower(minimizer),
								gsl_min_fminimizer_x_upper(minimizer), prec_interval, prec_interval
								) != GSL_SUCCESS; ++j )
						gsl_min_fminimizer_iterate( minimizer );
					if ( j < iter_max )
						addResults( data, i, minimizer );
					else
						if ( GlobalLimiter::the()->show( "GenericCFAnalyzer/IterationCount" ) )
							qWarning() << "GenericCFAnalyzer::analyze() reached maximum iteration count!";
				}
#ifndef NODEBUG
				else {
					qDebug() << "GenericCFAnalyzer::analyze() gsl_min_fminimizer_set returned" <<
                        GSL_EINVAL << "(Not a minimum in [" << s0 << "," << s1 << "]";
					qDebug() << "                             " << s << ":" << ch[s] << "cf =" << cf;
					for ( int j=0; j<order+1; ++j ) {
						qDebug() << "                             _ya[" << s-int((order+1)/2) + j << "] =" << _ya[j];
					}
				}
#endif
			}
		}
};


class CFAnalyzer : public GenericCFAnalyzer
{
	protected:
		double relativecf;

	public:
		CFAnalyzer( NSCore::QTreeItem* p )
			: GenericCFAnalyzer( p )
		{
			_config->addParameter( "fraction point", 0.3 );
			_config->getParameter( "fraction point" )->setRange( 0.0, 1.0 );

		}

		void configChanged() {
			//qDebug() << "GSL_CFAnalyzer::configChanged()";
			relativecf = _config->getParameter( "fraction point" )->toDouble();
			GenericCFAnalyzer::configChanged();
		}

		double getFractionLevel( NSCore::AcqisitionDataSet& data, int channel ) {
			QString extremum = "minimum";
			if ( !_negative )
				extremum = "maximum";
			//qDebug() << "extremum =" << qPrintable(extremum) << " relativecf =" << relativecf;
			double min = data.getResult( QString( "channel %1/interpolated %2" ).arg(channel).arg(extremum) );
			double zero = data.getResult( QString( "channel %1/zeroline" ).arg(channel) );
			double cf = zero + ( min - zero ) * relativecf;
			if ( isnan( cf ) && GlobalLimiter::the()->show("CFAnalyzer") ) {
				qWarning() << "Didn't find a valid cf-point.  min =" << min << "\tzero =" << zero << "\tcf =" << cf;
			}
			//qDebug() << "CFAnalyzer::getFractionLevel() will return" << cf;
			return cf;
		}
		void addResults( NSCore::AcqisitionDataSet& data, int channel, gsl_min_fminimizer* minimizer ) {
			data.addResult( QString( "channel %1/interpolated fraction point at" ).arg( channel ),
					gsl_min_fminimizer_x_minimum( minimizer ) );
			data.addResult( QString( "channel %1/interpolated %2 fraction point at" ).arg( channel ).arg( relativecf ),
					gsl_min_fminimizer_x_minimum( minimizer ) );
		}

		QString type() const { return "CFAnalyzer"; }
};



class DynamicConstantFraction : public GenericCFAnalyzer
{
	protected:
		double _weight, _skipinvalid;
	public:
		DynamicConstantFraction( NSCore::QTreeItem* p )
			: GenericCFAnalyzer( p )
		{
			_config->addParameter( "weight", 0.05 );
			_config->getParameter( "weight" )->setRange( 0.0, 1.0 );
		}

		void configChanged() {
			_weight = _config->getParameter( "weight" )->toDouble();
			GenericCFAnalyzer::configChanged();
		}

		double getFractionLevel( NSCore::AcqisitionDataSet& data, int channel ) {
			double integral = data.getResult( QString( "channel %1/integral" ).arg( channel ) );
			return integral * _weight;
		}

		void addResults( NSCore::AcqisitionDataSet& data, int channel, gsl_min_fminimizer* minimizer ) {
			data.addResult( QString( "channel %1/dynamic %2 fraction point at" ).arg( channel ).arg( _weight ),
					gsl_min_fminimizer_x_minimum( minimizer ) );
			data.addResult( QString( "channel %1/dynamic fraction point at" ).arg( channel ),
					gsl_min_fminimizer_x_minimum( minimizer ) );
		}

		QString type() const { return "DynamicConstantFraction"; }

};


class ThresholdAnalyzer : public GenericCFAnalyzer
{
	protected:
		double _threshold;
	public:
		ThresholdAnalyzer( NSCore::QTreeItem* p )
			: GenericCFAnalyzer( p )
		{
			_config->addParameter( "threshold", _threshold = -0.1 );
			_config->getParameter( "threshold" )->setRange( -100.0, 100.0 );
		}

		void configChanged() {
			_threshold = _config->getParameter( "threshold" )->toDouble();
			GenericCFAnalyzer::configChanged();
		}

		double getFractionLevel( NSCore::AcqisitionDataSet& data, int channel ) {
			if ( data.getResult( QString( "channel %1/min" ).arg(channel) ) < _threshold
					&& data.getResult( QString( "channel %1/max" ).arg(channel) ) > _threshold )
				return _threshold;
			return NAN;
		}

		void addResults( NSCore::AcqisitionDataSet& data, int channel, gsl_min_fminimizer* minimizer ) {
			//qDebug() << "ThresholdAnalyzer::addResults()";
			data.addResult( QString( "channel %1/interpolated threshold at" ).arg( channel ),
					gsl_min_fminimizer_x_minimum( minimizer ) );
			data.addResult( QString( "channel %1/interpolated %2 threshold at" ).arg( channel ).arg( _threshold ),
					gsl_min_fminimizer_x_minimum( minimizer ) );
		}

		QString type() const { return "ThresholdAnalyzer"; }
};

class FirstPhotonTiming : public GenericCFAnalyzer
{
	protected:
		double _noisefactor;
	public:
		FirstPhotonTiming( NSCore::QTreeItem* p ) : GenericCFAnalyzer( p ) {
			delete _config->getParameter( "cf on minimum" );
			delete _config->getParameter( "cf before extremum" );
			_config->addParameter( "noise factor", 1.1 );
			_config->getParameter( "noise factor" )->setRange( 0.0, 10.0 );
		}

		void configChanged() {
			_negative = true;
			_before = true;
			_noisefactor = _config->getParameter( "noise factor" )->toDouble();
			GenericCFAnalyzer::configChanged();
		}

		double getFractionLevel( NSCore::AcqisitionDataSet& data, int channel ) {
			double noisedev = data.getResult( QString( "channel %1/zeroline deviation" ).arg(channel) );
			double noise = data.getResult( QString( "channel %1/zeroline" ).arg(channel) );
			double level = noise - _noisefactor * noisedev;
			//qDebug() << "FirstPhotonTiming::getFractionLevel() will return" << noise - _noisefactor * noisedev;
			if ( level > -0.001 ) level = -0.001;
			data.addResult( QString( "channel %1/first photon level" ).arg( channel ), level );
			return level;
		}

		void addResults( NSCore::AcqisitionDataSet& data, int channel, gsl_min_fminimizer* minimizer ) {
			data.addResult( QString( "channel %1/first photon at" ).arg( channel ),
					gsl_min_fminimizer_x_minimum( minimizer ) );
		}

		QString type() const { return "FirstPhotonTiming"; }
};


#endif // ANALYZERS_GSL_CF_H

