/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "modifiers.h"
//#include "modifiers.moc"

#include <limits>
#include <math.h>
#include <QtCore/QDebug>
#include <analyzers.h>
#include <globallimiter.h>

AbsModifier::AbsModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
}

void AbsModifier::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	//qDebug() << "AbsModifier::modify( data )";
	for ( int i=0; i<data.channels(); ++i ) {
		double* in = data.getRawChannel( i, oldgeneration );
		double* out = data.getRawChannel( i );
		for ( int j=0; j<data.samples(); ++j )
			out[ j ] = fabs( in[ j ] );
	}
	data.updateMinMax();
}

NormalizeModifier::NormalizeModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	for ( int i=0; i<2; ++i ) {
		_config->addParameter( QString( "Channel %1/Height to Normalize" ).arg( i ), 1.0 );
		_config[ QString( "Channel %1/Height to Normalize" ).arg( i ) ]->setMin( -100.0 );
		_config[ QString( "Channel %1/Height to Normalize" ).arg( i ) ]->setMax( 100.0 );
		_config->addParameter( QString( "Channel %1/Offset after Normalize" ).arg( i ), 0.0 );
		_config[ QString( "Channel %1/Offset after Normalize" ).arg( i ) ]->setMin( -100.0 );
		_config[ QString( "Channel %1/Offset after Normalize" ).arg( i ) ]->setMax( 100.0 );
	}
}

void NormalizeModifier::configChanged() {
	for ( int i=0; i<2; i++ ) {
		_height_to_normalize[i] = _config[ QString( "Channel %1/Height to Normalize" ).arg( i%2 ) ]->toDouble();
		_offset_after_normalize[i] = _config[ QString( "Channel %1/Offset after Normalize" ).arg( i%2 ) ]->toDouble();
	}
}

void NormalizeModifier::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double* in = data.getRawChannel( i, oldgeneration );
		double* out = data.getRawChannel( i );

		double factor = 1/_height_to_normalize[i%2];

		for ( int j=0; j<data.samples(); ++j )
			out[ j ] = in[ j ] * factor + _offset_after_normalize[i%2];

		data.addResult( QString( "Normalized by factor (Channel %1)" ).arg( i ), factor );
		data.addResult( QString( "Added Offset (Channel %1)" ).arg( i ), _offset_after_normalize[i%2] );

	}
	data.updateMinMax();
}

void simpleAverager( int neighbours, int samplesize, const double* in, double* out ) {
	int vsize = neighbours*2 + 1;
	double* v = new double[ vsize ];

	for ( int i=0; i<vsize; ++i )
		v[ i ] = in[ i ];

	for ( int i=0; i<neighbours; ++i ) {
		out[ i ] = in[ i ];
		out[ samplesize - i - 1 ] = in[ samplesize - i - 1 ];
	}

	for ( int i=neighbours; i<samplesize-neighbours; ++i ) {
		// For each sample
		double value = 0.0;
		for ( int j=0; j<vsize; ++j )
			value += v[ j ];
		value /= vsize;

		out[ i ] = value;

		for ( int j=0; j<vsize-1; ++j )
			v[ j ] = v[ j+1 ];
		v[ vsize-1 ] = in[ i + neighbours + 1 ];
	}
}
void simpleAveragerV2( int neighbours, int samplesize, const double* in, double* out ) {
	int vsize = neighbours*2 + 1;
	double* v = new double[ vsize ];

	for ( int i=0; i<neighbours; ++i )
		v[ i ] = 0;
	for ( int i=neighbours; i<vsize; ++i )
		v[ i ] = in[ i-neighbours ];

	for ( int i=0; i<samplesize; ++i ) {
		// For each sample
		double value = 0.0;
		for ( int j=0; j<vsize; ++j )
			value += v[ j ];
		value /= vsize;

		out[ i ] = value;

		for ( int j=0; j<vsize-1; ++j )
			v[ j ] = v[ j+1 ];
		if ( i<samplesize-neighbours )
			v[ vsize-1 ] = in[ i + neighbours + 1 ];
		else
			v[ vsize-1 ] = 0.0;
	}
}
void weightedAverager( int neighbours, int samplesize, const double* in, double* out, double* weights ) {
	int vsize = neighbours*2 + 1;
	double* v = new double[ vsize ];
	if ( weights == 0 ) {
		weights = new double[ vsize ];
		for ( int i=0; i<vsize; ++i )
			weights[ i ] = 1.0/vsize;
	}

	for ( int i=0; i<neighbours; ++i )
		v[ i ] = 0;
	for ( int i=neighbours; i<vsize; ++i )
		v[ i ] = in[ i-neighbours ];

	for ( int i=0; i<samplesize; ++i ) {
		// For each sample
		double value = 0.0;
		for ( int j=0; j<vsize; ++j )
			value += weights[ j ] * v[ j ];

		out[ i ] = value;

		for ( int j=0; j<vsize-1; ++j )
			v[ j ] = v[ j+1 ];
		if ( i<samplesize-neighbours )
			v[ vsize-1 ] = in[ i + neighbours + 1 ];
		else
			v[ vsize-1 ] = 0.0;
	}
}

FloatingAverageModifier::FloatingAverageModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	qDebug() << "FloatingAverageModifier::FloatingAverageModifier()";
	_config->addParameter( "neighbours", 1 );
	_config[ "neighbours" ]->setMin( 0 );
	_config[ "neighbours" ]->setMax( 10 );
	_config[ "neighbours" ]->setToolTip( "The number of neighbours to the left and right that is looked at." );
}

void FloatingAverageModifier::configChanged() {
	_neighbours = _config[ "neighbours" ]->toInt();
}

void FloatingAverageModifier::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {

		double *_in = data.getRawChannel( i, oldgeneration );
		double *_out = data.getRawChannel( i );

		int size = _neighbours * 2 + 1;
		double *weights = new double[ size ];
		double s = 0.75;
		double sum = 0;
		for ( int i=0; i<size; i++ ) {
			weights[ i ] = exp( - pow( i-_neighbours, 2 ) / ( 2 * pow( s, 2 ) ) );
			sum += weights[  i ];
		}
		for ( int i=0; i<size; ++i )
			weights[ i ] /= sum;
		weightedAverager( _neighbours, data.samples(), _in, _out, weights );
		delete weights;
	}
	data.updateMinMax();
}

IntegrateModifier::IntegrateModifier( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
}

void IntegrateModifier::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double* in = data.getRawChannel( i, oldgeneration );
		double* out = data.getRawChannel( i );
		double old = 0;
		for ( int j=0; j<data.samples(); ++j ) {
			old += in[ j ];
			out[ j ] = old;
		}
	}
}

PulseIntegrate::PulseIntegrate( NSCore::QTreeItem* p )
	: NSEvaluation::Modifier( p )
{
	_config->addParameter( "samples before", 20 );
	_config->addParameter( "samples after", 30 );

	configChanged();
}
void PulseIntegrate::configChanged() {
	cache.before = _config[ "samples before" ]->toInt();
	cache.after  = _config[ "samples after"  ]->toInt();
}
void PulseIntegrate::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	//qDebug() << "PulseIntegrate::modify(" << &data << "," << oldgeneration << ")";
	bool allpassed = true;
	int c = data.channels( /*oldgeneration*/ );
	int s = data.samples( /*oldgeneration*/ );
	for ( int i=0; i<c; ++i ) {

		int min = data.getResult( QString( "../channel %1/min at" ).arg( i ) );
		if ( min - cache.before < 0 || min + cache.after > s )
			allpassed = false;

		int start = qMax( 1, min - cache.before );
		int stop  = qMin( min + cache.after, s );

		double* in = data.getRawChannel( i, oldgeneration );
		double* out = data.getRawChannel( i );

		if ( in != out )
			memcpy( out, in, sizeof( double ) * data.samples() );

		for ( int j=start; j<s && j<stop; ++j )
			out[j] = out[j-1] + in[j];

		for ( int j=stop; j<s; ++j )
			out[j] = out[j-1] - in[j];

	}
	data.addResult( "integral passed", allpassed?0.0:NAN );
	data.updateMinMax();
}

SignedIntegrate::SignedIntegrate( NSCore::QTreeItem* p )
	:NSEvaluation::Modifier( p )
{
}

void SignedIntegrate::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double* in = data.getRawChannel( i, oldgeneration );
		double* out = data.getRawChannel( i );
		double last = 0; //data.channel( i )->sample( 0, 0 );
		double now = 0, sum = 0;
		//data.channel( i )->setSample( 0, sum );
		for ( int j=0; j<data.samples(); ++j ) {

			now = in[ j ];

			if ( now < last )
				sum -= fabs( now );
			else
				sum += fabs( now );

			out[ j ] = sum;

			last = now;
		}
	}
}

Differentiate::Differentiate( NSCore::QTreeItem* p )
	: NSEvaluation::Modifier( p )
{
}
void Differentiate::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double* olddata = data.getRawChannel( i, oldgeneration );
		double* newdata = data.getRawChannel( i );

		double last = 0;
		for ( int j=0; j<data.samples(); ++j ) {
			double tmp = olddata[ j ];
			newdata[ j ] = tmp - last;
			last = tmp;
		}
	}
	data.updateMinMax();
}

CrossCorrelation::CrossCorrelation( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	_config->addParameter( "middle percentage", 100 );
	_config[ "middle percentage" ]->setMin( 0 );
	_config[ "middle percentage" ]->setMax( 100 );
	_config[ "middle percentage" ]->setToolTip( "The percentage of data that is to be evaluated. Will be aligned in the middle of the dataset..." );
}

void CrossCorrelation::configChanged() {
	_percent = _config->getParameter( "middle percentage" )->toInt() / 100.0;
}

void CrossCorrelation::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	if ( data.channels() < 2 ) {
		qDebug() << "Dataset doesn't contain enough channels.";
		return;
	}

	double firstamplitude = 1;
	double secondamplitude = 1;
	double tmp;
	if ( ( tmp = data.getResult( "../channel 0/interpolated minimum" ) ) != NAN )
		firstamplitude = tmp;
	if ( ( tmp = data.getResult( "../channel 1/interpolated minimum" ) ) != NAN )
		secondamplitude = tmp;

	int ibounds = data.samples() - int( data.samples() * _percent );
	ibounds /= 2;

	double* in1 = data.getRawChannel( 0, oldgeneration );
	double* in2 = data.getRawChannel( 1, oldgeneration );
	double* out = data.getRawChannel( 0 );
	for ( int i=0; i<ibounds; ++i )
		out[ i ] = 2;

	for ( int i=ibounds; i< data.samples() - ibounds; ++i ) {

		int joffset = qMax( data.samples() / 2 - i, 0 );
		int ioffset = qMax( i - data.samples()/2, 0 );
		int jend = data.samples()/2 + i;
		if ( jend > data.samples() )
			jend = data.samples()/2 + data.samples() - i;

		double sum = 0;
		double tmp;
		for ( int j=0; j<jend; ++j ) {
			tmp = in1[ ioffset + j ]/firstamplitude - in2[ joffset + j ]/secondamplitude;
			sum += tmp*tmp;
		}

		out[ i ] = sqrt( sum / jend );
	}

	for ( int i=data.samples() - ibounds; i<data.samples(); ++i )
		out[ i ] = 2;

}

AddAdditiveGaussianWhiteNoise::AddAdditiveGaussianWhiteNoise( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	_config->addParameter( "Sigma", 0.1 );
	_config[ "Sigma" ]->setMin( 0.0 );
	_config[ "Sigma" ]->setToolTip( "The sigma value (in Volts) used for the generation of additive gaussian white noise." );
	rand = gsl_rng_alloc (gsl_rng_taus);
	gsl_rng_set(rand, QDateTime::currentDateTime().toTime_t());
}

AddAdditiveGaussianWhiteNoise::~AddAdditiveGaussianWhiteNoise() {
	gsl_rng_free(rand);
}

void AddAdditiveGaussianWhiteNoise::configChanged() {
	_sigma = _config->getParameter( "Sigma" )->toDouble();
}

void AddAdditiveGaussianWhiteNoise::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {
	for ( int i=0; i<data.channels(); ++i ) {
		double* olddata = data.getRawChannel( i, oldgeneration );
		double* newdata = data.getRawChannel( i );

		for ( int j=0; j<data.samples(); ++j )
			newdata[ j ] = gsl_ran_gaussian(rand, _sigma) + olddata[ j ];
	}
	data.updateMinMax();
}


Quantize::Quantize( NSCore::QTreeItem* p ) : NSEvaluation::Modifier( p ) {
	_config->addParameter( "Fullscale", 5.0 );
	_config[ "Fullscale" ]->setMin( 0.0 );
	_config[ "Fullscale" ]->setToolTip( "The fullscale value of the quantization (largest possible value - smallest possible value)." );
	_config->addParameter( "Offset", 0.0 );
	_config[ "Offset" ]->setToolTip( "Position of the fullscale range's center (fs=5, offset=-2.5 => measurement range: [-5,0])." );
	_config->addParameter( "Resolution", 8 );
	_config[ "Resolution" ]->setMin( 0 );
	_config[ "Resolution" ]->setToolTip( "Number of used bits for the quantization." );

	_config->addParameter( "ENOB Simulation/LSB Moving Average Width", 1 );
	_config[ "ENOB Simulation/LSB Moving Average Width" ]->setMin( 1 );
	_config[ "ENOB Simulation/LSB Moving Average Width" ]->setToolTip( "Width of the moving average used to lowpass filter the LSB in order to simulate the frequency dependency of the ENOB.\nMust be an odd number!" );
	_config->addParameter( "ENOB Simulation/Randomize LSB", false );
	_config[ "ENOB Simulation/Randomize LSB" ]->setToolTip( "If Enabled the LSB is randomized by comparing a flat distributed random number with the moving average." );
	_config->addParameter( "ENOB Simulation/Weight of last sample's LSB result in the calculation of this one's", 0.0 );
	_config[ "ENOB Simulation/Weight of last sample's LSB result in the calculation of this one's" ]->setMin( 0.0 );
	_config[ "ENOB Simulation/Weight of last sample's LSB result in the calculation of this one's" ]->setMax( 1.0 );
	_config[ "ENOB Simulation/Weight of last sample's LSB result in the calculation of this one's" ]->setToolTip( "This sample's result is computed as unit step ( this setting * last sample's bit value + (1 - this setting) * this sample's bit value )." );

	_config->addParameter( "ENOB Simulation/2nd Bit Moving Average Width", 1 );
	_config[ "ENOB Simulation/2nd Bit Moving Average Width" ]->setMin( 1 );
	_config[ "ENOB Simulation/2nd Bit Moving Average Width" ]->setToolTip( "Width of the moving average used to lowpass filter the 2nd byte in order to simulate the frequency dependency of the ENOB.\nMust be an odd number!" );
	_config->addParameter( "ENOB Simulation/Randomize 2nd Bit", false );
	_config[ "ENOB Simulation/Randomize 2nd Bit" ]->setToolTip( "If Enabled the 2nd bit is randomized by comparing a flat distributed random number with the moving average." );
	_config->addParameter( "ENOB Simulation/Weight of last sample's 2nd bit result in the calculation of this one's", 0.0 );
	_config[ "ENOB Simulation/Weight of last sample's 2nd bit result in the calculation of this one's" ]->setMin( 0.0 );
	_config[ "ENOB Simulation/Weight of last sample's 2nd bit result in the calculation of this one's" ]->setMax( 1.0 );
	_config[ "ENOB Simulation/Weight of last sample's 2nd bit result in the calculation of this one's" ]->setToolTip( "This sample's result is computed as unit step ( this setting * last sample's bit value + (1 - this setting) * this sample's bit value )." );

	_config->addParameter( "ENOB Simulation/3rd Bit Moving Average Width", 1 );
	_config[ "ENOB Simulation/3rd Bit Moving Average Width" ]->setMin( 1 );
	_config[ "ENOB Simulation/3rd Bit Moving Average Width" ]->setToolTip( "Width of the moving average used to lowpass filter the 3rd byte in order to simulate the frequency dependency of the ENOB.\nMust be an odd number!" );
	_config->addParameter( "ENOB Simulation/Randomize 3rd Bit", false );
	_config[ "ENOB Simulation/Randomize 3rd Bit" ]->setToolTip( "If Enabled the 3rd bit is randomized by comparing a flat distributed random number with the moving average." );
	_config->addParameter( "ENOB Simulation/Weight of last sample's 3rd bit result in the calculation of this one's", 0.0 );
	_config[ "ENOB Simulation/Weight of last sample's 3rd bit result in the calculation of this one's" ]->setMin( 0.0 );
	_config[ "ENOB Simulation/Weight of last sample's 3rd bit result in the calculation of this one's" ]->setMax( 1.0 );
	_config[ "ENOB Simulation/Weight of last sample's 3rd bit result in the calculation of this one's" ]->setToolTip( "This sample's result is computed as unit step ( this setting * last sample's bit value + (1 - this setting) * this sample's bit value )." );

	rand = gsl_rng_alloc(gsl_rng_taus);
	gsl_rng_set(rand, QDateTime::currentDateTime().toTime_t());
}

Quantize::~Quantize() {
	gsl_rng_free(rand);
}

void Quantize::configChanged() {
	_mawidthbit[0] = _config->getParameter( "ENOB Simulation/LSB Moving Average Width" )->toInt();
	if (_mawidthbit[0]%2 != 1) {
		qDebug() << "WARNING: LSB Moving Average Width (" << _mawidthbit[0] << ") must be an odd number!";
		_config->getParameter( "ENOB Simulation/LSB Moving Average Width" )->setValue(_mawidthbit[0]-1);
	}
	_mawidthbit[1] = _config->getParameter( "ENOB Simulation/2nd Bit Moving Average Width" )->toInt();
	if (_mawidthbit[1]%2 != 1) {
		qDebug() << "WARNING: LSB Moving Average Width (" << _mawidthbit[1] << ") must be an odd number!";
		_config->getParameter( "ENOB Simulation/2nd Bit Moving Average Width" )->setValue(_mawidthbit[1]-1);
	}
	_mawidthbit[2] = _config->getParameter( "ENOB Simulation/3rd Bit Moving Average Width" )->toInt();
	if (_mawidthbit[2]%2 != 1) {
		qDebug() << "WARNING: LSB Moving Average Width (" << _mawidthbit[2] << ") must be an odd number!";
		_config->getParameter( "ENOB Simulation/3rd Bit Moving Average Width" )->setValue(_mawidthbit[2]-1);
	}
	_randomize[0] = _config->getParameter( "ENOB Simulation/Randomize LSB" )->toBool();
	_randomize[1] = _config->getParameter( "ENOB Simulation/Randomize 2nd Bit" )->toBool();
	_randomize[2] = _config->getParameter( "ENOB Simulation/Randomize 3rd Bit" )->toBool();
	_uselastratio[0] = _config->getParameter( "ENOB Simulation/Weight of last sample's LSB result in the calculation of this one's" )->toDouble();
	_uselastratio[1] = _config->getParameter( "ENOB Simulation/Weight of last sample's 2nd bit result in the calculation of this one's" )->toDouble();
	_uselastratio[2] = _config->getParameter( "ENOB Simulation/Weight of last sample's 3rd bit result in the calculation of this one's" )->toDouble();
	_fullscale = _config->getParameter( "Fullscale" )->toDouble();
	_offset = _config->getParameter( "Offset" )->toDouble();
	_resolution = _config->getParameter( "Resolution" )->toInt();
	_stepspervolt = pow(2.0, _resolution)/_fullscale;
}

void Quantize::modify( NSCore::AcqisitionDataSet& data, int oldgeneration ) {

	int samples = data.samples(); // Cache this to be faster


	for ( int i=0; i<data.channels(); ++i ) {
		double* olddata = data.getRawChannel( i, oldgeneration );
		double* newdata = data.getRawChannel( i );

		/*if( GlobalLimiter::the()->show( "Quantize/Debug" ) ) {
			qDebug() << "stepspervolt =" << _stepspervolt;
			unsigned int tmp;
			qDebug() << " " << olddata[20] << " becomes " << (tmp = floor((olddata[20]-offset+fullscale/2.0)*_stepspervolt));
		    qDebug() << "   and back again to " << double(tmp)/_stepspervolt+offset-fullscale/2.0;
		}*/

		// quantize
		unsigned int qdata[samples];
		for ( int j=0; j<samples; ++j )
		    if (olddata[j] < -_fullscale/2.0+_offset)
				qdata[j] = 0;
		    else if (olddata[j] > _fullscale/2.0+_offset)
				qdata[j] = (1U << _resolution)-1;
		    else
				qdata[j] = floor((olddata[j]-_offset+_fullscale/2.0)*_stepspervolt);

		// initialize moving averages
		int mavones[3] = {0,0,0};
		int mavcontainssamples[3] = {0,0,0};
		for (int k=0; k<3; k++)
		    for (int m=0; m<(_mawidthbit[k]-1)/2; m++) {
				if (m >= samples) {
					qDebug() << "WARNING: ENOB moving average range larger than waveform!!";
					break;
				}
				mavones[k] += ((qdata[m] & (1U << k)) > 0);
				mavcontainssamples[k]++;
		    }

		// adjust lower bits and save
		for ( int j=0; j<samples; ++j ) {

		    // update moving averages for lower bits
		    for (int k=0; k<3; k++)
			if (_mawidthbit[k] > 1) {
			    // add next sample if possible
			    if (j+(_mawidthbit[k]-1)/2 < samples) {
					mavones[k] += ((qdata[j+(_mawidthbit[k]-1)/2] & (1U << k)) > 0);
					mavcontainssamples[k]++;
			    }
			    // remove old sample if possible
			    if (j-(_mawidthbit[k]-1)/2 >= 0) {
					mavones[k] -= ((qdata[j-(_mawidthbit[k]-1)/2] & (1U << k)) > 0);
					mavcontainssamples[k]--;
			    }

			    // replace bit in the result with their mav-value
			    Q_ASSERT(mavcontainssamples[k] <= _mawidthbit[k]);
			    Q_ASSERT(mavcontainssamples[k] > 0);
			    bool setbit = _randomize[k] ? mavones[k] > gsl_ran_flat(rand, 0.0, double(mavcontainssamples[k]))
						       : mavones[k]*2 > mavcontainssamples[k];
			    if (j>0)
					setbit = qRound(_uselastratio[k] * double((qdata[j] & (1U << k)) > 0) + (1.0-_uselastratio[k]) * double(setbit));
			    if (setbit)
					qdata[j] += ((qdata[j] & (1U << k)) > 0) ? 0 : (1U << k);
			    else
					qdata[j] -= ((qdata[j] & (1U << k)) > 0) ? (1U << k) : 0;

			}

		    // save
		    newdata[j] = double(qdata[j])/_stepspervolt+_offset-_fullscale/2.0;
		}

		// add fullscale and offset info to channel
		data.addResult( QString( "channel %1/fullscale" ).arg( i ), _fullscale );
		data.addResult( QString( "channel %1/offset" ).arg( i ), -_offset );
	}
	data.updateMinMax();
}
