/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "scriptworkspace.h"
#include "scriptworkspace.moc"
#include "scriptenvironment.h"
#include "scriptsnippet.h"
#include "scriptintrospection.h"

#include <QtCore/QDebug>
#include <QtGui/QMdiArea>
#include <QtGui/QMdiSubWindow>
#include <QtGui/QDockWidget>
#include <QtGui/QMenu>
#include <QtGui/QMenuBar>
#include <QtGui/QToolBar>
#include <QtGui/QAction>
#include <QtGui/QFileDialog>
#include <QtCore/QFile>
#include <QtGui/QMessageBox>

using namespace NSScript;

ScriptWorkSpace::ScriptWorkSpace( ScriptEnvironment* env, QWidget* p )
	: QMainWindow( p )
	, _env( env )
	, _mdiarea( new QMdiArea( this ) )
	, _introspec( new ScriptIntrospection( _env, 0 ) )
{
	setCentralWidget( _mdiarea );
	_mdiarea->setVerticalScrollBarPolicy( Qt::ScrollBarAsNeeded );
	_mdiarea->setHorizontalScrollBarPolicy( Qt::ScrollBarAsNeeded );

	QMenu* filemenu = menuBar()->addMenu( "File" );
	QMenu* scriptmenu = menuBar()->addMenu( "Script" );
	QMenu* windowmenu = menuBar()->addMenu( "Windows" );
	QMenu* settings = windowmenu->addMenu( "Settings" );
	QToolBar* toolbar = addToolBar( "Actions" );

	QAction* newsnippet = new QAction( QIcon( ":/icons/kde/cr22-action-filenew.png" ), "New Code Snippet", this );
	connect( newsnippet, SIGNAL( triggered() ), this, SLOT( newCodeSnippet() ) );
	filemenu->addAction( newsnippet );
	toolbar->addAction( newsnippet );
	QAction* opensnippet = new QAction( QIcon( ":/icons/kde/cr22-action-fileopen.png" ), "Load Code Snippet...", this );
	connect( opensnippet, SIGNAL( triggered() ), this, SLOT( loadCodeSnippet() ) );
	filemenu->addAction( opensnippet );
	toolbar->addAction( opensnippet );
	QAction* closesnippet = new QAction( QIcon( ":/icons/kde/cr22-action-fileclose.png" ), "Close Active Snippet", this );
	connect( closesnippet, SIGNAL( triggered() ), this, SLOT( closeActiveSnippet() ) );
	filemenu->addAction( closesnippet );
	toolbar->addAction( closesnippet );

	QDockWidget* dockintro = new QDockWidget( "Script Environment Introspection", this );
	settings->addAction( dockintro->toggleViewAction() );
	dockintro->setWidget( _introspec );
	dockintro->setFeatures( dockintro->features() | QDockWidget::DockWidgetVerticalTitleBar );
	addDockWidget( Qt::LeftDockWidgetArea, dockintro );
	_introspec->on_btnUpdate_clicked();

	QAction* tileaction = new QAction( "Tile Windows", this );
	connect( tileaction, SIGNAL( triggered() ), _mdiarea, SLOT( tileSubWindows() ) );
	windowmenu->addAction( tileaction );
	QAction* cascadeaction = new QAction( "Cascade Windows", this );
	connect( cascadeaction, SIGNAL( triggered() ), _mdiarea, SLOT( cascadeSubWindows() ) );
	windowmenu->addAction( cascadeaction );

	toolbar->addSeparator();

	QAction* execaction = new QAction( QIcon( ":/icons/kde/cr22-action-player_play.png" ), "Execute", this );
	execaction->setToolTip( "Execute the active Script Snippet" );
	connect( execaction, SIGNAL( triggered() ), this, SLOT( executeActiveSnippet() ) );
	scriptmenu->addAction( execaction );
	toolbar->addAction( execaction );

	QAction* stopaction = new QAction( QIcon( ":/icons/kde/cr22-action-player_stop.png" ), "Stop!", this );
	stopaction->setToolTip( "Stop the current running script." );
	connect( stopaction, SIGNAL( triggered() ), this, SLOT( stopScript() ) );
	scriptmenu->addAction( stopaction );
	toolbar->addAction( stopaction );

	connect( _env, SIGNAL( snippetsToRestoreAvailable() ), this, SLOT( restoreSnippets() ) );
}

void ScriptWorkSpace::setVisible( bool n ) {
	QWidget::setVisible( n );
	emit visibilityChanged( n );
}

void ScriptWorkSpace::newCodeSnippet() {
	//qDebug() << "ScriptWorkSpace::newCodeSnippet()";
	newCodeSnippet( _env->newSnippet( 0 ) );
}
void ScriptWorkSpace::loadCodeSnippet() {
	//qDebug() << "ScriptWorkSpace::loadCodeSnippet()";
	QString filename = QFileDialog::getOpenFileName( this, "Open a script snippet", "", "*.js\n*.*" );
	if ( filename.isEmpty() )
		return;
	QFile file( filename );
	if ( file.open( QIODevice::ReadOnly ) ) {
		QString code = file.readAll();
		qDebug() << " Read:" << code;
		ScriptSnippet* tmp = _env->newSnippet( 0 );
		newCodeSnippet( tmp );
		tmp->setCode( code );
		tmp->setName( filename.replace( ".js","" ).split( "/" ).last() );
		tmp->setSaveLocation( ScriptSnippet::OwnJsFile );
		file.close();
	}
}

void ScriptWorkSpace::closeActiveSnippet() {
	QMdiSubWindow* tmp = _mdiarea->activeSubWindow();
	if ( QMessageBox::warning( this, "Close Script Snippet?", "<qt>Do you really want to close the active Script Snipped?<p>This can not be un-done.</qt>", QMessageBox::Yes|QMessageBox::No, QMessageBox::No ) == QMessageBox::Yes )
		_mdiarea->removeSubWindow( tmp );
}
void ScriptWorkSpace::closeAllSnippets() {
}

void ScriptWorkSpace::newCodeSnippet( ScriptSnippet* w ) {
	qDebug() << "ScriptWorkSpace::newCodeSnippet(" << w << ")";
	QMdiSubWindow* tmp = _mdiarea->addSubWindow( w );
	connect( w, SIGNAL( destroyed( ScriptSnippet* ) ), this, SLOT( removedSnippet( ScriptSnippet* ) ) );
	tmp->setWindowIcon( w->windowIcon() );
	tmp->show();
}
void ScriptWorkSpace::removedSnippet( ScriptSnippet* w ) {
	QMdiSubWindow* tmp = 0;
	foreach( QMdiSubWindow* w, _mdiarea->subWindowList() ) {
		if ( w->widget() == w )
			tmp = w;
	}
	if ( tmp ) {
		_mdiarea->removeSubWindow( tmp );
	}
}

void ScriptWorkSpace::executeActiveSnippet() {
	qDebug() << "ScriptWorkSpace::executeActiveSnippet()";
	QWidget* active = _mdiarea->activeSubWindow()->widget();
	ScriptSnippet* snippet = qobject_cast<ScriptSnippet*>( active );
	if ( snippet ) {
		qDebug() << "Active Window is a snippet";
		snippet->on_btnExec_clicked();
	}
}
void ScriptWorkSpace::stopScript() {
	qDebug() << "ScriptWorkSpace::stopScript() [defunct]";
}

void ScriptWorkSpace::restoreSnippets() {
	while ( _env->hasUnrestoredSnippets() )
		newCodeSnippet( _env->restoreFirstSnippet( 0 ) );
}

