/*
    Copyright 2005-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "resulttospectrum.h"
#include "resulttospectrum.moc"
#include "resulttospectrum_p.h"
#include "resulttospectrum_p.moc"

#include <QtCore/QDebug>
#include <QtCore/QTimer>
#include <QtXml/QDomDocument>
#include <QtGui/QIcon>

#include <math.h>
#include <mathexpression.h>
#include "spectrummodel.h"
#include <thread_template.h>
#include <tiaction.h>

using namespace NSSpectrum;

ResultAction::ResultAction( QString name, ResultsToSpectrum* parent )
	: NSCore::MathTreeRoot( name, parent, true )
	, _model( 0 )
{
	//qDebug() << "ResultAction::ResultAction(" << parent << ")";
	setSingleChild( false );
}

ResultAction::~ResultAction() {
	//qDebug() << "ResultAction::~ResultAction()";
	if ( _model != 0 ) {
		delete _model;
		_model = 0;
	}
	//debugPrint();
}

void ResultAction::data( NSCore::AcqisitionDataSet dataset ) {
	QList<double> results;
	foreach ( NSCore::MathExpression* exp, childobjects<NSCore::MathExpression>() ) {
		results << exp->value( dataset );
	}
	_model->data( results );
}

QDomNode ResultAction::saveToXML( QDomDocument doc ) {
	QDomElement node = doc.createElement( "spectrum" );
	node.setAttribute( "name", objectName() );
	node.setAttribute( "dimensions", _model->dimensions() );
	return node;
}
bool ResultAction::restoreFromXML( QDomNode node ) {
	if ( node.nodeName() != "spectrum" ) {
		return false;
	}
	QDomElement elem = node.toElement();
	setObjectName( elem.attributes().namedItem( "name" ).nodeValue() );
	int dimensions = 1;
	if ( elem.attributes().contains( "dimensions" ) ) {
		dimensions = elem.attributes().namedItem( "dimensions" ).nodeValue().toInt();
	}
	if ( dimensions == 2 )
		_model = new TwoDimSpectrumModel( this );
	else
		_model = new ClassicSpectrumModel( this );
	_model->setObjectName( objectName() );
	addChild( _model->config() );
	/*qDebug() << "this=" << this;
	qDebug() << " _model=" << _model;
	qDebug() << " _model->config()=" << _model->config();
	qDebug() << "\033[32m_model->config()->treeParent()\033[0m" << _model->config()->treeParent();
	qDebug() << "\033[33m" << children() << endl << _model->children() << "\033[0m";*/
	QDomElement path = elem.firstChildElement( "parameter" );
	if ( ! path.isNull() ) {
		_model->config()->fromXml( path );
	}

	return true;
}


Q_DECLARE_METATYPE( QModelIndex )
Q_DECLARE_METATYPE( NSCore::AcqisitionDataSet )

ResultsToSpectrum::ResultsToSpectrum( QObject* p )
	: NSCore::QTreeItem( "ResultsToSpectrum", p )
{
	/*qDebug() << "ResultsToSpectrum::ResultsToSpectrum() maxThreadCount ="
		<< QThreadPool::globalInstance()->maxThreadCount()
		<< "activeThreadCount ="
		<< QThreadPool::globalInstance()->activeThreadCount();*/
	init( QThreadPool::globalInstance()->maxThreadCount() - QThreadPool::globalInstance()->activeThreadCount() );
	/*qDebug() << "ResultsToSpectrum::ResultsToSpectrum() maxThreadCount ="
		<< QThreadPool::globalInstance()->maxThreadCount()
		<< "activeThreadCount ="
		<< QThreadPool::globalInstance()->activeThreadCount();*/
}
ResultsToSpectrum::ResultsToSpectrum( int nrThreads, QObject* p )
	: NSCore::QTreeItem( "ResultsToSpectrum", p )
{
	init( nrThreads );
}
void ResultsToSpectrum::init( int nrThreads ) {
	_threadcycler = 0;
	qRegisterMetaType<QModelIndex>();
	qRegisterMetaType<NSCore::AcqisitionDataSet>();
	nrThreads = qMax( 1, nrThreads );
	qDebug() << "ResultsToSpectrum::init(" << nrThreads << " (effective) )";
	for ( int i=0; i<nrThreads; ++i ) {
		QThreadPool::globalInstance()->reserveThread();
		_threads.append( new NSCore::Thread_Template<ResultThread>( this ) );
		_threads.last()->start();
		_threads.last()->waitForStarted();
		_threads.last()->item()->setFirst( i==0 );
		connect( _threads.last()->item(), SIGNAL( newSpectrum( NSSpectrum::AbstractSpectrumModel* ) ), this, SIGNAL( newSpectrum( NSSpectrum::AbstractSpectrumModel* ) ) );
		_threads.last()->item()->setParent( this );
	}
	{
		QDomDocument add( "epos-action" );
		add.setContent( QString( "<action name=\"Add Spectrum\">\
<argument type=\"string\"><title>Spectrum name</title><label>Please enter the name of the new spectrum.</label></argument>\
<argument type=\"int\"><title>Dimenstions</title><label>Please enter the number of (base) dimensions \
of the spectrum. Currently 1 and 2 are supported.</label><range min=\"1\" max=\"2\" /></argument> \
</action>\
" ) );
		connect( new NSCore::TiAction( add, this ), SIGNAL( activated( QList<QVariant> ) ),
				this, SLOT( localExecAdd( QList<QVariant> ) ) );
	}
}
ResultsToSpectrum::~ResultsToSpectrum() {
}

void ResultsToSpectrum::removeAllPaths() {
	foreach( NSCore::QTreeItem* item, childobjects<NSCore::QTreeItem>() )
		item->deleteLater();
}

void ResultsToSpectrum::addPath( QString xml ) {
	qDebug() << "ResultsToSpectrum::addPath(" << xml << ")";
	QDomDocument doc( "bla" );
	doc.setContent( xml );
	fromXml( doc.documentElement() );
}

bool ResultsToSpectrum::restoreFromXML( QDomNode node ) { return node.nodeName() == "resultstospectrum"; }
bool ResultsToSpectrum::createFromXML( QDomNode node ) {
	if( thread() != QThread::currentThread() )
		qFatal( "ResultsToSpectrum::createFromXML() Not in my own thread!" );
	//qDebug() << "ResultsToSpectrum::createFromXML(" << &node << ")";
	if ( node.nodeName() == "resultstospectrum" ) {

		if ( ! node.attributes().contains("guiadd") )
			foreach ( ResultAction* action, childobjects<ResultAction>() ) {
				action->deleteLater();
			}

		QDomNode elem = node.firstChildElement( "spectrum" );
		while ( ! elem.isNull() ) {
			_threadcycler %= _threads.size();
			_threads[ _threadcycler ]->item()->addXml( elem );
			_threadcycler += 1;
			elem = elem.nextSiblingElement( "spectrum" );
		}

	}
	//qDebug() << "ResultsToSpectrum::createFromXML()  finished!";
	return false;
}

QStringList ResultsToSpectrum::spectra() const {
	QStringList ret = childlist<ResultAction>();
	qDebug() << "ResultsToSpectrum::spectra() returning: " << ret.join( "," );
	return ret;
}

AbstractSpectrumModel* ResultsToSpectrum::spectrum( QString name ) {
	qDebug() << "ResultsToSpectrum::spectrum(" << name << ")";
	ResultAction* action = child<ResultAction>( name );
	if ( action )
		return action->_model;
	return 0;
}
void ResultsToSpectrum::saveSpectrumTo( QString name, QIODevice* dev ) {
	ResultAction* action = child<ResultAction>( name );
	if ( action ) {
		action->_model->saveTo( dev );
	}
}
void ResultsToSpectrum::clearAllSpectra() {
	foreach (ResultAction* action, childobjects<ResultAction>()) {
		action->_model->clearSpectrum();
	}
}

void ResultsToSpectrum::reset() {
	clearAllSpectra();
}

void ResultsToSpectrum::localExecAdd( QList<QVariant> args ) {
	qDebug() << "ResultsToSpectrum::localExecAdd(" << args << ")";
	if ( args.size() < 2 || args[0].type() != QVariant::String || args[1].type() != QVariant::Int ) {
		qWarning() << "Wrong arguments for ResultsToSpectrum::localExecAdd():" << args;
		return;
	}
	addPath( QString(
		"<resultstospectrum guiadd=\"True\"><spectrum name=\"%1\" dimensions=\"%2\"/></resultstospectrum>"
		).arg( args[0].toString() ).arg( args[1].toInt() ) );
}

QVariant ResultsToSpectrum::guiIdentifier( int role ) const {
	switch ( role ) {
		case Qt::DecorationRole:
			return QIcon( ":/icons/kde/cr48-apps-edu_mathematics.png" );
			break;
		default:
			return QTreeItem::guiIdentifier( role );
	}
}

QList<NSCore::DataSetChainMember*> ResultsToSpectrum::chainmembers() {
	QList<NSCore::DataSetChainMember*> ret;
	for (int i=0; i<_threads.size(); i++) {
		ret.append(_threads[i]->item());
	}
	return ret;
}
