/*
    Copyright 2005-2010 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RESULTTOSPECTRUM_H
#define RESULTTOSPECTRUM_H

#include <datasetchain.h>
#include <thread_template.h>

class QTimer;

namespace NSSpectrum {

class ResultsToSpectrum;
class AbstractSpectrumModel;
class ResultAction;
class ResultThread;

/**
 * @brief Sends results into a spectrummodel
 *
 * This class sends results into a SpectrumModel for further analysis and
 * storage.
 *
 * It will also be expanded to store into other models as well ( network-send,
 * sql-database, etc ).
 *
 * The results to be written into the storage can be defined by an own language
 * containing result,add,neg,smb,precision. Use and explore it online by
 * accessing this objects via TreeItemModel and QTreeView...
 */
class ResultsToSpectrum : public NSCore::QTreeItem
{
	Q_OBJECT
	public:
		/// constructor
		ResultsToSpectrum( QObject* =0 );
		/**
		 * Set the number of child-threads by hand
		 */
		ResultsToSpectrum( int nrThreads, QObject* );
		/// destructor
		~ResultsToSpectrum();

		/**
		 * @brief returns a list of the current spectra
		 */
		QStringList spectra() const;
		/**
		 * @brief Access the created SpectrumModels by name
		 */
		NSSpectrum::AbstractSpectrumModel* spectrum( QString name );
		/**
		 * @brief save the spectrum to the device
		 */
		void saveSpectrumTo( QString name, QIODevice* dev );

		//@{
		/**
		 * @brief name and actions for the gui
		 */
		QVariant guiIdentifier( int ) const;
		//@}

		/**
		 * @brief Re-implement from DataSetChainMember
		 *
		 * Reset all counters, clear all spectra.
		 */
		void reset();

		QList<NSCore::DataSetChainMember*> chainmembers();

	public slots:
		/**
		 * @brief Clear all spectra
		 */
		void clearAllSpectra();

		/**
		 * @brief Removes all paths
		 */
		void removeAllPaths();

		/**
		 * @brief Create path and spectrum by XML
		 */
		void addPath( QString xml );

	signals:
		/**
		 * @brief A new spectrum is created.
		 */
		void newSpectrum( NSSpectrum::AbstractSpectrumModel* );

	private slots:

		void localExecAdd( QList<QVariant> );

	private:

		void init( int nrThreads );

		/// debug for NSCore::QTreeItem
		QString debug() const { return "ResultsToSpectrum"; }

		QDomNode saveToXML( QDomDocument doc ) { return doc.createElement( "resultstospectrum" ); }

		bool restoreFromXML( QDomNode );
		bool createFromXML( QDomNode );

		QList< NSCore::Thread_Template<ResultThread>* > _threads;
		int _threadcycler;
};

}; // NSSpectrum

#endif // RESULTTOSPECTRUM_H

