/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef RESULTTOSPECTRUM_P_H
#define RESULTTOSPECTRUM_P_H

#include <mathexpression.h>
#include <QtCore/QThread>
#include "spectrummodel.h"

class QTimer;

namespace NSSpectrum {

class ResultsToSpectrum;

/**
 * @brief A path of expressions together with a spectrummodel
 */
class ResultAction : public NSCore::MathTreeRoot
{
	public:
		ResultAction( QString name, ResultsToSpectrum* parent =0 );
		~ResultAction();

		void data( NSCore::AcqisitionDataSet );

		QString guiString() const { return QString( "Spectrum: %1" ).arg( objectName() ); }

		AbstractSpectrumModel* _model;

	private:
		QDomNode saveToXML( QDomDocument );
		bool restoreFromXML( QDomNode );
		QString debug() const { return QString( "ResultAction: %1" ).arg( objectName() ); }
};


class ResultThread: public NSCore::DataSetChainMember
{
	Q_OBJECT
	public:
		ResultThread( bool first=false )
			: NSCore::DataSetChainMember("ResultThread", 0)
			, _parent( 0 )
			, _first( first )
		{
			//qDebug() << "ResultThread::ResultThread()";
			_timer = new QTimer( this );
			_timer->setInterval( 500 );
			connect( _timer, SIGNAL( timeout() ), this, SLOT( doWork() ) );
			_timer->setSingleShot( true );
			_timer->start();
		}
		~ResultThread() {
			//qDebug() << "ResultThread::~ResultThread()";
		}

		void addXml( QDomNode node ) {
			_tocreate.append( node );
		}

		void setParent( ResultsToSpectrum* p ) { _parent = p; }

		QDomNode saveToXML(QDomDocument) {
			return QDomNode();
		}
		bool restoreFromXML(QDomNode) {
			return false;
		}

		void setFirst( bool first ) { _first = first; }

	signals:
		void newSpectrum( NSSpectrum::AbstractSpectrumModel* );

	private slots:
		void doWork() {
			parseXml();
				_timer->start( 500 );
		}
		void parseXml() {
			if ( _tocreate.size() == 0 ) return;
			//qDebug() << "ResultThread::parseXml()";
			while ( _tocreate.size() > 0 ) {
				QDomNode node = _tocreate.takeFirst();
				ResultAction* tmp = new ResultAction( node.toElement().attribute( "name" ) );
				connect( tmp, SIGNAL( destroyed( QObject* ) ), this, SLOT( removeAction( QObject* ) ) );
				_actions.append( tmp );
				_parent->addChild( tmp );
				tmp->fromXml( node );
				emit newSpectrum( tmp->_model );
			}
		}

		void removeAction( QObject* obj ) {
			QMutexLocker lock(& _actions_mutex );
			ResultAction* action = static_cast<ResultAction*>( obj );
			_actions.removeAll( action );
		}

	private:

		bool dataArrived(NSCore::AcqisitionDataSet data) {
			//qDebug() << "ResultAction::dataArrived(" << data.samples() << ")";
			foreach ( ResultAction* action, _actions ) {
				action->data( data );
			}
			return _first;
		}

		void reset() {
			foreach ( ResultAction* action, _actions ) {
				action->_model->clearSpectrum();
			}
		}

		QString debug() const { return "ResultThread"; }

		QList<QDomNode> _tocreate;
		QList<ResultAction*> _actions;
		QMutex _actions_mutex;
		QTimer* _timer;
		ResultsToSpectrum* _parent;
		bool _first;
};

}; // NSSpectrum

#endif // RESULTTOSPECTRUM_P_H

