/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "spectrumgui.h"
#include "spectrumgui.moc"

#include "spectrumgui_base.h"

#include <QtGui/QTableView>
#include <QtGui/QFileDialog>
#include <QtCore/QDebug>
#include <QtGui/QTreeView>

#include "spectrummodel.h"
#include <chartview.h>
#include <xyspecview.h>
#include <flickcharm.h>
#include <qtreeitemmodel.h>

namespace NSSpectrum {

SpectrumGuiCreator::SpectrumGuiCreator( QObject* p ) : QObject( p ) {
	//qDebug() << "SpectrumGuiCreator::SpectrumGuiCreator(" << p << ")";
}
SpectrumGuiCreator::~SpectrumGuiCreator() {
	//qDebug() << "SpectrumGuiCreator::~SpectrumGuiCreator()";
}
void SpectrumGuiCreator::newSpectrum( AbstractSpectrumModel* model ) {
	//qDebug() << "SpectrumGuiCreator::newSpectrum(" << model << ")";
	SpectrumGui* tmp = new SpectrumGui();
	tmp->setModel( model );
	tmp->setObjectName( model->objectName() );
	connect( model, SIGNAL( destroyed() ), tmp, SLOT( deleteLater() ) );
	connect( this, SIGNAL( autoSave() ), model, SLOT( autoSave() ) );
	emit newSpectrumGui( tmp );
}

SpectrumGui::SpectrumGui( QWidget* p )
	: QFrame( p )
	, _model( 0 )
	, _table( new QTableView( this ) )
	, _chart( new ChartView( this ) )
	, _parameters( 0 )
	, _flick( new FlickCharm( this ) )
{
	_flick->activateOn( _chart );
	_flick->activateOn( _table );
	_base = new Ui::SpectrumGUI_Base();
	_base->setupUi( this );
	while ( _base->tabs->count() > 0 ) {
		delete _base->tabs->widget( 0 );
		_base->tabs->removeTab( 0 );
	}
	_base->tabs->addTab( _chart, "Graph" );
	_base->tabs->addTab( _table, "Table" );
	connect( _base->btnSave, SIGNAL( clicked() ), this, SLOT( saveToFile() ) );

	connect( _base->btnLoad, SIGNAL( clicked() ), this, SLOT( loadFromFile() ) );

	connect( _chart, SIGNAL( rangeAdded( Range* ) ), this, SLOT( rangeAdded( Range* ) ) );
}
SpectrumGui::~SpectrumGui() {
	//qDebug() << "SpectrumGui::~SpectrumGui()";
}

void SpectrumGui::setModel( AbstractSpectrumModel* model ) {
	//qDebug() << "SpectrumGui::setModel(" << model << ")";
	_model = model;
	if ( model != 0 ) {
		connect( _base->btnClear, SIGNAL( clicked() ), model, SLOT( clearSpectrum() ) );
		if ( _model->dimensions() == 2 ) {
			delete _chart;
			_chart = new XYSpecView( this );
			_flick->activateOn( _chart );
			_base->tabs->insertTab( 0, _chart, "Graph" );
		}
		if ( _parameters != 0 )
			delete _parameters;
		NSCore::QTreeItemModel* timodel = new NSCore::QTreeItemModel( this );
		timodel->addToRootItem( _model->config() );
		QTreeView* view = new QTreeView( this );
		view->setModel( timodel );
		_parameters = view;
		_base->tabs->addTab( _parameters, "Parameters" );
	} else {
		delete _parameters;
		_parameters = 0;
	}
	_table->setModel( _model );
	_chart->setModel( _model );
}

void SpectrumGui::saveToFile() {
	if ( _model ) {
		QString filename;
		filename = QFileDialog::getSaveFileName( this, "Save Spectrum to file", "", _model->supportedFilters() + "\nAll files (*)" );
		if ( !filename.isNull() )
			_model->saveToFile( filename );
	}
}

void SpectrumGui::loadFromFile() {
	if ( _model ) {
		QString filename = QFileDialog::getOpenFileName( this, "Load Spectrum from file", "",
				_model->supportedFilters() + "\nAll files (*)" );

		if ( !filename.isNull() )
			_model->loadFromFile( filename );
	}
}

void SpectrumGui::rangeAdded( Range* r ) {
	connect( r, SIGNAL( rangeChanged() ), this, SLOT( rangeChanged() ) );
}

void SpectrumGui::rangeChanged() {
	Range* r = dynamic_cast<Range*>( sender() );
	if ( r ) {
		QString out = QString( "Range: %1 - %2\n" ).arg( r->min ).arg( r->max );
		for ( int j=0; j<_model->columnCount(); ++j ) {
			double wholesum =0, partsum =0;
			for ( int i=0; i<_model->rowCount(); ++i ) {
				double y = dynamic_cast<QAbstractItemModel*>(_model)->data( _model->index( i, j ), Qt::DisplayRole ).toDouble();
				double x = _model->headerData( i, Qt::Vertical ).toDouble();
				wholesum += y;
				if ( r->min < x && r->max > x )
					partsum += y;
			}
			out += QString( "Col(%1): %2 (%3%)\n" ).arg( j ).arg( partsum ).arg( partsum / wholesum * 100 );
		}
		r->label = out;
	}
}

}; // NSSpectrum
