/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#include "spectrummodel.h"
#include "spectrummodel.moc"

#include <math.h>
#include <QtCore/QDebug>
#include <QtCore/QFile>
#include <QtCore/QTextStream>

#include <modeltest.h>

using namespace NSSpectrum;

AbstractSpectrumModel::AbstractSpectrumModel( QObject* p )
	: QAbstractItemModel( p )
	, _config( new NSCore::TiParameter( "options", QVariant::Invalid, 0 ) )
	, _changed_index_timer( new QTimer( this ) )
	, _autosavetimer( new QTimer( this ) )
	, _cps( 0 )
	, _cps_timer( new QTimer( this ) )
	, _total_in_spec( 0 )
{
	connect( _changed_index_timer, SIGNAL( timeout() ), this, SLOT( changedIndexTimerTimeout() ) );
	_changed_index_timer->start( 1000 );
	connect( _cps_timer, SIGNAL( timeout() ), this, SLOT( cpsTimerTimeout() ) );
	_cps_timer->start( 10000 );

	connect( _autosavetimer, SIGNAL( timeout() ), this, SLOT( autoSave() ) );

	_config->addParameter( "Autosave", QVariant::Bool );
	_config->addParameter( "Autosave/Filename", "autosave-DATE.csv" );
	_config->getParameter( "Autosave/Filename" )->setToolTip( "Variables to be extended: DATE - current date-time combination, MTIME - start time of the spectrum since last reset or since creation of the spectrum." );
	_config->addParameter( "Autosave/Interval", 10 );
	_config->getParameter( "Autosave/Interval" )->setRange( 1, 60 );
	connect( _config->getParameter( "Autosave" ), SIGNAL( guiValueChanged() ), this, SLOT( autoSaveChanged() ) );

	_config->addParameter( "In Spectrum", QVariant::Invalid );
	_config->addParameter( "In Spectrum/Total Counts", QVariant::Int );
    _config->getParameter( "In Spectrum/Total Counts" )->setGuiHint( NSCore::TiParameter::Counter );
	_config->addParameter( "In Spectrum/Countrate", QVariant::Double );
    _config->getParameter( "In Spectrum/Countrate" )->setGuiHint( NSCore::TiParameter::ReadAllowed );

	_mtime = QDateTime::currentDateTime();
}
AbstractSpectrumModel::~AbstractSpectrumModel() {
	//qDebug() << "AbstractSpectrumModel::~AbstractSpectrumModel()";
	_config->deleteLater();
}

QModelIndex AbstractSpectrumModel::parent( const QModelIndex& ) const {
	return QModelIndex();
}
void AbstractSpectrumModel::changedIndexTimerTimeout() {
	if ( !_changed_index.isEmpty() ) {
		QMutexLocker lock( &_changed_index_lock );
		qSort( _changed_index );
		emit dataChanged( _changed_index.first(), _changed_index.last() );
		_changed_index.clear();
	}
}
void AbstractSpectrumModel::cpsTimerTimeout() {
	_config->getParameter( "In Spectrum/Total Counts" )->setValue( _total_in_spec );
	_config->getParameter( "In Spectrum/Countrate" )->setValue( _cps / 10.0 );
	_cps = 0;
}

void AbstractSpectrumModel::autoSaveChanged() {
	//qDebug() << "AbstractSpectrumModel::autoSaveChanged()";
	//_config->getParameter( "Autosave" )->debugPrint();
	static bool lastactive = false;
	static int lastinterval = 0;
	if ( lastinterval != _config->getParameter( "Autosave/Interval" )->toInt() ) {
		lastinterval = _config->getParameter( "Autosave/Interval" )->toInt();
		_autosavetimer->stop();
		lastactive = false;
	}
	if ( lastactive != _config->getParameter( "Autosave" )->toBool() ) {
		_autosavetimer->start( lastinterval * 60 * 1000 );
	}
}
void AbstractSpectrumModel::autoSave() {
	//qDebug() << "AbstractSpectrumModel::autoSave()";
	if ( ! _config->getParameter( "Autosave" )->toBool() ) return;

	QString filename = _config->getParameter( "Autosave/Filename" )->toString();
	filename.replace( "DATE", QDateTime::currentDateTime().toString( "yyyyMMdd-hhmmss" ) );
	filename.replace( "MTIME", _mtime.toString( "yyyyMMdd-hhmmss" ) );

	saveToFile( filename );
}
void AbstractSpectrumModel::saveToFile( QString filename ) {
	if (QFile::exists(filename)) {
		QFile::remove(filename + "~");
		QFile::copy(filename, filename + "~");
	}
	QFile f(filename);
	if (f.open(QIODevice::WriteOnly)) {
		saveTo(&f);
	}
}
void AbstractSpectrumModel::loadFromFile( QString n ) {
	qWarning( "\n\033[01;31mSorry, loading spectra from file is not implemented in this spectrum-model.\033[0m\n" );
}


TwoDimSpectrumModel::TwoDimSpectrumModel( QObject* p )
	: AbstractSpectrumModel( p )
{
	_cache.rows = _cache.cols = 0;
	new ModelTest( this, this );

	_config->addParameter( "X-Axis", QVariant::Invalid );
	NSCore::TiParameter* xaxis = _config->getParameter( "X-Axis" );
	xaxis->addParameter( "Minimum", -2.0 );
	xaxis->getParameter( "Minimum" )->setRange( -INFINITY, INFINITY );
	xaxis->addParameter( "Maximum", 8.0 );
	xaxis->getParameter( "Maximum" )->setRange( -INFINITY, INFINITY );
	xaxis->addParameter( "Channels", 100 );
	xaxis->getParameter( "Channels" )->setRange( 1, pow( 2,16 ) );
	connect( xaxis, SIGNAL( guiValueChanged() ), this, SLOT( clearSpectrum() ) );
	_config->addParameter( "Y-Axis", QVariant::Invalid );
	NSCore::TiParameter* yaxis = _config->getParameter( "Y-Axis" );
	yaxis->addParameter( "Minimum", -2.0 );
	yaxis->getParameter( "Minimum" )->setRange( -INFINITY, INFINITY );
	yaxis->addParameter( "Maximum", 8.0 );
	yaxis->getParameter( "Maximum" )->setRange( -INFINITY, INFINITY );
	yaxis->addParameter( "Channels", 100 );
	yaxis->getParameter( "Channels" )->setRange( 1, pow( 2,16 ) );
	connect( yaxis, SIGNAL( guiValueChanged() ), this, SLOT( clearSpectrum() ) );
	clearSpectrum();
}
int TwoDimSpectrumModel::columnCount( const QModelIndex& parent ) const {
	if ( parent.isValid() ) return 0;
	return _cache.cols;
}
int TwoDimSpectrumModel::rowCount( const QModelIndex& parent ) const {
	if ( parent.isValid() ) return 0;
	return _cache.rows;
}
QModelIndex TwoDimSpectrumModel::index( int row, int col, const QModelIndex& parent ) const {
	if ( parent.isValid() || row < 0 || col < 0 || row >= rowCount() || col >= columnCount() ) {
		return QModelIndex();
	}
	return createIndex( row, col );
}
void TwoDimSpectrumModel::data( QList<double> items ) {
	//qDebug() << "TwoDimSpectrumModel::data(" << x << "," << y << ")";
	while ( items.size() > 1 ) {
		_cps++;
		double x = items.takeFirst();
		double y = items.takeFirst();
		if ( isnan( x ) || isnan( y ) )
			break;

		if ( !( x < _cache.xmax && x > _cache.xmin ) )
			break;
		if ( !( y < _cache.ymax && y > _cache.ymin ) )
			break;

		int _x = int( ( x - _cache.xmin ) / _cache.xstep );
		int _y = int( ( y - _cache.ymin ) / _cache.ystep );
		_data[ _x ][ _y ] += 1;
		_total_in_spec++;
		QMutexLocker lock( &_changed_index_lock );
		_changed_index.append( createIndex( _y,_x ) );
	}
}
QVariant TwoDimSpectrumModel::data( const QModelIndex& index, int role ) const {
	if ( role != Qt::DisplayRole )
		return QVariant();

	//qDebug() << "TwoDimSpectrumModel::data(" << index << "," << role << ")";

	if ( index.row() >= _cache.rows )
		return QVariant();
	if ( index.column() >= _cache.cols )
		return QVariant();

	if ( role == Qt::DisplayRole ) {
		data_t::const_iterator x = _data.find( index.column() );
		if ( x != _data.end() ) {
			std::map<int,int>::const_iterator y = x->second.find( index.row() );
			if ( y != x->second.end() )
				return y->second;
		}
	}

	return QVariant();
}
QVariant TwoDimSpectrumModel::headerData( int section, Qt::Orientation orientation, int role ) const {
	if ( role != Qt::DisplayRole )
		return QVariant();

	if ( orientation == Qt::Horizontal && section < _cache.cols )
		return ( _cache.xmax - _cache.xmin ) / _cache.cols * section + _cache.xmin;
	if ( orientation == Qt::Vertical && section < _cache.rows )
		return ( _cache.ymax - _cache.ymin ) / _cache.rows * section + _cache.ymin;

	return QVariant();
}
void TwoDimSpectrumModel::saveTo( QIODevice* dev ) {
	//qDebug() << "TwoDimSpectrumModel::saveTo( QString" << path << ")";
	double xmin = _cache.xmin;
	//double xmax = _cache.xmax;
	double ymin = _cache.ymin;
	//double ymax = _cache.ymax;
	double xstep = _cache.xstep;
	double ystep = _cache.ystep;
	if ( dev->isWritable() ) {
		QTextStream stream( dev );

		stream << "ValueX ValueY Count" << endl;
		//foreach( int x, _data.keys() ) {
		for ( data_t::iterator it=_data.begin(); it!=_data.end(); ++it ) {
			int x = ( *it ).first;
			//foreach( int y, _data[ x ].keys() ) {
			for ( std::map<int,int>::iterator it2=( *it ).second.begin(); it2!=( *it ).second.end(); ++it2 ) {
				int y = ( *it2 ).first;
				//stream << xmin + xstep*x << " " << ymin + ystep*y << " " << _data[ x ][ y ] << endl;
				stream << xmin + xstep*x << " " << ymin + ystep*y << " " << ( *it2 ).second << endl;
			}
		}
	}
}
void TwoDimSpectrumModel::clearSpectrum() {
	//qDebug() << "TwoDimSpectrumModel::clearSpectrum()";
	//beginRemoveRows( QModelIndex(), 0, __rows );
	/*QList<int> keysx = _data.keys();
	while ( ! _data.isEmpty() ) {
		_data.take( keysx.takeFirst() ).clear();
	}*/
	_data.clear();
	//endRemoveRows();
	_mtime = QDateTime::currentDateTime();
	//beginInsertRows( QModelIndex(), 0, __rows );
	//beginInsertColumns( QModelIndex(), 0, __cols );
	//endInsertRows();
	//endInsertColumns();
	emit dataChanged( index( 0,0, QModelIndex() ), index( rowCount()-1, columnCount()-1, QModelIndex() ) );
	updateCache();
	_total_in_spec = 0;
}
void TwoDimSpectrumModel::updateCache() {
	_cache.cols = _config->getParameter( "X-Axis/Channels" )->toInt();
	_cache.xmin = _config->getParameter( "X-Axis/Minimum" )->toDouble();
	_cache.xmax = _config->getParameter( "X-Axis/Maximum" )->toDouble();
	_cache.xstep = ( _cache.xmax - _cache.xmin ) / _cache.cols;
	_cache.rows = _config->getParameter( "Y-Axis/Channels" )->toInt();
	_cache.ymin = _config->getParameter( "Y-Axis/Minimum" )->toDouble();
	_cache.ymax = _config->getParameter( "Y-Axis/Maximum" )->toDouble();
	_cache.ystep = ( _cache.ymax - _cache.ymin ) / _cache.rows;
}


ClassicSpectrumModel::ClassicSpectrumModel( QObject* p )
	: AbstractSpectrumModel( p )
{
	new ModelTest( this, this );

	//qDebug() << "ClassicSpectrumModel::ClassicSpectrumModel(" << p << ")";
	_config->addParameter( "Minimum", -10.0 );
	_config->getParameter( "Minimum" )->setRange( -INFINITY, INFINITY );
	_config->addParameter( "Maximum", 10.0 );
	_config->getParameter( "Maximum" )->setRange( -INFINITY, INFINITY );
	_config->addParameter( "Channels", 200 );
	_config->getParameter( "Channels" )->setRange( 1, pow( 2, 24 ) );
	connect( _config->getParameter( "Minimum" ), SIGNAL( guiValueChanged() ), this, SLOT( redoChannels() ) );
	connect( _config->getParameter( "Maximum" ), SIGNAL( guiValueChanged() ), this, SLOT( redoChannels() ) );
	connect( _config->getParameter( "Channels" ), SIGNAL( guiValueChanged() ), this, SLOT( redoChannels() ) );
	_config->addParameter( "Columns", 1 );
	connect( _config->getParameter( "Columns" ), SIGNAL( guiValueChanged() ), this, SLOT( changeColumns() ) );
	_config->addParameter( "Nanoseconds per Sample", 0.25 );
	_config->getParameter( "Nanoseconds per Sample" )->setRange( 0.01, 5000 );
	_config->addParameter( "Key value for *.dat-file", "300" );

	/**
	 * Seems as if calling clearSpectrum during construction is not good on the positron-machine...
	 */
	//clearSpectrum();
	int channels = config()["Channels"]->toInt();
	_cache.min = config()[ "Minimum" ]->toDouble();
	_cache.max = config()[ "Maximum" ]->toDouble();
	_cache.step = ( _cache.max - _cache.min ) / channels;

	beginInsertRows( parent( QModelIndex() ), 0, channels-1 );
	_data.resize( 1 );
	_data[ 0 ].resize( channels );
	endInsertRows();

	//qDebug() << "ClassicSpectrumModel::ClassicSpectrumModel() finished";
}
ClassicSpectrumModel::~ClassicSpectrumModel() {
	//qDebug() << "ClassicSpectrumModel::~ClassicSpectrumModel()";
}

QModelIndex ClassicSpectrumModel::index( int row, int column, const QModelIndex& parent ) const {
	if ( parent.isValid() || row<0 || column<0 ) {
		return QModelIndex();
	}
	if ( row < rowCount() && column < columnCount() ) {
		return createIndex( row, column );
	}
	return QModelIndex();
}
int ClassicSpectrumModel::columnCount( const QModelIndex& ) const {
	return _data.size();
}
int ClassicSpectrumModel::rowCount( const QModelIndex& parent ) const {
	if ( parent.isValid() ) {
		return 0;
	}
	if ( _data.size() == 0 )
		return 0;
	return _data[ 0 ].size();
}
QVariant ClassicSpectrumModel::headerData( int section, Qt::Orientation orientation, int role ) const {
	if ( role != Qt::DisplayRole )
		return QVariant();

	if ( orientation == Qt::Vertical && _data.size() > 0 )
		if ( section < _data[ 0 ].size() )
			return section * _cache.step + _cache.min;
		return QVariant();

	if ( orientation == Qt::Horizontal ) {
		return "Count";
	}

	return QVariant();
}

void ClassicSpectrumModel::data( QList<double> items ) {
	//qDebug() << "ClassicSpectrumModel::data(" << n << ")";
	_cps += items.size();
	for ( int i=0; i<items.size(); ++i ) {
		double n = items[i];
		//qDebug() << n << "is" << int((n-_cache.min)/_cache.step) << "*" << _cache.step;
		if ( n > _cache.min && n < _cache.max ) {
			int row = int((n-_cache.min)/_cache.step);
			if ( row < _data[ i % _data.size() ].size() ) {
				_data[ i % _data.size() ][ row ] += 1;
				_total_in_spec++;
				QMutexLocker lock( &_changed_index_lock );
				_changed_index.push_back( createIndex( row, 1 ) );
			}
		}
	}
}

QVariant ClassicSpectrumModel::data( const QModelIndex& index, int role ) const {
	//qDebug() << "ClassicSpectrumModel::data( const QModelIndex& " << index << ", int " << role << ")";
	if ( index.isValid() && role == Qt::DisplayRole && index.column() < _data.size() && index.row()
			< _data[ index.column() ].size() ) {
		return _data[ index.column() ].at( index.row() );
	}
	return QVariant();
}
void ClassicSpectrumModel::saveTo( QIODevice* dev ) {
	qDebug() << "ClassicSpectrumModel::saveTo(" << dev << ")";
	int channels = config()["Channels"]->toInt();
	if ( dev->isWritable() ) {
		QTextStream stream( dev );

		QFile* f = dynamic_cast<QFile*>(dev);

		if ( f && f->fileName().endsWith( ".dat" ) ) {
			stream << QString( "; KeyValue is '%1'\r\n" ).arg( _config->getParameter( "Key value for *.dat-file" )->toString() );
			// First is bin-width
			stream << _config->getParameter( "Nanoseconds per Sample" )->toDouble() * _cache.step << "\r\n";
			// Second an arificial key value to identify the measurement.
			stream << _config->getParameter( "Key value for *.dat-file" )->toString() << "\r\n";
			// Third is the accuracy of the measurement
			stream << 0.200 << "\r\n";

			for ( int i=0; i<channels; ++i ) {
				for ( int j=0; j<_data.size(); ++j )
					stream << " " << _data[ j ][ i ];
				stream << "\r\n";
			}
		}
		else {
			for ( int i=0; i<channels; ++i ) {
				stream << _cache.min + i*_cache.step;
				for ( int j=0; j<_data.size(); ++j )
					stream << " " << _data[ j ][ i ];
				stream << endl;
			}
		}
	}
}
void ClassicSpectrumModel::loadFromFile( QString path ) {
	qDebug() << "ClassicSpectrumModel::loadFromFile(" << path << ")";
	QFile file( path );
	if ( file.open( QIODevice::ReadOnly ) ) {

		if ( path.endsWith( ".csv" ) ) {
			QList<double> keys;
			QList< QList<int> > d;
			QString line;
			while ( ! file.atEnd() ) {
				line = file.readLine().simplified();
				QStringList list = line.split(" ", QString::SkipEmptyParts);
				keys.append( list.takeFirst().toDouble() );
				//qDebug() << key << list;
				QList<int> values;
				foreach ( QString n, list )
					values.append( n.toDouble() );
				d.append( values );
			}
			qDebug() << d;
			double xmin = keys.first();
			double xmax = keys.last() + keys[1] - keys[0];
			_config->getParameter( "Minimum" )->setValue( xmin );
			_config->getParameter( "Maximum" )->setValue( xmax );
			_config->getParameter( "Channels" )->setValue( d.size() );
			_config->getParameter( "Columns" )->setValue( d[0].size() );
			_total_in_spec = 0;
			//qDebug() << "_data.size()" << _data.size() << " d.size()" << d.size();
			for ( int i=0; i<d.size(); ++i ) {
				for ( int j=0; j<d[i].size() && j<_data.size(); ++j ) {
					//qDebug() << i << j << _data[j][i] << d[i][j];
					_data[j][i] = d[i][j];
					_total_in_spec += d[i][j];
				}
			}
			emit dataChanged( createIndex( 0, 0 ), createIndex( d.size(), _data.size() ) );
		}
		else if ( path.endsWith( ".dat" ) ) {
			qFatal( "\033[01;31mOpening *.dat-files is not supported.\033[0m" );
		}

		file.close();
	}
}

void ClassicSpectrumModel::clearSpectrum() {
	redoChannels();
	_mtime = QDateTime::currentDateTime();
	//_config->getParameter( "In Spectrum/Total Counts" )->setValue( 0 );
	_total_in_spec = 0;
}

void ClassicSpectrumModel::redoChannels() {
	//qDebug() << "Something changed, redoing channels";

	int channels = config()["Channels"]->toInt();
	if (_cache.min != config()[ "Minimum" ]->toDouble()
		   || _cache.max != config()[ "Maximum" ]->toDouble()
		   || channels != _data[ 0 ].size() ) {
		_cache.min = config()[ "Minimum" ]->toDouble();
		_cache.max = config()[ "Maximum" ]->toDouble();
		_cache.step = ( _cache.max - _cache.min ) / channels;

		beginRemoveRows( QModelIndex(), 0, _data[ 0 ].size()-1 );
		for ( int i=0; i<_data.size(); ++i )
			_data[ i ].clear();
		endRemoveRows();

		beginInsertRows( QModelIndex(), 0, channels-1 );
		for ( int i=0; i<_data.size(); ++i )
			_data[ i ].resize( channels );
		endInsertRows();
	} else {
		for ( int i=0; i<_data.size(); ++i ) {
			_data[ i ].resize( 0 );
			_data[ i ].resize( channels );
			emit dataChanged( createIndex(0, 0), createIndex( channels, _data.size() ) );
		}
	}
}

void ClassicSpectrumModel::changeColumns() {
	int channels = config()["Channels"]->toInt();
	int columns = _config->getParameter( "Columns" )->toInt();
	if ( _data.size() > columns ) {
		beginRemoveColumns( QModelIndex(), columns, _data.size() - 1 );
		_data.resize( columns );
		endRemoveColumns();
	}
	if ( _data.size() < columns ) {
		beginInsertColumns( QModelIndex(), _data.size(), columns-1 );
		_data.resize( columns );
		for ( int i=0; i<_data.size(); ++i )
			_data[ i ].resize( channels );
		endInsertColumns();
	}
}

