/*
    Copyright 2005-2009 EPOS-Project
                        (FZ Dresden-Rossendorf & MLU Halle-Wittenberg)
                        Arnold Krille <arnold@arnoldarts.de>

    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This library is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; see the files lgpl-3.0.txt and
    gpl-3.0.txt. If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef SPECTRUMMODEL_H
#define SPECTRUMMODEL_H

#include <QtCore/QAbstractItemModel>
#include <QtCore/QMap>
#include <QtCore/QList>
#include <QtCore/QMutex>
#include <QtCore/QReadWriteLock>
#include <QtCore/QTimer>
#include <QtCore/QDateTime>

#include <tiparameter.h>

namespace NSSpectrum
{

/**
 * @brief Abstract baseclass for SpectrumModels
 *
 * This class provides the basic interface for SpectrumModels. That way several
 * models can benefit from one gui.
 *
 * @see SpectrumGui
 */
class AbstractSpectrumModel : public QAbstractItemModel
{
	Q_OBJECT
	public:
		/// simple default-constructor
		AbstractSpectrumModel( QObject* p );

		/// simple destructor
		~AbstractSpectrumModel();

		// QAbstractItemModel::virtual functions
		QModelIndex parent( const QModelIndex& ) const;

		/// Access the configuration parameters
		NSCore::TiParameterPointer config() const { return NSCore::TiParameterPointer( _config ); }

		/**
		 * @brief return the supported filters
		 *
		 * Implement this to return the valid file-endings your model can save.
		 * Is used in the file selection dialog.
		 */
		virtual QString supportedFilters() const =0;

		/**
		 * @brief return the number of dimensions of this spectrum
		 */
		virtual int dimensions() const =0;

		/**
		 * @brief Save the spectrum to an iodevice
		 */
		virtual void saveTo( QIODevice* ) =0;

	public slots:
		/**
		 * @brief Save the spectrum to a file
		 */
		void saveToFile( QString );

		/**
		 * @brief Clear the whole spectrum
		 */
		virtual void clearSpectrum() =0;

		/**
		 * @brief Do the autosave (if wanted)
		 */
		void autoSave();

		/**
		 * @brief Add data to the spectrum
		 *
		 * The list of double shoul get devided by the number of dimensions. Then it is possible to
		 * add data to multiple sub-spectra with just one call.
		 */
		virtual void data( QList<double> ) =0;

		/**
		 * @brief Load a spectrum from a file
		 */
		virtual void loadFromFile( QString );

	private slots:

		void changedIndexTimerTimeout();
		void cpsTimerTimeout();

		void autoSaveChanged();

	protected:

		NSCore::TiParameter* _config;

		QTimer* _changed_index_timer;
		QMutex _changed_index_lock;
		QList<QModelIndex> _changed_index;

		QTimer* _autosavetimer;
		QDateTime _mtime;

		int _cps;
		QTimer* _cps_timer;
		int _total_in_spec;
};


/**
 * @brief Two dimensional spectrum
 *
 * For example for doppler-coinc.
 *
 * Or for getting the correlation between energy-by-minimum and energy-by-integration. Or for
 * getting lifetime-vs-stop-energy...
 */
class TwoDimSpectrumModel : public AbstractSpectrumModel
{
	Q_OBJECT
	public:
		/// constructor
		TwoDimSpectrumModel( QObject* );

		/**
		 * @{
		 * @brief Model-functions
		 */
		QModelIndex index( int, int, const QModelIndex& = QModelIndex() ) const;
		int columnCount( const QModelIndex& = QModelIndex() ) const;
		int rowCount( const QModelIndex& = QModelIndex() ) const;
		QVariant headerData( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const;
		QVariant data( const QModelIndex&, int = Qt::DisplayRole ) const;
		// @}

		QString supportedFilters() const { return "Comma separated file (*.csv)"; }

		int dimensions() const { return 2; }

		void saveTo( QIODevice* );

	public slots:
		/**
		 * @brief Adds data to the model
		 */
		void data( QList<double> );

		void clearSpectrum();

	private slots:
		void updateCache();

	private:
		typedef std::map< int, std::map<int, int> > data_t;
		data_t _data;

		/// Cache for clearSpectrum
		struct {
			int rows, cols;
			double xmin, xmax, xstep;
			double ymin, ymax, ystep;
		} _cache;
};


/**
 * @brief A model for classic Spectra
 *
 * This class provides a model for classic spectra which have a min and max and
 * a number of channels.
 */
class ClassicSpectrumModel : public AbstractSpectrumModel
{
	Q_OBJECT
	public:
		ClassicSpectrumModel( QObject* );
		~ClassicSpectrumModel();

		/**
		 * @{
		 * @brief QAbstractItemModel::virtual functions
		 */
		QModelIndex index( int, int, const QModelIndex& = QModelIndex() ) const;
		int rowCount( const QModelIndex& = QModelIndex() ) const;
		QVariant data( const QModelIndex&, int = Qt::DisplayRole ) const;
		QVariant headerData( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const;
		int columnCount( const QModelIndex& = QModelIndex() ) const;
		// @}

		QString supportedFilters() const { return "Comma separated File (*.csv)\nDAT for LT (*.dat)"; }

		int dimensions() const { return 1; }

		void saveTo( QIODevice* );

	public slots:
		void data( QList<double> );

		void loadFromFile( QString );

		void clearSpectrum();

	private slots:
		void redoChannels();
		void changeColumns();

	private:
		struct
		{
			double min, max, step;
		} _cache;

		QVector< QVector<int> > _data;

};

} // namespace NSSpectrum

#endif // SPECTRUMMODEL_H

